#include "DetectionTestMainFrame.h"
#include "logger.h"
#include "SerialPortHelper.h"
#include "communication/clients/SerialPortScanner.h"
#include "events/SerialPortScannerEvent.h"
#include "events/CommunicationEvent.h"
#include "events/VemsDeviceControllerEvent.h"
#include "vemsprotocol/DeviceModeController.h"


//helper functions
enum wxbuildinfoformat {
    short_f, long_f };

wxString wxbuildinfo(wxbuildinfoformat format)
{
    wxString wxbuild(wxVERSION_STRING);

    if (format == long_f )
    {
#if defined(__WXMSW__)
        wxbuild << _T("-Windows");
#elif defined(__UNIX__)
        wxbuild << _T("-Linux");
#endif

#if wxUSE_UNICODE
        wxbuild << _T("-unicode build");
#else
        wxbuild << _T("-ANSI build");
#endif // wxUSE_UNICODE
    }

    return wxbuild;
}


/// toolbar items
int idButtonDevice              = wxNewId();
int idButtonDeviceMode          = wxNewId();
int idComboBoxBaudrates         = wxNewId();

int idButtonDetect              = wxNewId();


BEGIN_EVENT_TABLE(DetectionTestMainFrame, wxFrame)

    EVT_SERIAL_PORT_SCANNER( SerialPortScannerEvent::SPS_EVT_DEVICE_FOUND, DetectionTestMainFrame::OnDeviceFound )
    EVT_SERIAL_PORT_SCANNER( SerialPortScannerEvent::SPS_EVT_DEVICE_NOT_FOUND, DetectionTestMainFrame::OnDeviceNotFound )
    EVT_SERIAL_PORT_SCANNER( SerialPortScannerEvent::SPS_EVT_CANNOT_OPEN_PORT, DetectionTestMainFrame::OnCannotOpenPort )

    EVT_TIMER( wxID_OK, DetectionTestMainFrame::OnStartDetection )
    EVT_TIMER( wxID_CANCEL, DetectionTestMainFrame::OnChangeMode )

    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_DEVICE_MODE_CHANGED, DetectionTestMainFrame::OnDeviceModeChanged )
    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_DEVICE_DISCONNECTED, DetectionTestMainFrame::OnDeviceDisconnected )

    /// Toolbar items
    EVT_BUTTON( idButtonDetect, DetectionTestMainFrame::OnDetectButton )

END_EVENT_TABLE()

DetectionTestMainFrame::DetectionTestMainFrame(wxFrame *frame, const wxString& title)
    : wxFrame(frame, -1, title)
    , m_detectionTimer ( this, wxID_OK )
    , m_detectionSleep( 1000 )
    , m_changeModeTimer ( this, wxID_CANCEL )
{
    ///toolbar
    wxPanel* mainPanel = new wxPanel( this );
    mainPanel->SetSizer( new wxBoxSizer( wxVERTICAL ) );

    wxButton* deviceButton = new wxButton( mainPanel, idButtonDevice, _T("N/A") );
    deviceButton->Enable( false );

    wxButton* deviceModeButton = new wxButton( mainPanel, idButtonDeviceMode, _T("N/A") );
    deviceModeButton->Enable( false );
    deviceModeButton->Connect( wxID_ANY, wxEVT_RIGHT_DOWN, wxMouseEventHandler(DetectionTestMainFrame::OnDeviceMode), NULL, this );

    wxArrayString baudrates = SerialPortHelper::getSerialBaudrates();

    wxComboBox* baudrateComboBox = new wxComboBox(  mainPanel,
                                                    idComboBoxBaudrates,
                                                    _T("19200"),
                                                    wxDefaultPosition,
                                                    wxDefaultSize,
                                                    baudrates,
                                                    wxCB_READONLY );
    baudrateComboBox->Enable( false );

    wxBoxSizer* toolbar = new wxBoxSizer( wxHORIZONTAL );
    toolbar->Add( deviceButton, 0, wxRIGHT, 5 );
    toolbar->Add( deviceModeButton, 0, 0, 0 );
    toolbar->Add( baudrateComboBox, 0, wxLEFT | wxALIGN_CENTRE_VERTICAL, 5 );
    mainPanel->GetSizer()->Add( toolbar, 0, wxEXPAND | wxALL, 5 );

    wxBoxSizer* toolbar2 = new wxBoxSizer( wxHORIZONTAL );
    wxButton* detectButton = new wxButton( mainPanel, idButtonDetect, _T("Detect") );
//    detectButton->Enable( false );
    toolbar2->Add( detectButton, 0, wxRIGHT, 5 );

    mainPanel->GetSizer()->Add( toolbar2, 0, wxEXPAND | wxALL, 5 );


//    ///log window
//    UserLogWindow* logWindow = new UserLogWindow( mainPanel, wxID_ANY );
//    mainPanel->GetSizer()->Add( logWindow, 1, wxEXPAND );
//    UserLog::Init( logWindow );

    ///DeviceModeSelector Test
    m_modeSequence.Add( ECUInfo::MEGATUNE );
    m_modeSequence.Add( ECUInfo::MAN );
    m_modeSequence.Add( ECUInfo::BOOT );
    m_modeSequence.Add( ECUInfo::MEGATUNE );
    m_modeSequence.Add( ECUInfo::MAN );
    m_modeSequence.Add( ECUInfo::MEGATUNE );
    m_modeSequence.Add( ECUInfo::BOOT );
    m_modeSequence.Add( ECUInfo::MEGATUNE );

    m_modeSequenceIdx = 0;
    m_changeModeTestCount = 10;

    ///SerialPortScanner
    m_serialPortScanner = NULL;
    startDetecting();
}

DetectionTestMainFrame::~DetectionTestMainFrame()
{

}

void DetectionTestMainFrame::startDetecting()
{
    delete m_serialPortScanner;

    wxArrayString ports;
    ports.Add(_T("COM2"));
    IntArray baudrates;
    baudrates.Add(9600);
    baudrates.Add(19200);
    size_t timeout = 20;
    size_t interWriteDelay = 20;

    m_serialPortScanner = new SerialPortScanner( this, ports, baudrates, timeout, interWriteDelay );
}

void DetectionTestMainFrame::startDetectionTimer()
{
    m_detectionTimer.Start( m_detectionSleep, true );
}

void DetectionTestMainFrame::refreshToolbarButton( int buttonId, const wxString& label, const wxColour& color, bool isEnabled )
{
    wxButton* button = (wxButton*)wxWindow::FindWindowById( buttonId );

    if ( !button ) {
        LOGP( wxString::Format( _T("Button not found! - buttonId: %d"), buttonId ) );
    }

    button->SetForegroundColour( color );
    button->Enable( true );
    button->SetLabel( label );
}

void DetectionTestMainFrame::refreshBaudrateComboBox( bool isEnabled, size_t baudrate )
{
    wxComboBox* baudrateComboBox = (wxComboBox*)wxWindow::FindWindowById( idComboBoxBaudrates );
    baudrateComboBox->Enable( isEnabled );
    int itemIdx = baudrateComboBox->FindString( wxString::Format( _T("%d"), baudrate ) );
    if ( itemIdx != wxNOT_FOUND ) {
        baudrateComboBox->Select( itemIdx );
    }
}

void DetectionTestMainFrame::OnDeviceFound( SerialPortScannerEvent& event )
{
    wxString deviceString = ECUInfo::getDeviceAsString( event.getDevice() );
    wxString deviceModeString = ECUInfo::getStateAsString( event.getDeviceMode() );

    LOGP( _T("DEVICE_FOUND: ") + deviceString + _T(", MODE: ") + deviceModeString );

    m_deviceInfo = event.getDeviceInfo();


    wxColor connectedColor( 0, 128, 0 );


    refreshToolbarButton( idButtonDevice, deviceString, connectedColor, true );
    refreshToolbarButton( idButtonDeviceMode, deviceModeString, connectedColor, true );
    refreshBaudrateComboBox( true, event.getBaudrate() );



    delete m_serialPortScanner;
    m_serialPortScanner = NULL;

//
//    m_modeSequenceIdx = 0;
//
//    m_deviceModeController = new DeviceModeController( this, m_deviceInfo, 20, 20 );
//    m_deviceModeController->changeMode( (ECUInfo::EcuMode)m_modeSequence[m_modeSequenceIdx] );

    startDetectionTimer();
}

void DetectionTestMainFrame::OnDeviceNotFound( SerialPortScannerEvent& event )
{
    LOGP( _T("DEVICE_NOT_FOUND") );

    wxString deviceString = _T("N/A");
    wxString deviceModeString = _T("N/A");

    refreshToolbarButton( idButtonDevice, deviceString, *wxRED, true );
    refreshToolbarButton( idButtonDeviceMode, deviceModeString, *wxRED, true );
    refreshBaudrateComboBox( false, 0 );

    m_deviceInfo = DeviceInfo();

    if ( event.isScanningFinished() ) {
        startDetectionTimer();
    }
}

void DetectionTestMainFrame::OnCannotOpenPort( SerialPortScannerEvent& event )
{
    LOGP( _T("CANNOT_OPEN_PORT") );

    wxString deviceString = _T("N/A");
    wxString deviceModeString = _T("N/A");

    refreshToolbarButton( idButtonDevice, deviceString, *wxLIGHT_GREY, true );
    refreshToolbarButton( idButtonDeviceMode, deviceModeString, *wxLIGHT_GREY, true );
    refreshBaudrateComboBox( false, 0 );

    m_deviceInfo = DeviceInfo();

    startDetectionTimer();
}

void DetectionTestMainFrame::OnStartDetection( wxTimerEvent& event )
{
    startDetecting();
}

void DetectionTestMainFrame::OnChangeMode( wxTimerEvent& event )
{
    m_deviceModeController->changeMode( (ECUInfo::EcuMode)m_modeSequence[m_modeSequenceIdx]);
}

void DetectionTestMainFrame::OnCommunicationEvent( CommunicationEvent& event )
{
    LOGP( event.toString() );

    switch ( event.GetId() )
    {
        case CommunicationEvent::REMOVE_USB:
        {
            wxMessageBox( _T("CommunicationError"), _T("REMOVE_USB") );
        } break;
    }
}

void DetectionTestMainFrame::OnDeviceMode( wxMouseEvent& event )
{
    if ( m_deviceInfo.isUnknownMode() ) {
        return;
    }


}

void DetectionTestMainFrame::OnDeviceModeChanged( VemsDeviceControllerEvent& event )
{

    wxColor connectedColor( 0, 128, 0 );

    wxString deviceString = ECUInfo::getDeviceAsString( event.getDevice() );
    wxString deviceModeString = ECUInfo::getStateAsString( event.getDeviceMode() );

    refreshToolbarButton( idButtonDevice, deviceString, connectedColor, true );
    refreshToolbarButton( idButtonDeviceMode, deviceModeString, connectedColor, true );
    refreshBaudrateComboBox( true, event.getBaudrate() );

    if ( ++m_modeSequenceIdx < m_modeSequence.GetCount() ) {
        m_changeModeTimer.Start( 1000, true );

    } else if ( m_changeModeTestCount > 0 ) {
        m_modeSequenceIdx = 0;
        m_changeModeTimer.Start( 1000, true );
    } else {
        delete m_deviceModeController;
        m_deviceModeController = NULL;
    }
}

void DetectionTestMainFrame::OnDeviceDisconnected( VemsDeviceControllerEvent& event )
{
    LOG
}

void DetectionTestMainFrame::OnDetectButton( wxCommandEvent& event )
{
    LOG
}
