#include <math.h>
#include "MGraphics.h"

#include <wx/dcclient.h>
#include <wx/brush.h>

#include <wx/image.h>

#include "../logger.h"


//-----------------------------------------------------------------------
// MAJER JÁNOS
//-----------------------------------------------------------------------
// Super-sampling egyenes rajzolás
// S lépéssel hossz és keresztirányban mintavételezett egyenes rajzolás
// S lépés hossz és keresztirányban, 1 pixelen belül S*S scanposition így
// S^2 sampling (S=5 esetén 25 árnyalat), (S=10 esetén 100 árnyalat)
// (S=16 esetén 256 árnyalat), minél nagyobb S értéke annál több lépés
// (és több gépidő), de jobb a minőség.
// A bejárás a Perspective-Trans és Rotate-Final algoritmusokhoz hasonló pásztázással
//
// Működés
// -------
//   0) Találati Pixellista tömb létrehozása (H * MaxW) = ((L + 2) * (2<=N)), width = 1 px esetén
//      MaxH helyett H, így nem kell HOffset ami minden lépést lassít, kirajzoláskor
//      (SSHit[0].Hit == 0) estén a sor üres!
//   1) Pásztázó koordináták meghatározása
//   2) Pásztázó egyenesek bejárása, találati lista feltöltése
//   3) Találati lista alapján árnyalt pixelek kirajzolása
//   4) Drasztikus futásidő miatt csak LUT (e * X)&(f * X) használatával lehet elviselhető!
//   5) Találati listában keresztirányban nem csinálok rendezést, intuitive a rendezési
//      lépések és a hozzájuk kapcsolódó admin rosszabb választás mint az
//      "észnélküli végigrohanás"
//
//-----------------------------------------------------------------------
// Optimalizálási lehetőségek PDA-ra és PC-re
// ------------------------------------------
//  - lebegőpontos műveletek pontosságának csökkentése, ha lehet!!!
//  - kerekítés átvétele (felesleges léések elkerülése)
//  - (e)-vel és (f)-el való szorzás átvétele LUT-al
//      - Felhasználó megadja a Sampling értékét 2-15 -ig  (3-16)
//        Ennek megfelelően fel kell tölteni az előtérszínhez és háttérszínnek
//        megfelelő LUT-ot Ha van elég memória akár 256*256*256 elemű is lehet
//        Előtérszín - Háttérszín - Érték
//        Egyébként jó megoldás ha az előtér és háttérszínre külön-külön adom
//        meg az értékeket, a túlcsordulás lehetőségét mindig el kell kerülni!
//  -  A találati pixellista mérete PDA esetén 320x4 emiatt egy grafikus osztályban
//     létre lehet hozni állandóra is egy közöset, nem kell állandóan újra létrehozni
//     Csak a törlését kell megoldani.
//
//-----------------------------------------------------------------------
void TMGraphics::SSDrawLine(TRGB RGB, TDPoint P1, TDPoint P2, int S)
  {
    TSSHit    **SSHit;      // Super-Sampling hitlist
    double    dx,   dy;     // Length diferrencies
    double    ddx,  ddy;    // Cross differencies
    double    x,    y;      // LengthRun coordinates
    int       X,    Y;      // Rounded LengthRun coordinates
    TDPoint   P;            // RunCross endpoint
    BYTE      R, G, B;      // Colour elements for macros
    double    L;            // Length of line
    double    Alpha;        // Angle of line
    int       LS;           // Length corrected number of sampling
    double    SS;           // squared of sampling
    bool      Hit;          // we have a hit in a cycle?
    int       i, j;         // common cycle-variables
    double    e, f;         // distances
    BYTE      *pB;          // ScanLine pointer
    int       FailCt;       // Number of errors in a cycle

    FailCt = 0;

    if(15 < S)
      S = 15;

    if(S < 2)
      S = 2;

    //S = 1;

    if((P1.x == P2.x) && (P1.y == P2.y))
      {
        // Zero length line, this is not a point!!!
        return;
      }

    if(!BMP->IsOk())
      {
        // Background bitmap has a problem???
        return;
      }

    W = BMP->GetWidth();
    H = BMP->GetHeight();

    //-----------------------------------------------------------------------
    // End point correction move in the boundaries (pluss 1-1 pixel)

    // ToDo


    //-----------------------------------------------------------------------
    // Create empty HitList
    if(fabs(P2.x - P1.x) < fabs(P2.y - P1.y))
      {
        if(P2.y < P1.y)
          ChangePoint(&P2, &P1);

        dy = P2.y - P1.y;
        dx = P2.x - P1.x;

        Alpha = atan2(dy , dx);
        Alpha = Alpha + (M_PI / 2);

        L = sqrt((dy * dy) + (dx * dx));

        ddx = cos(Alpha) / (S + 0);
        ddy = sin(Alpha) / (S + 0);

        dx = dx / (L * (S + 0));
        dy = dy / (L * (S + 0));

        P.y = P1.y - (S/2.0 * ddy);
        P.x = P1.x - (S/2.0 * ddx);

        /**
        if(0 < ddx)
          P.x = P1.x - (S/2.0 * ddx);
        else
          P.x = P1.x + (S/2.0 * ddx);
        **/

        LS = (int)(L * S);

        SSHit = new TSSHit*[H];

        for(j=0; j<H; j++)
          SSHit[j] = new TSSHit[4];

        //-------------------------------------------------------------------
        // Search hits
        for(i=0; i<=S; i++)
          {
            y = P.y;
            x = P.x;

            for(j=0; j<=LS; j++)
              {
                // Komoly optimalizálási lehetőség a kerekítés átvételével és a pontosság sazbályozásával
                X = Round(x);
                Y = Round(y);

                Hit = false;
                if(SSHit[Y][0].Hit == 0)
                  {
                    SSHit[Y][0].Hit = 1;
                    SSHit[Y][0].XPos = X;
                    Hit = true;
                  }
                else if(SSHit[Y][0].XPos == X)
                  {
                    (SSHit[Y][0].Hit)++;
                    Hit = true;
                  }

                if(!Hit)
                  {
                    if(SSHit[Y][1].Hit == 0)
                      {
                        SSHit[Y][1].Hit = 1;
                        SSHit[Y][1].XPos = X;
                        Hit = true;
                      }
                    else if(SSHit[Y][1].XPos == X)
                      {
                        (SSHit[Y][1].Hit)++;
                        Hit = true;
                      }
                  }
                /**/
                if(!Hit)
                  {
                    if(SSHit[Y][2].Hit == 0)
                      {
                        SSHit[Y][2].Hit = 1;
                        SSHit[Y][2].XPos = X;
                        Hit = true;
                      }
                    else if(SSHit[Y][2].XPos == X)
                      {
                        (SSHit[Y][2].Hit)++;
                        Hit = true;
                      }
                  }

                if(!Hit)
                  {
                    if(SSHit[Y][3].Hit == 0)
                      {
                        SSHit[Y][3].Hit = 1;
                        SSHit[Y][3].XPos = X;
                        Hit = true;
                      }
                    else if(SSHit[Y][3].XPos == X)
                      {
                        (SSHit[Y][3].Hit)++;
                        Hit = true;
                      }
                  }
                /**/

                if(!Hit)
                  FailCt++;

                y += dy;
                x += dx;
              }

            P.y += ddy;
            P.x += ddx;
            /**
            if(0 < ddx)
              P.x += ddx;
            else
              P.x -= ddx;
            **/
          }

        //-------------------------------------------------------------------
        // Draw the hits with shading

        RGB.GetRGB(&R, &G, &B);

        SS = (S + 0) * (S + 0);

        for(Y=0; Y<H; Y++)
          {
            if(SSHit[Y][0].Hit != 0)
              {
                // RGB_RED
                X = SSHit[Y][0].XPos;
                e = SSHit[Y][0].Hit / SS;
                f = 1 - e;

                if(1 < e)
                  {
                    e = 1;
                    f = 0;
                  }

                // BEGIN_ALIAS
                END_ALIAS
                // SET_PIXEL

                if(SSHit[Y][1].Hit != 0)
                  {
                    // RGB_GREEN
                    X = SSHit[Y][1].XPos;
                    e = SSHit[Y][1].Hit / SS;
                    f = 1 - e;

                    if(1 < e)
                      {
                        e = 1;
                        f = 0;
                      }

                    // BEGIN_ALIAS
                    END_ALIAS
                    // SET_PIXEL

                    if(SSHit[Y][2].Hit != 0)
                      {
                        // RGB_BLUE
                        X = SSHit[Y][2].XPos;
                        e = SSHit[Y][2].Hit / SS;
                        f = 1 - e;

                        if(1 < e)
                          {
                            e = 1;
                            f = 0;
                          }

                        // BEGIN_ALIAS
                        END_ALIAS
                        // SET_PIXEL


                        if(SSHit[Y][3].Hit != 0)
                          {
                            // RGB_BLUE
                            X = SSHit[Y][3].XPos;
                            e = SSHit[Y][3].Hit / SS;
                            f = 1 - e;

                            if(1 < e)
                              {
                                e = 1;
                                f = 0;
                              }

                            // BEGIN_ALIAS
                            END_ALIAS
                            // SET_PIXEL
                          }
                      }
                  }
              }
            delete SSHit[Y];
          }

        delete SSHit;
      }
    //-----------------------------------------------------------------------
    //-----------------------------------------------------------------------
    //-----------------------------------------------------------------------
    //-----------------------------------------------------------------------
    else
      {
        // return;

        if(P2.x < P1.x)
          ChangePoint(&P2, &P1);

        dx = P2.x - P1.x;
        dy = P2.y - P1.y;


        Alpha = atan2(dy , dx);
        Alpha = Alpha + (M_PI / 2);

        L = sqrt((dy * dy) + (dx * dx));

        ddx = cos(Alpha) / (S + 0);
        ddy = sin(Alpha) / (S + 0);

        dx = dx / (L * (S + 0));
        dy = dy / (L * (S + 0));

        P.y = P1.y - (S/2.0 * ddy);
        P.x = P1.x - (S/2.0 * ddx);

        LS = (int)(L * S);

        SSHit = new TSSHit*[W];

        for(j=0; j<W; j++)
          SSHit[j] = new TSSHit[4];

        //-------------------------------------------------------------------
        // Search hits
        for(i=0; i<=S; i++)
          {
            y = P.y;
            x = P.x;

            for(j=0; j<=LS; j++)
              {
                // Komoly optimalizálási lehetőség a kerekítés átvételével és a pontosság sazbályozásával
                X = Round(x);
                Y = Round(y);
                if(Y < 0)
                  Y = -1 * Y;

                Hit = false;
                if(SSHit[X][0].Hit == 0)
                  {
                    SSHit[X][0].Hit = 1;
                    SSHit[X][0].XPos = Y;
                    Hit = true;
                  }
                else if(SSHit[X][0].XPos == Y)
                  {
                    (SSHit[X][0].Hit)++;
                    Hit = true;
                  }

                if(!Hit)
                  {
                    if(SSHit[X][1].Hit == 0)
                      {
                        SSHit[X][1].Hit = 1;
                        SSHit[X][1].XPos = Y;
                        Hit = true;
                      }
                    else if(SSHit[X][1].XPos == Y)
                      {
                        (SSHit[X][1].Hit)++;
                        Hit = true;
                      }
                  }
                /**/
                if(!Hit)
                  {
                    if(SSHit[X][2].Hit == 0)
                      {
                        SSHit[X][2].Hit = 1;
                        SSHit[X][2].XPos = Y;
                        Hit = true;
                      }
                    else if(SSHit[X][2].XPos == Y)
                      {
                        (SSHit[X][2].Hit)++;
                        Hit = true;
                      }
                  }

                if(!Hit)
                  {
                    if(SSHit[X][3].Hit == 0)
                      {
                        SSHit[X][3].Hit = 1;
                        SSHit[X][3].XPos = Y;
                        Hit = true;
                      }
                    else if(SSHit[X][3].XPos == Y)
                      {
                        (SSHit[X][3].Hit)++;
                        Hit = true;
                      }
                  }
                /**/

                if(!Hit)
                  FailCt++;

                y += dy;
                x += dx;
              }

            P.y += ddy;
            P.x += ddx;
            /**
            if(0 < ddx)
              P.x += ddx;
            else
              P.x -= ddx;
            **/
          }

        //-------------------------------------------------------------------
        // Draw the hits with shading

        RGB.GetRGB(&R, &G, &B);

        SS = (S + 0) * (S + 0);

        for(X=0; X<W; X++)
          {
            if(SSHit[X][0].Hit != 0)
              {
                // RGB_RED
                Y = SSHit[X][0].XPos;
                e = SSHit[X][0].Hit / SS;
                f = 1 - e;

                if(1 < e)
                  {
                    e = 1;
                    f = 0;
                  }

                // BEGIN_ALIAS
                END_ALIAS
                // SET_PIXEL

                if(SSHit[X][1].Hit != 0)
                  {
                    // RGB_GREEN
                    Y = SSHit[X][1].XPos;
                    e = SSHit[X][1].Hit / SS;
                    f = 1 - e;

                    if(1 < e)
                      {
                        e = 1;
                        f = 0;
                      }

                    // BEGIN_ALIAS
                    END_ALIAS
                    // SET_PIXEL

                    if(SSHit[X][2].Hit != 0)
                      {
                        // RGB_BLUE
                        Y = SSHit[X][2].XPos;
                        e = SSHit[X][2].Hit / SS;
                        f = 1 - e;

                        if(1 < e)
                          {
                            e = 1;
                            f = 0;
                          }

                        // BEGIN_ALIAS
                        END_ALIAS
                        // SET_PIXEL


                        if(SSHit[X][3].Hit != 0)
                          {
                            // RGB_BLUE
                            Y = SSHit[X][3].XPos;
                            e = SSHit[X][3].Hit / SS;
                            f = 1 - e;

                            if(1 < e)
                              {
                                e = 1;
                                f = 0;
                              }

                            // BEGIN_ALIAS
                            END_ALIAS
                            // SET_PIXEL
                          }
                      }
                  }
              }
            delete SSHit[X];
          }

        delete SSHit;
      }

    //-----------------------------------------------------------------------
    // Hibák számának kiírása
    Logger::log(LOG_MSG, wxString::Format(wxT("SSDrawALine -> Number of errors = %d"), FailCt));

  }//TMGraphics::SSDrawLine()
