#include "SymbolMgr.h"
#include <wx/wx.h>

#include "common/types/StringHashSet.h"
#include "vemsprotocol/SIPRTypes.h"
#include "logger.h"

SymbolMgr::SymbolMgr()
{
    //ctor
}

SymbolMgr::~SymbolMgr()
{
    //dtor
}

bool SymbolMgr::isExistSymbol(const wxString& symbolName)
{
    return m_allSymbolTable.find( symbolName ) == m_allSymbolTable.end() ? false : true;
}

const SymbolTable* SymbolMgr::getScalarTable(int type)
{
    if ( type == SiprSensorType::SIPR_RT_MT_EXT || type == SiprSensorType::SIPR_RT_ROUND )
    {
        type = SiprSensorType::SIPR_RT_MT_NORMAL;
    }

    SymbolTables::const_iterator it = m_scalarTables.find(type);
    if( it != m_scalarTables.end() ){
        return it->second;
    }
    return NULL;
}

void SymbolMgr::addOrReplaceSymbol( Symbol* symbol )
{
    if ( !symbol )
    {
        wxLogMessage(_T("Symbol is NULL!!"));
        return;
    }

    wxString key = symbol->getName();

    SymbolTable::iterator it = m_allSymbolTable.find( key );

    if ( it != m_allSymbolTable.end() )
    {
        Symbol* sym = it->second;

        SymbolTable::iterator it2;

        if ( sym->isExpr() )
        {
            it2 = m_expTable.find(key);
            m_expTable.erase( it2 );
        }
        else if ( sym->isBits() )
        {
            it2 = m_bitTable.find(key);
            m_bitTable.erase( it2 );
        }
        else if ( sym->isArray() )
        {
            it2 = m_arrayTable.find(key);
            m_arrayTable.erase( it2 );
        }
        else if ( sym->isScalar() )
        {
            int type = symbol->getInputType();
            SymbolTable* scalarTable = m_scalarTables[type];
            it2 = scalarTable->find(key);
            scalarTable->erase( it2 );
        }
        m_allSymbolTable.erase( it );

        delete sym;
    }

    addSymbol( symbol );
}

void SymbolMgr::addSymbol(Symbol* symbol){
    if(!symbol){
        wxLogMessage(_T("NULL Symbol!: ")); //TEST message?
        return;
    }

    wxString key = symbol->getName();

    if( m_allSymbolTable.find(key) != m_allSymbolTable.end() ){
        wxLogMessage(_T("Duplicate Symbol!: ")+ key );
    }
    m_allSymbolTable[key] = symbol;

    if( symbol->isExpr() ){
        m_expTable[ key ] = symbol;
    }else if( symbol->isBits() ){
        m_bitTable[ key ] = symbol;
    }else if( symbol->isArray() ){
        m_arrayTable[ key ] = symbol;
        addConfigArraySymbolInOrder( key, symbol );
    }else if( symbol->isScalar() ){
        int type = symbol->getInputType();

        SymbolTable*  scalarTable;

        if(m_scalarTables.find(type) == m_scalarTables.end()){
            scalarTable = new SymbolTable();
            m_scalarTables[ type ] = scalarTable;
        }else{
            scalarTable = m_scalarTables[type ];
        }
        (*(scalarTable))[key] = symbol;
    }
    //m_symbolOrder.Add(symbol);
}

Symbol* SymbolMgr::getSymbol(const wxString& symbolName)
{
    SymbolTable::iterator itSymTbl = m_allSymbolTable.find( symbolName );

    if( itSymTbl != m_allSymbolTable.end() ){
        return itSymTbl->second;
    }else{
        return NULL;
    }
}

int SymbolMgr::getSymbolPage(const wxString& symbolName)
{
    Symbol* symbol = getSymbol(symbolName);
    if(symbol == NULL) { return -1; }
    return symbol->getPage();
}

wxString SymbolMgr::getSymbolType(const wxString& symbolName)
{
    Symbol* symbol = getSymbol(symbolName);
    if(symbol == NULL) { return _T(""); }
    return symbol->getType();
}

void SymbolMgr::clear()
{
    WX_CLEAR_HASH_MAP( SymbolTable, m_allSymbolTable );
	m_bitTable.clear();
	m_arrayTable.clear();
	m_expTable.clear();

    WX_CLEAR_HASH_MAP( SymbolGroupHash, m_symbolGroups );
    m_configTableSymbolsInOrder.Clear();

	SymbolTables::iterator it = m_scalarTables.begin();

	for ( ;it != m_scalarTables.end(); ++it )
	{
		SymbolTable*  scalarTable = it->second;
		scalarTable->clear();
		delete scalarTable;
	}

	m_scalarTables.clear();

    m_usedStatic.clear();
    m_configTableSymbolH.clear();
    WX_CLEAR_HASH_MAP( Int2SymbolArrayHash, m_page2SymbolArrayHash );
     WX_CLEAR_HASH_MAP( SymbolTable, m_cfgSymbolTable );
}


void SymbolMgr::addGroup( const wxString& name )
{
    if( m_symbolGroups.find( name ) == m_symbolGroups.end() )
    {
        m_symbolGroups[ name ] = new SymbolGroup( name, 0 );
    }
}


void SymbolMgr::setGroupOffset( const wxString& name, int offset )
{
    if( m_symbolGroups.find( name ) != m_symbolGroups.end() )
    {
        m_symbolGroups[ name ]->offset = offset;
    }
}


SymbolGroup* SymbolMgr::getGroup( const wxString& groupName )
{
    if( m_symbolGroups.find( groupName ) == m_symbolGroups.end() )
    {
        return NULL;
    }
    else
    {
        return m_symbolGroups[ groupName ];
    }
}

const Int2SymbolArrayHash& SymbolMgr::getTheSymbolOrder()
{
    Symbol *symbol ;
    /** set the page order */
    for(SymbolTable::iterator it = m_allSymbolTable.begin(); it != m_allSymbolTable.end(); it++) {
        //wxString symbolName = it->first;
        symbol = it->second;
        int page = symbol->getPage();
        if(  m_page2SymbolArrayHash.find(page) == m_page2SymbolArrayHash.end() ){
            m_page2SymbolArrayHash[page] = new SymbolArray();
        }
        m_page2SymbolArrayHash[page]->Add(symbol);
    }

    /** set the offset order */
    for(Int2SymbolArrayHash::iterator it2 = m_page2SymbolArrayHash.begin(); it2 !=m_page2SymbolArrayHash.end(); it2++) {
        int page = it2->first;
        m_page2SymbolArrayHash[page]->Sort(&SymbolMgr::compareByPageAndOffset);
    }
    return m_page2SymbolArrayHash;
}

int SymbolMgr::compareByPageAndOffset(Symbol **A,Symbol **B)
{
    int pageA = (*A)->getPage();
    int pageB = (*B)->getPage();

    int AOffset = (*A)->getOffset();
    int BOffset = (*B)->getOffset();

    if(pageA > pageB){return 1; }
    if(pageA < pageB){return -1; }

    if(AOffset > BOffset){
        return 1;
    }else if(AOffset < BOffset){
        return -1;
    }else if(AOffset == BOffset){// I hope which one is array other is constant
        wxString Atype = (*A)->getType();
        wxString Btype = (*B)->getType();
        int relation = Atype.compare(Btype);
        if (relation > 0){
            return 1;
        }else if(relation < 0){
            return -1;
        }else{
            return 0 ;
        }
    }
    return 0 ;
}

bool SymbolMgr::isConst(const wxString& symbolName)
{
    SymbolTable::iterator itSymTbl = m_allSymbolTable.find( symbolName );
    if( itSymTbl == m_allSymbolTable.end() ){
        return false;
    }
    Symbol* symbol = itSymTbl->second;
    if( symbol->isScalar() && symbol->isConst() ) return true;
    return false;
}

//StringHashSet& variableSet = m_rpnParser->getVariables(expr );

SymbolTable& SymbolMgr::getUsedStaticAndBit()
{
    return m_usedStatic;
}

void SymbolMgr::setUsedStatic(StringHashSet& symbolNameSet)
{
    //m_usedStatic.clear();
    Symbol* symbol = NULL;
    for( StringHashSet::const_iterator itSet = symbolNameSet.begin(); itSet != symbolNameSet.end(); ++itSet ){
        wxString symbolName = *itSet;
        if( isConst(symbolName) ){
            symbol = m_allSymbolTable[symbolName];
        }else if( m_bitTable.find(symbolName) != m_bitTable.end() ){
            symbol = m_bitTable[symbolName];
        }else{
            continue;
        }

        m_usedStatic[symbolName] = symbol;
        //Logger::logf(LOG_MSG, _T("SymbolMgr::setUsedStatic(() add symbol: ")+symbolName);
    }
}

SymbolTable& SymbolMgr::getConfigTableSymbols()
{
    m_configTableSymbolH.clear();
    SymbolTable::iterator it = m_arrayTable.begin();
    for( ;it != m_arrayTable.end(); ++it ){

        wxString key = it->first;
        Symbol* symbol = it->second;

        if( symbol->isVemsCfgArray() ){
        //if( isConfigArraySymbol(key) ){
            //Symbol* symbol = it->second;
             m_configTableSymbolH[key] = symbol;
        }
    }
    return m_configTableSymbolH;
}


SymbolArray& SymbolMgr::getConfigTableSymbolsInOrder()
{
    return m_configTableSymbolsInOrder;
}


void SymbolMgr::addConfigArraySymbolInOrder( const wxString& name, Symbol* symbol )
{
    if ( symbol->isVemsCfgArray() )
    {
        m_configTableSymbolsInOrder.Add( symbol );
        Logger::logf( LOG_MSG, _T("-----SymbolMgr::addConfigArraySymbolInOrder() - ") + name );
    }
}

/*
bool SymbolMgr::isConfigArraySymbol( const wxString& name ) const
{
    //Logger::logf(LOG_MSG, _T("SymbolMgr::isConfigArraySymbol() symbol: ")+name);
    if ( name.Length() == 1 )                           /// j,l,n, etc tables
    {
        //Logger::logf(LOG_MSG, _T("SymbolMgr::isConfigArraySymbol() 1 symbol: ")+name);
        return true;
    }
    else if ( name.BeforeFirst('_').Length() == 1 )     /// j_r, j_k, etc tables
    {
       //Logger::logf(LOG_MSG, _T("SymbolMgr::isConfigArraySymbol() 2 symbol: ")+name);
        return true;
    }else if(name.Length() == 2){
        long idx
        if( name.Mid(1).ToLong(&idx) ){
            return true;
        }

    }

    return false;
}
*/

void SymbolMgr::addConfigTableSymbolsAsScalar(SymbolTable& symbolTable)
{
    SymbolTable& confgigArrayH = getConfigTableSymbols();
    SymbolTable::iterator it = confgigArrayH.begin();
    for( ;it != confgigArrayH.end(); ++it ){
        wxString baseKey = it->first;
        Symbol* symbol = it->second;
        int size = symbol->getArraySize();
        int page = symbol->getPage();
        int offset = symbol->getOffset();
        int i;
        for(i= 0; i < size; ++i){
            wxString key = baseKey + wxString::Format(_T("[%d]"), i);
            Symbol* symbolScalar  = new Symbol();
            symbolScalar->setPage(page);
            symbolScalar->setOffset(offset+i);
            symbolTable[key] = symbolScalar;
        }
    }
}

