#include "VemsAboutDialog.h"

#include <wx/sizer.h>
#include <wx/statbmp.h>
#include <wx/stattext.h>
#include <wx/button.h>
#include <wx/collpane.h>

#include "version.h"
#include "DisplayHelper.h"

bool StringIsURL(const wxString &str)
{
    wxArrayString types;
    types.Add(_T("http://"));
    types.Add(_T("https://"));
    types.Add(_T("ftp://"));

    for(unsigned short a=0;a<types.GetCount();a++)
        if( str.StartsWith(types.Item(a)) )
            return true;

    return false;

}

static wxString AllAsString(const wxArrayString& a)
{
    wxString s;
    const size_t count = a.size();
    s.reserve(20*count);
    for ( size_t n = 0; n < count; n++ )
    {
        s << a[n] << (n == count - 1 ? _T("\n") : _T(", "));
    }

    return s;
}

VemsAboutDialog::VemsAboutDialog(const wxAboutDialogInfo& info)
{
    //ctor
    m_sizerText = NULL;

    Create(info);
}

VemsAboutDialog::VemsAboutDialog()
{
    //ctor
    m_sizerText = NULL;

    Create(VemsAboutData());
    ShowModal();
}


VemsAboutDialog::~VemsAboutDialog()
{
    //dtor
}


bool VemsAboutDialog::Create(const wxAboutDialogInfo& info)
{

    if ( !wxDialog::Create(NULL, wxID_ANY, _("About ") + info.GetName(),
                           wxDefaultPosition, wxDefaultSize, wxRESIZE_BORDER|wxDEFAULT_DIALOG_STYLE|wxSTAY_ON_TOP) )
        return false;


    m_sizerText = new wxBoxSizer(wxVERTICAL);
    wxString nameAndVersion = info.GetName();
    if ( info.HasVersion() )
        nameAndVersion << _T(' ') << info.GetVersion();
    wxStaticText *label = new wxStaticText(this, wxID_ANY, nameAndVersion);
    wxFont fontBig(*wxNORMAL_FONT);
    fontBig.SetPointSize(fontBig.GetPointSize() + 2);
    fontBig.SetWeight(wxFONTWEIGHT_BOLD);
    label->SetFont(fontBig);

    m_sizerText->Add(label, wxSizerFlags().Centre().Border());
    m_sizerText->AddSpacer(5);

    AddText(info.GetCopyright());

    AddText(info.GetDescription());

    if ( info.HasWebSite() ) {

        AddControl(new wxHyperlinkCtrl(this, wxID_ANY,
                                       info.GetWebSiteDescription(),
                                       info.GetWebSiteURL()));
    }


    if(info.HasLicence()) AddCollapsiblePane(wxT("License"), info.GetLicence());

    if(info.HasDevelopers()) AddCollapsiblePane(wxT("Developers"), AllAsString(info.GetDevelopers()));

    if(info.HasDocWriters()) AddCollapsiblePane(wxT("Documentation writers"), AllAsString(info.GetDocWriters()));

    if(info.HasArtists()) AddCollapsiblePane(wxT("Artists"), AllAsString(info.GetArtists()));

    if(info.HasTranslators()) AddCollapsiblePane(wxT("Translators"), AllAsString(info.GetTranslators()));


    DoAddCustomControls();


    wxSizer *sizerIconAndText = new wxBoxSizer(wxHORIZONTAL);

    wxIcon icon = info.GetIcon();
    if ( icon.Ok() )
    {
        sizerIconAndText->Add(new wxStaticBitmap(this, wxID_ANY, icon),
                                wxSizerFlags().Border(wxRIGHT));
    }

    sizerIconAndText->Add(m_sizerText, wxSizerFlags(1).Expand());

    wxSizer *sizerTop = new wxBoxSizer(wxVERTICAL);
    sizerTop->Add(sizerIconAndText, wxSizerFlags(1).Expand().Border());

    wxSizer *sizerBtns = CreateButtonSizer(wxOK);
    if ( sizerBtns )
    {
        sizerTop->Add(sizerBtns, wxSizerFlags().Expand().Border());
    }

    SetSizerAndFit(sizerTop);

    CentreOnScreen();

    SetIcon(wxIcon(_T("vemsLogoIcon2")));

    return true;

}

void VemsAboutDialog::AddHyperlink(const wxString& text, int index)
{

    AddControl(new wxHyperlinkCtrl(this, wxID_ANY, text, text) , index);

}

void VemsAboutDialog::AddControl(wxWindow *win, int index)
{

    AddControl(win, wxSizerFlags().Border(wxDOWN).Centre(),index);

}

void VemsAboutDialog::AddControl(wxWindow *win, const wxSizerFlags& flags, int index)
{
    wxCHECK_RET( m_sizerText, _T("can only be called after Create()") );
    wxASSERT_MSG( win, _T("can't add NULL window to about dialog") );

    if(index == -1)
        m_sizerText->Add(win, flags);
    else
        m_sizerText->Insert(index,win, flags);

}

void VemsAboutDialog::AddText(const wxString& text, int index)
{

    if(text.empty()) return;

    if(StringIsURL(text))
        AddHyperlink(text);
    else
        AddControl(new wxStaticText(this, wxID_ANY, text),index);

}

void VemsAboutDialog::AddCollapsiblePane(const wxString& title, const wxString& text, int index)
{
    wxCollapsiblePane *pane = new wxCollapsiblePane(this, wxID_ANY, title);
    wxStaticText *txt = new wxStaticText(pane->GetPane(), wxID_ANY, text,
                                         wxDefaultPosition, wxDefaultSize,
                                         wxALIGN_CENTRE);

    // don't make the text unreasonably wide
    static const int maxWidth = DisplayHelper::GetDisplaySize().x/3;
    txt->Wrap(maxWidth);

    // NB: all the wxCollapsiblePanes must be added with a null proportion value
    if(index == -1)
        m_sizerText->Add(pane, wxSizerFlags(0).Expand().Border(wxBOTTOM));
    else
        m_sizerText->Insert(index, pane, wxSizerFlags(0).Expand().Border(wxBOTTOM));

}

wxAboutDialogInfo VemsAboutDialog::VemsAboutData()
{

    wxAboutDialogInfo info;

//    info.AddArtist(_T("Artist 1"));
//    info.AddArtist(_T("Artist 2"));
    info.SetCopyright(_T("Versatile Engine Management Systems"));
    info.SetDescription(_T("http://www.vems.hu/manual/vemstune/help2/htmls"));
   // info.AddDeveloper(_T("Developer 1"));
   // info.AddDeveloper(_T("Developer 2"));
   // info.AddDeveloper(_T("Developer 3"));

    //info.SetIcon(_T(""));
    info.SetLicence(_T("Versatile Engine Management Systems"));
    info.SetName(AutoVersion::getAppName() );
    wxString version = wxString::Format(_T("%u.%u.%u"), AutoVersion::MAJOR, AutoVersion::MINOR, AutoVersion::BUILD);
    wxString year = wxString::FromAscii(AutoVersion::YEAR);
    wxString month = wxString::FromAscii(AutoVersion::MONTH);
    wxString date = wxString::FromAscii(AutoVersion::DATE);

    wxString dateStr = _T(" (")+year +_T(".")+ month +_T(".") +date + _T(")");

    info.SetVersion(version+dateStr);
    info.SetWebSite(_T("http://www.vems.hu"));

    return info;

}
