#include "VemsTuneController.h"

#include <wx/filename.h>

#include "version.h"
#include "CommunicationBase.h"

#include "logger.h"
#include "MTConfigHelper.h"
#include "VemsFlagConfigHelper.h"
#include "configmanager/ConfigHelper2.h"
#include "ConfigBase.h"

#include "ThemesHelper.h"
#include "VemsMsgListCtrl.h"

#include "DirHelper.h"
#include "FirmwareInfo.h"
#include "RawLogHelper.h"
#include "AIMParser.h"
#include "ChannelCollision.h"

#include "customwidgets/InfoWindow.h"

#include "rtdatamanager/RTDataMgrFacade.h"
#include "rtdatamanager/RTDataInputSIPR.h"
#include "rtdatamanager/RealtimeDataChecker.h"
#include "rtdatamanager/ConfigSwitchViewer.h"
#include "rtdatamanager/RTDataAnalyzer.h"

#include "loghistory/LogHistoryMgrHelper.h"
#include "ringbuffer/ThreadLogPlayer.h"
#include "ringbuffer/CommunicationThread.h"

#include "stdatamanager/STDataModelSerial.h"
#include "stdatamanager/STDataIOSIPR.h"
#include "stdatamanager/STDataMgrFacade.h"
#include "stdatamanager/STDataViewerHash.h"
#include "stdatamanager/STDataIOMSQ.h"
#include "stdatamanager/STDataModelHelper.h"

#include "events/ECULogEvent.h"
#include "events/SelectInputEvent.h"
#include "events/VemsDeviceControllerEvent.h"
#include "events/InputEvent.h"
#include "events/CommunicationEvent.h"

#include "serialhelper.h"


#include "gps/GPSCommunicationController.h"
#include "vemsprotocol/SIPRFileLogger.h"
#include "vemsprotocol/VemsLogReader.h"
#include "vemsprotocol/VemsLogReaderHelper.h"
#include "vemsprotocol/VemsCommand.h"
#include "vemsprotocol/SIPRPacketHelper.h"
#include "vemsprotocol/SIPRTypes.h"
#include "gaugedesc/GaugeDescHelper.h"
#include "gaugedesc/GaugeDescManager.h"

#include "expression/RPNHelper.h"

#include "firmwareconfig/FirmwareConfigLoader.h"

#include "VemsTuneHelper.h"
#include "common/VemsContext.h"
#include "vemsprotocol/STDataModelWriter.h"
#include "loghistory/LogHistoryMgr.h"
#include "loghistory/QuickAnalyze.h"
#include "loghistory/FilePlayCommander.h"
#include "gui/FileOpenSQL.h"

/// refactor communication
#include "vemsprotocol/VemsDeviceController.h"
#include "configmanager/ConfigMgrCommunication.h"


BEGIN_EVENT_TABLE( VemsTuneController, wxEvtHandler )

    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_CANNOT_OPEN_PORT, VemsTuneController::OnCannotOpenPort )
    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_DEVICE_FOUND, VemsTuneController::OnDeviceFound )
    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_DEVICE_NOT_FOUND, VemsTuneController::OnDeviceNotFound )
    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_DEVICE_CONNECTED, VemsTuneController::OnDeviceConnected )
    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_DEVICE_DISCONNECTED, VemsTuneController::OnDeviceDisconnected )
    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_DEVICE_RECONNECTED, VemsTuneController::OnDeviceReconnected )
    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_STOPPED, VemsTuneController::OnDeviceControllerStopped )
    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_RESUMED, VemsTuneController::OnDeviceResumed )
//    EVT_VEMS_DEVICE_CONTROLLER( VDC_EVENT_DEVICE_PROTECTED, VemsTuneController::OnDeviceProtected )

    EVT_COMMUNICATION( wxID_ANY, VemsTuneController::OnCommunicationError )

END_EVENT_TABLE()


VemsTuneController::VemsTuneController( wxEvtHandler* pEventHandler )
{
    m_selectedInput         = INPUT_TYPE_SIPR;
    m_stInputType           = ST_INPUT_SIPR;
    m_stInputFileName       = _T("");
    m_logHistory            = NULL;
    m_loganalyzer           = NULL;
    m_rtanalyzer            = NULL;
    m_deviceController      = NULL;

    m_vemsLogFileVersion    = 0;

    Connect( wxEVT_INPUT, InputEventHandler( VemsTuneController::OnSIPRInput ) );

    MTConfigHelper::createMTConfigData();
    m_mtConfig = MTConfigHelper::getMTConfigData();

    m_siprFileLogger    =   new SIPRFileLogger( 512 );
    m_pEventHandler = pEventHandler;
    m_userDialog = NULL;

    m_fpc = new FilePlayCommander( this );

    VemsContext::setObject(&m_fwInfo);
    VemsContext::setObject( m_fpc );

    m_offlineState = VTC_OFFLINE_NONE;
}

VemsTuneController::~VemsTuneController()
{
    delete m_fpc;
}

void VemsTuneController::init()
{
    Logger::logf( LOG_MSG, _T("VemsTuneController::init() - START") );

    // init all gauge-descriptor-type handler
    GaugeDescHelper::CreateMgr()->Init();

    initCommunicationModule();
    initRTModule();
    initSTModule();
    initGPSModule();
    initVemsLogModule();

    Logger::logf( LOG_MSG, _T("VemsTuneController::init() - END") );
}

void VemsTuneController::startThreads()
{
    setGPSOnline(true);

#if not defined NO_AUTO_START_SERIAL_THREAD_IN_V3GUI_COMMODEL
    m_deviceController->start();
#endif
}

void VemsTuneController::initCommunicationModule()
{
    CommunicationConfig commConfig;
    ConfigBase* cfg = ConfigHelper2::getConfig(ConfigHelper2::CONFIG_MAIN);
    ConfigMgrCommunication cfgMgr(cfg);
    cfgMgr.load(&commConfig);

    m_deviceController = new VemsDeviceController( commConfig );

    m_deviceController->Connect( wxEVT_INPUT, InputEventHandler( VemsTuneController::OnSIPRInput ), NULL, this );
}

void VemsTuneController::initRTModule()
{
    m_rtDataMgr = new RTDataMgrFacade( &(m_mtConfig->getSymbolMgr()) );

    ConfigBase* mainCfg = ConfigHelper2::getConfig();

    mainCfg->setPath(_T("/"));
    int mininterval = mainCfg->read(_T("rtMinRefreshInterval"),0L);

    m_rtDataMgr->setMinRefreshInterval(mininterval);

    RTDataInputSIPR* inputSIPR = new RTDataInputSIPR( &(m_mtConfig->getSymbolMgr()), &m_rtViewerModel );

    m_rtDataMgr->registerInput( inputSIPR );
    m_rtDataMgr->selectInput( INPUT_TYPE_SIPR );

    RealtimeDataChecker* checker  = new RealtimeDataChecker( this );
    m_rtDataMgr->RegisterViewer( checker );
    checker->SetDataMgr( m_rtDataMgr );

    ConfigSwitchViewer* aConfigSwitchViewer = new ConfigSwitchViewer( m_deviceController );
    m_rtDataMgr->RegisterViewer( aConfigSwitchViewer );
    aConfigSwitchViewer->SetDataMgr( m_rtDataMgr );
}

void VemsTuneController::initSTModule()
{
    Logger::log(LOG_MSG,_T("VemsTuneController::initSTModule()"));

    m_stController = new STDataMgrFacade();
    STDataModelSerial*  myModelSerial = new STDataModelSerial( m_mtConfig->getSymbolTable(), m_mtConfig->getSIPRDescriptor() );
    m_stController->setModel( myModelSerial );
    VemsTuneHelper::staticVariables = myModelSerial;

    /// for SDCard logger get modelbase
    VemsContext::setObject(myModelSerial);

    STDataModelSerial*  deviceModel = new STDataModelSerial( m_mtConfig->getSymbolTable(), m_mtConfig->getSIPRDescriptor() );
    m_stController->setDeviceModel( deviceModel );

    STDataIOSIPR* siprInput  = new STDataIOSIPR( m_deviceController, &( m_mtConfig->getSymbolMgr().getSymbolTable() ) );

    m_stController->setInput( siprInput );
}


void VemsTuneController::initGPSModule()
{
    Logger::log(LOG_MSG,_T("VemsTuneController::initGPSModule()"));
    m_pCommunicationControllerGPS = new GPSCommunicationController( this );
}

void VemsTuneController::initVemsLogModule()
{
    m_vemsLogReader = VemsLogReaderHelper::getInstance();

    if ( m_logHistory == NULL )
    {
        if ( m_rtDataMgr == NULL || m_stController == NULL )
            Logger::logf(LOG_MSG, _T("VemsTuneController::init() -- error: can not create loghistory - either m_rtDataMgr or m_stController is null"));
        else
        {
            m_logHistory = LogHistoryMgrHelper::create( m_rtDataMgr, m_stController, &m_mtConfig->getSymbolMgr());
            Logger::logf(LOG_MSG, _T("VemsTuneController::init() - LogHistory created"));
        }
    }
}


void VemsTuneController::selectInputFile(const wxString& fileName)
{
    m_fpc->Stop();

    m_vemsLogFile = fileName;

//    if ( m_offlineState == VTC_OFFLINE_NONE ) {
        m_deviceController->stop( VDC_FLAG_NOEVENT );
//    }

    setGPSOnline(false);

    m_logHistory->emptyHistory();

    bool ret = m_logHistory->loadHistoryFromLog(m_vemsLogFile);

/*    if( m_vemsLogReader->SetInputFile( fileName ) < 0 ) {
        sendInputSourceEvent(INPUT_TYPE_ECU);
        Logger::logf(LOG_MSG,_T("VemsTuneController::selectInputFile SetInputFile error"));
        return;
    }
    if( m_vemsLogReader->ParseHeader() < 0 ) {
        sendInputSourceEvent(INPUT_TYPE_ECU);
        Logger::logf(LOG_MSG,_T("VemsTuneController::selectInputFile ParseHeader error"));
        return;
    }
    m_vemsLogReader->reset();*/

    if ( ret == false )
    {
        sendInputSourceEvent(INPUT_TYPE_ECU);
        return;
    }


    // const StringHash& headerHash = m_vemsLogReader->getHeaderHash();
    // RawLogHelper::fillFirmwareInfo(m_fwInfo, headerHash);

    FirmwareInfo prevFirmwareInfo = m_fwInfo;

    RawLogHelper::fillFirmwareInfo(m_fwInfo, m_logHistory->getHeaderHash());

    /// get log version
    // long logVersion = 0;
    // m_vemsLogReader->getHeaderValue( RawLogHelper::logVersion ).ToLong( &logVersion );
    // m_vemsLogFileVersion = (int)logVersion;
    m_vemsLogFileVersion = (int) m_logHistory->getLogVersion();

//    if ( !m_fwInfo.isVersionAndCompileOptionsIdentical( prevFirmwareInfo ) ) {

#ifdef COMM_MODEL_WITH_GUI
    MTConfigHelper::parseMTConfigDataV2( m_fwInfo );
#else
    MTConfigHelper::parseMTConfigData( m_fwInfo );
#endif


//    }

    //m_rtDataMgr->selectInput(INPUT_TYPE_RINGBUFFER_RAWFILE);
    //m_pCommunicationController->setDisabled(true);
//    setECUOnline(false);
//    setGPSOnline(false);

    //registerSTConst();

    m_stController->clear();

    m_logHistory->refresh(0);

    Logger::logf(LOG_MSG,_T("VemsTuneController::selectInputFile thread started"));

    sendInputSourceEvent(INPUT_TYPE_RINGBUFFER_RAWFILE);
    m_selectedInput = INPUT_TYPE_RINGBUFFER_RAWFILE;

    m_offlineState = VTC_OFFLINE_LOG;

//    /// analyze vemslog in the background
//   if ( m_loganalyzer != NULL && m_loganalyzer->isrunning() )
//    {
//        m_loganalyzer->stop();
//    }
//    if( m_loganalyzer != NULL ){
//        delete m_loganalyzer;
//        m_loganalyzer = NULL;
//    }
//
//    if ( QuickAnalyze::isAnalyzed( fileName ) == false )
//    {
//        m_loganalyzer = new QuickAnalyze( m_logHistory, fileName );
//        m_loganalyzer->start();

//    }

//    switch(m_fwInfo.getDevice() )
//    {
//        case ECUInfo::DEVICE_V3:
//        {
//            ChannelCollision cc;
//            wxArrayString res = cc.detectCollisions( m_stController->getModel(), &m_mtConfig->getSymbolTable() );
//
//            if ( res.GetCount() > 0 )
//            {
//                InfoWindow::Display( NULL, _T("Channel conflict"), _T("There are conflicts between output channels\nPress 'Validate config' button for further details"),
//                    IW_FADE_IN | IW_MOVE | IW_WARNING );
//            }
//        } break;
//    }
}


void VemsTuneController::selectInputECU()
{
    LOG
    m_logHistory->emptyHistory();
    m_fpc->Stop();

    sendInputSourceEvent( INPUT_TYPE_SIPR );
    m_selectedInput = INPUT_TYPE_SIPR;
    m_stInputType = ST_INPUT_SIPR;
    setECUOnline(true);
    setGPSOnline(true);

    m_fpc->enableTools( false );

    m_offlineState = VTC_OFFLINE_NONE;
}


void VemsTuneController::sendInputSourceEvent(RTDataInputType type)
{
    SelectInputEvent event;
    event.type = type;
    AddPendingEvent(event);
}


void VemsTuneController::setECUOnline( bool state, bool isBlocking )
{
    if(state){
            CommunicationConfig commConfig;
    ConfigBase* cfg = ConfigHelper2::getConfig(ConfigHelper2::CONFIG_MAIN);
    ConfigMgrCommunication cfgMgr(cfg);
    cfgMgr.load(&commConfig);

        m_deviceController->start(commConfig);
    }else{
        if ( m_deviceController->isWorking() ) {
            m_deviceController->stop();
        }
    }
}

void VemsTuneController::setGPSOnline( bool state, bool reloadCfg )
{
    if ( state ) {
        m_pCommunicationControllerGPS->start();
    } else {
        m_pCommunicationControllerGPS->stop();
    }
}

int VemsTuneController::getGPSState()
{
    return m_pCommunicationControllerGPS->getGpsState();
}

bool VemsTuneController::isECUOnline()
{
    //return m_ecuCommThread->isEnable();
//    return m_pCommunicationController->isRunning();
    return m_deviceController->isConnected();
}

bool VemsTuneController::isGPSOnline()
{
    return m_pCommunicationControllerGPS->isRunning();
}

//void VemsTuneController::setFilePlay(bool b)
//{
//    if(b){
//        //m_pVemslogCtrl->setLogHistory(m_logHistory);
//        m_pVemslogCtrl->start();
//    }else{
//        m_pVemslogCtrl->stop();
//    }
//}
//
//bool VemsTuneController::getFilePlay()
//{
//    return m_pVemslogCtrl->isRunning();
//}

void VemsTuneController::setECULog( bool b, bool saveSDCard )
{

    ///m_siprFileLogger->init(); //close old log if exist, set filename
    ///m_siprFileLogger->process(); // save all pages first
    m_siprFileLogger->enableLog(b);
    if( b && !saveSDCard ){
        m_vemslogFilename = fillLogHeader();
        //m_commProtocolBase->requestReadAllPage(); /// Save all page to vemslog!!
        logAllPages();
        //m_pCommunicationController->request( new ReadAllPageCommand() ); /// Save all page to vemslog!!

        if ( m_rtanalyzer != NULL )
        {
            m_rtDataMgr->UnRegisterViewer( m_rtanalyzer );
            delete m_rtanalyzer;
        }
        m_rtanalyzer = new RTDataAnalyzer();
        m_rtDataMgr->RegisterViewer( m_rtanalyzer );
        m_rtanalyzer->SetDataMgr( m_rtDataMgr );

    }
    if ( b == false )
    {
        FileInfo info;
        info.fileType = _T("VEMSLOG file");
        info.fw_major = m_fwInfo.getFirmwareVersion().major;
        info.fw_minor = m_fwInfo.getFirmwareVersion().minor;
        info.fw_build = m_fwInfo.getFirmwareVersion().build;

        DoubleHash* dh = m_rtanalyzer->getValues();
        info.rpmMax = (int)((*dh)[_T("rpmMax")]);
        info.mapMax = (int)((*dh)[_T("mapMax")]);
        info.triggerErr = (int)(*dh)[_T("triggerError")];
        info.hasGPSData = m_logHistory->hasGPSdata();

        FileAnalyzeSQL::addData( m_vemslogFilename, info );

        // unregister viewer & delete analyzer
        m_rtDataMgr->UnRegisterViewer( m_rtanalyzer );
        delete m_rtanalyzer;
        m_rtanalyzer = NULL;
    }

    //m_siprFileLogger->setSaveSDCard( saveSDCard );
    //m_inputGPS->enableLog( b );

    ECULogEvent event;
    if( b )
    {
        event.state = ECULogEvent::ON;
    }
    else
    {
        event.state = ECULogEvent::OFF;
    }
    AddPendingEvent( event );
}

bool VemsTuneController::isECULog()
{
    return m_siprFileLogger->isLog();
}

int VemsTuneController::getSelectedInput()
{
    return m_selectedInput;
}

void VemsTuneController::checkSignature()
{
    wxString ecuSign = m_fwInfo.getSignature();
    wxString expSignature = m_mtConfig->getSignature();
    if(ecuSign != expSignature){
        wxString msg = _T("Project signature warning!\n\n");
        msg += _T("Expected: ")+ expSignature+_T("\n");
        msg += _T("ECU uses: ")+ ecuSign;
        msg += _T("\n\nPlease Check Preferences/ini file config!");
        wxMessageBox( msg,
                _T("Project signature warinig!"), wxOK | wxICON_ERROR |wxSTAY_ON_TOP  );
    }
}



int VemsTuneController::getDevice()
{
    if( isECUOnline() ){
        //return m_ecuStateMgr->getDevice();
        return m_deviceController->getDevice();
    }else{
        return m_fwInfo.getDevice();
    }
}

FirmwareInfo* VemsTuneController::getFirmwareInfo()
{
    return &m_fwInfo;
}

wxString VemsTuneController::getFirmwareInfoStr()
{
    LOGP( m_fwInfo.getFirmwareVersionAsString() );
    return m_fwInfo.getFirmwareVersionAsString();
}


void VemsTuneController::selectInputVemsCfg( const wxString& filename )
{
    m_vemsLogFile = wxEmptyString;
    m_fpc->Reset();
    m_fpc->enableTools( false );

//    setECUOnline(false);
//    m_deviceController->stop( VDC_FLAG_NOEVENT );

//    if ( m_offlineState == VTC_OFFLINE_NONE ) {
        m_deviceController->stop( VDC_FLAG_NOEVENT );
//    }

    setGPSOnline(false);

    m_logHistory->emptyHistory();

    Logger::logf(LOG_MSG,_T("VemsTuneController::selectInputVemsCfg() - device: %d"), getDevice() );

    //if ( getDevice() == ECUInfo::DEVICE_V3 ) return;

    m_stInputType       = ST_INPUT_FILE;
    m_stInputFileName   = filename;
    FirmwareConfigLoader configLoader;
    configLoader.setMTConfigData(m_mtConfig);

    wxArrayString fileA;
    fileA.Add(filename);
    configLoader.parse(fileA);

	STDataModelBase* viewerModel = m_stController->getModel();
    //configLoader.load(viewerModel);

    STDataModelBase* deviceModel = m_stController->getDeviceModel();
    configLoader.load(deviceModel);
    deviceModel->copyTo(viewerModel);

    FirmwareConfigData& fwConfigData = configLoader.getFirmwareConfigData();
    RawLogHelper::fillFirmwareInfo( m_fwInfo, fwConfigData.header );
    Logger::logf(LOG_MSG,_T("VemsTuneController::selectInputVemsCfg() 100") );

    m_stController->updateViewers();

    m_offlineState = VTC_OFFLINE_CFG;

//    switch(m_fwInfo.getDevice() ){
//        case ECUInfo::DEVICE_V3:
//            {
//                ChannelCollision cc;
//                wxArrayString res = cc.detectCollisions( m_stController->getModel(), &m_mtConfig->getSymbolTable() );
//
//                if ( res.GetCount() > 0 )
//                {
//                    InfoWindow::Display( NULL, _T("Channel conflict"), _T("There are conflicts between output channels\nPress 'Validate config' button for further details"),
//                        IW_FADE_IN | IW_MOVE | IW_WARNING );
//                }
//            }
//            break;
//    }
}


void VemsTuneController::OnSIPRInput( InputEvent& event )
{
    SIPRPacket p = event.getPacket();

    m_siprFileLogger->process(p);
    m_logHistory->process(p);

    if ( m_pEventHandler ) { m_pEventHandler->AddPendingEvent( event ); }
}

wxString VemsTuneController::fillLogHeader( bool isSDSave )
{
    wxString deviceStr = _T("ECU");
    int device = m_fwInfo.getDevice();
    if(device != ECUInfo::DEVICE_UNKNOWN) deviceStr  = ECUInfo::getDeviceAsString( device );

    wxString baseFileName;

    if(device == ECUInfo::DEVICE_V3 || device == ECUInfo::DEVICE_ROUND ){
        baseFileName = m_fwInfo.getSerialNumber();
        if ( isSDSave ) { baseFileName += _("-SDCard"); }
        baseFileName += _T(".vemslog");
    }else{
        baseFileName = deviceStr + _T("-.vemslog");
    }

    wxString logdir = DirHelper::GetPath(dLOG);
    ConfigBase* mainCfg = ConfigHelper2::getConfig();
    logdir = mainCfg->getValue( _T("/environment"), _T("logdir"), logdir);

    if(! wxFileName::DirExists( logdir ))
        logdir = DirHelper::GetPath(dLOG);

    wxString fileName = DirHelper::CreateNewFileName(baseFileName,logdir,true,true);
    Logger::logf(LOG_MSG,_T("VemsTuneController::fillLogHeader() fileName: ")+ fileName);

    wxString header = RawLogHelper::getVemslogHeader(m_fwInfo, 2);
    m_siprFileLogger->setHeader(header);
    m_siprFileLogger->setFile( fileName);

    return fileName;
}

bool VemsTuneController::saveMSQ(wxString& fileName)
{
    wxString cfgFileName = getStaticSourceAsString();
    wxFileName fn(cfgFileName);
    if(fn.IsFileReadable() ){
        fileName = cfgFileName.BeforeLast('.');
        fileName += _T(".msq");
    }else{
        if( m_fwInfo.getDevice() != ECUInfo::DEVICE_V3 ) return false;
        wxString baseFileName;
        baseFileName = m_fwInfo.getSerialNumber()+_T(".msq");
        fileName = DirHelper::CreateNewFileName(baseFileName, DirHelper::GetPath(dLOG),true,true);
    }

    wxArrayString pathArray;
    pathArray.Add(fileName);
    STDataIOMSQ stDataIoMSQ( &(m_mtConfig->getSymbolMgr()) , m_mtConfig->getSIPRDescriptor(), pathArray);
    stDataIoMSQ.setModel( m_stController->getModel() );
    stDataIoMSQ.writeAllData();
    return true;
}

void VemsTuneController::registerSTConst()
{
    SymbolMgr symbolMgr = m_mtConfig->getSymbolMgr();
    RPNHelper::setUsedStatic(symbolMgr);
    const SymbolTable& symbolTable = symbolMgr.getUsedStaticAndBit();

    //register scalar viewers to connect RT and ST model

    for( SymbolTable::const_iterator it = symbolTable.begin(); it != symbolTable.end(); ++it )
    {
        Symbol* symbol = it->second;
        STDataViewerHash* viewer = new STDataViewerHash( &m_rtViewerModel );
        viewer->setSymbol( *symbol );
        m_stScalarViewerA.Add( viewer );
        m_stController->RegisterViewer( viewer );
        //Logger::logf(LOG_MSG, _T("VemsTuneController::registerSTConst() register symbol: ")+ symbol->getName() );
    }
    m_logHistory->initST(*m_mtConfig->getSIPRDescriptor() );
}

void VemsTuneController::request( VemsCommand* command )
{
    m_deviceController->request( command );
}

void VemsTuneController::setLogSaveSDCard( bool enable )
{
    // need stop?, must be stop
    m_siprFileLogger->setSaveSDCard( enable );
}


void VemsTuneController::logAllPages()
{
    STDataModelWriter::logPages( m_stController->getModel(), m_siprFileLogger, m_mtConfig->getSIPRDescriptor() );
}

wxString VemsTuneController::getRealtimeSourceAsString()
{
    if( m_selectedInput == INPUT_TYPE_SIPR && getDevice() == ECUInfo::DEVICE_V3 )
    {
        return _T("ECU");
    }
    else if ( m_selectedInput == INPUT_TYPE_RINGBUFFER_RAWFILE )
    {
        return m_vemsLogFile;
    }

    return _T("None");
}


wxString VemsTuneController::getStaticSourceAsString()
{
    if ( m_stInputType == ST_INPUT_SIPR )
    {
         if ( getDevice() == ECUInfo::DEVICE_V3 )
        {
            return _T("ECU");
        }
        else if ( m_selectedInput == INPUT_TYPE_RINGBUFFER_RAWFILE && m_vemsLogFileVersion >= 2 )
        {
            return m_vemsLogFile;
        }
    }
    else if ( m_stInputType == ST_INPUT_FILE )
    {
        return m_stInputFileName;
    }

    return _T("None");
}


//ProcessBase* VemsTuneController::getECUCtrl(){
////    return m_pCommunicationController;
//    return m_deviceController->getCommunicationController();
//    return NULL;
//}

VemsDeviceController* VemsTuneController::getECUController()
{
    return m_deviceController;
}

DeviceInfo VemsTuneController::getDeviceInfo() const
{
    return m_deviceController->getDeviceInfo();
}

LogHistoryBase* VemsTuneController::getLogHistory() {
    return m_logHistory;
}

Observable*     VemsTuneController::getLogHistoryObs() {
    return m_logHistory;
}

LogHistoryMgr*  VemsTuneController::getLogHistoryMgr()
{
    return m_logHistory;
}

RTDataMgrBase*  VemsTuneController::getRTDataMgr()
{
    return m_rtDataMgr;
}

STDataMgrBase*  VemsTuneController::getSTDataMgr(){
    return m_stController;

}

void VemsTuneController::OnCannotOpenPort( VemsDeviceControllerEvent& event )
{
}

///BOOT MODE FOUND
void VemsTuneController::OnDeviceFound( VemsDeviceControllerEvent& event )
{
}

void VemsTuneController::OnDeviceNotFound( VemsDeviceControllerEvent& event )
{
}

void VemsTuneController::OnDeviceConnected( VemsDeviceControllerEvent& event )
{
    m_offlineState = VTC_OFFLINE_NONE;

    m_logHistory->emptyHistory();

//    setFilePlay(false);

//    sendInputSourceEvent( INPUT_TYPE_SIPR );
    m_selectedInput = INPUT_TYPE_SIPR;
    m_stInputType   = ST_INPUT_SIPR;
//    setECUOnline(true);
//    setGPSOnline(true);

    m_stController->setOnline( true );
    m_stController->setConnected();

    m_fwInfoPrev = m_fwInfo;
    m_fwInfo = m_deviceController->getFirmwareInfo();

//    if ( isDeviceSameAsPrev() ) {
//        LOG
//        m_stController->setRefreshDeviceModelOnly(true);
//        sendDetectedPagesToLogHistory();
//        m_stController->setRefreshDeviceModelOnly(false);
//
//        STDataModelBase* deviceModel = m_stController->getDeviceModel();
//        STDataModelBase* viewerModel = m_stController->getModel();
//        const IntArray& pages = m_mtConfig->getSIPRDescriptor()->getPagesOrder();
//
//        m_stModelsDiff.clear();
//        m_stModelsDiff = STDataModelHelper::compare( deviceModel, viewerModel, pages );
//    } else {
        sendDetectedPagesToLogHistory();
//    }

    if( m_siprFileLogger->isLog() ){
        fillLogHeader();
        m_siprFileLogger->enableLog(true); /// create new file
        logAllPages();
    }

//    if ( isConfigSameAsPrev() ) {
        m_stController->updateRegisteredViewers();
        m_deviceController->enableRealtime();
//    }
}

void VemsTuneController::OnDeviceDisconnected( VemsDeviceControllerEvent& event )
{
    m_stController->setOnline( false );
    m_stController->setConnected(false);
    m_stController->updateRegisteredViewers();

//    setECULog( false );

//    ECUCommEvent ecuEvent( ECUCommEvent::OFFLINE );
//    AddPendingEvent( ecuEvent );

}

void VemsTuneController::OnDeviceReconnected( VemsDeviceControllerEvent& event )
{
    m_fwInfo = m_deviceController->getFirmwareInfo();

    LOGP( m_fwInfo.getFirmwareVersionAsString() );

    const SIPRPacketList& pages = m_deviceController->getDetectedPages();

    SIPRPacketList::const_iterator it = pages.begin();

    ///TODO - refresh deviceModel only
    m_stController->setRefreshDeviceModelOnly(true);
    for ( ; it != pages.end(); ++it ) {
        SIPRPacket* packet = *it;
//        m_siprFileLogger->process(*packet);
        m_logHistory->process(*packet);
    }
    m_stController->setRefreshDeviceModelOnly(false);

    if( m_siprFileLogger->isLog() ){
        fillLogHeader();
        m_siprFileLogger->enableLog(true); /// create new file
        logAllPages();
    }

    m_stController->setOnline( true );
    m_stController->setConnected();
    m_stController->updateRegisteredViewers();

    m_deviceController->enableRealtime();

//    IntToIntSetHash diff = STDataModelHelper::compare(  m_stController->getModel(),
//                                                        m_stController->getDeviceModel(),
//                                                        m_mtConfig->getSIPRDescriptor()->getPagesOrder() );
//
//    LOGP( wxString::Format( _T("diff.size(): %d"), diff.size() ) );



//    selectInputECU();

}


void VemsTuneController::OnDeviceControllerStopped( VemsDeviceControllerEvent& event )
{
    m_stController->setOnline( false );
    m_stController->updateRegisteredViewers();

    if ( isECULog() ) {
        setECULog( false );
    }

}

void VemsTuneController::OnDeviceResumed( VemsDeviceControllerEvent& event )
{
    LOG
    m_logHistory->emptyHistory();

//    setFilePlay(false);

//    sendInputSourceEvent( INPUT_TYPE_SIPR );
    m_selectedInput = INPUT_TYPE_SIPR;
    m_stInputType   = ST_INPUT_SIPR;
//    setECUOnline(true);
//    setGPSOnline(true);

    m_stController->setOnline( true );
    m_stController->setConnected();

    m_fwInfo = m_deviceController->getFirmwareInfo();

    const SIPRPacketList& pages = m_deviceController->getDetectedPages();

    LOGP( wxString::Format( _T("pages.GetCount(): %d"), pages.GetCount() ) );

    SIPRPacketList::const_iterator it = pages.begin();

    for ( ; it != pages.end(); ++it ) {
        SIPRPacket* packet = *it;
        m_logHistory->process(*packet);
    }

    if( m_siprFileLogger->isLog() ){
        fillLogHeader();
        m_siprFileLogger->enableLog(true); /// create new file
        logAllPages();
    }

    m_stController->updateRegisteredViewers();

    m_deviceController->enableRealtime();
}

void VemsTuneController::OnDeviceProtected( VemsDeviceControllerEvent& event )
{

}

void VemsTuneController::OnCommunicationError( CommunicationEvent& event )
{

}

bool VemsTuneController::isOnline() const
{
    return m_deviceController->isOnline();
}

void VemsTuneController::setOffline()
{
    m_deviceController->stop();
}

void VemsTuneController::setOnline()
{
    m_fpc->Stop();
    m_fpc->enableTools(false);
    m_deviceController->start( VDC_FLAG_NO_RECONNECT );
}

void VemsTuneController::sendDetectedPagesToLogHistory()
{
    const SIPRPacketList& pages = m_deviceController->getDetectedPages();

    SIPRPacketList::const_iterator it = pages.begin();

    if ( !pages.empty() ) {
        uint8_t packetData = SIPR_REFRESH_START;
        SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
        m_logHistory->process(packet);
    }

    for ( ; it != pages.end(); ++it ) {
        SIPRPacket* packet = *it;
        m_logHistory->process(*packet);
    }

    if ( !pages.empty() ) {
        uint8_t packetData = SIPR_REFRESH_STOP;
        SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &packetData );
        m_logHistory->process(packet);
    }
}

bool VemsTuneController::isDeviceSameAsPrev() const
{
    return ( ( m_fwInfoPrev.getSerialNumberAsLong() == m_fwInfo.getSerialNumberAsLong() ) &&
                ( m_fwInfoPrev.getFirmwareVersion() == m_fwInfo.getFirmwareVersion() ) );

}

bool VemsTuneController::isConfigSameAsPrev() const
{
    return m_stModelsDiff.empty();
}

void VemsTuneController::writeConfigDifferences()
{
    LOG
    IntToIntSetHash::const_iterator pageIter = m_stModelsDiff.begin();

    STDataModelBase* viewerModel = m_stController->getModel();
    STDataModelBase* deviceModel = m_stController->getDeviceModel();

    for ( ; pageIter != m_stModelsDiff.end(); ++pageIter ) {
        int pageNo = pageIter->first;
        const IntSet& page = pageIter->second;

        IntSet::const_iterator offsetIter = page.begin();

        for ( ; offsetIter != page.end(); ++offsetIter ) {
            int pageOffset = *offsetIter;
            uint8_t value = viewerModel->getRawData( pageNo, pageOffset );
            deviceModel->setRawData( value, pageNo, pageOffset );
            m_deviceController->request( new WritePageCommand( pageNo, pageOffset , value ) );
            LOGP( wxString::Format( _T("pageNo: %d, pageOffset: %d, value: %d"), pageNo, pageOffset , value ) );
        }
    }

    m_deviceController->request( new BurnCommand() );

    m_stController->updateRegisteredViewers();

}

void VemsTuneController::enableRealtime()
{
    LOG
    m_deviceController->enableRealtime();
}

FilePlayCommander* VemsTuneController::getFilePlayCommander()
{
    return m_fpc;
}
