#include "wxAgg.h"

#include "agg_path_storage.h"
#include "agg_conv_stroke.h"
#include "agg_ellipse.h"
#include "agg_arc.h"
#include "agg_bezier_arc.h"
#include "agg_span_gradient.h"
#include "agg_span_interpolator_linear.h"
#include "agg_span_allocator.h"
#include "agg_conv_curve.h"
#include "agg_gsv_text.h"
#include "agg_bounding_rect.h"
#include <wx/wx.h>

#include "agg_conv_dash.h"
#include "agg_conv_smooth_poly1.h"
#include "agg_vcgen_markers_term.h"
#include "../logger.h"
#include "../UnicodeHelper.h"
#include "../common/typedef.h"



#include <wx/image.h>

#define M_PI_2		1.57079632679489661923

#include "../bcWidgets/BitmapCanvasWidgetBase.h"

wxAgg::wxAgg( uint32_t width, uint32_t height )
{
    Create( width, height );
}

wxAgg::wxAgg( const wxSize& size )
{
    Create( size.x, size.y );
}

wxAgg::wxAgg( wxWindow* win )
{
    Create( win->GetSize().x, win->GetSize().y );
}

wxAgg::wxAgg( BitmapCanvasWidgetBase* bcWidget )
{
    wxSize s = bcWidget->GetSize().Ceil();
    Create( s.x, s.y );
}

wxAgg::~wxAgg()
{
    delete m_ren_sl;
    delete m_ren_base;
    delete m_pixf;
    delete m_rb;
    delete[] m_buffer;

    delete m_rena;
    delete m_rba;
    delete m_pixfg;
    delete[] m_alphaBuffer;
}

void wxAgg::Create( uint32_t width, uint32_t height )
{
    m_width     = width;
    m_height    = height;

    m_font_weight   = 0.0;

    m_buffer        = new unsigned char[ m_width * m_height * 3 ];
    m_rb            = new agg::rendering_buffer( m_buffer, m_width, m_height, m_width * 3 );
    m_pixf          = new pixfmt( *m_rb );
    m_ren_base      = new renderer_base_type( *m_pixf );
    m_ren_sl        = new renderer_scanline_type( *m_ren_base );

    m_ren_base->clear( agg::rgba( 0, 0, 0, 0 ) );

    m_alphaBuffer   = new unsigned char[ m_width * m_height ];
    m_alpha_mask_rbuf.attach( m_alphaBuffer, m_width, m_height, m_width );
    m_pixfg         = new agg::pixfmt_gray8( m_alpha_mask_rbuf );
    m_rba           = new ren_base_alpha( *m_pixfg );
    m_rena          = new renderer_alpha( *m_rba );

    m_rba->clear( agg::gray8( 0 ) );

    m_penColor      = agg::rgba( 1.0, 1.0, 1.0 );
    m_brushColor    = agg::rgba( 1.0, 1.0, 1.0 );

    m_clippingRect = wxDefaultRect;

}

agg::rgba wxAgg::convertToRGBA( const wxColour& c )
{
    double r = (double)c.Red() / 255;
    double g = (double)c.Green() / 255;
    double b = (double)c.Blue() / 255;
    double a = (double)c.Alpha() / 255;
    return agg::rgba( r, g, b, a );
}

wxColor wxAgg::convertToColor( const agg::rgba& c )
{
    unsigned char r = (unsigned char)( c.r * 255 );
    unsigned char g = (unsigned char)( c.g * 255 );
    unsigned char b = (unsigned char)( c.b * 255 );
    unsigned char a = (unsigned char)( c.a * 255 );
    return wxColor( r, g, b, a );
}

/*
void wxAgg::draw()
{
    m_ren_base->clear( agg::rgba( 0, 0, 0 ) );

    //raw_anti_aliased
    agg::path_storage path;
    path.move_to( m_width / 2, 0 );
    path.line_to( m_width , m_height );
    path.line_to( 0, m_height );
    path.close_polygon();

    //m_ren_sl->color( agg::rgba( 0.0, 1.0, 0.0, 0.5 ) );
    //m_ras.gamma( agg::gamma_power( 0.5 * 2.0 ) );
    m_ras.add_path( path );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    wxImage  myImage( m_width, m_height, m_buffer, true );
    wxBitmap myBitmap( myImage, 3 );
    m_dc.DrawBitmap( myBitmap, 0, 0, true );
}
*/

void wxAgg::setPen( const wxColour& c )
{
    m_penColor = convertToRGBA( c );
}


void wxAgg::setBrush( const wxColour& c )
{
    m_brushColor = convertToRGBA( c );
}

void wxAgg::clear( const wxColor& c )
{
    agg::rgba cRGBA = convertToRGBA( c );

    m_ren_base->clear( cRGBA );
    m_rba->clear( agg::gray8( c.Alpha() ) );
    //Logger::logf( LOG_MSG, _T(" alpha : %f"), cRGBA.a );
}

void wxAgg::drawLine( double x1, double y1, double x2, double y2, double width, agg::line_cap_e cap )
{
    agg::path_storage path;

    path.move_to( x1, y1 );
    path.line_to( x2, y2 );
    //path.close_polygon();

    agg::conv_stroke< agg::path_storage > stroke( path );
    stroke.line_join( agg::miter_join );
    //stroke.miter_limit(m_miter_limit.value());
    stroke.line_cap(cap);
    stroke.width( width );

    m_ras.reset();
    applyClipping();

    m_ras.add_path( stroke );
    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );
    //Logger::logf( LOG_MSG, _T("wxAgg::drawLine() - alpha : %f "), m_penColor.a );
}

void wxAgg::drawDashedLine( double x1, double y1, double x2, double y2, wxPen& pen )
{
    agg::rgba oldcolor = m_penColor;
    m_penColor = convertToRGBA( pen.GetColour() );

    int style = pen.GetStyle();
    double mw = ( pen.GetWidth() < 1 ) ? 1.0 : pen.GetWidth();

    double* pattern = NULL;
    int patternN = 0;

    switch ( style )
    {
        case wxDOT: patternN = 1; pattern = new double[2]; pattern[0] = mw; pattern[1] = mw + 2.0; break;
        case wxLONG_DASH: patternN = 1; pattern = new double[2]; pattern[0] = 19; pattern[1] = 9; break;
        case wxSHORT_DASH: patternN = 1; pattern = new double[2]; pattern[0] = 9; pattern[1] = 6; break;
        case wxDOT_DASH: patternN = 2; pattern = new double[4]; pattern[0] = 9; pattern[1] = 6;
            pattern[2] = 3.0; pattern[3] = 3.0; break;
        case wxUSER_DASH: {
            wxDash* dash; int d = pen.GetDashes(&dash); patternN = d / 2;

            if ( dash != NULL && d > 0 )
            {
                pattern = new double[d] ;
                for ( int i = 0 ; i < d ; ++i )
                {
                    pattern[i] = dash[i] * mw ;
                    if ( i % 2 == 1 && pattern[i] < mw + 2.0 )
                        pattern[i] = mw + 2.0 ;
                    else if ( i % 2 == 0 && pattern[i] < mw )
                        pattern[i] = mw ;
                }
            }
            break;
        }
        case wxSOLID:
        default: patternN = 1; pattern = new double[2]; pattern[0] = 1; pattern[1] = 0; break;
    }
    drawDashedLine( x1, y1, x2, y2, pen.GetWidth(), patternN, pattern );
    m_penColor = oldcolor;
    delete[] pattern;
}

void wxAgg::drawDashedLine( double x1, double y1, double x2, double y2, double width, int patternN, double* pattern, agg::line_cap_e cap )
{
    typedef agg::path_base<
        agg::vertex_stl_storage<
            agg::pod_auto_vector<
                agg::vertex_d, 5> > > path_storage_type;

    path_storage_type path;

    path.move_to( x1, y1 );
    path.line_to( x2, y2 );

    m_ras.add_path(path);
    m_ren_sl->color( m_penColor );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    agg::conv_smooth_poly1<path_storage_type> smooth(path);
    smooth.smooth_value( 5.0 );

    m_ras.add_path(smooth);
    m_ren_sl->color( agg::rgba( 0.8, 0.8, 0.8, 0.9 ) );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    agg::conv_stroke<agg::conv_smooth_poly1<path_storage_type> > smooth_outline(smooth);
    m_ras.add_path(smooth_outline);

    m_ren_sl->color( agg::rgba( 0.9, 0.9, 0.9, 0.0 ) );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    agg::conv_curve<agg::conv_smooth_poly1<path_storage_type> > curve(smooth);
    agg::conv_dash<agg::conv_curve<agg::conv_smooth_poly1<path_storage_type> >, agg::vcgen_markers_term> dash(curve);
    agg::conv_stroke<agg::conv_dash<agg::conv_curve<agg::conv_smooth_poly1<path_storage_type> >, agg::vcgen_markers_term> > stroke(dash);
    stroke.line_cap(cap);
    stroke.width( width );

    if ( pattern != NULL )
    {
        for ( int j = 0; j < patternN; ++j )
            dash.add_dash( pattern[2*j], pattern[2*j+1] );
        dash.dash_start( pattern[0] + pattern[1] / 2);
    }
    else
    {
        dash.add_dash( 10.0, 10.0 );
        dash.dash_start( 15 );
    }

    m_ras.reset();
    applyClipping();

    m_ras.add_path(stroke);

    m_ren_sl->color( m_penColor );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );
}

void wxAgg::drawLine( RealPointArray coords, double width, agg::line_cap_e cap)
{
    if( coords.GetCount() < 2 ) return;

    agg::path_storage path;

    path.move_to( coords[0].x, coords[0].y );

    for( size_t a = 1; a < coords.GetCount(); a++ ) {

        path.line_to( coords[a].x, coords[a].y );

    }

    agg::conv_stroke< agg::path_storage > stroke( path );
    stroke.line_join( agg::miter_join );
    //stroke.miter_limit(m_miter_limit.value());
    stroke.line_cap(cap);
    stroke.width( width );

    m_ras.reset();
    applyClipping();

    m_ras.add_path( stroke );

    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );

}

void wxAgg::drawEllipse( double x, double y, double r1, double r2, double width )
{
    agg::ellipse ellipse( x, y, r1, r2 );
    agg::conv_stroke< agg::ellipse > stroke( ellipse );


    stroke.width( width );

    m_ras.reset();
    applyClipping();

    m_ras.add_path( stroke );

    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    renderAlphaMask( m_penColor.a );
}

void wxAgg::drawFilledEllipse( double x, double y, double r1, double r2, double a1, double a2 )
{
    agg::bezier_arc arc( x, y, r1, r2, a1, a2 );

    agg::path_storage path;

    path.concat_path( arc, 0 );
    path.line_to( x, y );

    agg::conv_curve<agg::path_storage> curve( path );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( curve );

    m_ren_sl->color( m_brushColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );

/*

    agg::ellipse ellipse( x, y, r1, r2 );
    m_ras.reset();
    m_ras.add_path( ellipse );

    m_ren_sl->color( m_brushColor );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_brushColor.a );

    agg::conv_stroke< agg::ellipse > stroke( ellipse );
    stroke.width( width );
    m_ras.add_path( stroke );
    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    renderAlphaMask( m_penColor.a );*/
}

void wxAgg::drawCircle( double x, double y, double r, double width )
{
    agg::ellipse ellipse( x, y, r, r );
    agg::conv_stroke< agg::ellipse > stroke( ellipse );

    //stroke.line_join(join);
    //stroke.line_cap(cap);
    //stroke.miter_limit(m_miter_limit.value());
    stroke.width( width );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( stroke );
    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    renderAlphaMask( m_penColor.a );
}


void wxAgg::drawFilledCircle( double x, double y, double r )
{
    agg::ellipse ellipse( x, y, r, r );
    m_ras.reset();

    applyClipping();

    m_ras.add_path( ellipse );

    m_ren_sl->color( m_brushColor );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    renderAlphaMask( m_brushColor.a );
}


void wxAgg::drawGradientFilledCircle( const wxColour& bc, const wxColour& ec, double deg, double cx, double cy, double r )
{

    // The gradient objects declarations
    //----------------
    gradient_func_type  gradient_func;                   // The gradient function
    agg::trans_affine   gradient_mtx;                    // Affine transformer
    interpolator_type   span_interpolator(gradient_mtx); // Span interpolator
    //interpolator_type   span_interpolator_trans((gradient_mtx); // Span interpolator

    //interpolator_type   span_interpolator(); // Span interpolator
    span_allocator_type span_allocator;                  // Span Allocator
    color_array_type    color_array;                     // Gradient colors


    // Declare the gradient span itself.
    // The last two arguments are so called "d1" and "d2"
    // defining two distances in pixels, where the gradient starts
    // and where it ends. The actual meaning of "d1" and "d2" depands
    // on the gradient function.
    //----------------
    span_gradient_type span_gradient(span_interpolator,
                                     gradient_func,
                                     color_array,
                                     0, 100);

    // The gradient renderer
    //----------------
    renderer_gradient_type ren_gradient( *m_ren_base, span_allocator, span_gradient);

    fillColorArray( color_array,
                     convertToRGBA( bc ),
                     convertToRGBA( ec ) );

    agg::ellipse ell( cx, cy, r, r );

    deg = 180.0 - deg;

    double x = r * cos( deg * 3.1415926 / 180.0 );
    double y = r * sin( deg * 3.1415926 / 180.0 );

    applyClipping();

    m_ras.add_path( ell );

    gradient_mtx.reset();
    gradient_mtx *= agg::trans_affine_scaling( 2 * r / 100.0 );
    gradient_mtx *= agg::trans_affine_rotation( deg * 3.1415926 / 180.0 );
    gradient_mtx *= agg::trans_affine_translation(  cx - x , cy - y  );
    gradient_mtx.invert();

    agg::render_scanlines( m_ras, m_sl, ren_gradient );

    renderAlphaMask( m_brushColor.a );
}


void wxAgg::drawArc( double centerx, double centery, double r, double startAngle, double deltaAngle, double w )
{
    agg::bezier_arc arc( centerx, centery, r, r, startAngle, deltaAngle );

    agg::path_storage path;

    path.concat_path( arc, 0 );
    //path.line_to( x, y );

    agg::conv_curve<agg::path_storage> curve( path );

    agg::conv_stroke< agg::conv_curve<agg::path_storage > > stroke( curve );
    //stroke.line_join(join);
    //stroke.line_cap(cap);
    //stroke.miter_limit(m_miter_limit.value());
    stroke.width( w );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( stroke );

    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    renderAlphaMask( m_penColor.a );
}


void wxAgg::drawFilledArc( double x, double y, double r, double a1, double a2 )
{
    agg::bezier_arc arc( x, y, r, r, a1, a2 );

    agg::path_storage path;

    path.concat_path( arc, 0 );
    path.line_to( x, y );

    agg::conv_curve<agg::path_storage> curve( path );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( curve );

    m_ren_sl->color( m_brushColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );
}


void wxAgg::drawFilledCircleSection( double x, double y, double r, double startAngle, double deltaAngle )
{
    agg::bezier_arc arc( x, y, r, r, startAngle, deltaAngle );

    agg::path_storage path;

    path.concat_path( arc, 0 );

    agg::conv_curve<agg::path_storage> curve( path );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( curve );

    m_ren_sl->color( m_brushColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );
}


void wxAgg::drawFilledRectangle( double x, double y, double w, double h, double s )
{
    agg::path_storage path;

    path.move_to( x, y );
    path.line_rel( w, 0 );
    path.line_rel( 0, h );
    path.line_rel( -w, 0 );
    path.close_polygon();

    m_ras.reset();

    applyClipping();

    m_ras.add_path( path );
    m_ren_sl->color( m_brushColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_brushColor.a );


    agg::conv_stroke< agg::path_storage > stroke( path );

    stroke.width( s );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( stroke );
    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );
}

void wxAgg::drawSolidRectangle( double x, double y, double w, double h )
{
    agg::path_storage path;

    path.move_to( x, y );
    path.line_rel( w, 0 );
    path.line_rel( 0, h );
    path.line_rel( -w, 0 );
    path.close_polygon();

    //agg::conv_stroke< agg::path_storage > stroke( path );
    //agg::conv

    //stroke.width( s );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( path );
    m_ren_sl->color( m_brushColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_brushColor.a );
}

void wxAgg::drawRectangle( double x, double y, double w, double h, double s )
{
    agg::path_storage path;

    path.move_to( x, y );
    path.line_rel( w, 0 );
    path.line_rel( 0, h );
    path.line_rel( -w, 0 );
    path.close_polygon();

    agg::conv_stroke< agg::path_storage > stroke( path );

    stroke.width( s );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( stroke );
    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );
}

void wxAgg::drawFilledRoundedRectangle( double x, double y, double w, double h, double r, double s)
{
    double d = 2.0*r;

/// RoundedRectangle

  /// create paths

    agg::path_storage path;

    path.move_to( x+r, y );
    path.line_rel( w-d, 0 );
    path.move_to( x+w, y+r );
    path.line_rel( 0, h-d );
    path.move_to( x+r, y+h );
    path.line_rel( w-d, 0 );
    path.move_to( x, y+r );
    path.line_rel( 0, h-d );

    path.close_polygon();

    agg::conv_stroke< agg::path_storage > stroke1( path );
    stroke1.width( s );

    agg::bezier_arc arc1( x+r, y+r, r, r, M_PI, M_PI_2 );
    agg::bezier_arc arc2( x+w-r, y+r, r, r, M_PI + M_PI_2 , M_PI_2 );
    agg::bezier_arc arc3( x+w-r, y+h-r, r, r, 0, M_PI_2 );
    agg::bezier_arc arc4( x+r, y+h-r, r, r, M_PI_2, M_PI_2 );
    agg::path_storage path2;
    path2.concat_path( arc1, 0 );
    path2.concat_path( arc2, 0 );
    path2.concat_path( arc3, 0 );
    path2.concat_path( arc4, 0 );
    agg::conv_curve<agg::path_storage> curve( path2 );
    agg::conv_stroke< agg::conv_curve<agg::path_storage > > stroke2( curve );
    stroke2.width( s );

/// fill in

 /// create paths

    agg::path_storage path_i;

    path_i.move_to( x+r, y );
    path_i.line_rel( w-d, 0 );
    path_i.line_rel( r, r );
    path_i.line_rel( 0, h-d );
    path_i.line_rel( -r, r );
    path_i.line_rel( -(w-d), 0 );
    path_i.line_rel( -r, -r );
    path_i.line_rel( 0, -(h-d) );

    path_i.close_polygon();


/// render

    m_ras.reset();

    applyClipping();

    m_ras.add_path( curve );
    m_ras.add_path( path_i );
    m_ren_sl->color( m_brushColor );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_brushColor.a );


    m_ras.reset();

    applyClipping();

    m_ras.add_path( stroke1 );
    m_ras.add_path( stroke2 );
    m_ren_sl->color( m_penColor );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );


}

void wxAgg::drawRoundedRectangle( double x, double y, double w, double h, double r, double s )
{
    agg::path_storage path;

    double d = 2.0*r;

    path.move_to( x+r, y );
    path.line_rel( w-d, 0 );
    path.move_to( x+w, y+r );
    path.line_rel( 0, h-d );
    path.move_to( x+r, y+h );
    path.line_rel( w-d, 0 );
    path.move_to( x, y+r );
    path.line_rel( 0, h-d );

    path.close_polygon();

    agg::conv_stroke< agg::path_storage > stroke1( path );
    stroke1.width( s );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( stroke1 );


    agg::bezier_arc arc1( x+r, y+r, r, r, M_PI, M_PI_2 );
    agg::bezier_arc arc2( x+w-r, y+r, r, r, 0, -M_PI_2 );
    agg::bezier_arc arc3( x+w-r, y+h-r, r, r, 0, M_PI_2 );
    agg::bezier_arc arc4( x+r, y+h-r, r, r, M_PI_2, M_PI_2 );
    agg::path_storage path2;
    path2.concat_path( arc1, 0 );
    path2.concat_path( arc2, 0 );
    path2.concat_path( arc3, 0 );
    path2.concat_path( arc4, 0 );
    agg::conv_curve<agg::path_storage> curve( path2 );
    agg::conv_stroke< agg::conv_curve<agg::path_storage > > stroke2( curve );
    stroke2.width( s );


    m_ras.add_path( stroke2 );

    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );
}

void wxAgg::drawRotatedRectangle( double x, double y, double w, double h, double s, double d )
{
    agg::path_storage path;

    path.move_to( x, y );
    path.line_rel( w, 0 );
    path.line_rel( 0, h );
    path.line_rel( -w, 0 );
    path.close_polygon();


    agg::trans_affine mtx;

    mtx *= agg::trans_affine_translation( -( x + w / 2 ) , -( y + h / 2 ) );
    mtx *= agg::trans_affine_rotation( d * 3.1415926 / 180.0 );
    mtx *= agg::trans_affine_translation( x + w / 2 , y + h / 2 );

    path.transform( mtx );

    agg::conv_stroke< agg::path_storage > stroke( path );

    stroke.width( s );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( stroke );
    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );

}

void wxAgg::drawText( const wxString& text, double x, double y, double w, double h, double s )
{
    m_ras.reset();

    applyClipping();

    agg::gsv_text t;
    t.size( h );
    t.text( text.mb_str( wxConvUTF8 ) );
    t.start_point( x, y );
    t.flip( true );

    agg::conv_stroke<agg::gsv_text> stroke(t);
    stroke.width( s );
    m_ras.add_path(stroke);
    //agg::conv_contour< agg::gsv_text> contour(t);
    //contour.width( 0.5 );
    //m_ras.add_path(contour);

    m_ren_sl->color( m_penColor );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );
}

double wxAgg::getTextWidth( const wxString& text, double h )
{
    agg::gsv_text t;
    t.size( h );
    t.text( text.mb_str( wxConvUTF8 ) );
    t.start_point( 0, 0 );
    t.flip( true );
    //agg::conv_stroke<agg::gsv_text> stroke( t );
    //stroke.width(  );

    double x1,x2,y1,y2;
	agg::bounding_rect_single( t, 0, &x1, &y1, &x2, &y2);

    //return t.text_width();
    return  x2 - x1;
}


void wxAgg::drawPolygon( const RealPointArray& points )
{
    agg::path_storage path;

    for( size_t i = 0; i < points.GetCount(); ++i )
    {
        double x = points.Item( i ).x;
        double y = points.Item( i ).y;

        if( i == 0 )
        {
            path.move_to( x, y );
            continue;
        }

        path.line_to( x, y );
    }

    path.close_polygon();

    m_ras.reset();

    applyClipping();

    m_ras.add_path( path );
    m_ren_sl->color( m_brushColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_brushColor.a );
}

void wxAgg::drawSolidPolygon( const RealPointArray& points, double s )
{
    agg::path_storage path;

    for( size_t i = 0; i < points.GetCount(); ++i )
    {
        double x = points.Item( i ).x;
        double y = points.Item( i ).y;

        if( i == 0 )
        {
            path.move_to( x, y );
            continue;
        }

        path.line_to( x, y );
    }

    path.close_polygon();

    m_ras.reset();

    applyClipping();

    m_ras.add_path( path );
    m_ren_sl->color( m_brushColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_brushColor.a );

    agg::conv_stroke< agg::path_storage > stroke( path );

    stroke.width( s );

    m_ras.reset();

    applyClipping();

    m_ras.add_path( stroke );
    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );

}

void wxAgg::drawInnerRectShadow(int width, int alpha, const wxRealRect& rect, const wxColor& color)
{
    int alphastep = alpha / width;

    // draw vertical lines
    for(int a = 0; a < width; a++) {
        wxColor c(color.Red(), color.Green(), color.Blue(),  alpha-(a*alphastep) );
        setPen( c );
        drawLine(rect.GetLeft()+a-1, rect.GetTop() + a, rect.GetRight(), rect.GetTop() + a, 1.0 );
    }

    // draw horiyontal lines
    for(int a = 0; a < width; a++) {
        wxColor c(color.Red(), color.Green(), color.Blue(),  alpha-(a*alphastep) );
        setPen( c );
        drawLine(rect.GetLeft() + a, rect.GetTop()+a, rect.GetLeft()+a, rect.GetBottom(), 1.0 );
    }

}

void wxAgg::drawInnerRectShadow(int width, int alpha, const wxRect& rect, const wxColor& color)
{
    wxRealRect rrect( (double)rect.x + 0.5, (double)rect.y + 0.5, (double)rect.width - 1.0 , (double)rect.height - 1.0 );
    drawInnerRectShadow(width,alpha,rrect,color);
}

/*
void wxAgg::getTextExtent( const wxString& text )
{
    agg::gsv_text t;
    t.size( 30 );
    t.text( text.mb_str( wxConvUTF8 ) );
    t.start_point( 30, 30 );
    t.flip( true );
    agg::conv_stroke<agg::gsv_text> stroke(t);
    stroke.width(3.0);

    double x1,x2,y1,y2;

    bounding_rect_single( stroke, 0, &x1, &y1, &x2, &y2);

    //wxLogMessage( _T(" x1 : %.2f, y1 : %.2f, x2 : %.2f, y2 : %.2f "), x1, y1, x2, y2 );
    //return ()

    //bool bounding_rect_single(VertexSource& vs, unsigned path_id,
    //                          CoordT* x1, CoordT* y1, CoordT* x2, CoordT* y2)
}

*/


void wxAgg::render( wxDC& dc )
{
    wxImage  myImage( m_width, m_height, m_buffer, true );
    //myImage.SetAlpha( m_alphaBuffer, false );
    myImage.SetAlpha( m_alphaBuffer, true );
    if(myImage.IsOk()){
#if defined(__UNIX__)
        wxBitmap myBitmap( myImage );
#else
        wxBitmap myBitmap( myImage,32);
#endif
        dc.DrawBitmap( myBitmap, 0, 0, true );
    }
}


void wxAgg::renderAlphaMask( const double a )
{
    m_rena->color( agg::gray8( (int)(a * 255.0) ) );
    agg::render_scanlines( m_ras, m_sl, *m_rena );
}

wxBitmap* wxAgg::renderToBitmap()
{
    wxImage  myImage( m_width, m_height, m_buffer, true );
    //myImage.SetAlpha( m_alphaBuffer, false );
    myImage.SetAlpha( m_alphaBuffer, true );
    wxBitmap* myBitmap = new wxBitmap( myImage );
    return myBitmap;
}

wxImage* wxAgg::renderToImage()
{
    wxImage *myImage = new wxImage( m_width, m_height, m_buffer, true );
    //myImage.SetAlpha( m_alphaBuffer, false );
    myImage->SetAlpha( m_alphaBuffer, true );

    return myImage;
}

/// necessary to copy, if the image 'life' is longer then the canvas 'life'
wxImage* wxAgg::renderCopyToImage(const wxRect& renderRect)
{
    wxRect renderBox = renderRect;

    if(renderBox == wxDefaultRect) renderBox = wxRect(0,0,m_width,m_height);

    renderBox.Intersect(wxRect(0,0,m_width,m_height));

    wxImage *image = new wxImage( renderBox.width, renderBox.height, false );
    image->InitAlpha();

    unsigned char* data = image->GetData();
    unsigned char* alpha = image->GetAlpha();

    long src_x = 0, src_y = 0, src_idx = 0, target_y = 0, target_idx = 0, src_w = 0, target_w = 0;

    for(src_y = renderBox.GetTop(), target_y=0; src_y <= renderBox.GetBottom(); src_y++, target_y++ ) {

        src_x = renderBox.GetLeft();
        src_w = m_width;
        target_w = renderBox.GetWidth();

        src_idx = src_y * m_width + src_x;
        target_idx = target_y * target_w;

        memcpy(data + target_idx * 3, m_buffer + src_idx * 3, target_w * 3);

        memcpy(alpha + target_idx, m_alphaBuffer + src_idx, target_w);
    }

    return image;
}


// A simple function to form the gradient color array
// consisting of 3 colors, "begin", "middle", "end"
//---------------------------------------------------
void wxAgg::fillColorArray( color_array_type& array, agg::rgba8 begin, agg::rgba8 end )
{
    unsigned i;

    for( i = 0; i < array.size(); ++i )
    {
        array[i] = begin.gradient( end, i / double( array.size() ) );
    }
}

/*TEST*/
wxColor wxAgg::gradient( const wxColor& bc, const wxColor& ec, double k )
{
    agg::rgba  bcRGBA = convertToRGBA( bc );
    agg::rgba  ecRGBA = convertToRGBA( ec );

    wxColor     gc = convertToColor( bcRGBA.gradient( ecRGBA, k ) );

    return gc;
}

wxRealRect wxAgg::getTTFTextBound(const wxString& text, double angleDeg, double x, double y)
{
    wxRealRect box;

    font_manager_type fman( m_font_engine );

    bool kerning = true;

    double height = m_font_engine.height();


    m_font_engine.gamma( agg::gamma_power( 1.0 ) );

    agg::trans_affine mtx;
    mtx *= agg::trans_affine_rotation( agg::deg2rad( angleDeg ) );
    m_font_engine.transform(mtx);

    double x_ = x;
    double y_ = y;

    box.x = x;
    box.y = y;

    const wxChar* convertedTxt = text.c_str();

    int maxh = 0;

    for ( unsigned int j = 0; j < text.Len(); ++j )
    {
        unsigned char_code = (unsigned)convertedTxt[j];

        const agg::glyph_cache* glyph = fman.glyph( char_code );

        if( glyph )
        {
            if( kerning )
            {
                fman.add_kerning( &x_, &y_ );
            }

            fman.init_embedded_adaptors( glyph, x_, y_ );

            // increment pen position
            x_ += glyph->advance_x;
            y_ += glyph->advance_y;

            int glyph_h = abs(glyph->bounds.y2 - glyph->bounds.y1);
            if( glyph_h >= maxh  ) maxh = glyph_h;
        }
    }

    box.w = x_ - x + maxh * sin( agg::deg2rad( angleDeg ) ) + m_font_weight;
    box.h = y_ - y + height * cos( agg::deg2rad( angleDeg ) ) + m_font_weight;

    return box;
}

bool wxAgg::loadTTFont( const wxString& fontfilename, double height, double weight, double width )
{
    const char* fname = UnicodeHelper::U2C( fontfilename );
    agg::glyph_rendering gren = agg::glyph_ren_outline;

    m_font_engine.gamma( agg::gamma_power( 1.0 ) );

    bool ret = false;

    if ( width == 0 ) width = height;

    if ( m_font_engine.load_font( fname, 0, gren ) )
    {
        m_font_engine.hinting( true );
        m_font_engine.height( height);
        m_font_engine.width( width );
        m_font_engine.flip_y( true );
        m_font_engine.char_map( FT_ENCODING_UNICODE );
        m_font_weight = weight;

        ret = true;
    }

    return ret;
}

wxRealRect wxAgg::drawTTFText( double x, double y, const wxString& text, double angleDeg )
{
    wxRealRect box;


    font_manager_type fman( m_font_engine );
    agg::conv_curve<font_manager_type::path_adaptor_type> curves(fman.path_adaptor());
    agg::conv_contour<agg::conv_curve<font_manager_type::path_adaptor_type> > contour(curves);

    m_ras.reset();

    applyClipping();

    bool kerning = true;

    curves.approximation_scale( 2.0 );
    contour.auto_detect_orientation( false );

    double height = m_font_engine.height();

    contour.width( m_font_weight * height * 0.05 );
    m_ras.gamma( agg::gamma_power( 1.0 ) );
    m_font_engine.gamma( agg::gamma_power( 1.0 ) );

    agg::trans_affine mtx;
    mtx *= agg::trans_affine_rotation( agg::deg2rad( angleDeg ) );
    m_font_engine.transform(mtx);

    double x_ = x;
    double y_ = y;

    box.x = x;
    box.y = y;

    int maxh = 0;

    const wxChar* convertedTxt = text.c_str();

    for ( unsigned int j = 0; j < text.Len(); ++j )
    {
        unsigned char_code = (unsigned)convertedTxt[j];

        const agg::glyph_cache* glyph = fman.glyph( char_code );

        if( glyph )
        {
            if( kerning )
            {
                fman.add_kerning( &x_, &y_ );
            }

            fman.init_embedded_adaptors( glyph, x_, y_ );

            if(fabs( m_font_weight ) <= 0.01)
                m_ras.add_path( curves );
            else
                m_ras.add_path( contour );

            // increment pen position
            x_ += glyph->advance_x;
            y_ += glyph->advance_y;

            int glyph_h = abs(glyph->bounds.y2 - glyph->bounds.y1);
            if( glyph_h >= maxh  ) maxh = glyph_h;

            if( x_ > (double)m_width-0.5 ) break;
            if( y_ > (double)m_height-0.5 ) break;
        }
    }



    m_ren_sl->color( m_penColor );
    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
    renderAlphaMask( m_penColor.a );


    box.w = x_ - x + maxh * sin( agg::deg2rad( angleDeg ) ) + m_font_weight;
    box.h = y_ - y + height * cos( agg::deg2rad( angleDeg ) ) + m_font_weight;


    return box;
}

wxRealRect wxAgg::drawTTFText( double x, double y, const wxString& text, double height, double weight, const wxString& fontfilename, double angleDeg, double width )
{
    wxRealRect box;

    font_manager_type fman( m_font_engine );
    agg::conv_curve<font_manager_type::path_adaptor_type> curves(fman.path_adaptor());
    agg::conv_contour<agg::conv_curve<font_manager_type::path_adaptor_type> > contour(curves);

    m_ras.reset();

    applyClipping();

    agg::glyph_rendering gren = agg::glyph_ren_outline;

    bool kerning = true;

    if ( fabs( width + 1 ) < 0.05 ) width = height;

    unsigned num_glyphs = 0;

    curves.approximation_scale( 2.0 );
    contour.auto_detect_orientation( false );

    contour.width( weight * height * 0.05);

    const char* fname = UnicodeHelper::U2C( fontfilename );

    m_ras.gamma( agg::gamma_power( 1.0 ) );
    m_font_engine.gamma( agg::gamma_power( 1.0 ) );

    box.x = x;
    box.y = y;

    int maxh = 0;


    if( m_font_engine.load_font(fname, 0, gren) )
    {
        m_font_engine.hinting( true );
        m_font_engine.height( height);
        m_font_engine.width( width );
        m_font_engine.flip_y( true );
        m_font_engine.char_map( FT_ENCODING_UNICODE );

        agg::trans_affine mtx;
        mtx *= agg::trans_affine_rotation( agg::deg2rad( angleDeg ) );
        m_font_engine.transform(mtx);

        double x_ = x;
        double y_ = y;

        const wxChar* convertedTxt = text.c_str();

        for ( unsigned int j = 0; j < text.Len(); ++j )
        {
            unsigned char_code = (unsigned)convertedTxt[j];

            const agg::glyph_cache* glyph = fman.glyph( char_code );

            if( glyph )
            {
                if( kerning )
                {
                    fman.add_kerning( &x_, &y_ );
                }

                fman.init_embedded_adaptors( glyph, x_, y_ );

                if ( fabs( weight ) <= 0.01 )
                {
                    m_ras.add_path( curves );
                }
                else
                {
                    m_ras.add_path( contour );
                }

                // increment pen position
                x_ += glyph->advance_x;
                y_ += glyph->advance_y;
                ++num_glyphs;

                int glyph_h = abs(glyph->bounds.y2 - glyph->bounds.y1);
                if( glyph_h >= maxh  ) maxh = glyph_h;

                if( x_ > (double)m_width-0.5 ) break;
                if( y_ > (double)m_height-0.5 ) break;
            }
        }

        m_ren_sl->color( m_penColor );
        agg::render_scanlines( m_ras, m_sl, *m_ren_sl );
        renderAlphaMask( m_penColor.a );

        box.w = x_ - x + maxh * sin( agg::deg2rad( angleDeg ) ) + m_font_weight;
        box.h = y_ - y + height * cos( agg::deg2rad( angleDeg ) ) + m_font_weight;
    }


    return box;
}

void wxAgg::aggTest1()
{

}

void wxAgg::aggTest2()
{

}

void wxAgg::setFontHeight( double h )
{
    m_font_engine.height(h);
    m_font_engine.width(h);
}

double wxAgg::getFontHeight()
{
    return m_font_engine.height();
}

void wxAgg::applyClipping()
{
    if ( m_clippingRect != wxDefaultRect )
        m_ras.clip_box( m_clippingRect.x, m_clippingRect.y, m_clippingRect.GetRight()+1, m_clippingRect.GetBottom()+1 );
    else
        m_ras.clip_box( 0, 0, m_width , m_height );
}

/*
void wxAgg::drawCircleWithClipping( double x, double y, double r, wxRealRect rect, double s )
{
    agg::ellipse ellipse( x, y, r, r );
    agg::conv_stroke< agg::ellipse > stroke( ellipse );

    //stroke.line_join(join);
    //stroke.line_cap(cap);
    //stroke.miter_limit(m_miter_limit.value());
    stroke.width( s );

    m_ras.reset();
//    m_ras.reset_clipping(true);
    m_ras.clip_box( rect.x, rect.y, rect.x + rect.w, rect.y + rect.h );
    m_ras.add_path( stroke );
    m_ren_sl->color( m_penColor );

    agg::render_scanlines( m_ras, m_sl, *m_ren_sl );

    renderAlphaMask( m_penColor.a );
//    m_ras.reset_clipping(true);
}*/
