#include "VemsMatrixHelper.h"
#include <wx/wx.h>
#include "../Math.h"
#include "../expression/RPNParser.h"
#include "../expression/VariableData.h"

/**
*/
void VemsMatrixHelper::print( const VemsMatrix<uint8_t>& m )
{
    wxString line;
    wxString table;

    for( size_t i = 0; i < m.rows(); ++i )
    {
        for ( size_t j = 0; j < m.cols(); ++j )
        {
            line.Append( wxString::Format( _T("%02X"), m.get( i, j ) ) );
            if ( j < ( m.cols() - 1 ) ) { line.Append( _T(" ") ); }
        }

        //wxLogMessage( line );
        table.Append( _T("\t\t") + line + _T("\n") );
        line.Clear();

    }

    wxString info = wxString::Format( _T("VemsMatrix<uint8_t>[%d][%d] dump\n"), m.rows(), m.cols() );
    wxLogMessage( info + table );
}

/**
*/
void VemsMatrixHelper::print( const VemsMatrix<double>& m )
{
    wxString line;
    wxString table;

    for( size_t i = 0; i < m.rows(); ++i )
    {
        for ( size_t j = 0; j < m.cols(); ++j )
        {
            line.Append( wxString::Format( _T("%5.2f"), m.get( i, j ) ) );
            if ( j < ( m.cols() - 1 ) ) { line.Append( _T(" ") ); }
        }

        //wxLogMessage( line );
        table.Append( _T("\t\t") + line + _T("\n") );
        line.Clear();

    }

    wxString info = wxString::Format( _T("VemsMatrix<double>[%d][%d] dump\n"), m.rows(), m.cols() );
    wxLogMessage( info + table );
}

/**
*/
bool VemsMatrixHelper::convert( const VemsMatrix<uint8_t>& from, VemsMatrix<double>& to, double scale, double translate )
{
    //check size match
    if( from.rows() != to.rows() || from.cols() != to.cols() ) { return false; }

    //copy and convert data
    for ( size_t i = 0; i < from.size(); ++i )
    {
        to.set( i, ( from.get( i ) + translate ) * scale );
//        Logger::logf( LOG_MSG, _T("VemsMatrixHelper::convert() - idx: %d, %2X -> %f"), i, from.get(i), to.get(i) );
    }

    return true;
}


/**
*/
bool VemsMatrixHelper::convert( const VemsMatrix<uint8_t>& from, VemsMatrix<double>& to, const wxString& expr )
{
    //check size match
    if( from.rows() != to.rows() || from.cols() != to.cols() ) { return false; }

    RPNParser       rpn;
    VariableDataDH  variables;

    rpn.setVariableData( &variables );

    //copy and convert data
    for ( size_t i = 0; i < from.size(); ++i )
    {

        variables.addValue( _T("x"), (double)from.get( i ) );
        double converted = rpn.calculate( expr );
        to.set( i, converted );
    }

    return true;
}


/**
*/
bool VemsMatrixHelper::convert( const VemsMatrix<double>& from, VemsMatrix<uint8_t>& to, double scale, double translate )
{
    //check size match
    if( from.rows() != to.rows() || from.cols() != to.cols() ) { return false; }

    //copy and convert data
    double d = 0.0;

    for ( size_t i = 0; i < from.size(); ++i )
    {
        d = from.get( i ) / scale - translate;
        if ( ceil(d) > 255 ) { d = 255; }
        else if ( ceil(d) < 0 ) { d = 0; }

        to.set( i, (uint8_t)d );
    }

    return true;
}

bool VemsMatrixHelper::convert( const VemsMatrix<int>& from, VemsMatrix<double>& to )
{
    //check size match
    if( from.rows() != to.rows() || from.cols() != to.cols() ) { return false; }

    //copy and convert data
    for ( size_t i = 0; i < from.size(); ++i )
    {
        to.set( i, from.get( i ) );
    }

    return true;
}

bool VemsMatrixHelper::convertLambda( VemsMatrix<uint8_t>& m )
{
    double mtValue = 0.0;

    for ( size_t i = 0; i < m.size(); ++i )
    {
        mtValue = 65535.0 / ( (double)m.get(i) + 200.0 ) - 73.0 + 0.55;
//        Logger::logf( LOG_MSG, _T("VemsMatrixHelper::convertLambda() - idx: %d, %2X -> %2X"), i, m.get(i), (uint8_t)mtValue );
        m.set( i, (uint8_t)mtValue );

    }

    return true;
}


/**

*/
void VemsMatrixHelper::interpolate_col( VemsMatrix<double>* data,
                                        const VemsMatrix<int>& cols,
                                        size_t col )

{
    if ( cols.cols() < col || data->cols() < col ) { return; }

    if ( col == 0 || col == data->cols() - 1 ) { return; }

    double p0[2];
    double p1[2];
    double px[2];

    for ( size_t i = 0; i < data->rows(); ++i )
    {
        p0[0] = cols.get( 0, col - 1 );
        p0[1] = data->get( i, col - 1 );

        p1[0] = cols.get( 0, col + 1 );
        p1[1] = data->get( i, col + 1 );

        px[0] = cols.get( 0, col );

        bool res = Math::linearInterpolation2D( p0, p1, px );

        if ( res ) { data->set( i, col, px[1] ); }
    }
}


/**

*/
void VemsMatrixHelper::interpolate_row( VemsMatrix<double>* data,
                                        const VemsMatrix<int>& rows,
                                        size_t row )

{
    if ( rows.cols() < row || data->rows() < row ) { return; }

    if ( row == 0 || row == data->rows() - 1 ) { return; }

    double p0[2];
    double p1[2];
    double px[2];

    for ( size_t i = 0; i < data->cols(); ++i )
    {
        p0[0] = rows.get( 0, row - 1 );
        p0[1] = data->get( row - 1, i );

        p1[0] = rows.get( 0, row + 1 );
        p1[1] = data->get( row + 1, i );

        px[0] = rows.get( 0, row );

        bool res = Math::linearInterpolation2D( p0, p1, px );

        if ( res ) { data->set( row, i, px[1] ); }
    }
}


/**

*/
void VemsMatrixHelper::interpolate( VemsMatrix<double>* data,
                                    const VemsMatrix<double>& row,
                                    const VemsMatrix<double>& col,
                                    const MatrixSelection& s )

{
    /// verify dimensions
    if ( data->rows() > row.size() || data->cols() > col.size() ) { return; }

    /// verify selection

    /// interpolate
    double topleft      = data->get( s.top, s.left );            /// z0
    double topright     = data->get( s.top, s.right );           /// z1
    double bottomleft   = data->get( s.bottom, s.left );         /// z2
    double bottomright  = data->get( s.bottom, s.right );        /// z3

    double p0[3];
    double p1[3];
    double p2[3];
    double p3[3];
    double px[3];
    p0[0] = p2[0] = col[ s.left ];      ///x0 x2
    p0[1] = p1[1] = row[ s.top ];      ///y0 y1
    p0[2] = topleft;

    p1[0] = p3[0] = col[ s.right ];     ///x1 x3
    p1[2] = topright;

    p2[1] = p3[1] = row[ s.bottom ];    ///y2 y3
    p2[2] = bottomleft;

    p3[2] = bottomright;

    for ( int i = s.top; i <= s.bottom; ++i )
    {
        for ( int j = s.left; j <= s.right; ++j )
        {
            px[0] = col[j]; ///x
            px[1] = row[i]; ///y
            px[2] = -1; /// invalid value

            bool res = Math::linearInterpolation3D(p0, p1, p2, p3, px);

            if ( res ) { data->set( i, j, px[2] ); }
        }
    }

}

bool VemsMatrixHelper::resample1D(  const VemsMatrix<uint8_t>& oldX,
                                    const VemsMatrix<uint8_t>& newX,
                                    VemsMatrix<uint8_t>* table )
{
    VemsMatrix<uint8_t> oldTable8 ( *table );

    VemsMatrix<double> oldXDouble(oldX.rows(), oldX.cols());
    VemsMatrix<double> newXDouble(newX.rows(), newX.cols());
    VemsMatrix<double> oldTable( oldTable8.rows(), oldTable8.cols());

    convert(oldX, oldXDouble, 1.0, 0.0);
    convert(newX, newXDouble, 1.0, 0.0);
    convert(oldTable8, oldTable, 1.0, 0.0);

    /// resize oldTable to->  table->setSize(  newYDouble.size(), newXDouble.size() );
    bool res = resample1D(oldXDouble, newXDouble, &oldTable);
    table->resize(1, newX.cols() ); /// !!!
    //table->setSize(  oldTable.rows(), oldTable.cols() );
    convert(oldTable, *table, 1.0, 0.0);
    return res;
}

bool VemsMatrixHelper::resample1D(  const VemsMatrix<double>& oldX,
                                    const VemsMatrix<double>& newX,
                                    VemsMatrix<double>* table )
{
    VemsMatrix<double> oldData( *table );
    table->setSize(1, newX.size() );
    table->fill( 0.0 );

    double p0[2],p1[2], px[2];
    int idxX0 = 0, idxX1 =0;
    //bool res = false;


    size_t xSize = newX.size();
    for(size_t x =0; x < xSize; ++x){
        px[0] = newX[x]; /// px.x
        p0[0] = getLeftNeighbour(px[0], oldX, idxX0);
        p1[0] = getRightNeighbour(px[0], oldX, idxX1);
        //Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() --- col:%d,x=%f,(%d,%d) (%f,%f)"),x,px[0], idxX0, idxX1, p0[0], p1[0] );
        p0[1] = oldData.get(0, idxX0);
        p1[1] = oldData.get(0, idxX1);
        //Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - old(%i,%i)=%f"),idxX0, idxY0, p0[2]);
        Math::linearInterpolation2D(p0, p1, px);
        table->set(0,x, px[1] );
        //idx++;
    }
    return true;
}

/////////////////////////////////////////////////////////////////////

bool VemsMatrixHelper::resample2D(  const VemsMatrix<uint8_t>& oldX,
                                    const VemsMatrix<uint8_t>& oldY,
                                    const VemsMatrix<uint8_t>& newX,
                                    const VemsMatrix<uint8_t>& newY,
                                    VemsMatrix<uint8_t>* table )
{
    VemsMatrix<uint8_t> oldTable8 ( *table );

    VemsMatrix<double> oldXDouble(oldX.rows(), oldX.cols());
    VemsMatrix<double> oldYDouble(oldY.rows(), oldY.cols());
    VemsMatrix<double> newXDouble(newX.rows(), newX.cols());
    VemsMatrix<double> newYDouble(newY.rows(), newY.cols());
    VemsMatrix<double> oldTable( oldTable8.rows(), oldTable8.cols());

    convert(oldX, oldXDouble, 1.0, 0.0);
    convert(oldY, oldYDouble, 1.0, 0.0);
    convert(newX, newXDouble, 1.0, 0.0);
    convert(newY, newYDouble, 1.0, 0.0);
    convert(oldTable8, oldTable, 1.0, 0.0);

    /// resize oldTable to->  table->setSize(  newYDouble.size(), newXDouble.size() );
    bool res = resample2D(oldXDouble, oldYDouble, newXDouble, newYDouble, &oldTable);
    table->resize(newY.cols(), newX.cols() ); /// !!!
    //table->setSize(  oldTable.rows(), oldTable.cols() );
    convert(oldTable, *table, 1.0, 0.0);
    return res;
}


/**
*/
bool VemsMatrixHelper::resample2D(  const VemsMatrix<double>& oldX,
                                    const VemsMatrix<double>& oldY,
                                    const VemsMatrix<double>& newX,
                                    const VemsMatrix<double>& newY,
                                    VemsMatrix<double>* table )
{
//    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - table rows: %d, cols: %d"),
//                    newX.size(), newY.size() );


    //VemsMatrixHelper::print(*table);

    VemsMatrix<double> oldData( *table );
/*
    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() -     old(1,1) %f"), table->get(1,1) );
    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - copyied(1,1) %f"), oldData.get(1,1) );

    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() -     old(10,10) %f"), table->get(10,10) );
    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - copyied(10,10) %f"), oldData.get(10,10) );
*/
    table->setSize(  newY.size(), newX.size() );
    table->fill( 0.0 );

//    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - table rows: %d, cols: %d"),
//                    table->rows(), table->cols() );

    double p0[3],p1[3],p2[3],p3[3],px[3];
    int idxY0 =0, idxY1 =0, idxX0 = 0, idxX1 =0;
    //int idx =0;
    //bool res = false;

    size_t xSize = newX.size();
    for( size_t y =0; y < newY.size(); ++y )
    {
        px[1] = newY[y]; /// px.y
        p0[1] = p1[1] = getLeftNeighbour( px[1], oldY, idxY0 );
        p2[1] = p3[1] = getRightNeighbour(px[1], oldY, idxY1);

        //Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - row:%d,y=%f,(%d,%d) (%f,%f)"),y,px[1], idxY0, idxY1, p0[1], p2[1] );
        for(size_t x =0; x < xSize; ++x){
                px[0] = newX[x]; /// px.x
                p0[0] = p2[0] = getLeftNeighbour(px[0], oldX, idxX0);
                p1[0] = p3[0] = getRightNeighbour(px[0], oldX, idxX1);
                //Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() --- col:%d,x=%f,(%d,%d) (%f,%f)"),x,px[0], idxX0, idxX1, p0[0], p1[0] );

                /// (row,col)
                p0[2] = oldData.get(idxY0, idxX0); /// p0.z
                p1[2] = oldData.get(idxY0, idxX1); /// p1.z
                p2[2] = oldData.get(idxY1, idxX0); /// p2.z
                p3[2] = oldData.get(idxY1, idxX1); /// p3.z

                //Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - old(%i,%i)=%f"),idxX0, idxY0, p0[2]);

                px[2] = Math::linearInterpolation3D(p0,p1,p2,p3 ,px[0], px[1]);
                //Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() --- old(%i,%i)=%f => new(%i,%i)%f"),idxX0, idxY0, p0[2], x,y, px[2]);
/*
                if (!res)
                {
                    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - res = false") );
                    return false;
                }
*/
//                Logger::logf( LOG_MSG, _T("VemsMatrixHelper::resample2D() - idx = %d, px[2] = %f"), idx, px[2] );

                table->set(y,x, px[2] );
                //idx++;
        }
    }

    return true;
}


double VemsMatrixHelper::getLeftNeighbour( double x, const VemsMatrix<double>& oldA, int& oldIdx )
{
    for( size_t i = 0; i < oldA.size()-1; ++i )
    {
        if ( oldA[i] == x ){
            oldIdx = i;
            return x;
        }

        if( oldA[i] <= x && x < oldA[i+1] ){
            oldIdx = i;
            return oldA[i];
        }
    }

    if ( oldA[ oldA.size()-1 ] == x )
    {
        oldIdx = oldA.size()-1;
        return x;
    }

    if ( oldA.size() > 0 ) {
        if ( x < oldA[0] ) {
            oldIdx = 0;
            return oldA[0];
        } else if ( x > oldA[oldA.size() - 1] ) {
            oldIdx = oldA.size() - 1;
            return oldA[oldA.size() - 1];
        }

    } else {
        oldIdx = 0;
        return 0;
    }
    return 0;
}


double VemsMatrixHelper::getRightNeighbour(double x, const VemsMatrix<double>& oldA, int& oldIdx)
{
    for(size_t i = 0; i < oldA.size()-1; ++i)
    {
        if ( oldA[i] == x ){
            oldIdx = i;
            return x;
        }

        if(oldA[i] < x && x <= oldA[i+1]){
            oldIdx = i+1;
            return oldA[i+1];
        }
    }

    if ( oldA.size() > 0 ) {
        if ( x < oldA[0] ) {
            oldIdx = 0;
            return oldA[0];
        } else if ( x > oldA[oldA.size() - 1] ) {
            oldIdx = oldA.size() - 1;
            return oldA[oldA.size() - 1];
        }

    } else {
        oldIdx = 0;
        return 0;
    }
    return 0;
}


double  VemsMatrixHelper::interpolate( const VemsMatrix<double>& row, const VemsMatrix<double>& col, double x )
{
    double p0[2];
    double p1[2];
    double px[2];

    int right   = 0;
    int left    = 0;

    p0[0] = getLeftNeighbour( x, col, left );
    p1[0] = getRightNeighbour( x, col, right );

    p0[1] = row.get( left );
    p1[1] = row.get( right );

    px[0] = x;

//    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::interpolate() left: %d, right: %d"), left, right );
//    Logger::logf( LOG_MSG, _T("VemsMatrixHelper::interpolate() p0.x = %f, p0.y = %f, p1.x = %f, p1.y = %f"), p0[0], p0[1], p1[0], p1[1] );

    Math::linearInterpolation2D( p0, p1, px );

    return px[1];
}


/// p0(x0,y0,z0) | p1(x1,y1,z1)
/// p2(x2,y2,z2) | p3(x3,y3,z3)
/// assume: x0 = x2, x1 = x3
/// assume: y0 = y1, y2 = y3
double VemsMatrixHelper::interpolate3D( const VemsMatrix<double>& col,
                                        const VemsMatrix<double>& row,
                                        const VemsMatrix<double>& table,
                                        double x, double y )
{
	int right = 0;

	if ( col.get ( col.cols() - 1 ) < x )
	{
	    right = col.cols() - 1;
	}
	else
	{
        getRightNeighbour( x , col, right );
	}

    int bottom = 0;

    if ( row.get ( row.cols() - 1 ) < y )
	{
	    bottom =  row.cols() - 1;
	}
	else
	{
        getRightNeighbour( y, row, bottom );
	}


    int left = right ? right - 1 : 0;
    int top = bottom ? bottom - 1 : 0;

    double p0[3] = { col[left], row[top], table.get( top, left ) };
    double p1[3] = { col[right], row[top], table.get( top, right ) };
    double p2[3] = { col[left], row[bottom], table.get( bottom, left ) };
    double p3[3] = { col[right], row[bottom], table.get( bottom, right) };

	return Math::linearInterpolation3D( p0, p1, p2, p3, x, y );
}

