#include "EepromReaderClient.h"

#include <wx/file.h>

#include "DeviceModeSelector.h"
#include "../CommunicationClientEventHandler.h"
#include "../../events/CommunicationEvent.h"
#include "../../events/GeneralEvent.h"
#include "../../common/TxData.h"
#include "../../logger.h"
#include "../../Math.h"
#include "../../RawLogHelper.h"
#include "../../triggerlog/Unescaper.h"
#include "../../FirmwareInfo.h"
#include "../../common/VemsContext.h"
#include "../../DirHelper.h"
#include "../../events/EepromRawDataEvent.h"

EepromReaderClient::EepromReaderClient( CommunicationClientEventHandler* clientEventHandler,
                                        wxEvtHandler* eventHandler,
                                        CommunicationBase* commBase,
                                        const IntArray& baudrates,
                                        ECUInfo::EcuMode sourceMode )
{
    m_clientEventHandler    = clientEventHandler;
    m_eventHandler          = eventHandler;

    m_fileaddress           = 0;

    m_sourceMode = sourceMode;

    m_client            = NULL;

    m_crcErrors         = 0;

    m_address = 0;

    if ( m_sourceMode == ECUInfo::MAN )
    {
        m_state = SEND_COMMAND;
    }
    else
    {
        m_client    = new DeviceModeSelector( this, this, commBase, baudrates, m_sourceMode, ECUInfo::MAN );
        m_state     = SELECT_MAN_MODE;
    }

    m_membuffer = new wxMemoryBuffer( 240 * 20 ); // allocate extra space
}

EepromReaderClient::~EepromReaderClient()
{
    if ( m_client )
        delete m_client;

    delete m_membuffer;
}

void EepromReaderClient::stop()
{
    /// TODO
}

//void EepromReaderClient::handleEvent( CommunicationClientEvent::Type eventType )
//{
//    switch ( eventType )
//    {
//        case CommunicationClientEvent::DEVICE_MODE_SELECTED:
//            LOGP(_T("MAN Mode selected"))
//            m_state = SEND_COMMAND;
//            break;
//
//        case CommunicationClientEvent::TIMEOUT:
//            m_state = TIMEDOUT;
//            break;
//
//        default:
//            break;
//    }
//}

void EepromReaderClient::OnDeviceModeSelected( ECUInfo::EcuMode deviceMode )
{
    m_state = SEND_COMMAND;
}

void EepromReaderClient::OnTimeout()
{
    m_state = TIMEDOUT;
}

void EepromReaderClient::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case SELECT_MAN_MODE:
            selectManMode( buffer, size );
            break;

        case SEND_COMMAND:
            sendCommand( buffer, size );
            break;

        case TIMEDOUT:
            LOGP(_T("Timeout!!!"))
            //m_state = FINISHED;
            sendNoValidData();
            break;

        case CRC_FAILED:
            sendNoValidData();
            break;

        default:
            break;
    }
}

void EepromReaderClient::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case WAIT_MAN_MODE:
            waitManMode( buffer, size );
            break;

        case WAIT_DATA:
            waitData( buffer, size );
            break;

        default:
            break;
    }
}

void EepromReaderClient::finish()
{
    m_clientEventHandler->handleEvent( CommunicationClientEvent::TIMEOUT );
}

bool EepromReaderClient::isFinished()
{
    m_clientEventHandler->handleEvent( CommunicationClientEvent::EEPROM_DOWNLOADED );
    return m_state == FINISHED;
}


void EepromReaderClient::selectManMode( uint8_t* buffer, int& size )
{
    m_client->setTimeout( getTimeout() );
    m_client->setInterWriteDelay( getInterWriteDelay() );
    m_client->write( buffer, size );

    m_state = WAIT_MAN_MODE;
}

void EepromReaderClient::waitManMode( uint8_t* buffer, int size )
{
    m_client->read( buffer, size );

    if ( m_client->isFinished () )
    {
        m_client->finish();
    }
}

void EepromReaderClient::sendCommand( uint8_t* buffer, int& size )
{
    uint8_t data[4];
    data[0] = 'm';
    data[1] = 'c';
    data[2] = 'R';
    data[3] = m_address;

    TxData command( data, 4 );

    size = command.size;

    memcpy( buffer, command.data, command.size );

    m_rxDataLength          = 0;
    m_timeoutCount          = 0;

    LOGP( command.toString() +_T(" sent"))

    m_state = WAIT_DATA;
}

void EepromReaderClient::waitData( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {
        m_rxBuffer[ m_rxDataLength++ ] = buffer[0];

        if ( buffer[0] == HDLC_END )
        {
            uint8_t* escaped = new uint8_t[256+10];
            size_t unescLen = unescapeData( m_rxBuffer, m_rxDataLength-4, &escaped );

            // LITTLE ENDIAN
            uint16_t ecu_crc = m_rxBuffer[m_rxDataLength-1-2] + 256 * m_rxBuffer[m_rxDataLength-1-1]; // last one is the 0x7e

            // calc crc
            uint16_t calc_crc = 0xffff;

            for( size_t i = 0; i < unescLen; ++i )
            {
                calc_crc = Math::crc16_sipr( calc_crc, escaped[i] );
            }
            calc_crc = Math::crc16_sipr( calc_crc, m_rxBuffer[m_rxDataLength-4]); // type 0x11

            if ( calc_crc != ecu_crc )
            {
                LOGP( wxString::Format( _T("CRC error: calculated=%0x, received=%0x"), calc_crc, ecu_crc ) );
                LOGP( wxString::Format( _T("Reading address 0x%x again"), m_address ))

                ++m_crcErrors;

                if ( m_crcErrors >= MAX_CRCERR )
                {
                    m_state = CRC_FAILED;
                }
                else
                    m_state = SEND_COMMAND;

                delete[] escaped;

                return ;
            }
            else
            {
                LOGP( wxString::Format( _T("Addr 0x%x CRC OK"), m_address ) )

                if ( m_writeToFile )
                    writeToFile( escaped, unescLen );

                m_crcErrors = 0;

                addToBuffer( escaped, unescLen );
            }

            delete[] escaped;

            m_address+=15;

            if ( m_address > 255 )
            {
                if ( m_writeToFile )
                    m_file.Close();

                sendEvent(4096);

                m_state = FINISHED;
            }
            else
                m_state = SEND_COMMAND;
        }
    }
    else
    {
        wxThread::Sleep(2);

        m_timeoutCount++;

        if ( m_timeoutCount < 1000 )
        {
            return;
        }

        m_state = TIMEDOUT;
    }
}

size_t EepromReaderClient::unescapeData( uint8_t* buffer, size_t buflen, uint8_t** outbuffer )
{
    Unescaper u;

    size_t outlen = 0;
    for ( size_t i = 0; i < buflen; ++i )
    {
        uint8_t in = buffer[i];
        uint8_t out = 0;

        bool ret = u.parse( in, out );

        if ( ret )
            (*outbuffer)[outlen++] = out;
    }

    return outlen;
}

void EepromReaderClient::writeToFile( uint8_t* data, size_t datalen )
{
    if ( m_fileaddress >= 4096 )
        return;

    if ( m_file.IsOpened() == false )
    {
        wxString fname = _T("unknown-");
        wxString hdr = _T("");

        wxObject* fwobj = VemsContext::getObject( _T("FirmwareInfo") );
        FirmwareInfo* fwInfo = dynamic_cast<FirmwareInfo*>(fwobj);

        char chdr[512];
        memset( chdr, 32, 512 );

        if ( fwInfo != NULL )
        {
            hdr = RawLogHelper::getVemsCfgHeader( *fwInfo );
            memcpy( chdr, hdr.mb_str(), hdr.Length() );
            fname = fwInfo->getSerialNumber() + _T("-");
        }

        wxDateTime now = wxDateTime::Now();
        fname += now.FormatISODate() + _T("_") + now.FormatISOTime();
        fname.Replace( _T(":"),_T("."),true );
        fname += _T(".raweeprom");

        m_file.Create( DirHelper::GetPath(dLOG) + fname, true );

        if ( hdr.Length() > 0 )
        {
            LOGP(_T("Header written"))
            m_file.Write( chdr, hdr.Length() );
        }

    }

    for ( size_t i=0; i < datalen; i+=16 )
    {
        char cb[10]; sprintf( cb, "%08x: ", m_fileaddress+i);
        m_file.Write( cb, 10 );
        for ( size_t j = i; j < i+16; ++j )
        {
            char val[3] = { 32, 32, 32 };
            sprintf( val, "%02X ", *(data+j));
            m_file.Write( val, 3 );
        }

        char enter[2] = { 13, 10 };
        m_file.Write( enter, 2 );

        if ( m_fileaddress + i+16 >= 4096 )
        {
            m_fileaddress+=datalen;
            return;
        }
    }
    m_fileaddress+=datalen;
}

void EepromReaderClient::sendEvent(size_t WXUNUSED(size))
{
    m_membuffer->SetDataLen( 4096 );
    if ( m_eventHandler )
    {
        EepromRawDataEvent event;

        event.SetDataBuffer( *m_membuffer );

        m_eventHandler->AddPendingEvent( event );
    }
}

void EepromReaderClient::addToBuffer( uint8_t* data, size_t datalen )
{
    m_membuffer->AppendData( data, datalen );

    if ( m_eventHandler )
    {
        GeneralEvent evt;

        evt.SetVoidPtr( this );

        evt.SetDouble( m_membuffer->GetDataLen() / 4096.0 * 100 );

        evt.SetString( _T("EEPROM_DL_PROGRESS") );

        m_eventHandler->AddPendingEvent( evt );
    }
}

void EepromReaderClient::sendNoValidData()
{
    if ( m_writeToFile )
        m_file.Close();

    // not valid data
    m_membuffer->SetDataLen( 0 );

    if ( m_eventHandler )
    {
        EepromRawDataEvent event;

        event.SetDataBuffer( *m_membuffer );

        m_eventHandler->AddPendingEvent( event );
    }

    m_state = FINISHED;
}
