#include "FirmwareUploader.h"
#include "FirmwareUploaderListener.h"
#include "TimeoutListener.h"
#include "../../common/TxData.h"
#include "../../logger.h"
#include "../../Math.h"
#include "../../vemsprotocol/FirmwareData.h"
#include "../../events/FirmwareUploadEvent.h"


FirmwareUploader::FirmwareUploader( FirmwareUploaderListener* uploaderListener,
                                    TimeoutListener* timeoutListener,
                                    FirmwareData* fwData )
{
    m_timeoutListener   = timeoutListener;
    m_uploaderListener  = uploaderListener;
    m_firmwareData      = fwData;

    if ( m_firmwareData->getType() == FirmwareData::HEX ) {
        m_state = ERASE_ALL_PAGES; //V3
    } else {
        m_state = SEND_PAGE_ADDRESS; //ROUND
    }


    m_currentPage = 0;
    m_erasedTryCount = 0;

    m_currentPageCommandSent = false;
    m_currentPageBlockIdx = 0;

    m_clearCount    = 0;
    m_clearCountMax = 0;
}

FirmwareUploader::~FirmwareUploader()
{
}

void FirmwareUploader::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case ERASE_ALL_PAGES:
            eraseAllPages( buffer, size );
            break;

        case SEND_LEAVE_BOOT_MODE_COMMAND:
            sendLeaveBootModeCommand( buffer, size );
            break;

        case SEND_PAGE_ADDRESS:
            sendPageAddress( buffer, size );
            break;

        case SEND_PAGE:
            sendPage( buffer, size );
            break;

        default:
            break;
    }
}

void FirmwareUploader::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case WAIT_ALL_PAGE_ERASED:
            waitAllPageErased( buffer, size );
            break;

        case WAIT_AUTO_ENTER_BOOT_MODE:
            waitAutoEnterBootMode( buffer, size );
            break;

        case CLEARING:
            clearing( buffer, size );
            break;

        case WAIT_PAGE_ADDRESS_OK:
            waitPageAddressOk( buffer, size );
            break;

        case WAIT_SEND_PAGE_OK:
            waitSendPageOk( buffer, size );
            break;

        default:
            break;
    }
}

void FirmwareUploader::finish()
{
    if  ( m_state == TIMEDOUT ) {
        LOG
        m_timeoutListener->OnTimeout();
    }
}

bool FirmwareUploader::isFinished()
{
    return ( ( m_state == FINISHED ) || ( m_state == TIMEDOUT ) );
}

/**
int BootModeV3::eraseFlash()
{
    //Logger::logf(LOG_MSG, _T("FirmwareMgr::eraseFlash() START"));
    int MAX_TRY=30;
    observerInit(_T("Erasing all pages"), _T(""), MAX_TRY+1);
    observerNotify(_T("Erasing all pages...(be patient)") );
    //Logger::logf(LOG_MSG, _T("FirmwareMgr::eraseFlash() START 2"));
    m_commBase->clear();
    m_commBase->SendString((uint8_t*)"e");



    int tryCount =0;
    do{
        wxThread::Sleep(100); // 100 mS
        int result1 = m_commBase->ReadString(m_buff, 10);
        if (result1 != 1 || m_buff[0] != '\r')
        {
            Logger::logf(LOG_MSG, _T("BootModeV3::eraseFlash() waiting for erase completed size=%i"), result1);
            //throw PageException(PAGE_ERASE_FAILED, 0);
            tryCount++;
        }else{
            tryCount = MAX_TRY+1;
        }
        observerNotify(tryCount, tryCount);
    }
    while(tryCount < MAX_TRY);
    observerNotify(MAX_TRY, MAX_TRY);
    if(tryCount == MAX_TRY){
        observerNotify(_T("Erase FAILED!"));
        return -1;
    }
    observerNotify(_T("Erase OK!"));
    return 0;
}
*/
void FirmwareUploader::eraseAllPages( uint8_t* buffer, int& size )
{
    sendFirmwareUploadEvent( FirmwareUploadEvent::ERASING_PAGES_STARTED );

    m_timeoutCount = 0;
    m_erasedTryCount = 0;

    TxData command = TxData( _T("e") );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = WAIT_ALL_PAGE_ERASED;
}

void FirmwareUploader::waitAllPageErased( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {
        if( buffer[0] == '\r' ) {
            sendFirmwareUploadEvent( FirmwareUploadEvent::ERASING_PAGES_SUCCEEDED );

            if ( m_firmwareData->getPages() >= MAX_PAGE_NUMBER ) {
                m_state = SEND_LEAVE_BOOT_MODE_COMMAND;
            } else {
                m_state = SEND_PAGE_ADDRESS;
            }
        } else {
            m_state = TIMEDOUT;
        }
    }
    else
    {
        wxThread::Sleep( ERASED_SLEEP_TIME );

        if ( ++m_erasedTryCount >= ERASED_MAX_TRY_COUNT ) {
            sendFirmwareUploadEvent( FirmwareUploadEvent::ERASING_PAGES_FAILED );
            m_state = TIMEDOUT;
        } else {
            sendFirmwareUploadEvent( FirmwareUploadEvent::ERASING_PAGES );
        }
    }
}

/**
    /// Leave bootmode and write page 503 (due to erase)
    /// auto enter bootMode (firmware erased) and without erase command page 503 wont be destrolyed by bootloader
    size_t pagesNum = m_firmwareData->getPages();
    if(pagesNum >= 503){
        m_commBase->SendString((uint8_t*)"E");
        /// needs some wait??
        /// just for sure wait 20 ms
    }
*/
void FirmwareUploader::sendLeaveBootModeCommand( uint8_t* buffer, int& size )
{
    TxData command = TxData( _T("E") );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = WAIT_AUTO_ENTER_BOOT_MODE;
}

void FirmwareUploader::waitAutoEnterBootMode( uint8_t* buffer, int size )
{
    LOG
    wxThread::Sleep( AUTO_ENTER_BOOT_MODE_SLEEP_TIME );
    LOG
    m_clearCount = 0;
    m_clearCountMax = 64;
    m_state = CLEARING;
}

 void FirmwareUploader::clearing( uint8_t* buffer, int size )
 {
    if ( ++m_clearCount >= m_clearCountMax ) {
        m_state = SEND_PAGE_ADDRESS;
    }
 }

void FirmwareUploader::sendPageAddress( uint8_t* buffer, int& size )
{
    if ( m_currentPage == 0 ) {
        LOGP( _T("UPLOADING_PAGES_STARTED") );
        sendFirmwareUploadEvent( FirmwareUploadEvent::UPLOADING_PAGES_STARTED );
        m_interWriteDelay = 0;
    }

    m_timeoutCount = 0;

    uint32_t pageIdx = m_firmwareData->getPageIndexbyOrderIdx(m_currentPage);
    uint32_t addressW = m_firmwareData->getPageAddressW(pageIdx);

    //skip V3_MAX_PAGE_IDX
    if ( ( m_firmwareData->getType() == FirmwareData::HEX ) && ( pageIdx > V3_MAX_PAGE_IDX ) ) {
        LOGP( wxString::Format( _T("Page write error in page: %i. (Page write to idx > %i is not allowed!)"), pageIdx, V3_MAX_PAGE_IDX) );
        if ( ++m_currentPage >= m_firmwareData->getPages() ) {
            sendFirmwareUploadEvent( FirmwareUploadEvent::UPLOADING_PAGES_SUCCEEDED );
            m_state = FINISHED;
        }
        return;
    }


    /// addresW is actually 2 bytes

    ///TODO - handle command size > 1

    uint8_t str[3];
    str[0] = 'A';
    str[1] = (addressW >> 8) & 0xFF;
    str[2] = addressW & 0xFF;

    TxData command = TxData( str, 3 );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = WAIT_PAGE_ADDRESS_OK;
}

void FirmwareUploader::waitPageAddressOk( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {

        if( buffer[0] == '\r' ) {
//            LOGP( _T("SEND_PAGE;") );
            m_state = SEND_PAGE;
        } else {
            m_state = TIMEDOUT;
        }
    }
    else
    {
        ++m_timeoutCount;
        wxThread::Sleep( m_sleep );

        if ( m_timeoutCount >= m_timeout ) {
            LOGP( _T("TIMEDOUT") );
            m_state = TIMEDOUT;
        }
    }
}

/**
int BootModeBase::uploadPageCommon(const char* cmd, uint32_t pageIdx)
{

    m_commBase->SendString((uint8_t*)cmd);
    bool found;
    int result;
    uint8_t *buf = m_firmwareData->getPageData(pageIdx, found);
    uint32_t pageSize = m_firmwareData->getPageSize();
    result = m_commBase->sendStringBlock(buf, pageSize, 1);
    if (result != (int)pageSize)
    {
        throw PageException(PAGE_WRITE_ERROR, pageIdx);
    }

    Math::getBytes(m_firmwareData->getPageCheckSum(pageIdx), m_buff);

    result = m_commBase->SendString(m_buff, 4); // send checksum
    if(result != 4){
        Logger::logf(LOG_MSG, _T("BootModeBase::uploadPageCommon() send checksum byte: Failed") );
    }
#ifdef SFILE_LIB
	wxMilliSleep((1+10)*10*1000/(m_commBase->getBaudRate()));
#endif
    result = m_commBase->ReadString(m_buff,1, 64); // we except 1 byte in fact
    if(result == 1){
        if( m_buff[0] == '\r' ){ // OK
            //Logger::logf(LOG_MSG, _T("Firmware::uploadPage() after checksum read in SYNC OK!") );
        }else{ // typically 'N' for bad checksum or nothing for serious out of sync error
            Logger::logf(LOG_MSG, _T("BootModeBase::uploadPageCommon() page write checksum failed! Received Byte: %c"), m_buff[0] );
            throw PageException(PAGE_WRITE_CHECKSUM_ERROR, pageIdx);
        }
    }else{
        Logger::logf(LOG_MSG, _T("BootModeBase::uploadPageCommon failed, received bytecount: %d"), result );
         throw PageException(PAGE_WRITE_SYNC_ERROR, pageIdx);
    }

    return PAGE_OK;
}
*/

/**
int SerialPortHandlerCTB::sendStringBlock(uint8_t* value, size_t len, size_t blockLen)
{
    //char * p = (char *) value;
    int result = 0;
    size_t i = 0;
    size_t currBlock;
    for(i=0; i< len; i += blockLen){
        currBlock = len - i;
        if( currBlock > blockLen ){
            currBlock = blockLen;
        }
        result += SendString( (&(value[i])), currBlock);
    }
    return result;
}
*/

void FirmwareUploader::sendPage( uint8_t* buffer, int& size )
{
    if ( !m_currentPageCommandSent ) {
        TxData command;

        if ( m_firmwareData->getType() == FirmwareData::HEX ) {
            command = TxData( _T("Dkk") ); //V3
        } else {
            command = TxData( _T("Dcc") ); //ROUND
        }

        size = command.size;
        memcpy( buffer, command.data, command.size );
        m_currentPageCommandSent = true;
    } else {
        uint32_t pageSize = m_firmwareData->getPageSize();

        if ( m_currentPageBlockIdx < pageSize ) {
            uint32_t pageIdx = m_firmwareData->getPageIndexbyOrderIdx(m_currentPage);
            bool found;
            uint8_t* pageBuffer = m_firmwareData->getPageData(pageIdx, found);

            size = 1;
            memcpy( buffer, &pageBuffer[m_currentPageBlockIdx], 1 );

            ++m_currentPageBlockIdx;
        } else {
            uint32_t pageIdx = m_firmwareData->getPageIndexbyOrderIdx(m_currentPage);
            uint32_t filePageCheckSum = m_firmwareData->getPageCheckSum(pageIdx);
            uint8_t checksumBuffer[4];
            Math::getBytes( filePageCheckSum, checksumBuffer );
            size = 4;
            memcpy( buffer, checksumBuffer, 4 );

            m_timeoutCount = 0;
            m_state = WAIT_SEND_PAGE_OK;
        }
    }
}

void FirmwareUploader::waitSendPageOk( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {
        if( buffer[0] == '\r' ) {
            sendFirmwareUploadEvent( FirmwareUploadEvent::PAGE_UPLOADED );

            if ( ++m_currentPage < m_firmwareData->getPages() ) {
                //++m_currentPage;
                m_currentPageCommandSent = false;
                m_currentPageBlockIdx = 0;
                m_state = SEND_PAGE_ADDRESS;
            } else {
                sendFirmwareUploadEvent( FirmwareUploadEvent::UPLOADING_PAGES_SUCCEEDED );
                m_state = FINISHED;
            }
        } else {
            LOGP( _T("TIMEDOUT") );
            m_state = TIMEDOUT;
        }
    }
    else
    {
        ++m_timeoutCount;
        wxThread::Sleep( m_sleep );

        if ( m_timeoutCount >= m_timeout + 100 ) {
            LOGP( _T("TIMEDOUT") );
            m_state = TIMEDOUT;
        }
    }
}

void FirmwareUploader::sendFirmwareUploadEvent( int eventType )
{
    FirmwareUploadEvent event( eventType );
    event.setFirmwarePageCount( m_firmwareData->getPages() );
    event.setFirmwarePageIdx( m_currentPage );
    m_uploaderListener->OnFirmwareUploaderEvent( event );
}
