#include "MegatuneSyncer.h"
#include "DeviceDetector.h"
#include "MegatuneSyncerListener.h"
#include "TimeoutListener.h"
#include "../../vemsprotocol/SIPRDescriptor.h"
#include "../../vemsprotocol/SiprDescHelper.h"
#include "../../vemsprotocol/SIPRPacketHelper.h"
#include "../../vemsprotocol/MegatuneConfig.h"
#include "../../common/types/Int2IntHash.h"
#include "../../common/TxData.h"
#include "../../Math.h"
#include "../../logger.h"

/// use for faster testing
//#define NO_SYNC_PAGES


MegatuneSyncer::MegatuneSyncer( MegatuneSyncerListener* listener,
                                TimeoutListener* timeoutListener,
                                SIPRDescriptor* descriptor,
                                MegatuneConfig* config )
    : RT_LEN_TRY_COUNT_MAX (3)
    , CRC_CHECK_TRY_COUNT_MAX (3)
{
    m_listener          = listener;
    m_timeoutListener   = timeoutListener;
    m_descriptor        = descriptor;
    m_config            = config;

    WX_CLEAR_LIST( SIPRPacketList, m_config->pages );
    m_config->detectedSiprTableCount = 0;

    Int2IntHash& pageSize = m_config->pageSize;
    pageSize.clear();

    m_state             = REQUEST_RT;
    m_isFirstRequest    = true;
    m_rxDataLength      = 0;
    m_timeoutCount      = 0;
    m_clearCount        = 0;
    m_realtimeLength    = 0;
    m_expectedRxSize    = 0;

    m_currentPageIdx    = 0;
    m_currentPageSize   = 0;

    m_detectCount       = 0;
    m_maxDetectCount    = 3;

    m_client            = NULL;

    m_device            = ECUInfo::DEVICE_UNKNOWN;
    m_deviceMode        = ECUInfo::UNKNOWN;

    m_rtLenTryCount     = 0;
    m_crcCheckTryCount  = 0;
}

MegatuneSyncer::~MegatuneSyncer()
{

}

void MegatuneSyncer::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case REQUEST_RT:
            requestRt( buffer, size );
            break;

        case REQUEST_PAGE:
            requestPage( buffer, size );
            break;

        case CHECK_DEVICE:
            checkDevice( buffer, size );

        default:
            break;
    }
}

void MegatuneSyncer::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case WAIT_RT:
            waitRt( buffer, size );
            break;

        case WAIT_PAGE:
            waitPage( buffer, size );
            break;

        case WAIT_CHECK_DEVICE:
            waitCheckDevice( buffer, size );
            break;

        case CLEAR:
            clear( buffer, size );
            break;

        default:
            break;
    }
}

void MegatuneSyncer::finish()
{
    if ( m_state == TIMEDOUT ) {
        m_timeoutListener->OnTimeout();
    } else {
        m_listener->OnFirmwareDataSynced();
    }
}


bool MegatuneSyncer::isFinished()
{
    return m_state == FINISHED || m_state == TIMEDOUT;
}

void MegatuneSyncer::requestRt( uint8_t* buffer, int& size )
{
    m_expectedRxSize = m_isFirstRequest ? 256 : m_realtimeLength;

    m_rxDataLength = 0;
    m_timeoutCount = 0;

    TxData command( m_config->readRealtimeCommand.To8BitData() );
    size = command.size;

    memcpy( buffer, command.data, command.size );

    m_state = WAIT_RT;
//    LOGP( _T("COMMAND SENT - ") +  m_realtimeCommand );
}

void MegatuneSyncer::waitRt( uint8_t* buffer, int size )
{
    switch ( m_config->rtCheckType )
    {
        case MegatuneConfig::TRIGGER:
            waitRtTriggerFrame( buffer, size );
            break;

        case MegatuneConfig::CRC:
            waitRtCrc( buffer, size );
            break;

        case MegatuneConfig::LENGTH:
            waitRtLength( buffer, size );
            break;
    }
}

void MegatuneSyncer::waitRtTriggerFrame( uint8_t* buffer, int size )
{
    if ( size > 0 ) {
        if( m_triggerFrameFSM.parse( buffer[0] ) ){
            const IntArray& trFrame = m_triggerFrameFSM.getFrame();
            // TODO: check type
            //if(m_triggerFrameFSM.type != 32){

            //}
            m_config->realtimeLength = trFrame.GetCount() - 3;;
            m_state = REQUEST_PAGE;
        }
    }

    if (  ++m_timeoutCount < m_timeout ) {
        wxThread::Sleep( m_sleep );
        return;
    }

    m_state = TIMEDOUT;
}

void MegatuneSyncer::waitRtCrc( uint8_t* buffer, int size )
{
    if ( size > 0 ) {
        if ( ( m_rxDataLength >= MS_RX_BUFFER_SIZE ) || ( m_rxDataLength >= m_config->maxRealtimeLength ) ) {
            m_rxDataLength = 0;
        }
        m_rxBuffer[m_rxDataLength] = *buffer;
        ++m_rxDataLength;
        return;

    }

    if (  ++m_timeoutCount < m_timeout ) {
        wxThread::Sleep( m_sleep );
        return;
    }

    if ( m_rxDataLength < 2 ) {
        m_state = TIMEDOUT;
    } else {
        m_realtimeLength = m_rxDataLength;

        if ( checkRtCrc() ) {
            m_config->realtimeLength = m_rxDataLength;
            m_state = REQUEST_PAGE;
        } else if ( m_crcCheckTryCount < CRC_CHECK_TRY_COUNT_MAX ) {
            ++m_crcCheckTryCount;
            m_state = REQUEST_RT;
        } else {
            m_state = TIMEDOUT;
        }

    }
}

void MegatuneSyncer::waitRtLength( uint8_t* buffer, int size )
{
    if ( m_isFirstRequest ) {
        waitRtLengthFirst( buffer, size );
    } else {
        waitRtLengthSecond( buffer, size );
    }
}

void MegatuneSyncer::waitRtLengthFirst( uint8_t* buffer, int size )
{
    if ( size > 0 ) {
        if ( ( m_rxDataLength >= MS_RX_BUFFER_SIZE ) || ( m_rxDataLength >= m_config->maxRealtimeLength ) ) {
            m_rxDataLength = 0;
        }
        m_rxBuffer[m_rxDataLength] = *buffer;
        ++m_rxDataLength;
        return;

    }

    if (  ++m_timeoutCount < m_timeout ) {
        wxThread::Sleep( m_sleep );
        return;
    }

    if ( m_rxDataLength == 0 ) {
        if ( m_rtLenTryCount < RT_LEN_TRY_COUNT_MAX ) {
            ++m_rtLenTryCount;
            m_state = REQUEST_RT;
        } else {

            if ( m_config->device == ECUInfo::DEVICE_V3 ) {
                if ( m_config->rtType == MegatuneConfig::NORMAL ) {
                    m_state = TIMEDOUT;
                } else {
                    m_config->rtType = MegatuneConfig::NORMAL;
                    m_config->readRealtimeCommand = wxString::From8BitData("A");
                    m_rtLenTryCount = 0;
                    m_state = REQUEST_RT;
                }
            } else {
                m_state = TIMEDOUT;
            }
        }
    } else if ( m_rxDataLength < m_config->minRealtimeLength ) {
        if ( m_rtLenTryCount < RT_LEN_TRY_COUNT_MAX ) {
            ++m_rtLenTryCount;
            m_state = REQUEST_RT;
        } else {
            m_state = TIMEDOUT;
        }
    } else {
        m_realtimeLength = m_rxDataLength;
        m_isFirstRequest = false;
        m_state = REQUEST_RT;
    }
}

void MegatuneSyncer::waitRtLengthSecond( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {
        if ( ++m_rxDataLength == m_realtimeLength )
        {
            m_isFirstRequest = true;
//            m_config->readRealtimeCommand = m_realtimeCommand;
            m_config->realtimeLength = m_realtimeLength;
#ifdef NO_SYNC_PAGES
            m_state = FINISHED;
#else
            m_state = REQUEST_PAGE;
#endif
        }
        return;
    }

    if ( ++m_timeoutCount < m_timeout ) {
        wxThread::Sleep( m_sleep );
        return;
    }

    ///TODO - add try count
    LOGP( wxString::Format( _T("TIMEOUT - m_rxDataLength: %d"), m_rxDataLength ) );
    m_isFirstRequest = true;
    m_state = TIMEDOUT;
}

void MegatuneSyncer::requestPage( uint8_t* buffer, int& size )
{
    ///TEST ONLY
    if ( m_currentPageIdx < m_descriptor->getPageCount() )
    {
        const int MAX_PAGE_SIZE = m_config->pageCheckType == MegatuneConfig::CRC ? 257 : 255;
        m_expectedRxSize = m_isFirstRequest ? MAX_PAGE_SIZE : m_currentPageSize;
        m_rxDataLength = 0;
        m_timeoutCount = 0;
        m_page = m_descriptor->getPageNumber( m_currentPageIdx );

        if ( m_page == -1 ) {
            LOGP( wxString::Format( _T("PAGE NOT FOUND - page idx: %d "), m_currentPageIdx ) );
            ++m_currentPageIdx;
        } else {
            wxString readCommand = m_descriptor->getPageReadCommand( m_page );

            TxData* command = SiprDescHelper::getPageReadCommand( readCommand, m_config->device );

            if ( !command ) {
                LOGP( _T("INVALID COMMAND - ") + readCommand );
                ++m_currentPageIdx;
                return;
            }

//            TxData command( readCommand );
            size = command->size;

            memcpy( buffer, command->data, command->size );

            delete command;

            m_detectedLength = 0;
            m_detectCount = 0;

            m_timeoutTimer.start( m_timeout + TIMEOUT );

            m_state = WAIT_PAGE;
        }
    } else {
        m_state = FINISHED;
    }
}

void MegatuneSyncer::waitPage( uint8_t* buffer, int size )
{
    switch ( m_config->pageCheckType )
    {
        case MegatuneConfig::CRC:
            waitPageCrc( buffer, size );
            break;

        default:
            waitPageLength( buffer, size );
            break;
    }
}


void MegatuneSyncer::waitPageCrc( uint8_t* buffer, int size )
{
    if ( size > 0 && m_rxDataLength < MS_RX_BUFFER_SIZE ) {
        m_rxBuffer[ m_rxDataLength++ ] = buffer[0];
        return;
    }

    if ( !m_timeoutTimer.isTimedout() ) {
            wxThread::Sleep( m_sleep );
            return;
    }

    if ( checkCrc( m_rxBuffer, m_rxDataLength ) ) {
                sendPagePacket();
                m_config->pageSize[m_page] = m_rxDataLength;
                ++m_currentPageIdx;
                m_state = REQUEST_PAGE;
    } else {
        m_state = CHECK_DEVICE;
    }
}

void MegatuneSyncer::waitPageLength( uint8_t* buffer, int size )
{
    if ( m_isFirstRequest ) {
        waitPageFirst( buffer, size );
    } else {
        waitPageSecond( buffer, size );
    }
}


void MegatuneSyncer::waitPageFirst( uint8_t* buffer, int size )
{
    if ( size > 0 ) {
        m_rxDataLength++;
        return;
    }

    if ( !m_timeoutTimer.isTimedout() ) {
            wxThread::Sleep( m_sleep );
            return;
    }

    if ( m_rxDataLength > 0 ) {
        int pageSize = m_descriptor->getPageSize( m_page );
        m_currentPageSize   = m_rxDataLength;
        m_isFirstRequest    = false;
        m_state = REQUEST_PAGE;
    } else {
        m_state = CHECK_DEVICE;
//        m_state = TIMEDOUT;
    }
}

void MegatuneSyncer::waitPageSecond( uint8_t* buffer, int size )
{
    if ( size > 0 ) {
        m_rxBuffer[ m_rxDataLength++ ] = buffer[0];

        if ( m_rxDataLength == m_currentPageSize )
        {
            LOGP( wxString::Format( _T("PAGE SIZE VERIFIED! - page: %d, size: %d"), m_page, m_currentPageSize ) );
            m_descriptor->getPageReadCommand( m_page );
            sendPagePacket();
            m_config->pageSize[m_page] = m_currentPageSize;

            m_isFirstRequest = true;
            ++m_currentPageIdx;
            m_state = REQUEST_PAGE;
        }
        return;
    }

    if ( ++m_timeoutCount < ( m_timeout + TIMEOUT ) ) {
        wxThread::Sleep( m_sleep );
        return;
    }

    m_isFirstRequest = true;
    m_state = TIMEDOUT;
}

void MegatuneSyncer::checkDevice( uint8_t* buffer, int& size  )
{
    m_client = new DeviceDetector( this );

    m_client->setTimeout( getTimeout() );
    m_client->setInterWriteDelay( getInterWriteDelay() );
    m_client->write( buffer, size );
}

void MegatuneSyncer::waitCheckDevice( uint8_t* buffer, int size )
{
    m_client->read( buffer, size );

    if ( m_client->isFinished () ) {
        m_client->finish();
    }
}

void MegatuneSyncer::clear( uint8_t* buffer, int size )
{
    if ( ++m_clearCount > MAX_CLEAR_COUNT ) {
        m_state = REQUEST_RT;
    } else {
        wxThread::Sleep( m_sleep );
    }
}

void MegatuneSyncer::sendPagePacket()
{
    SIPRPacketList packets = SIPRPacketHelper::createSIPRPackets( *m_descriptor, m_rxBuffer, m_page );

    for ( SIPRPacketList::iterator it = packets.begin(); it != packets.end(); ++it ){
        SIPRPacket* packet = *it;
        m_config->pages.Append( packet );
        ++m_config->detectedSiprTableCount;
    }
}

//void MegatuneSyncer::handleEvent( CommunicationClientEvent::Type eventType )
//{
//    switch ( eventType )
//    {
//        case CommunicationClientEvent::DEVICE_FOUND:
//        {
//            LOG
////            if ( m_detectedMode == m_targetMode ) {
////                LOGP( _T("DEVICE_FOUND") );
////                m_clearCount = 0;
////                m_clearCountMax = 20;
////                m_state = CLEAR;
////            } else {
////                ///TODO
////                LOGP( _T("NO TARGET MODE SELECTED") );
////                m_state = TIMEDOUT;
////            }
//        } break;
//
//        case CommunicationClientEvent::DEVICE_NOT_FOUND:
//        {
//            LOGP( _T("DEVICE_NOT_FOUND") );
//            m_state = TIMEDOUT;
//        } break;
//
//        default:
//            LOGP( wxString::Format( _T("INVALID EVENT! ( %d )"), eventType ) );
//            break;
//    }
//}

void MegatuneSyncer::OnDeviceFound( ECUInfo::Device device, ECUInfo::EcuMode mode )
{
//    if ( mode == m_targetMode ) {
////                LOGP( _T("DEVICE_FOUND") );
////        m_clearCount = 0;
////        m_clearCountMax = 20;
////        m_state = CLEAR;
//    } else {
//        m_state = TIMEDOUT;
//    }
}

void MegatuneSyncer::OnDeviceNotFound()
{
    /// TODO
    m_state = TIMEDOUT;
}

bool MegatuneSyncer::checkRtCrc()
{
    if ( ( m_realtimeLength < 2 ) || ( m_realtimeLength > MS_RX_BUFFER_SIZE ) ) {
        LOGP( wxString::Format( _T("ERROR - m_realtimeLength: %d is invalid!"), m_realtimeLength ) );
        return false;
    }

    return checkCrc( m_rxBuffer, m_realtimeLength );

//    if ( ( m_realtimeLength < 2 ) || ( m_realtimeLength > MS_RX_BUFFER_SIZE ) ) {
//        LOGP( wxString::Format( _T("ERROR - m_realtimeLength: %d is invalid!"), m_realtimeLength ) );
//        return false;
//    }
//
//    uint16_t crc = 0xFFFF;
//
//    for ( size_t i = 0; i < ( (size_t)m_realtimeLength - 2 ); ++i ) {
//        crc = Math::crc16_sipr( crc, m_rxBuffer[i] );
//    }
//
//    uint16_t rxCrc = 0;
//    rxCrc = m_rxBuffer[m_realtimeLength - 2];
//    rxCrc |= ((uint16_t)m_rxBuffer[m_realtimeLength - 1]<<8);
//
//    return ( crc == rxCrc );
}

///TODO - set correct fw version
bool MegatuneSyncer::isTriggerFrame()
{
    FirmwareVersion fw(1,1,86);
    return (m_config->fwVersion >= fw);
}


bool MegatuneSyncer::checkCrc( uint8_t* buffer, size_t length )
{
    if ( length < 2 ) return false;

    uint16_t crc = 0xFFFF;

    for ( size_t i = 0; i < ( length - 2 ); ++i ) {
        crc = Math::crc16_sipr( crc, buffer[i] );
    }

    uint16_t rxCrc = buffer[length - 2];
    rxCrc |= ((uint16_t)buffer[length - 1]<<8);

    return ( rxCrc == crc );
}
