#include "OffsetReaderClient.h"

#include <wx/textfile.h>

#include "DeviceModeSelector.h"
#include "../../DirHelper.h"
#include "../../logger.h"
#include "../../Math.h"
#include "../../common/TxData.h"
#include "../../events/OffsetReaderEvent.h"

#include "../../vemsprotocol/SIPRTypes.h"
#include "../../vemsprotocol/SIPRPacket.h"
#include "../../events/InputEvent.h"

/**
    OffsetReaderClient constructor - csvwriter mode

    Params:
        deviceMode  : original device mode we need to switch back to
        offsets     : offsets we want to dump

    Tasks:
        initializaton
        create readbuffer
        set initial state

    Return value:
        none
**/
OffsetReaderClient::OffsetReaderClient( CommunicationBase* commBase, const IntArray& baudrates,
    ECUInfo::EcuMode deviceMode, wxArrayString& offsets, bool hasCRC )
{
    m_commBase = commBase;
    m_baudrates = baudrates;

    m_evtHandler = NULL;

    m_crc = hasCRC;

    // original devicemode
    m_deviceMode = deviceMode;

    // offsets we need dump
    m_offsets = offsets;

    // write result to csv file
    m_mode = CSV_Writer;

    // 1 mdx command -> 32 / 34 byte result

    if ( m_crc )
        m_maxBufLen = offsets.GetCount() * 34;
    else
    m_maxBufLen = offsets.GetCount() * 32;

    if ( m_crc )
        m_maxBufLen = offsets.GetCount() * 34;
    else
        m_maxBufLen = offsets.GetCount() * 32;

    // create buffer
    m_buffer = new unsigned char[ m_maxBufLen + 2 ];
    m_bufLen = 0;

    // actual offset counter
    m_offsetCounter = 0;

    m_csvfile = NULL;

    // if already in man mode -> send mdx command
    if ( m_deviceMode == ECUInfo::MAN )
    {
        m_state = SEND_COMMAND;
    }
    // else switch to man mode
    else
    {
        m_state = GO_MAN;
        m_client = new DeviceModeSelector( this, this, m_commBase, m_baudrates, m_deviceMode, ECUInfo::MAN );
    }
}

OffsetReaderClient::OffsetReaderClient( ECUInfo::EcuMode deviceMode, wxArrayString& addresses, wxEvtHandler* evtHandler, bool hasCRC )
{
//    for ( unsigned int j = 0; j < addresses.GetCount(); ++j )
//        LOGP( addresses[j] )
    m_evtHandler = evtHandler;

    m_crc = hasCRC;

    // original devicemode
    m_deviceMode = deviceMode;

    // offsets we need dump
    m_offsets = addresses;

    m_mode = SIPRPacket_Creator;

    // 1 mdx command -> 32/34 byte result

    if ( m_crc )
        m_maxBufLen = 34;
    else
    m_maxBufLen = 32;

    if ( m_crc )
        m_maxBufLen = 34;
    else
        m_maxBufLen = 32;

    // create buffer
    m_buffer = new unsigned char[ m_maxBufLen + 2 ];
    m_bufLen = 0;

    // actual offset counter
    m_offsetCounter = 0;

    m_csvfile = NULL;

    // if already in man mode -> send mdx command
    if ( m_deviceMode == ECUInfo::MAN )
    {
        m_state = SEND_COMMAND;
    }
    // else switch to man mode
    else
    {
        m_state = GO_MAN;
        m_client = new DeviceModeSelector( this, this, m_commBase, m_baudrates, m_deviceMode, ECUInfo::MAN );
    }
}

/**
    OffsetReaderClient constructor - sendevent mode

    Params:
        deviceMode  : original device mode we need to switch back to
        variable    : variable names & offsets we want to dump
        evtHandler  : send events to

    Tasks:
        initializaton
        create readbuffer
        set initial state

    Return value:
        none
**/
OffsetReaderClient::OffsetReaderClient( ECUInfo::EcuMode deviceMode, StringHash& variable, wxEvtHandler* evtHandler, bool hasCRC )
{
    m_evtHandler = evtHandler;

    m_crc = hasCRC;

    // previous device mode
    m_deviceMode = deviceMode;

    // variables (varname => offset)
    m_variable = variable;

    // send offsetreaderevent
    m_mode = Evt_Sender;

    m_csvfile = NULL;

    // only the first variable
    StringHash::iterator it = variable.begin();

    // offset: 2 + 1 chars
    wxString addr = it->second.Left(2);
    wxString off = it->second.Right(1);

    m_offsets.Add( addr );
    m_varOffset = Math::hex2int( off ); // 0-15

    // 32/34 (+extra) bytes buffer
    if ( m_crc )
        m_maxBufLen = m_offsets.GetCount() * 34;
    else
    m_maxBufLen = m_offsets.GetCount() * 32;


    m_buffer = new unsigned char[ m_maxBufLen + 2 ];
    m_bufLen = 0;

    // offset counter
    m_offsetCounter = 0;

    // if in man mode -> send command
    if ( m_deviceMode == ECUInfo::MAN )
    {
        m_state = SEND_COMMAND;
    }
    // or switch to man
    else
    {
        m_state = GO_MAN;
        m_client = new DeviceModeSelector( this, this, m_commBase, m_baudrates, m_deviceMode, ECUInfo::MAN );
    }
}

/**
    OffsetReaderClient - destructor

    Params:
        none
    Tasks:
        clean up
    Return value:
        none
**/
OffsetReaderClient::~OffsetReaderClient()
{
    delete[] m_buffer;

    if ( m_csvfile )
        delete m_csvfile;
}

/**
    write() - CommunicationControllerClient virtual method

    Params:
        buffer  : command to send to commport (output)
        size    : size of the command

    Tasks:
        create command according to the actual state

    Return value:
        none
**/
void OffsetReaderClient::write( uint8_t* buffer, int& size )
{
    m_timeout = TIME_OUT;

    /// TODO
    switch ( m_state )
    {
        case GO_MAN:
            goMan( buffer, size );
            break;

        case SEND_COMMAND:
            sendCommand( buffer, size );
            break;

        case WRITETO_CSV:
            writeToCsv();
            break;

        case SEND_EVENT:
            sendEvent();
            break;

        case CREATE_SIPR:
            createSipr();
            break;

        case ABORTED:

            if ( m_mode == CSV_Writer )
                closeFile();
            else if ( m_mode == Evt_Sender )
                m_state = RESET_DEVMODE;
            break;

        case RESET_DEVMODE:
            delete m_client;
            // if it was MAN => nothing to do
            if ( m_deviceMode == ECUInfo::MAN )
                m_state = FINISHED;
            else
            {
                // select the appropriate mode
                m_client = new DeviceModeSelector( this, this, m_commBase, m_baudrates, ECUInfo::MAN, m_deviceMode );
                resetDevMode( buffer, size );
            }
            break;

        default:
            break;
    }
}


/**
    read()  - CommunicationControllerClient virtual method, incoming data

    Params:
        buffer  : the incoming data
        size    : the size of the data in bytes

    Task:
        interprets the data according to the current state

    Return value:
        none
**/
void OffsetReaderClient::read( uint8_t* buffer, int size )
{
    /// TODO
    switch ( m_state )
    {
        case WAIT_MAN:
            waitMan( buffer, size );
            break;

        case GET_RESPONSE:
            getResponse( buffer, size );
            break;

        case WAIT_RESET:
            waitResetDevMode( buffer, size );
            break;

        default:
            break;
    }
}

/**
    finish() - CommunicationControllerClient virtual method, will be called if isFinished() is true

    Params:
        none

    Tasks:
        currently none

    Return value:
        none
**/
void OffsetReaderClient::finish()
{
    if ( m_state == FINISHED )
    {
        LOGP( _T("Done") )
    }
    else
    {
        LOGP( _T("TIMEOUT!!") )
    }
}

/**
    isFinished() - CommunicationControllerClient virtual method

    Params:
        none

    Tasks:
        check whether client has done its job

    Return value:
        bool: if true => thread should be stopped
**/
bool OffsetReaderClient::isFinished()
{
    return ( m_state == FINISHED || m_state == TIMEOUT );
}

/**
    Abort() - Interruptable virtual method

    Params:
        none

    Tasks:
        set state to aborted

    Return value:
        none
**/
void OffsetReaderClient::Abort()
{
    m_aborted = true;
    m_state = ABORTED;
}

/**
    handleEvent - CommunicationClientEventHandler virtual method

    Params:
        eventType   : ..

    Tasks:
        if device mode changed to MAN -> send command -> select the next state
        if device mode changed back to original mode -> finished
        timeout -> timeout, finished

    Return value:
        none
**/
void OffsetReaderClient::handleEvent( CommunicationClientEvent::Type eventType )
{
    switch ( eventType )
    {
        case CommunicationClientEvent::DEVICE_MODE_SELECTED:
        {
            switch ( m_state )
            {
                case WAIT_MAN:
                {
                    m_state = SEND_COMMAND;
                    break;
                } // waitman

                case WAIT_RESET:
                {
                    m_state = FINISHED;
                }

                default:
                    break;
            }
        }
        break;

        case CommunicationClientEvent::TIMEOUT:
            m_state = TIMEOUT;
            break;

        default:
            break;
    }
}

void OffsetReaderClient::OnDeviceModeSelected( ECUInfo::EcuMode deviceMode )
{

}

///-------------------------
/**
    goMan()

    Params:
        buffer, size passed to DeviceModeselector

    Task:
        go to Man mode, set state to WAIT_MAN

    Return value:
        none
**/
void OffsetReaderClient::goMan( uint8_t* buffer, int& size )
{
    LOGP( _T("Going to MAN mode, please wait ") )

    m_client->setTimeout( getTimeout() );
    m_client->setInterWriteDelay( getInterWriteDelay() );
    m_client->write( buffer, size );

    m_state = WAIT_MAN;
}

/**
    waitMan()

    Params:
        buffer, size filled by DeviceModeSelector client

    Tasks:
        wait for changing to man mode, see handleEvent method!

    Return value:
        none
**/
void OffsetReaderClient::waitMan( uint8_t* buffer, int size )
{
    m_client->read( buffer, size );

    if ( m_client->isFinished () ) {
        m_client->finish();
    }
}

/**
    sendCommand()

    Params:
        buffer  : mdx[#offset] command to send to commport
        size    : size of the command

    Task:
        send command to ECU

    Return value:
        none
**/
void OffsetReaderClient::sendCommand( uint8_t* buffer, int& size )
{
    wxString cmd = wxString::Format(_T("mdx") + m_offsets[m_offsetCounter] );

    //LOGP( _T("Command sent:") + cmd )

    TxData command( cmd );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    // cyclic
    if ( ++m_offsetCounter >= m_offsets.GetCount() )
    {
        m_offsetCounter = 0;
    }

    m_state = ( m_aborted ) ? ABORTED : GET_RESPONSE;
}

/**
    getResponse()

    Params:
        buffer  : incoming data
        size    : size of the data
    Tasks:
        read 32 bytes from commport and write it to file (csvwriter mode) or send and event with the values of the requested variables (sendevent mode)

    Return value:
        none
**/
void OffsetReaderClient::getResponse( uint8_t* buffer, int size )
{
    if ( size == 0 )
    {
        ++m_timeoutCtr;

        wxThread::Sleep( m_sleep );

        if ( m_timeoutCtr > m_timeout )
        {
            m_state = TIMEOUT;
            return;
        }
    }
    else m_timeoutCtr = 0;

    memcpy( m_buffer + m_bufLen, buffer, size );
    m_bufLen += size;

    if ( m_mode == SIPRPacket_Creator )
    {
        unsigned int maxlen = ( m_crc ) ? 34 : 32;
        if ( m_bufLen >= maxlen )
        {
            // do the crc check if m_crc
            if ( m_crc )
            {
                uint16_t crc = 0xffff;
                uint16_t crc_rec;

                for ( int i = 0; i < 32; ++i )
                {
                    crc = Math::crc16_sipr( crc, m_buffer[i] );
                }

                crc_rec = m_buffer[33] << 8;
                crc_rec += m_buffer[32];

                if ( crc != crc_rec )
                {
                    m_state = ABORTED;
                    LOGP( _T("crc error") );
                    return;
                }
            }
            m_state = ( m_aborted ) ? ABORTED : CREATE_SIPR;
        }
        return;
    }
    // wait for n * 32 bytes (where n is the # of offsets)
    else if ( m_bufLen >= m_maxBufLen )
    {
        // TODO: crc check if needed

        if ( m_mode == CSV_Writer )
            m_state = ( m_aborted ) ? ABORTED : WRITETO_CSV;

        if ( m_mode == Evt_Sender )
            m_state = ( m_aborted ) ? ABORTED : SEND_EVENT;

        return;
    }
    // if 32 bytes have been read in -> change read next offset
    int maxlen = ( m_crc ) ? 34 : 32;
    if ( m_bufLen % maxlen == 0 )
    {
        m_state = ( m_aborted ) ? ABORTED : SEND_COMMAND;
    }
}

/**
    writeToCsv()

    Params:
        none

    Tasks:
        write dumped values into file in csv format (with additional header)

    Return value:
        none
**/
void OffsetReaderClient::writeToCsv()
{
    if ( m_csvfile == NULL )
    {
        // create filename
        wxDateTime now = wxDateTime::Now();
        wxString fileName = DirHelper::GetPath( dLOG ) + _T("offset-") + now.Format( _T("%Y.%m.%d-%H.%M.%S") ) + _T(".csv") ;

        m_csvfile = new wxTextFile( fileName );

        // open / create file
        if ( m_csvfile->Exists() )
        {
            m_csvfile->Open();
            m_csvfile->Clear();
        }
        else
            m_csvfile->Create();

        // create header
        wxString header = _T("");
        for ( size_t i = 0; i < m_offsets.GetCount(); ++i )
        {
            unsigned int offset = Math::hex2int( m_offsets[i] ) * 16;
            for ( size_t j = 0; j < 32; ++j, offset++ )
            {
                wxString strOffset = wxString::Format( _T("\"%03X\";"), offset );
                header += strOffset;
            }
        }
        header = header.RemoveLast();

        m_csvfile->AddLine( header );

        LOGP( _T("CSV File created as ") + fileName )
    }

    // write data
    wxString line;
    for ( size_t i = 0; i < m_maxBufLen; ++i )
    {
        unsigned int val = (unsigned int)m_buffer[i];
        line += wxString::Format( _T("%d;"), val );
        //Logger::logf(LOG_MSG, _T("buffer15=%d"), m_buffer[15]);
    }
    line = line.RemoveLast();

    m_csvfile->AddLine( line );

    // reset buffer
    m_bufLen = 0;
    memset( m_buffer, 0, m_maxBufLen );

    m_state = ( m_aborted ) ? ABORTED : SEND_COMMAND;
}

/// TODO: document this
void OffsetReaderClient::createSipr()
{
    if ( m_offsetCounter == 0 ) // create a frame_ended siprpacket
    {
        SIPRPacket packet( 0, SIPR_MDX_FRAMES_END );
        InputEvent event;
        event.setPacket( packet );
        m_evtHandler->ProcessEvent( event );
    }

    //
    uint8_t* siprbuffer = new uint8_t[32+4];

    int counter = m_offsetCounter - 1;
    if ( counter < 0 )
        counter = m_offsets.GetCount() - 1;

    long laddr = Math::hex2int( m_offsets[counter] );
    laddr *= 16;

    siprbuffer[0] = ( laddr >> 24 ) & 0xff;
    siprbuffer[1] = ( laddr >> 16 ) & 0xff;
    siprbuffer[2] = ( laddr >> 8 ) & 0xff;
    siprbuffer[3] = laddr & 0xff;

    memcpy( siprbuffer + 4, m_buffer, 32 );
    /*for (int j = 0; j<32;++j)
        siprbuffer[4+j] = m_buffer[j];
    if ( laddr==0x140 )Logger::logf(LOG_MSG, _T("addr=%x, buffer[15]=%d"), laddr, m_buffer[15]);*/

    SIPRPacket packet( 0, SIPR_READ_MEMORY_32, 32 + 4 , siprbuffer );
    InputEvent event;
    event.setPacket( packet );
    m_evtHandler->ProcessEvent( event );

    m_bufLen = 0;
    memset( m_buffer, 0, m_maxBufLen );
    m_state = ( m_aborted ) ? ABORTED : SEND_COMMAND;
}

/**
    sendEvent()

    Params:
        none

    Tasks:
        send an OffsetReaderEvent to a handler with values of the requestes variables

    Return value:
        none
**/
void OffsetReaderClient::sendEvent()
{
    OffsetReaderEvent event;

    DoubleHash var;

    StringHash::iterator it = m_variable.begin();
    wxString name = it->first;

    double value = (double)m_buffer[m_varOffset];

    var[name] = value;

    event.values = var;

    if ( m_evtHandler )
        m_evtHandler->AddPendingEvent( event );

    m_bufLen = 0;
    memset( m_buffer, 0, m_maxBufLen );

    m_state = ( m_aborted ) ? ABORTED : SEND_COMMAND;
}

/**
    closeFile()

    Params:
        none

    Tasks:
        close csv file
        set state to reset original devmode

    Return value:
        none
**/
void OffsetReaderClient::closeFile()
{
    if ( m_csvfile )
    {
        m_csvfile->Write();
        m_csvfile->Close();
    }

    m_state = RESET_DEVMODE;
}

/**
    resetDevMode()

    Params:
        buffer, size passed to DeviceModeselector

    Task:
        restore original device mode, set state to WAIT_RESET

    Return value:
        none
**/
void OffsetReaderClient::resetDevMode( uint8_t* buffer, int& size )
{
    LOGP( _T("Going back to original device mode, please wait") )
    m_client->setTimeout( getTimeout() );
    m_client->setInterWriteDelay( getInterWriteDelay() );
    m_client->write( buffer, size );

    m_state = WAIT_RESET;
}

/**
    waitResetDevMode()

    Params:
        buffer, size filled by DeviceModeSelector client

    Tasks:
        wait for changing to original device mode, see handleEvent method!

    Return value:
        none
**/
void OffsetReaderClient::waitResetDevMode( uint8_t* buffer, int size )
{
    m_client->read( buffer, size );

    if ( m_client->isFinished () )
    {
        m_client->finish();
    }
}

