#include "SDCardV3Logger.h"
#include "SDCardFileSearch.h"
#include "SDCardFileHelper.h"
#include "../../logger.h"
#include "../../common/TxData.h"
#include "../../events/SDCardEvent.h"
#include "../../common/VemsContext.h"
#include "../../common/VemsContext.h"
#include "../../FirmwareInfo.h"

/**
    SDCardV3Logger() - constructor

    Params:
        deviceMode  : current device mode ( MAN / Megatune... )
        command     : SD_START_LOG = 0, SD_STOP_LOG = 1, SD_ERASE_CARD = 2
        evthandler  : optional eventhandler

    Tasks:
        initialization
        go to man mode if neccessary / process the selected command

    Return value:
        none
**/
SDCardV3Logger::SDCardV3Logger( SDLogCommand command, wxEvtHandler * evthandler )
{
    // get fw info from vemscontext
    m_fwInfo = dynamic_cast<FirmwareInfo*>( VemsContext::getObject( _T("FirmwareInfo") ) );

    m_evthandler = evthandler;
    m_command = command;

    m_timeoutCtr = 0;

    m_buffer = NULL;
    m_buffersize = 0;

    m_checkSectTryCount = 0;

    switch ( m_command )
    {
        case SD_GET_SD_FIRMWARE:
            LOG
            m_state = SEND_GET_FWVER;
            break;

        case SD_START_LOG:
            m_state = SEND_START;
            break;
        case SD_STOP_LOG:
            m_state = SEND_STOP;
            break;
        case SD_ERASE_CARD:
            m_state = SEND_ERASE;
            break;

        default:
            m_state = FINISHED;
            break;
    }

}

/**
    SDCardV3Logger() Destructor

    Params:
        none

    Tasks:
        free memory

    Return value:
        none
**/
SDCardV3Logger::~SDCardV3Logger()
{
    if ( m_buffer != NULL )
    {
        delete[] m_buffer;
    }
}

/**
    write() - CommunicationControllerClient() virtual method, Tx, outgoing commands

    Params:
        buffer  : character buffer, store command which will be sent to the comm.port
        size    : size of the command

    Tasks:
        set timeout
        creates the appropriate command (according to the current state)

    Return value:
        none
**/
void SDCardV3Logger::write( uint8_t* buffer, int& size )
{
    m_timeout = iTIMEOUT;

    switch ( m_state )
    {
        case SEND_START:
            setSDInfoDirty();
            sendStart( buffer, size );
            break;

        case SEND_STOP:
            setSDInfoDirty();
            sendStop( buffer, size );
            break;

        case SEND_STATUS:
            sendStatus( buffer, size );
            break;

        case SEND_ERASE:
            setSDInfoDirty();
            sendErase( buffer, size );
            break;

        case SEND_GETBLOCK0:
            sendGetBlock0( buffer, size );
            break;

        // wait after erase (~3secs)
        case WAIT_AFTER_ERASE:
            waitAfterErase();
            break;

        case SEND_ERASE_STATUS:
            sendEraseStatus( buffer, size );
            break;

        case SEND_GET_FWVER:
            sendGetFwVersion( buffer, size );
            break;

        default:
            break;
    }
}

/**
    read() - CommunicationControllerClient virtual method, incoming data from comm.port

    Params:
        buffer  : the incoming data
        size    : size of the data

    Tasks:
        interpret the data according to the current state

    Return value:
        none
**/
void SDCardV3Logger::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case CHECK_STOP:
            checkStop( buffer, size );
            break;

        case CHECK_START:
            checkStart( buffer, size );
            break;

        case CHECK_GETBLOCK0:
            checkGetBlock0( buffer, size );
            break;

        case WAIT_AFTER_ERASE:
            waitAfterErase();
            break;

        case CHECK_ERASE_STATUS:
            checkEraseStatus( buffer, size );
            break;

        case CHECK_GET_FWVER:
            checkGetFwVersion( buffer, size );
            break;

        default:
            break;
    }
}

/**
    isFinished() - CommunicationControllerClient virtual method

    Params:
        none

    Tasks:
        check whether command has been processed fully

    Return value:
        bool: true => if finished or timeout => thread should be stopped
**/
bool SDCardV3Logger::isFinished()
{
    return ( m_state == FINISHED || m_state == TIMEOUT );
}

/**
    finish() - CommunicationControllerClient virtual method

    Params:
        none

    Tasks:
        called if client finished its job, actually it does nothing

    Return value:
        none
**/
void SDCardV3Logger::finish()
{
    // nop
}


/**
    sendStart()

    Params:
        buffer  : the 'msemss' command ( start logging, get status )
        size    : size of the command

    Tasks:
        start logging, get status

    Return value:
        none
**/
void SDCardV3Logger::sendStart( uint8_t* buffer, int& size )
{
//    LOGP( _T("START"))
    m_checkStatCounter = 0;
    wxString cmd = _T("msemss");

    TxData command(cmd);

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = CHECK_START;

}

/**
    checkStart()

    Params:
        buffer  : incoming byte
        size    : size of the data

    Tasks:
        interprets the result of mss
        if it is '0' -> comm. failed
        if it is 'u' -> undefined -> wait a little bit -> check status again
        it it is '1' -> ok

    Return value:
        none
**/
void SDCardV3Logger::checkStart( uint8_t* buffer, int size )
{
    if ( size == 1 )
    {
        char c = buffer[0];
        // Logger::logf(LOG_MSG, _T("size=%d, char=%c"), size,c );
        // result is '0'
        if ( c == '0' )
        {
            // comm failed -> reset original devmode -> finished
            if ( m_evthandler )
            {
                SDCardEvent event;
                event.SetType( SDCardEventTypes::COMM_FAILED );
                m_evthandler->AddPendingEvent( event );
            }

            //LOGP( _T("COMM FAILED") )
            m_state = FINISHED;
            return;
        }

        // it is undefined
        if ( c == 'u' )
        {
            m_checkStatCounter++;

            // if counter reached max. -> startlog failed, reset devmode -> finished
            if ( m_checkStatCounter > 250 )
            {
                sendEvent( SDCardEventTypes::STARTLOG_FAILED );
                //LOGP( _T("INIT FAILED (START)"))
                m_state = FINISHED;
            }
            // waiting
            else
            {
                sendEvent( SDCardEventTypes::STARTLOG_WAITING, (int)( (double)m_checkStatCounter * 100.0 / 250.0 ) );

                //LOGP( wxString::Format( _T("START waiting: %d"), m_checkStatCounter ) )
                m_state = SEND_STATUS;
            }
            // wxSleep( 1000 );
            return;
        }

        // OK, loggin started, send event, reset mode -> finished
        if ( c == '1' )
        {
            sendEvent( SDCardEventTypes::LOGGING_STARTED );

            //LOGP( _T("START started"))
            m_state = FINISHED;
        }
    }
    // timeout?
    if ( size == 0 )
    {
        m_sleep = 2;
        m_timeoutCtr++;
        wxSleep( m_sleep );

        if ( m_timeoutCtr > 2500 )
        {
            sendEvent( SDCardEventTypes::TIMEOUT );

            //LOGP( _T("TIMEOUT"))
            m_state = TIMEOUT;
        }
    }
}

/**
    sendStatus()

    Params:
        buffer  : the 'mss' command
        size    : size of the command

    Task:
        check status, send mss to comm.port

    Return value:
        none
**/
void SDCardV3Logger::sendStatus( uint8_t* buffer, int& size )
{
    wxString cmd = _T("mss");

    TxData command(cmd);

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = CHECK_START;
}

/// -------------------------------

/**
    sendStop()

    Params:
        buffer  : the 'msdmss' command
        size    : size of the command

    Tasks:
        disable logging, check status

    Return value:
        none
**/
void SDCardV3Logger::sendStop( uint8_t* buffer, int& size )
{
    /// TODO
    LOGP( _T("STOP"))
    wxString cmd = _T("msdmss");

    TxData command(cmd);

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = CHECK_STOP;

    m_timeoutTimer.start( 1000 );
}

/**
    checkStop()

    Params:
        buffer  : incoming data
        size    : size of the data

    Tasks:
        check wheter status is 0

    Return value:
        none
**/

void SDCardV3Logger::checkStop( uint8_t* buffer, int size )
{
    if ( size == 1 )
    {
        char c = buffer[0];
        // result is '0' => OK
        if ( c == '0' )
        {
            sendEvent( SDCardEventTypes::LOGGING_STOPPED );
            // LOGP( _T("STOP OK") )
        }
        // anything else => stoplog failed
        else
        {
            sendEvent( SDCardEventTypes::STOPLOG_FAILED );
            //LOGP( _T("NOT OK"))
        }
    }
    else
    {
        if ( m_timeoutTimer.isTimedout() )
        {
            sendEvent( SDCardEventTypes::COMM_FAILED );
        }
        else
        {
            return;
        }
        //LOGP( _T("COMM FAILED"))
    }

    // reset dev.mode
    m_state = FINISHED;
}

/// -------------------------------

/**
    sendErase()

    Params:
        buffer  : outgoing command (msc)
        size    : size of the command

    Tasks:
        send erase command (msc) to the commport

    Return value:
        none
**/
void SDCardV3Logger::sendErase( uint8_t* buffer, int& size )
{
    LOGP( _T("ERASE"))
    wxString cmd = _T("msc");

    TxData command( cmd );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    // wait couple of seconds after command is sent
    m_state = WAIT_AFTER_ERASE;

    sendEvent( SDCardEventTypes::ERASE_WAITING );

    m_sp.Start();
}

/**
    waitAfterErase()

    Params:
        none

    Tasks:
        wait 3 sec
        after this interval set state to check status

    Return value:
        none
**/
void SDCardV3Logger::waitAfterErase()
{
    if ( m_sp.Time() > 10000 ) // wait 10s after erase command was sent
    {
        m_sp.Pause();

        if ( m_buffer != NULL )
        {
            delete[] m_buffer;
        }

        m_buffer = new uint8_t[514 + 2];

        m_tryCount = 0;
        m_timeoutCtr = 0;

        m_mssCounter = 0;

        // lets check status (mss), start msg only if status = 0
        m_state = SEND_ERASE_STATUS;
    }
}

/**
    sendEraseStatus()

    Params:
        buffer  : the mss command
        size    : size of the command

    Tasks:
        send mss to the comm.port

    Return value:
        none
**/
void SDCardV3Logger::sendEraseStatus( uint8_t* buffer, int & size )
{
    wxString cmd = _T("mss");
    TxData command( cmd );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = CHECK_ERASE_STATUS;

    m_timeoutTimer.start( 1000 );
}

/**
    checkEraseStatus()

    Params:
        buffer  : the incoming data
        size    : the size of the data

    Tasks:
        check the response of the mss command
            -> 0: OK - lets check the 0. block whether it is empty
            -> u: undefined, wait a little bit and check status again

    Return value:
        none
**/
void SDCardV3Logger::checkEraseStatus( uint8_t* buffer, int size )
{
    // timeout ?
    if ( size == 0 )
    {
        if ( m_timeoutTimer.isTimedout() )
        {
            sendEvent( SDCardEventTypes::TIMEOUT );
            m_state = TIMEOUT;
        }
//        ++m_timeoutCtr;
//        wxThread::Sleep( m_sleep );
//
//        if ( m_timeoutCtr > m_timeout )
//        {
//            sendEvent( SDCardEventTypes::TIMEOUT );
//            m_state = TIMEOUT;
//        }
    }

    if ( size > 0 )
    {
        m_timeoutCtr = 0;

        char c = buffer[0];

        //
        if ( c == '0' )
        {
            // lets check sector 0 - it should be empty
            m_state = SEND_GETBLOCK0;
        }
        if ( c == 'u' )
        {
            LOG
            ++m_mssCounter;
            // check status again after 100ms
            m_state = SEND_ERASE_STATUS;

            wxThread::Sleep( 100 );

            if ( m_mssCounter > 10 )
            {
                LOGP(_T("ERROR: mss still responding 'u' after 10 tries") )
                sendEvent( SDCardEventTypes::ERASE_FAILED );

                // reset the original devmode
                m_state = FINISHED;
            }

        }
    }
}

/**
    sendGetBlock0()

    Params:
        buffer  : the msg000000
        size    : size of the command

    Tasks:
        send msg000000 command to check the emptiness of block #0

    Return value:
        none
**/
void SDCardV3Logger::sendGetBlock0( uint8_t* buffer, int& size )
{
    wxString cmd = _T("msg000000");

    TxData command(cmd);

    size = command.size;
    memcpy( buffer, command.data, command.size );

    LOGP( cmd );

    m_state = CHECK_GETBLOCK0;

    m_timeoutTimer.start( 1000 );
}

/**
    checkGetBlock0()

    Params:
        buffer  : the incoming data
        size    : size of the data

    Tasks:
        read the first sector, check whether it is empty
        if not -> send erase command again

    Return value:
        none
**/
void SDCardV3Logger::checkGetBlock0( uint8_t* buffer, int size )
{
    // timeout?
    if ( size == 0 )
    {
        LOG
//        if ( m_timeoutTimer.isTimedout() ) {
//            sendEvent( SDCardEventTypes::TIMEOUT );
//            m_state = TIMEOUT;
//        }

//        ++m_timeoutCtr;
//        wxThread::Sleep( m_sleep );
//
//        if ( m_timeoutCtr > m_timeout )
        if ( m_timeoutTimer.isTimedout() )
        {
            //Logger::logf(LOG_MSG, _T("m_timeOutctr=%d, m_timeout=%d"), m_timeoutCtr, m_timeout);
            if ( ++m_tryCount < iMAX_TRY_READCOUNT )
            {
                // send mss again
                m_state = SEND_ERASE_STATUS;
                m_timeoutCtr = 0;
            }
            else
            {
                // LOGP( _T("Timeout"))
                // timeout
                sendEvent( SDCardEventTypes::TIMEOUT );
                m_state = TIMEOUT;
            }
        }
        return;
    }
    else m_timeoutCtr = 0;

    memcpy( m_buffer + m_buffersize, buffer, size );
    m_buffersize += size;

    // first block read in
    if ( m_buffersize >= 514 )
    {
        // erase was successful
        if ( SDCardFile::checkSectorIsEmpty( m_buffer, 512 ) == true )
        {
            // restore original device mode & send event
            m_state = FINISHED;
            sendEvent( SDCardEventTypes::ERASE_SUCCESS );
            //LOGP( _T("ERASE WAS SUCCESSFUL"))
        }
        else
        {
            //LOGP( wxString::Format( _T("ERASE: SECTOR 0 IS NOT EMPTY! # OF TRIES = %d"), ++m_checkSectTryCount ) );
            // first sector is not empty -> send msc again

            // reached the max try count
            if ( m_checkSectTryCount > ( iMAX_TRY_ERASECOUNT - 1 ) )
            {
                //LOGP( _T("MAX TRYCOUNT REACHED"))

                sendEvent( SDCardEventTypes::ERASE_FAILED );

                m_state = FINISHED;
            }
            else
            {
                sendEvent( SDCardEventTypes::ERASE_NOTEMPTY, (int)( (double) m_checkSectTryCount * 100.0 / (double)iMAX_TRY_ERASECOUNT ) );
                m_state = SEND_ERASE;
            }
        }
    }
}

/**
    sendGetFwVersion()

    Params:
        buffer  : the msg000000 command
        size    : size of the command

    Tasks:
        send the msdmsg000000 command to obtaint the firmware version

    Return value:
        none
**/
void SDCardV3Logger::sendGetFwVersion( uint8_t* buffer, int& size )
{
    wxString cmd = _T("msdmsg000000");
    TxData command( cmd );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_timeoutTimer.start( 1000 );

    m_buffersize = 0;
    if ( m_buffer == NULL )
    {
        m_buffer = new unsigned char[514 + 2];
    }

    m_state = CHECK_GET_FWVER;
}

/**
    checkGetFwVersion

    Params:
        buffer  : the incoming data
        size    : size of the data

    Tasks:
        read the first sector, get firmware version from the 4th and 5th bytes of the buffer
        send event with firmware information to the event handler

    Return value:
        none
**/
void SDCardV3Logger::checkGetFwVersion( uint8_t* buffer, int size )
{
    // timeout?
    if ( size == 0 )
    {
        if ( m_timeoutTimer.isTimedout() )
        {
            if ( ++m_tryCount < iMAX_TRY_READCOUNT )
            {
                m_state = SEND_GET_FWVER;
                m_timeoutCtr = 0;
            }
            else
            {
                sendEvent( SDCardEventTypes::TIMEOUT );
                m_state = TIMEOUT;
            }
        }
        return;
    }
    else m_timeoutCtr = 0;

    memcpy( m_buffer + m_buffersize, buffer, size );
    m_buffersize += size;

    // first block has been read in
    if ( m_buffersize >= 514 )
    {
        // get FW Version
        FirmwareVersion fwVersion = SDCardFile::getFwVersion( m_buffer );
        unsigned int fwVer_int = SDCardFile::getFWVersionAsInt( m_buffer );
        wxString fwVerS = fwVersion.toString();

        if ( m_evthandler )
        {
            SDCardEvent event;
            event.SetType( SDCardEventTypes::FIRMWARE_VERSION );
            event.SetString( fwVerS );
            event.SetInt( fwVer_int );

            m_evthandler->AddPendingEvent( event );
        }

        delete[] m_buffer;
        m_buffer = NULL;

        m_state = FINISHED;
    }
}

///------------------
/// HELPERS

/**
    setSDInfoDirty()

    Params:
        none

    Tasks:
        mark the static sdinfo as dirty ( see SDStaticInfo in SDCardFileSearch.h )

    Return value:
        none
**/
void SDCardV3Logger::setSDInfoDirty()
{
    SDStaticInfo::setDirty( true );
}

/**
    sendEvent()

    Params:
        eventType   : see SDCardEvent.h
        intvalue    : set the int member of the event if it is needed

    Tasks:
        send event to the m_evthandler

    Return value:
        none
**/
void SDCardV3Logger::sendEvent( int eventType, int intvalue )
{
    if ( m_evthandler )
    {
        SDCardEvent event;

        event.SetType( eventType );

        if ( intvalue != -1 )
        {
            event.SetInt( intvalue );
        }

        m_evthandler->AddPendingEvent( event );
    }
}
