#include "AggToolTip.h"

#include "../Math.h"
#include <wx/wx.h>
#include <wx/tokenzr.h>
#include <wx/regex.h>

#include "../agg/wxAgg.h"
#include "../logger.h"
#include "../wxPointEx.h"

#ifndef M_PI
	#define M_PI		3.14159265358979323846
#endif
#define M_PI_2		1.57079632679489661923
#define M_PI_4		0.78539816339744830962

WX_DECLARE_HASH_MAP( int, RealPointArray, wxIntegerHash, wxIntegerEqual, RealPointArrayHash);

AggToolTip::AggToolTip()
{
    //ctor
    m_currentProfile = DefaultAggToolTipProfile;

    InitProfile(m_currentProfile);
}

AggToolTip::~AggToolTip()
{
    //dtor
}

void AggToolTip::InitProfile(const wxString &name)
{
    wxVariantIntHash props;

    props[ ATT_TEXT_CONTENT ] = _T("tooltip line 1\ntooltip line 2");
    props[ ATT_TEXT_ISWRAP ] = true;
    props[ ATT_TEXT_COLOR ] << wxColor(0,0,0,255);
    props[ ATT_TEXT_HEIGHT ] = 12.0;
    props[ ATT_TEXT_STROKE ] = 1.0;
    props[ ATT_TEXT_PADDING ] = 2.0;
    props[ ATT_TEXT_ALIGN ] = (long)ATT_TA_LEFT;
    props[ ATT_BACKGROUNDCOLOR ] << wxColor(255,255,255,250);
    props[ ATT_BORDER_WIDTH ] = 1.0;
    props[ ATT_BORDER_COLOR ] << wxColor(0,0,0,255);
    props[ ATT_ARROW_ALIGMENT ] = (long)ATT_AA_TOP;
    props[ ATT_ARROW_SIZE ] = 10.0;
    props[ ATT_ARROW_REFPOSITION ] = 0.5;
    props[ ATT_ROUND_RADIUS ] = 5.0;
    props[ ATT_ROUND_TYPE ] = (long)ATT_RT_NONE;
    props[ ATT_BOX_MARGIN ] = 0.0;

    m_profileProps[name] = props;
}

void AggToolTip::SetProperties(wxVariantIntHash hash)
{
    for(wxVariantIntHash::iterator it = hash.begin(); it != hash.end(); it++) {
        m_profileProps[m_currentProfile][ it->first ] = it->second;
    }
}

void AggToolTip::SetProperty(int prop, const wxVariant& value)
{
    m_profileProps[m_currentProfile][ prop ] = value;
}

void AggToolTip::SetColorProp(int prop, const wxColor& value)
{
    wxVariant v;
    v << value;
    m_profileProps[m_currentProfile][ prop ] = v;
}

wxSize AggToolTip::CalcSize()
{
    wxArrayString lines;
    DoubleArray line_widths;

    double contentwidth = CalcContentWidth(lines, line_widths);
    double arrowheight = GetArrowSize();
    double padding = GetDouble( ATT_TEXT_PADDING );
    double text_height = GetDouble( ATT_TEXT_HEIGHT );
    double border_width = GetDouble( ATT_BORDER_WIDTH );
    int    arrow_align = GetLong( ATT_ARROW_ALIGMENT );
    double margin = GetDouble( ATT_BOX_MARGIN );

    double boxw = contentwidth + 2 * padding + 2 * border_width;
    double boxh = (double)lines.GetCount() * text_height * 1.6 + 2 * padding + 2 * border_width;

    int width = (int)ceil( boxw + ((arrow_align == ATT_AA_LEFT || arrow_align == ATT_AA_RIGHT) ? arrowheight : 0.0 ));
    int height = (int)ceil( boxh + ((arrow_align == ATT_AA_BOTTOM || arrow_align == ATT_AA_TOP) ? arrowheight : 0.0 ));

    return wxSize( width + (int)ceil(margin*2), height + (int)ceil(margin*2) );
}

double AggToolTip::CalcContentWidth(wxArrayString& lines, DoubleArray& line_widths)
{
    double contentwidth = -1;
    double text_height = GetDouble( ATT_TEXT_HEIGHT );

    if(GetBool( ATT_TEXT_ISWRAP )) {
        lines = wxStringTokenize( GetStr( ATT_TEXT_CONTENT ), _T("\n") );

        for(size_t a=0; a<lines.GetCount(); a++){
            double d = wxAgg::getTextWidth(lines[a], text_height);
            if( d > contentwidth ) contentwidth = d;
            line_widths.Add(d);
        }

    } else {
        lines.Add( GetStr(ATT_TEXT_CONTENT) );
        contentwidth = wxAgg::getTextWidth( lines[0], text_height );
        line_widths.Add(contentwidth);
    }

    return contentwidth;
}

wxBitmap* AggToolTip::DrawBox(const wxVariant& value, int key)
{
    if( value != wxNullVariant ) {
       SetProperty( key, value );
    }

    wxArrayString lines;
    DoubleArray line_widths;

    bool   isRender = key == ATT_SPECIAL_NO_RENDER ? false : true;

    double contentwidth = CalcContentWidth(lines, line_widths);
    double arrowheight = GetArrowSize();
    double padding = GetDouble( ATT_TEXT_PADDING );
    double text_height = GetDouble( ATT_TEXT_HEIGHT );
    double border_width = GetDouble( ATT_BORDER_WIDTH );
    double arrowwidth = arrowheight;
    double text_stroke = GetDouble( ATT_TEXT_STROKE );
    int    text_align = GetLong( ATT_TEXT_ALIGN );
    int    arrow_align = GetLong( ATT_ARROW_ALIGMENT );
    int    round_type = GetLong(ATT_ROUND_TYPE);
    double round_radius = round_type == ATT_RT_NONE ? 0.0 : GetDouble( ATT_ROUND_RADIUS );
    double margin = GetDouble( ATT_BOX_MARGIN );


    double boxw = contentwidth + 2 * padding + 2 * border_width;
    double boxh = (double)lines.GetCount() * text_height * 1.6 + 2 * padding + 2 * border_width;

    int width = (int)ceil( boxw + ((arrow_align == ATT_AA_LEFT || arrow_align == ATT_AA_RIGHT) ? arrowheight : 0.0 ));
    int height = (int)ceil( boxh + ((arrow_align == ATT_AA_BOTTOM || arrow_align == ATT_AA_TOP) ? arrowheight : 0.0 ));

    wxAgg aggCanvas(width + (int)ceil(margin*2), height + (int)ceil(margin*2));

    RealPointArray arr;

    double arrow_width_ref = (arrow_align == ATT_AA_TOP || arrow_align == ATT_AA_BOTTOM) ? boxw : boxh;

    double arrow_ref_pos = GetDouble( ATT_ARROW_REFPOSITION );

    if(arrow_align == ATT_AA_BOTTOM || arrow_align == ATT_AA_RIGHT) arrow_ref_pos = 1 - arrow_ref_pos;

    double arrowpoint = arrow_width_ref * arrow_ref_pos;

    if(arrowpoint < (arrowwidth/2 + round_radius)) arrowpoint = arrowwidth / 2.0 + round_radius;

    if(arrowpoint > arrow_width_ref - (arrowwidth / 2.0) - round_radius) arrowpoint = arrow_width_ref - (arrowwidth / 2.0) - round_radius;

    double top_left_x = (arrow_align==ATT_AA_LEFT ? arrowheight : 0.0) + border_width/2.0 + margin;
    double top_left_y = (arrow_align==ATT_AA_TOP ? arrowheight : 0.0) + border_width/2.0 + margin;

    double sx = top_left_x + round_radius;
    double sy = top_left_y;

    arr.Add( wxRealPoint( sx, sy ) );

    RealPointArrayHash sides;

    for(char a = ATT_AA_TOP; a <= ATT_AA_LEFT; a++) {
        double px = arr.Last().x;
        double py = arr.Last().y;
        double x = cos(a*M_PI_2) * (boxw - border_width - 2*round_radius);
        double y = sin(a*M_PI_2) * (boxh - border_width - 2*round_radius);

        sides[a].Add(arr.Last());

        if(a == arrow_align) {

            double apx = px;
            double apy = py;

            apx += (cos(a*M_PI_2) * (arrowpoint-round_radius)) - cos(a*M_PI_2) * arrowwidth/2.0;
            apy += (sin(a*M_PI_2) * (arrowpoint-round_radius)) - sin(a*M_PI_2) * arrowwidth/2.0;
            arr.Add( wxRealPoint(apx,apy) ); sides[a].Add(arr.Last());

            apx += sin(a*M_PI_2) * arrowwidth + cos(a*M_PI_2) * arrowwidth/2.0;
            apy += cos(a*(M_PI_2)) * -1 * arrowwidth + sin(a*M_PI_2) * arrowwidth/2.0 ;
            arr.Add( wxRealPoint(apx,apy) ); sides[a].Add(arr.Last());

            wxVariant v; v << wxPointEx( (int)apx, (int)apy ); SetProperty( ATT_ARROW_PXPOSITION, v );

            apx += sin(a*(M_PI_2)) * -1 * arrowwidth + cos(a*M_PI_2) * arrowwidth/2.0;
            apy += cos(a*M_PI_2) * arrowwidth + sin(a*M_PI_2) * arrowwidth/2.0;
            arr.Add( wxRealPoint(apx,apy) ); sides[a].Add(arr.Last());
        }

        arr.Add( wxRealPoint(px+x,py+y) );
        sides[a].Add(arr.Last());

        if( round_type != ATT_RT_NONE ) {
            px = arr.Last().x;
            py = arr.Last().y;
			px += Math::round(cos((a*M_PI_2) + M_PI_4)) * round_radius;
            py += Math::round(sin((a*M_PI_2) + M_PI_4)) * round_radius;
            arr.Add( wxRealPoint(px,py) );
        }

    }

    aggCanvas.setBrush( GetColor( ATT_BACKGROUNDCOLOR ) );
    aggCanvas.setPen( GetColor( ATT_BORDER_COLOR ) );



    switch( round_type ){
        case ATT_RT_LINE:
        case ATT_RT_NONE:
            if(isRender) aggCanvas.drawSolidPolygon(arr,border_width);
            break;

        case ATT_RT_ARC:
            if(isRender) aggCanvas.drawPolygon(arr);
            for(char a = ATT_AA_TOP; a <= ATT_AA_LEFT; a++) {
                RealPointArray& rparr = sides[a];
                if(rparr.GetCount()<1) continue;
                if(border_width > 0.01 && isRender) aggCanvas.drawLine( rparr, border_width );
                wxRealPoint lp = rparr[rparr.GetCount()-1];
                double rx = lp.x + sin( (a*M_PI_2)+M_PI ) * round_radius;
                double ry = lp.y + cos( a*M_PI_2 ) * round_radius;
                if(isRender) aggCanvas.drawFilledCircleSection(rx, ry, round_radius, a * M_PI_2 - M_PI_2 , M_PI_2);
                if(border_width > 0.01 && isRender) aggCanvas.drawArc( rx, ry, round_radius, a * M_PI_2 - M_PI_2 , M_PI_2, border_width );
            }
            break;
    }



    aggCanvas.setPen( GetColor( ATT_TEXT_COLOR ) );

    for(size_t a = 0; a < lines.GetCount(); a++) {
        wxString str = lines[a];

        double y = top_left_y + padding + a * text_height * 1.6 + text_height * 1.3 + border_width/2.0;
        double w = line_widths[a];
        double x = top_left_x + padding + text_align * ((contentwidth - w) / 2.0) + border_width/2.0;

        if(isRender) aggCanvas.drawText(str, x, y, 0, text_height, text_stroke);
    }

    wxBitmap * bmp = NULL;

    if(isRender) bmp = aggCanvas.renderToBitmap();

    return bmp;
}

wxPoint AggToolTip::GetArrowPoint()
{
    if( m_profileProps[m_currentProfile][ATT_ARROW_PXPOSITION].GetType() != _T("wxPointEx") ) return wxPoint(0,0);

    wxPointEx p; p << m_profileProps[m_currentProfile][ATT_ARROW_PXPOSITION];

    return p;
}

double AggToolTip::GetDouble(int key)
{
    return m_profileProps[m_currentProfile][key].GetDouble();
}

bool AggToolTip::GetBool(int key)
{
    return m_profileProps[m_currentProfile][key].GetBool();
}

long AggToolTip::GetLong(int key)
{
    return m_profileProps[m_currentProfile][key].GetLong();
}

wxString AggToolTip::GetStr(int key)
{
    return m_profileProps[m_currentProfile][key].GetString();
}

wxColor AggToolTip::GetColor(int key)
{
    if( m_profileProps[m_currentProfile][key].GetType() != _T("wxColour") ) return wxNullColour;
    wxColor color;
    color << m_profileProps[m_currentProfile][key];
    return color;
}

double AggToolTip::GetArrowSize()
{
    if(m_profileProps[m_currentProfile][ATT_ARROW_SIZE].GetType() == _T("string")) {
        wxString str = m_profileProps[m_currentProfile][ATT_ARROW_SIZE].GetString();
        wxRegEx re;
        re.Compile(_T("([0-9]{1,3})%$"));
        if(re.Matches(str) == false) return 10.0;
        double tmp;
        re.GetMatch(str,0).ToDouble(&tmp);
        return m_profileProps[m_currentProfile][ATT_TEXT_HEIGHT].GetDouble() * tmp / 100.0;
    } else {
        return m_profileProps[m_currentProfile][ATT_ARROW_SIZE].GetDouble();
    }

}

wxVariant AggToolTip::GetProperty(int prop, const wxVariant& defvalue)
{
    if( m_profileProps[m_currentProfile].find(prop) == m_profileProps[m_currentProfile].end() ) return defvalue;

    return m_profileProps[m_currentProfile][prop];
}

