#include "IHexParser.h"

#include "../logger.h"
#include "../Utils.h"
#include "../Math.h"

/*
 *the intel index file format
 *:10035000A4A3A2A1A0A09F9E9D9C9C9B9A999998C2
 * ^^|   | |	 :0(LEN_IN_LINE) length (0x10)
 *   ^^^^| |	 :1-2(ADDRESS_IN_LINE) address: 0x0350
 *       ^^|	 :3(TYPE_IN_LINE) type
 *         ^^	 :4(DATA_IN_LINE) begin of data
 *	   			 : last 2 byte checksum:   ^^
 *
 * reads from line, checks chksum,
 * and writes result to lineBuffer (which must have been initiated by caller!)
 * (finds out address by itself)
 *
 * returns number of bytes received (negative for error)
 */

IHexParser::IHexParser(ECUInfo::Device device)
{
    //ctor
    m_addressOffset = 0; //0x1000; ///!!!!!!!!!!!!!!!!!!!!! DEBUG !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    m_firmwareData = NULL;
    m_firmwareData = new FirmwareData(device);

    switch(device){
        case ECUInfo::DEVICE_ARM:
        case ECUInfo::DEVICE_V3:
            m_pageSize = 256;
            break;
        case ECUInfo::DEVICE_ROUND:
            m_pageSize = 128;
            break;
        default:
            m_pageSize = 256;
    }
    //m_firmwareData.setPageSize(m_pageSize);
    //m_firmwareData.setType(HEX);
}

IHexParser::~IHexParser()
{
    //dtor
}

bool IHexParser::detectType()
{
    if( m_line.StartsWith(_T(":")) ){
        m_firmwareData->setType(FirmwareData::HEX);
        return true;
    }else if(m_line.StartsWith(_T("C:")) ){
        m_firmwareData->setType(FirmwareData::CHEX);
        return true;
    }
    return false;
}

void IHexParser::getDataAndCheksum(wxString& hexData, wxString& hexChksum)
{
    wxArrayString tmparray;
    Utils::tokenizeToArray( m_line, _T(":"), tmparray);

    if(tmparray.GetCount() < 2) return;

    //Logger::logf(LOG_ERROR, _T("IHexParser() count:%i")+tmparray[1], tmparray.GetCount() );

    hexData = tmparray[1];
    if(tmparray.GetCount() > 2){
        hexChksum = tmparray[2];
    }

// return m_line.Mid(m_leadingByteCount);
 /*
    switch(m_firmwareData.getType()){
        case FirmwareData::HEX:
             return m_line.Mid(1);                   //remove the first ":" char
        case FirmwareData::CHEX:
             return m_line.Mid(2);                   //remove the first "C:" char
        default:
             return m_line;
    }
*/
}


void IHexParser::parseLine()
{

   // if ( isComment() ) return;
    m_line.Trim(false); // removes leading spaces;
    if(  m_line.StartsWith(_T("#")) ){ return; }    // ignore comment
    if(  m_line.StartsWith(_T("\n")) ){ return; }    // ignore new  line
    if(  m_line.StartsWith(_T("\r")) ){ return; }    //
    if(  m_line.Length()  < 2){ return; }    // ignore  length < 2

    if( !detectType() ) return; //unknown type

    wxString hexData;
    wxString hexChksum;
    getDataAndCheksum(hexData, hexChksum);



    if(hexData.Length() < 10){ return; } //data: too short



    //ByteArray binaryLineA;
    hexString2binaryArray(hexData);


    uint8_t type = m_lineBuff[TYPE_IN_LINE];
    switch(type)
    {
     case (0x00): // data record;
        if( ! isLineChecksumError() ){
            //Logger::logf(LOG_MSG, _T("IHexParser() parse line: %d"), m_lineNumber );
            uint16_t pageIdx;
            if( fillFirmwareData(pageIdx)
                && m_firmwareData->getType() == FirmwareData::CHEX
                && hexChksum.Length() == 8){
                    uint32_t chksum = (uint32_t) Math::hex2int(hexChksum);
                    m_firmwareData->setPageCheckSum(pageIdx, chksum);
            }

        }else{
            Logger::logf(LOG_ERROR, _T("IHexParser() hexfile line neglected (type2)(checksum error): ")+m_line );
            //return -3;
        }
        break;
     case (0x02):
     // Extended Segment address record.
     // This indicates segment base address when 16 bits is not enough for addressing memory;
        m_addressOffset += 0x10000;
        break;
    case (0x04): // Extended Linear Address Record  allows 32 bit addressing.
        //m_addressOffset = 0x1000000;
        ///m_addressOffset += 0x1000;  bad idea the correct written in:
        ///http://homepages.cae.wisc.edu/~ece554/new_website/docs/intel_4.pdf
        //break;
     default:
        Logger::logf(LOG_ERROR, _T("IHexParser() hexfile line neglected (type=%02x): ") + m_line, type );
		//return -2;
    };
}

void IHexParser::hexString2binaryArray(const wxString& str)
{
    m_lineBuffLen = str.Length() >> 1;
    uint8_t val;
    wxChar c;
    for( int i = 0; i < m_lineBuffLen; i++ ){
        c = str.GetChar(2*i);
        val = hex2bin(c) * 16;
        c = str.GetChar(2*i+1);
        val += hex2bin(c);
        m_lineBuff[i] = val;
    }
}

/*
 * return true if checksum error
 * (0 if OK)
 */
uint8_t IHexParser::isLineChecksumError()
{
	uint8_t chksum = 0;
	uint8_t i = m_lineBuffLen;
	do{
		i--;
		chksum += m_lineBuff[i];
	} while ( i );
	return chksum;
}

/*
 * retreives big-endian address from rawBytes
 */
uint32_t IHexParser::getLineAddress()
{
	return ( m_lineBuff[ ADDRESS_IN_LINE ] << 8 )
		+  ( m_lineBuff[ ADDRESS_IN_LINE + 1] );

}

uint8_t IHexParser::getLineDataLength()
{
    return m_lineBuff[ LEN_IN_LINE ] ;
}

uint16_t IHexParser::getPageIdx(uint32_t addressB)
{
	return addressB / m_pageSize;
}

bool IHexParser::fillFirmwareData(uint16_t& pageIdx)
{
     uint32_t addressB = m_addressOffset + getLineAddress();
     pageIdx = getPageIdx(addressB);
     uint32_t offsetB = addressB - m_pageSize * pageIdx;

     uint32_t dataLen = getLineDataLength();

    //Logger::logf(LOG_ERROR, _T("IHexParser::fillFirmwareData() %04X, %u, %u, %u"),addressB, pageIdx, offsetB, dataLen);
    if(offsetB + dataLen > m_pageSize)
    {
        Logger::logf(LOG_ERROR, _T("IHexParser::fillFirmwareData() offset out of bounds: page: %u, offset%u, len: %u"), pageIdx, offsetB, dataLen);
        Logger::logf(LOG_ERROR, _T("IHexParser::fillFirmwareData() address: %04X"), addressB);

        int nextPageOffset = m_pageSize - offsetB;
        int nextPageDataLength = (offsetB+dataLen) - m_pageSize;
        uint8_t * lineDataPtr = &(m_lineBuff[DATA_IN_LINE]) + nextPageOffset;
        m_firmwareData->setPageData(pageIdx+1,0, lineDataPtr, nextPageDataLength);
        Logger::logf(LOG_ERROR, _T("IHexParser::fillFirmwareData() nextPageOffset: %i, nextPageDataLength: %i"), nextPageOffset, nextPageDataLength);
    }

    uint8_t * lineDataPtr = &(m_lineBuff[DATA_IN_LINE]);
    m_firmwareData->setPageData(pageIdx,offsetB, lineDataPtr, dataLen); //checksum do not need
    return true;

}
FirmwareData * IHexParser::getFirmwareData()
{
    return m_firmwareData;
}

uint8_t IHexParser::hex2bin( uint8_t hexchr )
{
    uint8_t res = 0;

	if( hexchr >= '0' && hexchr <= '9' ){
		res = hexchr - '0';
	}else if( hexchr >= 'A' && hexchr <= 'F' ){
		res = hexchr - 'A' + 10;
	}else if( hexchr >= 'a' && hexchr <= 'f' ){
		res = hexchr - 'a' + 10;
	}else{
	    fprintf( stdout, "ERROR: not a hexa digit: %c\n", hexchr );
	}
	return res;
}
