#include "ConfigValidatorMgr.h"

#include "ConfigCriteriaData.h"
#include "ConfigCriteria.h"
#include "ConfigCriteriaRange.h"
#include "../ChannelValidationHelper.h"

#include "../common/VemsMessage.h"
#include "../common/MenuLinkHelper.h"

#include "../stdatamanager/STDataModelBase.h"
#include "../stdatamanager/STDataHelper.h"

#include "../mtconfigdata.h"
#include "../MTConfigHelper.h"
#include "../FirmwareInfo.h"
#include "../logger.h"
#include "../Utils.h"
#include "../ChannelCollision.h"
#include "../common/types/DoubleHash.h"

ConfigValidatorMgr::ConfigValidatorMgr(ConfigCriteriaData* criteriaData,  MTConfigData* mtconfigData)
{
    //ctor
    m_criteriaData = criteriaData;
    m_mtconfigData = mtconfigData;
    reset();
}

ConfigValidatorMgr::ConfigValidatorMgr(ConfigCriteriaData* criteriaData, MTConfigData* mtconfigData, STDataModelBase* stdata_model,
    FirmwareInfo* fwInfo )
{
    m_criteriaData  = criteriaData;
    m_mtconfigData  = mtconfigData;
    m_stModel       = stdata_model;
    m_fwInfo        = fwInfo;

    SymbolTable& cfgSymbolH = m_mtconfigData->getSymbolMgr().getCfgSymbolTable();

    STDataHelper::fillVariableHashFromModel( *m_stModel, cfgSymbolH, m_variableDH );

    reset();
}

ConfigValidatorMgr::~ConfigValidatorMgr()
{
    //dtor
}

void ConfigValidatorMgr::registerVemsMessageViewer(VemsMessageViewerBase* vemsMessageViewer){
    m_viewerA.Add(vemsMessageViewer);
}

void ConfigValidatorMgr::checkInvalidBits( StringHash& symbols )
{
    m_rpn.setShowExceptions( false );

    SymbolTable* symTable = &m_mtconfigData->getSymbolTable();

    SymbolTable::iterator it = symTable->begin();

    m_rpn.setVariableData( m_stModel );

    for ( ; it != symTable->end(); ++it )
    {
        Symbol* sym = it->second;
        if ( sym != NULL )
        {
            if ( sym->isBits() )
            {
                if ( !sym->getValidationExpr().IsEmpty() && ( m_rpn.calculate( sym->getValidationExpr() ) == 0.0 ) )
                    continue;

                wxArrayString bitFields = sym->getBits();

                unsigned int bitlength = ( 1 << (sym->getBitHi() - sym->getBitLo() + 1) );
                if ( bitlength > bitFields.GetCount() )
                {
                    for ( size_t i = bitFields.GetCount(); i < bitlength; ++i )
                        bitFields.Add( wxString::Format(_T("%d"), i ) );
                }

                int symVal = (int)(m_stModel->getViewerData( *sym ) );

                wxString varname = sym->getName();
                varname.Replace(_T("["),_T(""));
                varname.Replace(_T("]"),_T(""));

                if ( symVal >= (int)bitFields.GetCount() )
                {
                    symbols[ varname ] = wxString::Format( _T("Warning: allowed range is 0 - %d"), bitlength );
                }
                else if ( bitFields[symVal] == _T("INVALID") )
                {
                    symbols[ varname ] = _T("Error: INVALID value");
                    m_errorCount++;
                }
            }
        }
    }
}

void ConfigValidatorMgr::validate( StringHash& symbols )
{
    reset();
    m_rpn.setShowExceptions( false );
    m_rpn.setVariableData( &m_variableDH );

    //m_criteriaTypeA.Clear(); //?should be cleared
    m_criteriaRangeIdxA.Clear();

    setVariableData( m_variableDH, m_criteriaData->variableHash );

    ConfigCriteriaArray& criteriaA =  m_criteriaData->criteriaArray;

    int resultType, rangeIdx;

    for ( size_t i = 0; i < criteriaA.GetCount(); ++i )
    {
        const ConfigCriteria* criteria = criteriaA[i];

        ConfigCriteria::ConfigCriteriaType type = criteria->getType();

        if ( type == ConfigCriteria::CC_RULE_FIRST_MATCH )
        {
            wxString expression = validate( *criteria, resultType, rangeIdx );
            m_criteriaTypeA.Add( resultType );
            m_criteriaRangeIdxA.Add( rangeIdx );

            wxString msg = wxEmptyString;

            switch ( resultType )
            {
                case CCR_TYPE_WARNING:
                    msg = _T("Warning: ") + criteria->getHtmlErrorMsg();
                    break;

                case CCR_TYPE_ERROR:
                    msg = _T("Error: ") + criteria->getHtmlErrorMsg();
                    break;

                default:
                    break;
            }

            if ( !msg.IsEmpty() )
            {
                const ConfigRangeArray& rangeArray = criteria->getRangeArray();
                ConfigCriteriaRange* r = rangeArray[rangeIdx];
                wxArrayString variablesA = r->getVariables();

                if ( variablesA.GetCount() == 0 )
                    variablesA = m_rpn.getVariablesArray( expression );
                for ( size_t j = 0; j < variablesA.GetCount(); ++j )
                {
                    wxString varname = variablesA[j];
                    varname.Replace(_T("["),_T(""));
                    varname.Replace(_T("]"),_T(""));
                    symbols[varname] = msg;
                }
            }
        }
        else
        {
            IntArray resultA;
            validate( *criteria, resultA );

            const ConfigRangeArray& rangeArray = criteria->getRangeArray();
            for( size_t j = 0; j < resultA.GetCount(); j++ )
            {
                int result = resultA.Item(i);
                const ConfigCriteriaRange* range = rangeArray.Item( j );
                int itemType = range->getType();
                if ( ( result == CCR_VALUE_IN_RANGE ) && ( itemType == CCR_TYPE_ERROR ) )
                {
                    wxString msg = wxEmptyString;

                    switch ( itemType )
                    {
                        case CCR_TYPE_WARNING:
                            msg = _T("Warning: ") + criteria->getWarningMsg();
                            break;

                        case CCR_TYPE_ERROR:
                            msg = _T("Error: ") + criteria->getErrorMsg();
                            break;

                        default:
                            break;
                    }

                    if ( !msg.IsEmpty() )
                    {
                        wxArrayString variablesA;

                        if ( range->getVariables().GetCount() > 0 )
                            variablesA = range->getVariables();
                        else
                            variablesA = m_rpn.getVariablesArray( range->getExpression() );

                        for ( size_t j = 0; j < variablesA.GetCount(); ++j )
                        {
                            wxString varname = variablesA[j];
                            varname.Replace(_T("["),_T(""));
                            varname.Replace(_T("]"),_T(""));
                            symbols[varname] = msg;
                        }
                    }
                }
            }
        }
    }
}

void ConfigValidatorMgr::validate(VariableDataDH& data)
{
    reset();
    m_rpn.setVariableData(&data);
    setVariableData(data, m_criteriaData->variableHash);
    ConfigCriteriaArray& criteriaA =  m_criteriaData->criteriaArray;
    //validate data
    int resultType, rangeIdx;
    for ( size_t i = 0; i < criteriaA.GetCount(); ++i ){
        const ConfigCriteria* criteria = criteriaA[ i ];
        ConfigCriteria::ConfigCriteriaType type = criteria->getType();
        if(type == ConfigCriteria::CC_RULE_FIRST_MATCH){
            wxString expression = validate(*criteria, resultType, rangeIdx);
            m_criteriaTypeA.Add(resultType);
            m_criteriaRangeIdxA.Add(rangeIdx);
            show(criteria, resultType, expression);
        }else{ /// ALL_ERROR
            IntArray resultA;
            validate(*criteria, resultA);
            show(criteria, resultA);
        }
        Logger::logf( LOG_MSG,_T("ConfigValidatorMgr::show():2 errorCount:%d"), m_errorCount);
    }

    int rulesCount  = m_okCount+ m_warningCount+ m_errorCount;
    wxString msg = wxString::Format(_T("Number of rules: %d, "),  rulesCount);
    msg += wxString::Format(_T("Number of passed rules: %d, "), m_okCount);
    msg += wxString::Format(_T("Number of warning: %d, "), m_warningCount);
    msg += wxString::Format(_T("Number of error: %d"), m_errorCount);
/*
    wxString htmlMsg = wxString::Format(_T("Number of rules: %d<br>"), criteriaA.GetCount() );
    htmlMsg += wxString::Format(_T("Number of passed rules: %d<br>"), m_okCount);
    htmlMsg += wxString::Format(_T("Number of warning: %d<br>"), m_warningCount);
    htmlMsg += wxString::Format(_T("Number of error: %d<br>"), m_errorCount);
*/
    wxString htmlMsg = wxString::Format(_T("<table border=\"1px\"><tr><td>Number of rules: </td><td>%d</td></tr>"), rulesCount );
    htmlMsg += wxString::Format(_T("<tr><td>Number of passed rules: </td><td>%d</td></tr>"), m_okCount);
    htmlMsg += wxString::Format(_T("<tr><td>Number of warning: </td><td>%d</td></tr>"), m_warningCount);
    htmlMsg += wxString::Format(_T("<tr><td>Number of error: </td><td>%d</td></tr></table>"), m_errorCount);


 //   VemsMessage vemsMsg1(VMT_INFO, _T("ok"), msg1);
 //   show(vemsMsg1);

 //   wxString msg2 = wxString::Format(_T("Number of warning found in config: %d"), m_warningCount);
 //   VemsMessage vemsMsg2(VMT_INFO, _T("warning"), msg2);
 //   show(vemsMsg2);

//    wxString msg3 = wxString::Format(_T("Number of error found in config: %d"), m_errorCount);
    VemsMessage vemsMsg3(VMT_INFO, _T("Validation summary"), msg, htmlMsg);
    show(vemsMsg3);
}

void ConfigValidatorMgr::checkInvalidBits( STDataModelBase * stdata_model )
{
    STDataModelBase* model = stdata_model;

    if ( model == NULL )
        model = m_stModel;
    SymbolTable* symTable = &m_mtconfigData->getSymbolTable();

    SymbolTable::iterator it = symTable->begin();

    wxString htmlMsg = _T("<font size='+2'><b>Following symbols have invalid values</b></font><br><br>");

    wxString htmlMsg2 = _T("<font size='+2'><b>None of the following symbols are used in any dialog</b></font><br><br>");

    bool foundUnused = false;
    bool foundInvalid = false;
    wxString warnings = wxEmptyString;
    m_rpn.setVariableData( model );

    for ( ; it != symTable->end(); ++it )
    {
        Symbol* sym = it->second;
        if ( sym != NULL )
        {
            if ( sym->isBits() )
            {
                wxArrayString bitFields = sym->getBits();
                size_t blen = bitFields.GetCount();

                unsigned int bitlength = ( 1 << (sym->getBitHi() - sym->getBitLo() + 1) );
                if ( bitlength > blen )
                {

                    warnings += _T("Some possible values of the <b>") + it->first + _T("</b> symbol are not enumerated in the INI file");
                    warnings += wxString::Format( _T("<br>It should contain %d values but contains only %d"), bitlength, blen );
                    warnings += _T("<br><br>");

                    for ( size_t i = bitFields.GetCount(); i < bitlength; ++i )
                        bitFields.Add( wxString::Format(_T("%d"), i ) );
                }

                if ( !sym->getValidationExpr().IsEmpty() && ( m_rpn.calculate( sym->getValidationExpr() ) == 0.0 ) )
                {
                    #ifdef VEMSDEBUG
                        LOGP( _T("Symbol") + sym->getName() + _T(" is disabled because of the validation expression"));
                    #endif
                    continue;
                }

                int symVal = (int)(model->getViewerData( *sym ) );

                blen = bitFields.GetCount();

                if ( symVal >= (int)blen )
                {
                    // is it used?
                    const StringSet* dlgs_using_symbol = m_mtconfigData->getDialogNamesBySymbol( *sym );
                    if ( dlgs_using_symbol == NULL )
                    {
                        foundUnused = true;
                        htmlMsg2 += _T("Unused symbol: <b>") + it->first + wxString::Format( _T("</b>, value of the symbol is %d<br><br>"), symVal );
                        continue;
                    }

                    htmlMsg += _T("<b>") + it->first + _T("</b> is not in the allowed range<br>");

                    /*if ( limitmax >= 0 )
                        htmlMsg += wxString::Format( _T("Current value is %d, allowed range is: 0 - %d <br>"), symVal, bitlength );
                    else
                        htmlMsg += wxString::Format( _T("Current value is %d, allowed range is: 0 - %d <br>"), symVal, bitlength );*/

                    htmlMsg += wxString::Format( _T("Current value is %d, allowed range is: 0 - %d <br>"), symVal, bitlength - 1 );


                    StringSet::const_iterator sit = dlgs_using_symbol->begin();
                    wxString dlgs = _T("Following dialogs are usings this symbol: ");
                    for ( ; sit != dlgs_using_symbol->end(); ++sit )
                    {
                        wxString dlgId = *sit;
                        wxString dialogName = m_mtconfigData->getMenuAndDialogNameByDialogId( dlgId );

                        dlgs += MenuLinkHelper::getLink( dlgId, dialogName, sym->getName() ) + _T(", ");
                    }
                    dlgs = dlgs.RemoveLast( 2 );

                    htmlMsg += dlgs;

                    htmlMsg += _T("<br><br>");

                    foundInvalid = true;
                }
                else if ( bitFields[symVal] == _T("INVALID") )
                {
                    m_errorCount++;

                    htmlMsg += _T("<b>") + it->first + wxString::Format( _T("</b> has INVALID value (%d) <br>"), symVal );
                    foundInvalid = true;

                    wxString dlgs;
                    const StringSet* dlgs_using_symbol = m_mtconfigData->getDialogNamesBySymbol( *sym );
                    if ( dlgs_using_symbol != NULL )
                    {
                        StringSet::const_iterator sit = dlgs_using_symbol->begin();
                        dlgs = _T("Following dialogs are usings this symbol: ");
                        for ( ; sit != dlgs_using_symbol->end(); ++sit )
                        {
                            wxString dlgId = *sit;
                            wxString dialogName = m_mtconfigData->getMenuAndDialogNameByDialogId( dlgId );

                            dlgs += MenuLinkHelper::getLink( dlgId, dialogName, sym->getName() ) + _T(", ");
                        }
                        dlgs = dlgs.RemoveLast( 2 );
                    }
                    else {
                        dlgs = _T("This symbol is not used by any dialog");
                    }

                    htmlMsg += dlgs + _T("<br><br>");
                }
            }
        }
    }

    if ( foundInvalid )
    {
        VemsMessage vemsMsg3( VMT_ERROR, _T("Invalid values"), _T("Invalid values"), htmlMsg );
        show( vemsMsg3 );
    }

#ifdef VEMSDEBUG
    if ( foundUnused )
    {
        VemsMessage vemsMsg3( VMT_WARNING, _T("Symbol bits check"), _T("Symbol bits check"), htmlMsg2 + warnings + _T("<br><br>"));
        show( vemsMsg3 );
    }
    if ( warnings != _T("") )
    {
        warnings.Prepend( _T("<font size='+2'><b>Check INI file</b></font><br><br>") );
        VemsMessage vemsMsg3( VMT_WARNING, _T("Symbol bits check"), _T("Symbol bits check"), warnings + _T("<br><br>"));
        show( vemsMsg3 );
    }
#endif
}

void ConfigValidatorMgr::checkChannelCollisions(STDataModelBase * stdata_model)
{
    STDataModelBase* model = stdata_model;

    if ( model == NULL )
        model = m_stModel;

    ChannelCollision cc;
    wxArrayString collisions = cc.detectCollisions(model, &m_mtconfigData->getSymbolTable() );

    m_channelCollisionCount = (int)collisions.GetCount();

    if ( ! collisions.GetCount() )
        return;

    wxString htmlMsg = ChannelValidationHelper::createCollisionHtmlMessage( collisions );

    VemsMessage vemsMsg3(VMT_ERROR, _T("Channel collisions"), wxString::Format(_T("%d channels are in conflicting state"), collisions.GetCount()), htmlMsg);
    m_errorCount++;
    show(vemsMsg3);
}

void ConfigValidatorMgr::checkInvalidChannels(STDataModelBase * stdata_model)
{
    STDataModelBase* model = stdata_model;

    if ( model == NULL )
        model = m_stModel;

    ChannelCollision cc;
    wxArrayString invalids = cc.detectInvalidChannels(model, &m_mtconfigData->getSymbolTable() );

    if ( ! invalids.GetCount() )
        return;

    m_invalidChannelCount = (int)invalids.GetCount();

    wxString htmlMsg = _T("<font size='+2'><b>Following channels are invalid:</b></font><br><br>");

    wxString target = _T("");

    for ( size_t i = 0; i < invalids.GetCount(); ++i )
    {
        wxString channel = invalids[i].BeforeFirst('|');

        wxString symName = invalids[i].AfterFirst('|');
        if ( symName != _T("") )
            target +=  symName + _T(",");

        htmlMsg += _T("<b>") + channel + _T("</b><br>");
    }

    if ( target.IsEmpty() == false )
        target.RemoveLast( 1 );

    htmlMsg += _T("Check <a href='outputChannels' target='") + target + _T("'>Output channels dialog</a><br>");

    VemsMessage vemsMsg3(VMT_ERROR, _T("Invalid channels"), wxString::Format(_T("%d channels are not set correctly"), invalids.GetCount()), htmlMsg);
    m_errorCount++;
    show(vemsMsg3);
}

void ConfigValidatorMgr::setVariableData(VariableDataDH& vd, SymbolTable& variableHash)
{
    SymbolTable::const_iterator IT = variableHash.begin();
    for ( ; IT != variableHash.end(); ++IT )
    {
        Symbol* symbol = IT->second;
        double value = m_rpn.calculate( symbol->getExpr() );
        vd.addValue( IT->first, value );
    }
}



wxString ConfigValidatorMgr::validate(const ConfigCriteria& criteria, int& resultType, int& rangeIdx)
{
    const ConfigRangeArray& rangeArray = criteria.getRangeArray();
    for( size_t i =0; i < rangeArray.GetCount(); i++ )
    {
        const ConfigCriteriaRange* range = rangeArray.Item( i );
        int result = validate(*range);
        if ( result == CCR_VALUE_IN_RANGE )
        {
            resultType = range->getType();
            rangeIdx= i;
            return range->getExpression();
        }
    }
    resultType = -1;
    rangeIdx = -1;

    return wxEmptyString;
}

void ConfigValidatorMgr::validate(const ConfigCriteria& criteria, IntArray& resultA)
{
    const ConfigRangeArray& rangeArray = criteria.getRangeArray();
    for( size_t i =0; i < rangeArray.GetCount(); i++ )
    {
        const ConfigCriteriaRange* range = rangeArray.Item( i );
        int itemType = range->getType();
        int result = validate(*range);
        if( (result == CCR_VALUE_IN_RANGE) && (itemType == CCR_TYPE_OK) ){
            return;
        }
        resultA.Add(result);
    }
}

int ConfigValidatorMgr::validate(const ConfigCriteriaRange& criteriaRange)
{
    //TODO catch rpn error
    wxString expression = criteriaRange.getExpression();
    double value = m_rpn.calculate( expression );
    //Logger::logf(LOG_MSG, _T("ConfigValidatorMgr::validate:%f ")+expression, value);
    if( value > 0.0 )
    {
         return CCR_VALUE_IN_RANGE; /// true
    }
    return CCR_VALUE_OUT_OF_RANGE; /// false
}

void ConfigValidatorMgr::show( const ConfigCriteria* criteria, const IntArray& resultA)
{
    const ConfigRangeArray& rangeArray = criteria->getRangeArray();
    for( size_t i =0; i < resultA.GetCount(); i++ )
    {
        int result = resultA.Item(i);
        const ConfigCriteriaRange* range = rangeArray.Item( i );
        int itemType = range->getType();
        /// show all inrange(true) error items
        if( (result == CCR_VALUE_IN_RANGE) && (itemType == CCR_TYPE_ERROR) ){
            show(criteria, itemType, range->getExpression() );
        }
    }
}

void ConfigValidatorMgr::show( const ConfigCriteria* criteria, int type, const wxString& expression)
{
    wxString name = criteria->getName();
    wxString msg;
    //wxString htmlMsg = criteria->getHtmlMsg();
    VemsMessageType msgType;
    wxString htmlMsg;
    switch ( type )
    {
        case CCR_TYPE_OK :
            m_okCount++;
            msgType = VMT_INFO;
            return;
            //break;
        case CCR_TYPE_WARNING:
            //strLine = _T("WARINING!:");

            msg = criteria->getWarningMsg();
            htmlMsg = criteria->getHtmlErrorMsg();
            m_warningCount++;
            msgType = VMT_WARNING;
            break;
        case CCR_TYPE_ERROR:
            msg = criteria->getErrorMsg();
            htmlMsg = criteria->getHtmlErrorMsg();
            m_errorCount++;
            msgType = VMT_ERROR;
            break;
        default:
            Logger::logf( LOG_MSG, criteria->getName() + _T(": matching rule not found!") );
            m_okCount++; /// ???
            return;
    }
    wxArrayString variablesA = m_rpn.getVariablesArray(expression);
    StringHash paramH;
    setParam(variablesA, paramH);

    Logger::logf( LOG_MSG,_T("ConfigValidatorMgr::show(): errorCount:%d"), m_errorCount);
    Logger::logf( LOG_MSG, criteria->getName() + _T(" msg:")+msg );

    wxString htmlMenu = MenuLinkHelper::getHtmlFromExpression(*m_mtconfigData, expression);
    paramH[_T("MENU_LINK")] = htmlMenu;
    //htmlMsg += htmlMenu;
    ///
    VemsMessage vemsMsg(msgType, name, msg, htmlMsg);
    vemsMsg.parseMessageTemplates(paramH);
    vemsMsg.parseHtmlTemplates(&paramH);
    show(vemsMsg);
}

void ConfigValidatorMgr::show(const VemsMessage& vemsMsg)
{
    for(size_t i = 0; i < m_viewerA.GetCount(); ++i){
        VemsMessageViewerBase* viewer = m_viewerA.Item(i);
        viewer->show(vemsMsg);
    }
}

void ConfigValidatorMgr::reset()
{
    m_okCount = 0;
    m_warningCount = 0;
    m_errorCount =0;
    m_invalidChannelCount = 0;
    m_channelCollisionCount = 0;
}


int ConfigValidatorMgr::compareInts(int* n1, int* n2)
{
    return *n1 - *n2;
}

void ConfigValidatorMgr::setParam(const wxArrayString& variablesA, StringHash& paramH)
 {
    for(size_t i = 0; i < variablesA.GetCount(); ++i )
    {
        wxString key = wxString::Format(_T("%d"),i+1);
        paramH[key] = variablesA.Item(i);
         //Logger::logf( LOG_MSG, _T("ConfigValidatorMgr::setParam():")+ key+_T(":")+  variablesA.Item(i) );
    }
 }

 /*
 void ConfigValidatorMgr::setParam(const StringHashSet& variablesSet, const wxString& expression, StringHash& paramH)
 {
    Int2StringHash int2String;
    StringHashSet::const_iterator IT = variablesSet.begin();
    for(; IT != variablesSet.end(); ++IT )
    {
        wxString variable = *IT;
        int idx = expression.Find(variable);
        int2String[idx] = variable;
    }

    Logger::logf( LOG_MSG, _T("ConfigValidatorMgr::setParam() 10"));

    //SortedIntArray  sortedIntA( &(ConfigValidatorMgr::compareInts) );
    IntArray sortedIntA;
    Int2StringHash::const_iterator it = int2String.begin();
    for(; it != int2String.end(); ++it )
    {
       int idx = it->first;
       Logger::logf( LOG_MSG, _T("ConfigValidatorMgr::setParam() 20"));
       sortedIntA.Add(idx);
    }

    Logger::logf( LOG_MSG, _T("ConfigValidatorMgr::setParam() 30"));
    for(int i = 0; i < sortedIntA.GetCount(); ++i){
        int idx = sortedIntA.Item(i);
        wxString variable = int2String[idx];
        wxString key = wxString::Format(_T("$%d"),i);
        paramH[key] = variable;
    }
    sortedIntA.Sort( (ConfigValidatorMgr::compareInts) );

 }
*/


ConfigValidatorMgr::ButtonState ConfigValidatorMgr::getButtonState() const
{
    LOGP( wxString::Format( _T("m_errorCount: %d, m_warningCount: %d"), m_errorCount, m_warningCount ) );

    if ( ( m_errorCount + m_warningCount ) == 0 ) {
        return CFV_GREEN;
    } else if ( m_errorCount > 1 ) {
        return CFV_RED_BLINKING;
    } else if ( m_errorCount >  0 ) {
        return CFV_RED;
    } else if ( m_warningCount > 3 ) {
        return CFV_YELLOW_BLINKING;
    } else if ( m_warningCount > 2 ) {
        return CFV_YELLOW;
    } else {
        return CFV_GREEN;
    }
}
