#include "FirmwareConfigData.h"

#include <wx/textfile.h>

#include "FirmwareConfigDifferences.h"
#include "../Math.h"
#include "../Symbol.h"
#include "../SymbolMgr.h"
#include "../logger.h"

#include "../common/VemsMatrixHelper.h"


FirmwareConfigData::~FirmwareConfigData()
{
    clear();
}

bool FirmwareConfigData::is12x12()
{
    VemsMatrix<uint8_t>* matrix = getTable(_T("j"));
    if(!matrix){
        Logger::logf( LOG_MSG, _T("FirmwareConfigData::is12x12() - j table not found!") );
        return true; /// default
    }
    if ( matrix->cols() == 12 ) {
        Logger::logf( LOG_MSG, _T("FirmwareConfigData::is12x12() - j table size 12x12!") );
        return true;
    }
    return false;
}

bool FirmwareConfigData::isHighBoost() const
{
    String2IntHash::const_iterator IT = m_scalarValueH.find( _T("kpa_conf") );

    //default: highboost
    if ( IT == m_scalarValueH.end() ) { return true; }
    //found highboost
    if ( IT->second == 34 ) { return true; }

    return false;
}

const BoolHash& FirmwareConfigData::getDetectedParams()
{
    m_boolHash.clear();
    if( is12x12() ){
        //m_boolHash[_T("12x12")] = true;
        m_boolHash[_T("12X12")] = true;
    }else{
        //m_boolHash[_T("12x12")] = false;
        m_boolHash[_T("12X12")] = false;
    }

    if( isHighBoost() ){
        m_boolHash[_T("HIGHBOOST")] = true;
    }else{
        m_boolHash[_T("HIGHBOOST")] = false;
    }
    return m_boolHash;
}

void FirmwareConfigData::addScalar( const wxString& name, int value, int lineNumber )
{
    if( m_scalarValueH.find(name) == m_scalarValueH.end() ){
    m_scalarOrderA.Add( name );
        m_scalarLineH[ name ] = lineNumber;
    }
    m_scalarValueH[ name ] = value;

    //Logger::logf( LOG_MSG, _T("\tFirmwareConfigData::addScalar() : %d"), m_scalarOrderA.GetCount() );
}

bool FirmwareConfigData::getScalar(const wxString& name, int& value)
{
    if( m_scalarValueH.find(name) == m_scalarValueH.end() ) return false;
    value = m_scalarValueH[ name ];
    return true;
}

void FirmwareConfigData::addTableLineMSQ( const wxString& name, size_t rowIndex, const wxArrayString& line, int lineNumber )
{
    size_t uColCount = line.GetCount();
    VemsMatrixByteHash::const_iterator IT = m_tableHash.find( name );
    // create VemsMatrix
    if ( IT == m_tableHash.end() )
    {
        m_tableHash[ name ] = new VemsMatrix<uint8_t>( 1, uColCount, 0x00 );
        m_tableOrderA.Add( name );
    }
    // fill VemsMatrix
    VemsMatrix<uint8_t>* matrix = m_tableHash[ name ];
    matrix->setRowCount( rowIndex + 1 );

    for ( size_t i = 0; i < uColCount; ++i )
    {
        wxString valueStr = line.Item( i );
        long value;
        valueStr.ToLong(&value);
        matrix->set( rowIndex, i, value );
    }
}

void FirmwareConfigData::addTableLine( const wxString& name, size_t rowIndex, const wxArrayString& line, int lineNumber )
{
    size_t uColCount = line.GetCount();
    VemsMatrixByteHash::const_iterator IT = m_tableHash.find( name );
    // create VemsMatrix
    if ( IT == m_tableHash.end() )
    {
        m_tableHash[ name ] = new VemsMatrix<uint8_t>( 1, uColCount, 0x00 );
        m_tableOrderA.Add( name );
    }
    // fill VemsMatrix
    VemsMatrix<uint8_t>* matrix = m_tableHash[ name ];
    matrix->setRowCount( rowIndex + 1 );

    for ( size_t i = 0; i < uColCount; ++i )
    {
        wxString hexValue = line.Item( i ).Mid( 0, 2 );
        uint8_t value = Math::hex2int( hexValue );
        matrix->set( rowIndex, i, value );
    }
}

void FirmwareConfigData::filter(const FirmwareConfigData& cfgData)
{
    String2IntHash::const_iterator IT = cfgData.m_scalarValueH.begin();
    for(; IT != cfgData.m_scalarValueH.end(); ++IT){
        wxString name = IT->first;
        filterScalar( name );
    }
}

void FirmwareConfigData::filterScalar( const wxString& name)
{
    removeScalar(name);
    m_filteredScalarH.insert(name);
}

void FirmwareConfigData::remove(const FirmwareConfigData& cfgData)
{
    String2IntHash::const_iterator IT = cfgData.m_scalarValueH.begin();
    for(; IT != cfgData.m_scalarValueH.end(); ++IT){
        wxString name = IT->first;
        removeScalar( name );
    }

}

void FirmwareConfigData::removeScalar( const wxString& name )
{
    m_scalarValueH.erase(  name );
    m_scalarOrderA.Remove( name );
}

void FirmwareConfigData::removeTable( const wxString& name )
{
    m_tableHash.erase( name );
    m_tableOrderA.Remove( name );
}


void FirmwareConfigData::compareScalars(FirmwareConfigDifferences* fcDiff, const FirmwareConfigData& fcData)
{
     //search missing scalars or different values
    const String2IntHash& scalarValueH = fcData.getScalarValuesH();

    for( size_t i = 0; i < m_scalarOrderA.GetCount(); ++i )
    {
        wxString strScalarName = m_scalarOrderA[ i ];

        String2IntHash::const_iterator SVH_IN_IT = scalarValueH.find( strScalarName );

        if ( SVH_IN_IT == scalarValueH.end() ) {
            const StringSet& filteredScalarH = fcData.getFilteredScalarH();
            const StringSet::const_iterator IT_FILTERED = filteredScalarH.find(strScalarName);
            if(IT_FILTERED != filteredScalarH.end() ) {
                /// Intentionaly removed
                fcDiff->filteredScalars.Add( strScalarName );
            }else{
                fcDiff->missingScalars.Add( strScalarName );
            }
        }else
        {
            int valueThis   = m_scalarValueH[ strScalarName ];
            int valueIn     = SVH_IN_IT->second;
/*
            Logger::logf( LOG_MSG, _T("FirmwareConfigData::compare(): ") +
                          wxString::Format( strScalarName + _T(": ") + m_name +
                          _T(" %d; ") + fcData.getName() + _T(" %d"),
                          valueThis, valueIn ) );
*/
            if ( valueThis != valueIn )
            {
                fcDiff->differentScalars.Add( wxString::Format( strScalarName + _T(": ") + m_name +
                                                                    _T(" %d != ") + fcData.getName() + _T(" %d"),
                                                                    valueThis, valueIn ) );
            }
        }
    }

    //search extra scalars
    const wxArrayString& SOA_IN = fcData.getScalarOrderA();

    for ( size_t i = 0; i < SOA_IN.GetCount(); ++i )
    {
        wxString strScalarName = SOA_IN[ i ];

        if ( m_scalarValueH.find( strScalarName ) == m_scalarValueH.end() )
        {
            fcDiff->extraScalars.Add( strScalarName );
        }
    }
}

void FirmwareConfigData::compareTables( const wxString& configName,
                                        const wxString& tableName,
                                        VemsMatrix<uint8_t>* matrix1,
                                        VemsMatrix<uint8_t>* matrix2,
                                        wxArrayString& differentTables )
{
    if ( ( matrix1->rows() != matrix2->rows() ) || ( matrix1->cols() != matrix2->cols() ) ) {

        differentTables.Add(    wxString::Format(   tableName + _T(" [%2dx%2d] != [%2dx%2d]"),
                                matrix1->rows(), matrix1->cols(), matrix2->rows(), matrix2->cols() ) );

        return;
    }

    for ( size_t i = 0; i < matrix1->size(); ++i ) {
        if ( matrix1->get(i) != matrix2->get(i) ) {
            differentTables.Add( wxString::Format( tableName + _T("[%i]: ") + m_name +
                                                    _T(" %u != ") + configName + _T(" %u"),
                                                    i, matrix1->get(i), matrix2->get(i) ) );
        }
    }
}

void FirmwareConfigData::compareTables(FirmwareConfigDifferences* fcDiff, const FirmwareConfigData& fcData)
{

    for ( size_t i = 0; i < m_tableOrderA.GetCount(); ++i ){
        wxString tableName = m_tableOrderA[ i ];
        VemsMatrix<uint8_t>* matrix2 = fcData.getTable(tableName);
        if(!matrix2){
            fcDiff->missingTables.Add( tableName );
            continue;
        }
        VemsMatrix<uint8_t>* matrix = getTable(tableName);
        if(!matrix){
            /// IMPOSSIBLE
            continue;
        }

        compareTables( fcData.getName(), tableName, matrix, matrix2, fcDiff->differentTables );
    }


    //search extra tables
    const wxArrayString& TOA_IN = fcData.getTableOrderA();
    for ( size_t i = 0; i < TOA_IN.GetCount(); ++i ){
        wxString tableName = TOA_IN[ i ];
        VemsMatrix<uint8_t>* matrix = getTable(tableName);
        if(!matrix){
            fcDiff->extraTables.Add( tableName );
        }
    }
}

FirmwareConfigDifferences* FirmwareConfigData::compare(const FirmwareConfigData& fcData )
{
    FirmwareConfigDifferences* fcDiff = new FirmwareConfigDifferences();

    compareScalars(fcDiff, fcData);
    compareTables(fcDiff, fcData);

    return fcDiff;
}

FirmwareConfigDifferences* FirmwareConfigData::compare( const SymbolTable& sTable )
{
    FirmwareConfigDifferences* fcDiff = new FirmwareConfigDifferences();

    searchMissingScalars( sTable, fcDiff );
    searchExtraScalars( sTable, fcDiff );
    searchMissingTables( sTable, fcDiff );
    searchExtraTables( sTable, fcDiff);

    return fcDiff;
}

void FirmwareConfigData::searchMissingScalars( const SymbolTable& sTable, FirmwareConfigDifferences* fcDiff )
{
    SymbolArray symbolA;

    SymbolTable::const_iterator IT_ST = sTable.begin();

    for ( ; IT_ST != sTable.end(); ++IT_ST )
    {
        Symbol* ptrSymbol = IT_ST->second;
        symbolA.Add( ptrSymbol );
    }

    symbolA.Sort(& SymbolMgr::compareByPageAndOffset );

    for ( size_t i = 0; i < symbolA.GetCount(); ++i )
    {
        Symbol* ptrSymbol       = symbolA[ i ];
        wxString strSymbolName  = ptrSymbol->getName();

        if ( ptrSymbol->isScalar() &&
             m_scalarValueH.find( strSymbolName ) == m_scalarValueH.end() )
        {
            fcDiff->missingScalars.Add( strSymbolName );
        }
    }

/*
    SymbolTable::const_iterator IT_ST = sTable.begin();

    for ( ; IT_ST != sTable.end(); ++IT_ST )
    {
        wxString strSymbolName  = IT_ST->first;
        Symbol* ptrSymbol       = IT_ST->second;

        if ( ptrSymbol->isScalar() &&
             m_scalarValueH.find( strSymbolName ) == m_scalarValueH.end() )
        {
            fcDiff->addMissingScalar( strSymbolName );
        }
    }
*/
}

void FirmwareConfigData::searchMissingTables( const SymbolTable& sTable, FirmwareConfigDifferences* fcDiff )
{
    SymbolTable::const_iterator IT_ST = sTable.begin();

    for ( ; IT_ST != sTable.end(); ++IT_ST ){
        wxString strSymbolName  = IT_ST->first;
        Symbol* ptrSymbol       = IT_ST->second;

        if ( ptrSymbol->isArray() && !getTable(strSymbolName) ){
            fcDiff->missingTables.Add( strSymbolName );
        }
    }
}

void FirmwareConfigData::searchExtraScalars( const SymbolTable& sTable, FirmwareConfigDifferences* fcDiff )
{
    //search extra scalars
    for ( size_t i = 0; i < m_scalarOrderA.GetCount(); ++i )
    {
        wxString strScalarName = m_scalarOrderA[ i ];

        if ( sTable.find( strScalarName ) == sTable.end() )
        {
            fcDiff->extraScalars.Add( strScalarName );
        }
    }
}

void FirmwareConfigData::searchExtraTables( const SymbolTable& sTable, FirmwareConfigDifferences* fcDiff )
{
    //search extra tables
    for ( size_t i = 0; i < m_tableOrderA.GetCount(); ++i )
    {
        wxString strTableName = m_tableOrderA[ i ];

        if ( sTable.find( strTableName ) == sTable.end() )
        {
            fcDiff->extraTables.Add( strTableName );
        }
    }
}

//DEBUG
void FirmwareConfigData::printScalarsToFile( const wxString& fileName )
{
    wxTextFile file( fileName );

    if ( file.Exists() )
    {
        file.Open();
        file.Clear();
    }
    else { file.Create(); }

    for ( size_t i = 0; i < m_scalarOrderA.Count(); ++i )
    {
        wxString strName    = m_scalarOrderA[ i ];
        int iValue = m_scalarValueH[ strName ];
        wxString strValue = wxString::Format( _T("%02X"), iValue );
        wxString strLine = strName + _T("=") + strValue;

        file.AddLine( strLine );
        file.Write();
    }

    file.Close();
}

void FirmwareConfigData::printTablesToFile( const wxString& fileName )
{
    wxTextFile file( fileName );
    file.Create();

    for ( size_t i = 0; i < m_tableOrderA.Count(); ++i )
    {
        wxString strName        = m_tableOrderA[ i ];
        VemsMatrix<uint8_t>*matrix   = getTable( strName );
        size_t uColCount        = matrix->cols();
        size_t uRowCount        = matrix->rows();

        for ( size_t r = 0; r < uRowCount; ++r )
        {
            wxString strTableName = strName + _T("[") + wxString::Format( _T("%X"), r ) + _T("]");
            wxString strTableLine = wxEmptyString;

            for( size_t c = 0; c < uColCount; ++c )
            {
                uint8_t uData = matrix->get(r,c);
                strTableLine += wxString::Format( _T("%02X "), uData );
            }
            strTableLine.RemoveLast();
            file.AddLine( strTableName + _T("=") + strTableLine );
            file.Write();
        }
    }

    file.Close();
}


wxString FirmwareConfigData::tableDataToString()
{
    return tableDataToString( _T("\n"), _T("\t") );
}

wxString FirmwareConfigData::tableDataToHtml()
{
    return tableDataToString( _T("<br>"), _T("&nbsp;&nbsp;&nbsp;&nbsp;") );
}

wxString FirmwareConfigData::tableDataToString( const wxString& eol, const wxString& tab )
{
    wxString result = _T("FirmwareConfigData - Parsed tables info") + eol;

    for ( size_t i = 0; i < m_tableOrderA.Count(); ++i )
    {
        wxString strName        = m_tableOrderA[ i ];
        VemsMatrix<uint8_t>* matrix = getTable(strName);

        size_t uColCount        = matrix->cols();
        size_t uRowCount        = matrix->rows();
        size_t uTableSize       =  matrix->size();

        result.Append( wxString::Format( tab + strName + _T(": %ux%u; size: %u") + eol,
                       uColCount, uRowCount, uTableSize ) );
    }

    return result;
}

wxString FirmwareConfigData::scalarDataToString()
{
    return scalarDataToString( _T("\n"), _T("\t") );
}

wxString FirmwareConfigData::scalarDataToHtml()
{
    return scalarDataToString( _T("<br>"), _T("&nbsp;&nbsp;&nbsp;&nbsp;") );
}

wxString FirmwareConfigData::scalarDataToString( const wxString& eol, const wxString& tab )
{
    wxString result = _T("FirmwareConfigData - Parsed scalar info") + eol;
    result.Append( wxString::Format( _T("\t parse %d scalar") + eol, m_scalarOrderA.GetCount() ) );

    for ( size_t i = 0; i < m_scalarOrderA.GetCount(); ++i )
    {
        wxString strName    = m_scalarOrderA[ i ];
        int iValue = m_scalarValueH[ strName ];
        wxString strValue = wxString::Format( _T("%02X"), iValue );
        result.Append( tab + strName + _T("=") + strValue + eol );
    }
    return result;
}


void FirmwareConfigData::clear()
{
    m_configLineCount = 0;
    WX_CLEAR_HASH_MAP( String2CharRefHash, m_tableDataH );
    m_tableSizeH.clear();
    m_tableColCountH.clear();
    m_tableStartLineNo.clear();
    m_scalarValueH.clear();
    m_scalarLineH.clear();
    m_scalarOrderA.Clear();
    m_tableOrderA.Clear();
    m_name.Clear();
    m_boolHash.clear();
    WX_CLEAR_HASH_MAP( VemsMatrixByteHash, m_tableHash );
    m_strFWVersion.Clear();
    WX_CLEAR_HASH_MAP( VemsMatrixByteHash, m_ConvertedTableH );
    m_XaxisNameH.clear();
    m_YaxisNameH.clear();
    WX_CLEAR_HASH_MAP( SymbolTable, m_tableSymbolH );
    m_filteredScalarH.clear();
}

VemsMatrix<uint8_t>* FirmwareConfigData::getTable( const wxString& name ) const
{
    VemsMatrixByteHash::const_iterator IT = m_tableHash.find( name );
    if ( IT == m_tableHash.end() ) { return NULL; }
    return IT->second;
}

void FirmwareConfigData::fillVariableData(VariableDataDH& vd)
{
    ///create and fill VariableDataDH with scalar values
    String2IntHash::const_iterator IT = m_scalarValueH.begin();
    for ( ; IT != m_scalarValueH.end(); ++IT )
    {
        vd.addValue( IT->first, IT->second);
    }
{
    //fill double hash with table data tablename[index]
    VemsMatrixByteHash::const_iterator IT = m_tableHash.begin();
    for(; IT != m_tableHash.end(); ++IT){
        wxString tableName = IT->first;
        VemsMatrix<uint8_t>* matrix = IT->second;
        for(size_t i = 0; i < matrix->size(); ++i){
            uint8_t value = matrix->get(i);
            wxString name = tableName + wxString::Format( _T("[%d]"), i );
            vd.addValue(name, value );
        }
    }
}
}

/*
int FirmwareConfigData::getScalarLineNumber( const wxString& name ) const
{
    String2IntHash::const_iterator it = m_scalarLineH.find( name );

    if ( it == m_scalarLineH.end() )
    {
        return -1;
    }

    return it->second;
}

*/
//int FirmwareConfigData::getTableStartLineNumber( const wxString& name ) const
//{
    /*
    String2IntHash::const_iterator it = m_tableStartLineNo( name );

    if ( it == m_tableStartLineNo.end() )
    {
        return -1;
    }

    return it->second;
*/
//	return 0;
//}

void FirmwareConfigData::tableDataToArrayString( wxArrayString& data )
{
    for ( size_t i = 0; i < m_tableOrderA.Count(); ++i )
    {
        wxString name        = m_tableOrderA[ i ];
        VemsMatrix<uint8_t>* matrix = getTable(name);

        size_t uColCount        = matrix->cols();
        size_t uRowCount        = matrix->rows();
        size_t uTableSize       = matrix->size();

        data.Add( wxString::Format( name + _T(": %ux%u; size: %u"), uColCount, uRowCount, uTableSize ) );
    }
}

void    FirmwareConfigData::addTable(const wxString& name, VemsMatrix<uint8_t>* table )
{
    uint8_t* data = table->getData();
    size_t rows = table->rows();
    size_t cols = table->cols();
    addTable(name, data, rows, cols);
}

void    FirmwareConfigData::addTable(const wxString& name, const uint8_t* data, size_t rows, size_t cols)
{
    //if found a new table, allocate memory for raw data
//    size_t size = rows * cols;
    VemsMatrix<uint8_t>* matrix = getTable(name);
    if ( matrix ) return;

    m_tableOrderA.Add( name );
    matrix = new VemsMatrix<uint8_t>( rows, cols, 0x00 );
    m_tableHash[ name ] =matrix;

    for(size_t i = 0; i < rows*cols; ++i){
        matrix->set(i, data[i] );
    }

/*
    // create VemsMatrix
    if ( m_tableHash.find( name ) == m_tableHash.end() )
    {
        m_tableHash[ name ] = new VemsMatrix<uint8_t>( rows, cols, 0x00 );
    }
*/
}

void FirmwareConfigData::setConvertedTable(const Symbol& symbol, const VemsMatrix<uint8_t>& table)
{
    Symbol* symbolStored = new Symbol(symbol);
    m_tableSymbolH[symbolStored->getName()] = symbolStored;

    VemsMatrix<uint8_t>* tableStored = new  VemsMatrix<uint8_t>(table);
    m_ConvertedTableH[symbolStored->getName()] = tableStored;
    Logger::logf( LOG_MSG, _T("\tFirmwareConfigData::setConvertedTable(): ")+ symbolStored->getName());
    VemsMatrixHelper::print(table);
    VemsMatrixHelper::print(*tableStored);



}

void FirmwareConfigData::setXaxisName(const wxString& tableName, const wxString& xAxisName)
{
    m_XaxisNameH[tableName] = xAxisName;
}

void FirmwareConfigData::setYaxisName(const wxString& tableName, const wxString& yAxisName)
{
    m_YaxisNameH[tableName] = yAxisName;
}

const VemsMatrix<uint8_t>* FirmwareConfigData::getConvertedTable(const wxString name) const
{
    VemsMatrixByteHash::const_iterator IT =  m_ConvertedTableH.find(name);
    if( IT == m_ConvertedTableH.end() ) return NULL;
    return IT->second;

}
wxString FirmwareConfigData::getXaxisName(const wxString& tableName) const
{
    StringHash::const_iterator IT = m_XaxisNameH.find(tableName);
    if(IT == m_XaxisNameH.end() ) return wxEmptyString;
    return IT->second;
}
wxString FirmwareConfigData::getYaxisName(const wxString& tableName) const
{
    StringHash::const_iterator IT = m_YaxisNameH.find(tableName);
    if(IT == m_YaxisNameH.end() ) return wxEmptyString;
    return IT->second;
}

bool FirmwareConfigData::isPartialCfg()
{
    bool isPartial = (m_configLineCount < 230) ? true : false;
    return isPartial;
}
