#include "FirmwareConfigLoader.h"

#include <wx/filename.h>

#include "../stdatamanager/STDataModelBase.h"
#include "../FirmwareInfo.h"
#include "../RawLogHelper.h"
#include "../fileparser/FirmwareConfigParser.h"
#include "../fileparser/ConfigParserMSQ.h"
#include "../fileparser/parseheader.h"
#include "../MTConfigHelper.h"
#include "../common/VemsMatrixHelper.h"
#include "../logger.h"


FirmwareConfigLoader::FirmwareConfigLoader()
{
    //ctor
//    m_fwInfo = fwInfo;
    m_currMTConfigData = NULL;
    m_fcData = new FirmwareConfigData();

    m_currFwInfo = &m_cfgFwInfo;
    m_currMTConfigData = &m_mtConfigData;

}

FirmwareConfigLoader::~FirmwareConfigLoader()
{
    delete m_fcData;
	WX_CLEAR_HASH_MAP( SymbolTable, m_cfgSymbolH );
}

/// Set in case of Upload to ECU (online)
void FirmwareConfigLoader::setDestinationFwInfo(FirmwareInfo* descFwInfo)
{
    m_currFwInfo = descFwInfo;
}

void FirmwareConfigLoader::setMTConfigData(MTConfigData* mtConfigData)
{
    m_currMTConfigData = mtConfigData;
}

bool FirmwareConfigLoader::parse(const wxArrayString& configFiles)
{
    if( isMSQ(configFiles) ){
        if( parseMSQ(configFiles) ){
             Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parse() msql parsed successfuly") );
        }else{
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parse() msql parse failed") );
            return false; ///
        }
    }else{
        if( parseCfg(configFiles) ){
            if( m_cfgFwInfo.getDevice() != m_currFwInfo->getDevice() ){
                /// If VT knows the device. Disable v3 -> round, round-> v3 config upload!
                return false;
            }
        }else{
            /// Seems vemscfg doesnt contain header
            /// So we dont know the device, version, etc
            /// So user sould decide what to do we dont want to disabe anything.
        }
    }

    parseHeader();
    parseIni();
    return true;

}

bool FirmwareConfigLoader::isMSQ(const wxArrayString& configFiles)
{
    if(configFiles.GetCount() == 0) return false;
    wxString fileName = configFiles.Item(0);
    wxFileName fn( fileName );
    if( fn.GetExt() == _T("msq") ) return true;
    return false;
}

bool FirmwareConfigLoader::parseMSQ(const wxArrayString& configFiles)
{
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parseMSQ() START") );
    //parse config files to configdata
    ConfigParserMSQ configParser( m_fcData, m_currFwInfo->isHighBoost() );
    bool result = false;
    for ( size_t i = 0; i < configFiles.Count(); ++i ){
         int res = configParser.parse( configFiles[ i ] );
         if(res == 0){
            result = true;
         }
    }
    m_cfgFwInfo = *m_currFwInfo;

    size_t configLineCount = m_fcData->m_configLineCount;
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parse() getConfigLineCount %i"), configLineCount);
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parse() header: ")+ m_cfgFwInfo.getHeaderFile() );

    return result;
}

bool FirmwareConfigLoader::parseCfg(const wxArrayString& configFiles)
{
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parse() START") );
    //parse config files to configdata
    FirmwareConfigParser configParser( m_fcData );

    for ( size_t i = 0; i < configFiles.Count(); ++i ){
        configParser.parse( configFiles[ i ] );
    }
    RawLogHelper::fillFirmwareInfo( m_cfgFwInfo, m_fcData->header);
    const BoolHash& detectedParams = m_fcData->getDetectedParams();
    m_cfgFwInfo.forceParams( detectedParams );

    size_t configLineCount = m_fcData->m_configLineCount;
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parse() getConfigLineCount %i"), configLineCount);
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parse() header: ")+ m_cfgFwInfo.getHeaderFile() );

    return m_cfgFwInfo.isValid();
}


void FirmwareConfigLoader::parseHeader()
{
    wxString headerFile = m_currFwInfo->getHeaderFile();
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parseHeader() header: ")+ headerFile);
    ///parse header to symboltable
	WX_CLEAR_HASH_MAP( SymbolTable, m_cfgSymbolH );
    ParseHeader headerParser( m_cfgSymbolH );
    headerParser.parse( headerFile );
}

bool FirmwareConfigLoader::parseIni()
{
    bool res = MTConfigHelper::parseMTConfigData(*m_currFwInfo, m_currMTConfigData);
    m_currFwInfo->forceParams( m_currMTConfigData->getParams() );
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::parseIni() res=%i"), res);
    if(!res){
        return res;
    }
    tableAxisHack();
    return res;
}

bool FirmwareConfigLoader::isAlienConfig()
{

    if( !(m_cfgFwInfo.isValid() && m_currFwInfo->isValid()) ){
        return true;
    }
    wxString cfgSerialNumber = m_cfgFwInfo.getSerialNumber();
    wxString currSerialNumber = m_currFwInfo->getSerialNumber();
    Logger::logf(LOG_MSG, _T("FirmwareConfigLoader::isAlienConfig() serialnum cfg: ")+cfgSerialNumber);
    Logger::logf(LOG_MSG, _T("FirmwareConfigLoader::isAlienConfig() serialnum ECU: ")+currSerialNumber);
    if(cfgSerialNumber.IsSameAs(currSerialNumber) ){
        Logger::logf(LOG_MSG, _T("FirmwareConfigLoader::isAlienConfig() serialnum is equal!"));
        return false;
     }
     return true;
}

void FirmwareConfigLoader::tableAxisHack()
{
    /// add tables to symboltable (it comes form the appropriate ini)
    SymbolTable& neededTables = m_currMTConfigData->getSymbolMgr().getConfigTableSymbols();
    SymbolTable::const_iterator IT = neededTables.begin();
    for ( ; IT != neededTables.end(); ++IT ){
        wxString name = IT->first;
        Symbol* neededSymbol = IT->second;
        m_cfgSymbolH[name] = new Symbol( *neededSymbol );

    /// fix missing table hack
        if( !name.Contains(_T("_")) ) continue;
        /// x,y bins
        VemsMatrix<uint8_t>* matrix = m_fcData->getTable(name);
        if(matrix) continue;
        /// not found
        const wxString mainTable = name.Before('_');
        const wxString axisName = name.After('_');
        if( ! hasRpmKpaHeader(mainTable) ) continue;
        /// has rpmkpa shared header
        matrix = m_fcData->getTable(axisName);
        if(!matrix) continue;
        /// found r or k
        addTable2Config(name, matrix, *m_fcData);

    }
}

void FirmwareConfigLoader::load(STDataModelBase* modelBase)
{
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::load()") );
    calculateTables();
    /// Load scalars
    loadScalars(modelBase);
    /// Tables
    loadTables(modelBase);
}


void FirmwareConfigLoader::loadScalars(STDataModelBase* modelBase)
{
    FirmwareConfigData& fcData = *m_fcData;
    const String2IntHash& scalarValueH = fcData.getScalarValuesH();
    SymbolTable::const_iterator IT = m_cfgSymbolH.begin();
    for ( ; IT != m_cfgSymbolH.end(); ++IT )
    {
        Symbol* ptrSymbol = IT->second;
        const wxString symbolName = ptrSymbol->getName();
        if ( ptrSymbol->isScalar() )
        {
            const int page      = ptrSymbol->getPage();
            const int offset    = ptrSymbol->getOffset();
            String2IntHash::const_iterator scalarValuesIT = scalarValueH.find( symbolName );

            if( scalarValuesIT == scalarValueH.end() ) {
                Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::loadScalars scalarValueH.find not found: ")+ symbolName );
                continue;
            }

            ptrSymbol->setValid( true );
            uint8_t value = scalarValuesIT->second;
            modelBase->setRawData( value, page, offset, STDataModelBase::USER_FORCE_WRITE );
            //Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel() write: ")+ symbolName+_T("page:%i offset:%i, value:%i"),page, offset, value );
        }
    }
}

void FirmwareConfigLoader::loadTables(STDataModelBase* modelBase)
{
    //SymbolTable& neededTables = m_currMTConfigData->getSymbolMgr().getConfigTableSymbols();
    FirmwareConfigData& fcData = *m_fcData;
    const SymbolTable& tablesymbolH = fcData.getConvertedTableSymbols();
    const VemsMatrixByteHash& tableH = fcData.getConvertedTableH();
    VemsMatrixByteHash::const_iterator IT = tableH.begin();

    for(; IT != tableH.end(); ++IT ){
        wxString name = IT->first;
        SymbolTable::const_iterator ITsymbol = tablesymbolH.find(name);
        if(ITsymbol == tablesymbolH.end()){
            /// error
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::loadTables() table not found: ")+ name );
            continue;
        }

        Symbol* tablesymbol= ITsymbol->second;
        Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::loadTables() table: ")+ name+_T(" page: %i offset: %i"), tablesymbol->getPage(), tablesymbol->getOffset() );
        VemsMatrix<uint8_t>* table =  IT->second;
        modelBase->setRawData(*tablesymbol, table);
        //VemsMatrixHelper::print(*table);
    }
}


VemsMatrix<double>* FirmwareConfigLoader::getUserTable( const wxString& tableName )
{
    Symbol* tableSymbol = m_currMTConfigData->getSymbol( tableName );

    if ( !tableSymbol ) { return 0; }

    const VemsMatrix<uint8_t>* table = m_fcData->getConvertedTable( tableSymbol->getName() );

    if ( !table ) { return 0; }

    if ( tableSymbol->useDefaultTransformation() )
    {
        double trans    = tableSymbol->getTranslation();
        double scale    = tableSymbol->getScale();

        VemsMatrix<double>* dTable = new VemsMatrix<double>( table->rows(), table->cols(), 0.0 );
        VemsMatrixHelper::convert( *table, *dTable, scale, trans );

        return dTable;
    }
    else
    {
        wxString expr = tableSymbol->getExprFromRaw();

        VemsMatrix<double>* dTable = new VemsMatrix<double>( table->rows(), table->cols(), 0.0 );
        VemsMatrixHelper::convert( *table, *dTable, expr );

        return dTable;
    }
}

void FirmwareConfigLoader::calculateTables()
{
    FirmwareConfigData& fcData = *m_fcData;
    SymbolTable& neededTables = m_currMTConfigData->getSymbolMgr().getConfigTableSymbols();
    SymbolTable::const_iterator IT = neededTables.begin();
    for ( ; IT != neededTables.end(); ++IT ){
        wxString name = IT->first;

        Symbol* neededSymbol = IT->second;
        const int page      = neededSymbol->getPage();
        const int offset    = neededSymbol->getOffset();
        size_t size         = neededSymbol->getArraySize();

        Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() name: ")+ name +_T(" (p:%i,o:%i, size:%i)"),page, offset, size);
        /// parse just the main tables ( which do not contains "_" )
        /// x,y bin headers will be filled together with main table
        if( name.Contains(_T("_")) ) continue;

        VemsMatrix<uint8_t>* tmpTable = fcData.getTable(name);
        if(!tmpTable){
            /// Missing table
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() missing table: ")+ neededSymbol->getName() );
            continue;
        }
        VemsMatrix<uint8_t> mTable = *tmpTable; // copy: do not want to ruin fcData.getTable(name)
        VemsMatrix<uint8_t>* mainTable = &mTable;


        /// HACK!!!!!
        //if(size == 1) size = mainTable->size();

        ///
        /// Exeptions: h, t tables without headers (small tables <  size 32, row or col = 1 )
        /// old "b" table has rpm-spark header only if fw < 1.1.67
        /// from fw 1.1.77 p table is 3*17=51 without axis the smallest table that has rpm and kpa header is 8x8
        if(mainTable->size() < 52){
            /// fill
            if( mainTable->size() == size ){ /// h, t tables allways have same size
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() upload small table: ")+ neededSymbol->getName()+_T(" p=%i,o=%i"),page,offset );
                //m_modelBase->setRawData(*neededSymbol, mainTable);
                fcData.setConvertedTable(*neededSymbol, *mainTable);
            }else{

                if( ! hasRpmKpaHeader(name) ){
                    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() no axis found (do not convert): ")+ neededSymbol->getName() );
                    continue;
                }
                /// b table exception if fw < 1.1.67
                VemsMatrix<uint8_t>* xBin = getXAxis(fcData, name);
                if(!xBin){
                    /// Missing bins tables
                    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() missing table bins (small tables): ")+ neededSymbol->getName() );
                    continue;
                }
                Symbol xBinSymbol = getXBinSymbol(neededTables, name);
                if( !xBinSymbol.isValid() ){
                    /// ERROR missing symbol???
                    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() missing table bins2 (small tables): ")+ neededSymbol->getName() );
                    continue;
                }
                Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() upload and convert (small table): ")+ neededSymbol->getName() );
                VemsMatrix<uint8_t>* newXBin = getBinMatrix(xBin, xBinSymbol.getArraySize() );
                VemsMatrixHelper::resample1D(*xBin, *newXBin, mainTable);
                fcData.setConvertedTable(*neededSymbol, *mainTable);
            }
            continue;
        }

        /// Normal (big tables must have header bins)
        VemsMatrix<uint8_t>* xBin = getXAxis(fcData, name); //m_fcData.getTable( getXAxisKey(name) );
        VemsMatrix<uint8_t>* yBin = getYAxis(fcData, name); //m_fcData.getTable( getYAxisKey(name) );

        if(!xBin){
            /// Missing X axis
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() missing X bins: ")+ neededSymbol->getName() );
            continue;
        }

        /// new big b table only have X axis
        /// b table exception if fw >= 1.1.67
        /// z table hack it is actually 2* 8x8 table + 2*8 common headers

        if( (name == _T("b")) || (name == _T("z")) ){
            Symbol xBinSymbol = getXBinSymbol(neededTables, name);
            if( mainTable->size() == size ){
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() upload  big \"")+ name +_T("\" table: ")+ neededSymbol->getName()+_T(" p=%i,o=%i"),page,offset );
            /// main table
            /// x
            fcData.setConvertedTable(*neededSymbol, *mainTable);
            fcData.setConvertedTable(xBinSymbol, *xBin);
            fcData.setXaxisName(neededSymbol->getName(), xBinSymbol.getName() );
            }else{
                Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() big \"")+ name +_T("\"  table different size (needed,found):%d,%d"), size, mainTable->size() );
            }
            continue;
        }
        if(!yBin){
             /// Missing Y axis
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() missing Y bins: ")+ neededSymbol->getName() );
            continue;
        }

        /// Big table has been found!!
        /// Try to find the needed x,y bins
        /// Exception: unfortunately y still use shared headers with ignition
        Symbol xBinSymbol = getXBinSymbol(neededTables, name);
        //xBinSymbol.setPage(page);
        Symbol yBinSymbol = getYBinSymbol(neededTables, name);
        //yBinSymbol.setPage(page);

        //Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::load(111): ")+ xBinSymbol.getName() );
        if( !xBinSymbol.isValid() || !yBinSymbol.isValid() ){
            /// ERROR missing symbol???
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() missing table bins2: ")+ neededSymbol->getName() );
            continue;
        }

        /// Exception: lambda (l)
        if(name == _T("l") && !m_currMTConfigData->isRawLambdaUsed()){
            VemsMatrixHelper::convertLambda( *mainTable );
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() CONVERT LAMBDA") );
        }

        if( mainTable->size() == size ){
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() upload main big table: ")+ neededSymbol->getName()+_T(" p=%i,o=%i"),page,offset );
            /// main table
            //m_modelBase->setRawData(*neededSymbol, mainTable);
            /// x,y bins
            //m_modelBase->setRawData(xBinSymbol, xBin);
            //m_modelBase->setRawData(yBinSymbol, yBin);

            fcData.setConvertedTable(*neededSymbol, *mainTable);
            fcData.setConvertedTable(xBinSymbol, *xBin);
            fcData.setConvertedTable(yBinSymbol, *yBin);
            fcData.setXaxisName(neededSymbol->getName(), xBinSymbol.getName() );
            fcData.setYaxisName(neededSymbol->getName(), yBinSymbol.getName() );


            /// add needed tables to fwconfig (needed: n_r, n_k, add these to fwconfigdata)
            /// this will fix mising tables
            ///addTable2Config(xBinSymbol.getName(), xBin, fcData);
            ///addTable2Config(yBinSymbol.getName(), yBin, fcData);
        }else{
            /// Interpolate!
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::calculateTables() upload Interpolate needed: ") +
            neededSymbol->getName() +_T(" size needed:%i found:%i"), size,  mainTable->size() );

            VemsMatrix<uint8_t>* newXBin = getBinMatrix(xBin, xBinSymbol.getArraySize() );
            VemsMatrix<uint8_t>* newYBin = getBinMatrix(yBin, yBinSymbol.getArraySize() );
            //Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::load() resize: %i,%i"),newXBin->cols(), newYBin->cols()  );
            //mainTable->resize(newYBin->cols(), newXBin->cols() );
            VemsMatrixHelper::resample2D(*xBin, *yBin, *newXBin, *newYBin, mainTable);
            /// main table
            //m_modelBase->setRawData(*neededSymbol, mainTable);
            /// x,y bins
            //m_modelBase->setRawData(xBinSymbol, newXBin);
            //m_modelBase->setRawData(yBinSymbol, newXBin);

            //VemsMatrixHelper::print(*newXBin);

            fcData.setConvertedTable(*neededSymbol, *mainTable);
            fcData.setConvertedTable(xBinSymbol, *newXBin);
            fcData.setConvertedTable(yBinSymbol, *newYBin);

            fcData.setXaxisName(neededSymbol->getName(), xBinSymbol.getName() );
            fcData.setYaxisName(neededSymbol->getName(), yBinSymbol.getName() );

            delete newXBin;
            delete newYBin;
        }
    }
}

void FirmwareConfigLoader::addTable2Config(const wxString& key, VemsMatrix<uint8_t>* matrix, FirmwareConfigData& fcData)
{
    VemsMatrix<uint8_t>* tmp = fcData.getTable(key);
    if(!tmp){
        fcData.addTable(key, matrix);
    }
}

VemsMatrix<uint8_t>* FirmwareConfigLoader::getBinMatrix(VemsMatrix<uint8_t>* oldBin, size_t neededSize)
{
    size_t oldSize = oldBin->size();
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::getBinMatrix() old: %i, new: %i size"),oldSize, neededSize  );

    VemsMatrix<uint8_t>* newBin = new VemsMatrix<uint8_t>( 1, neededSize, 0x00 );
    if(neededSize > oldSize){
        uint8_t value;
        for(size_t i = 0; i < neededSize; ++i){
            if(i < oldSize ){
                value = oldBin->get(i);
            }else{
                value = oldBin->get( oldSize -1 );
            }
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::getBinMatrix()1 (%i,%i)"),i, value  );
            newBin->set(i, value);
        }
    }else{
        uint8_t value;
        double deltaIdx = (double)oldSize/(double)neededSize;
        double idx =0;
        for(size_t i = 0; i < neededSize; ++i){
            value = oldBin->get( (int) idx );
            Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::getBinMatrix()2 (%i,%i)"),i, value  );
            newBin->set(i, value);
            idx += deltaIdx;
        }
    }
    return newBin;
}

Symbol FirmwareConfigLoader::getXBinSymbol(const SymbolTable& neededTables, const wxString& name)
{
    Symbol s;
    wxString axisName = name +_T("_r");
    SymbolTable::const_iterator IT = neededTables.find(axisName);
    if(IT == neededTables.end() ){
        if(name == _T("y") || name == _T("b") ){
            axisName = _T("n_r");
            IT = neededTables.find(axisName);
        }
    }
    if(IT == neededTables.end() ){
        if(!hasRpmKpaHeader(name)){
            //s.setValid(false);
            return s;
        }
        IT = neededTables.find( _T("r") );
        if(IT == neededTables.end()){
            //s.setValid(false);
            return s;
        }
    }
    s = *(IT->second);
    s.setValid(true);
    Logger::logf( LOG_MSG, _T("FirmwareConfigLoader::getXBinSymbol() ")+s.getName()+_T(" %d"), s.getPage()  );
    return s;
}

Symbol FirmwareConfigLoader::getYBinSymbol(const SymbolTable& neededTables, const wxString& name)
{
    Symbol s;
    wxString axisName = name +_T("_k");
    SymbolTable::const_iterator IT = neededTables.find(axisName);
    if( IT == neededTables.end() ){
        if(name == _T("y") ){
            axisName = _T("n_k");
            IT = neededTables.find(axisName);
        }
    }
    if(IT == neededTables.end() ){
        if(!hasRpmKpaHeader(name)){
            s.setValid(false);
            return s;
        }
        IT = neededTables.find( _T("k") );
        if(IT == neededTables.end()){
             s.setValid(false);
            return s;
        }
    }
    /// Copy the symbol;
    s = *(IT->second);
    s.setValid(true);
    return s;
    //return IT->second;
}

VemsMatrix<uint8_t>* FirmwareConfigLoader::getXAxis(FirmwareConfigData& fcData, const wxString& name)
{
    wxString axisName = name +_T("_r");
    VemsMatrix<uint8_t>* t = fcData.getTable(axisName);
    if(!t && (name == _T("y") || name == _T("b"))  ){
        axisName = _T("n_r");
        t = fcData.getTable(axisName);
    }
    if(!t){
        if(!hasRpmKpaHeader(name)) return NULL;
        t = fcData.getTable( _T("r") );
    }
    return t;
}

VemsMatrix<uint8_t>* FirmwareConfigLoader::getYAxis(FirmwareConfigData& fcData, const wxString& name)
{
    wxString axisName = name +_T("_k");
    VemsMatrix<uint8_t>* t = fcData.getTable(axisName);
    if(!t && name == _T("y") ){
        axisName = _T("n_k");
        t = fcData.getTable(axisName);
    }
    if(!t){
        if(!hasRpmKpaHeader(name)) return NULL;
        t = fcData.getTable( _T("k") );
    }
    return t;
}

bool FirmwareConfigLoader::hasRpmKpaHeader(const wxString& name)
{
    if(name == _T("j") )return true;
    if(name == _T("n") )return true;
    if(name == _T("l") )return true;
    if(name == _T("y") )return true;
    if(name == _T("b") )return true; /// actually it only has rpm header
    if(name == _T("z") )return true; /// it has rpm and kpa header but in the config
                                     /// only z[r] exists (kpa values from +8 offset)

    return false;
}

bool FirmwareConfigLoader::hasCommonAxis(const wxString& xAxisName)
{
    //if( xAxisName.isEmpty() ) { }

    if( xAxisName.Find(_T("_")) == wxNOT_FOUND ){
        return true;
    }
    return false;
}

