#include "FirmwareConfigManager.h"
#include <wx/filename.h>
#include <wx/progdlg.h>
#include <wx/textfile.h>
#include "FirmwareManager.h"
#include "FirmwareConfigDifferences.h"
#include "../VemsTuneController.h"
#include "../ObserverBase.h"
#include "../UserLog.h"
#include "../logger.h"
#include "../DirHelper.h"
#include "../unidlg2test.h"
#include "../vemsprotocol/SIPRTypes.h"
#include "../vemsprotocol/ConfigDownloadController.h"
#include "../vemsprotocol/VemsDeviceControllerBase.h"
#include "../project/ProjectManager.h"
#include "../project/ProjectHelper.h"
#include "../events/InputEvent.h"
#include "../events/CommunicationEvent.h"
#include "../events/VemsDeviceControllerEvent.h"
#include "../bcWidgets/bcProgressDialog.h"
#include "../vemsprotocol/VemsCommand.h"


const wxEventType VEMS_FIRMWARE_CONFIG_UPLOADED = wxNewEventType();


FirmwareConfigManager::FirmwareConfigManager(   wxWindow* parent,
                                                VemsDeviceControllerBase* deviceController,
                                                STDataModelBase* deviceModel,
                                                ObserverBase* observer,
                                                UserDialogBase* userDialog )

{
    m_parent                = parent;
    m_observer              = observer;

    m_pProgressDialog       = NULL;
    m_pUploadManager        = new FWConfigUploadManager( userDialog, &m_deviceInfo.firmwareInfo );
    m_iLineCount            = 0;
    m_bIsFilteredUpload     = false;

    m_deviceController = deviceController ;

    m_dirtyCount = 0;
    m_sentCount = 0;
    m_burnCount = 0;

    m_deviceModel = deviceModel;
    m_loadCmd = NULL;
}

FirmwareConfigManager::~FirmwareConfigManager()
{
    if ( m_pUploadManager ) { delete m_pUploadManager; }
    if ( m_pProgressDialog ) { delete m_pProgressDialog; }
}

void FirmwareConfigManager::uploadConfigFile( const wxString& filename, int currentConfigType, int uploadConfigType )
{
    m_deviceInfo = m_deviceController->getDeviceInfo();

    if ( !isEnabled() ) {
        showErrorMessage();
        return;
    }

    if ( !isHeaderFileExist() ) {
        showHeaderFileMissingMessage();
        return;
    }

    reset();

    wxArrayString configFiles;
    configFiles.Add( filename );

    uploadConfig( configFiles, currentConfigType, uploadConfigType );
}

void FirmwareConfigManager::filteredConfigUpload( int currentConfigType, int uploadConfigType )
{
    m_deviceInfo = m_deviceController->getDeviceInfo();

    if ( !isEnabled() ) {
        showErrorMessage();
        return;
    }

    if ( !isHeaderFileExist() ) {
        showHeaderFileMissingMessage();
        return;
    }

    if ( !getConfigFile() ) {
        return;
    }

    if ( !getFilterFile() ) {
        return;
    }

    reset();

    wxArrayString configFiles;
    configFiles.Add( m_cfgfilename );
    m_pUploadManager->parseConfigFromFiles( configFiles );

    StringSet filters;
    parseConfigFilterFile( m_filterFile, filters );

    wxArrayString missing;
    m_pUploadManager->applyFilter( filters, missing );

    if ( !missing.IsEmpty() )
    {
        wxString htmlLog;

        if ( missing.GetCount() == filters.size() )
        {
            htmlLog += _T("All filter keys missing!<br>");
            m_observer->notify( htmlLog );
            reset();
            wxMessageBox( _T("All filter keys missing!"), _T("Filtered Upload Config") );
            return;
        }

        htmlLog = _T("Missing filter keys found!<br>");

        for ( size_t i = 0; i < missing.GetCount(); ++i )
        {

            htmlLog += _T("&nbsp;&nbsp;&nbsp;&nbsp;") + missing[i] + _T("<br>");

        }

        m_observer->notify( htmlLog );
    }

    m_pUploadManager->updateModel();
    bool removeSameBytes = ( ( uploadConfigType == CFG_BOTH ) || ( currentConfigType == uploadConfigType ) );
    m_pUploadManager->createUploadData( *m_deviceModel, removeSameBytes );
    m_dirtyCount = m_pUploadManager->getUploadCount();
    m_sentCount = 0;

    m_burnCount = getUploadRequestCount( currentConfigType, uploadConfigType );
    m_dirtyCount += m_burnCount;

    if ( m_dirtyCount == 0 ) {
        wxMessageBox( _T("Configs are identical!"), _T("Upload config") );
        return;
    }

    setLoadCmd( currentConfigType, uploadConfigType );

    LOGP( wxString::Format( _T("m_dirtyCount: %d, m_sentCount: %d, removeSameBytes: %d"), m_dirtyCount, m_sentCount, removeSameBytes ) );

    createProgressDialog( _T("Config Upload"), _T("Write config in progress...") );

    connectEventHandlers();

    m_pUploadManager->updateECU( m_deviceController, currentConfigType, uploadConfigType );

}

void FirmwareConfigManager::easyUploadConfig( int currentConfigType, int uploadConfigType  )
{
    m_deviceInfo = m_deviceController->getDeviceInfo();

    if ( !isEnabled() ) {
        showErrorMessage();
        return;
    }

    if ( !isHeaderFileExist() ) {
        showHeaderFileMissingMessage();
        return;
    }

    if ( !getConfigFile() ) {
        return;
    }

    reset();

    wxArrayString configFiles;
    configFiles.Add( m_cfgfilename );

    uploadConfig( configFiles, currentConfigType, uploadConfigType );
}


void FirmwareConfigManager::uploadConfig( const wxArrayString& configFiles,  int currentConfigType, int uploadConfigType )
{
    bool error = m_pUploadManager->createUploadConfig( configFiles );

    /// create report to html log
    wxString htmlLog;
    htmlLog += _T("-------Upload config--------<br>");

    /// add parsed table info
    const wxArrayString& tables = m_pUploadManager->parsedTables;

    for ( size_t i = 0; i< tables.GetCount(); ++i )
    {
        htmlLog += _T("&nbsp;&nbsp;&nbsp;&nbsp;") + tables[i] + _T("<br>");
    }

    /// add detailed parsing info
    const FirmwareConfigDifferences& diff = m_pUploadManager->configsVsHeader;

    if ( diff.isEmpty() )
    {
        htmlLog += _T("Files parsing was successful!<br>");
    }
    else
    {
        htmlLog += _T("File parsing summary<br>");
        htmlLog += diff.summary( _T("<br>") );
        htmlLog += _T("File parsing details<br>");
        htmlLog += diff.details( _T("&nbsp;&nbsp;&nbsp;&nbsp;"), _T("<br>") );
        htmlLog += m_pUploadManager->menuLinks;
    }

    m_observer->notify( htmlLog );

    /// create message box info
    //wxString message = m_pUploadManager->getUploadSummary();
    wxString message;

    if ( diff.isEmpty() )
    {
        message += _T("Files parsing was successful!\n");
    }
    else
    {
        message += _T("File parsing summary\n");
        message += diff.summary( _T("\n") );
    }

    wxString configTableSize = _T("12x12");
    if ( !m_pUploadManager->isFile12x12() ) { configTableSize = _T("16x14"); }

    wxString ecuTableSize = _T("12x12");
    if ( !m_deviceInfo.firmwareInfo.is12x12() ) { ecuTableSize = _T("16x14"); }

    message += _T("Config table size:") + configTableSize + _T("\n") +
        _T("ECU table size:") + ecuTableSize + _T("\n");

    if ( error )
    {
        message += _T("More details can be found in debug tab.");
    }

    wxMessageBox( message, _T("Upload config") );

///// TODO    if ( m_vemsController->getDevice() != ECUInfo::DEVICE_V3 && m_vemsController->getDevice() != ECUInfo::DEVICE_ROUND )
//    {
//        wxMessageBox( _T("ECU not connected!"), _T("Upload config"), wxICON_ERROR | wxOK );
///// TODO       stop();
//        return;
//    }


    m_pUploadManager->updateModel();

    bool removeSameBytes = ( ( uploadConfigType == CFG_BOTH ) || ( currentConfigType == uploadConfigType ) );

    m_pUploadManager->createUploadData( *m_deviceModel, removeSameBytes );

    m_dirtyCount = m_pUploadManager->getUploadCount();
    m_sentCount = 0;

    m_burnCount = getUploadRequestCount( currentConfigType, uploadConfigType );
    m_dirtyCount += m_burnCount;

    LOGP( wxString::Format( _T("m_dirtyCount: %d, m_sentCount: %d, removeSameBytes: %d"), m_dirtyCount, m_sentCount, removeSameBytes ) );

    if ( m_dirtyCount == 0 ) {
        wxMessageBox( _T("Configs are identical!"), _T("Upload config") );
        return;
    }

    //m_deviceController->Connect( wxEVT_INPUT, InputEventHandler( FirmwareConfigManager::OnSIPR ), NULL, this );

    setLoadCmd( currentConfigType, uploadConfigType );

    createProgressDialog( _T("Config Upload"), _T("Write config in progress...") );

    LOGP( wxString::Format( _T("uploadConfigType: %d, currentConfigType: %d"), uploadConfigType, currentConfigType ) );

    connectEventHandlers();

    m_pUploadManager->updateECU( m_deviceController, currentConfigType, uploadConfigType );
}


void FirmwareConfigManager::verifyConfig( bool connectInputEvent )
{
    ///TODO TEST
    if ( connectInputEvent ) {
        //m_deviceController->Connect( wxEVT_INPUT, InputEventHandler( FirmwareConfigManager::OnSIPR ), NULL, this );
        connectEventHandlers();
    }
    m_deviceController->request( new ReadAllPageCommand() );
    ///OLD
//    m_pUploadManager->verifyConfig();
//    m_configDownloadController->start();
}


void FirmwareConfigManager::showConfigDownloadResult()
{
    ///TODO TEST
    FirmwareConfigDifferences* fcDiff = m_pUploadManager->compareConfig( m_deviceModel );

    ///create html log
    wxString htmlLog;
    wxString message;
    wxString title = _T("Verify config");
    int dialogStyle = wxOK;

    if ( fcDiff->isEmpty() ||
            ( m_bIsFilteredUpload && fcDiff->extraScalars.IsEmpty() && fcDiff->extraTables.IsEmpty() ) )
    {
        htmlLog += _T("Verify config was successful!<br>");
        message += _T("Verify config was successful!\n");
    }
    else
    {
        title = _T("Verify config failed!");
        dialogStyle = wxOK | wxICON_WARNING;

        if ( m_bIsFilteredUpload )
        {
            for ( size_t i = 0; i < fcDiff->extraScalars.GetCount(); ++i )
            {
                htmlLog += _T("&nbsp;&nbsp;&nbsp;&nbsp;") + fcDiff->extraScalars[i] + _T("<br>");
            }
            //fcDiff->extraTables();
        }
        else
        {
            htmlLog += ( _T("Verify config summary<br>") );
            htmlLog += fcDiff->summary( _T("<br>") );
            htmlLog += _T("Verify config details<br>");
            htmlLog += fcDiff->details( _T("&nbsp;&nbsp;&nbsp;&nbsp;"), _T("<br>") );
            message += fcDiff->summary( _T("\n") );
            message += _T("\n More details can be found in debug tab.");
        }
    }

    m_observer->notify( htmlLog );

    wxMessageBox( message, title, dialogStyle );

    wxCommandEvent event(VEMS_FIRMWARE_CONFIG_UPLOADED);
    m_parent->AddPendingEvent( event );
}

void FirmwareConfigManager::showFilterParsingResult()
{

}

void FirmwareConfigManager::showConfigUploadResult()
{
    wxString message = _T("Upload config was successful!\n");
    wxMessageBox( message, _T("Upload config") );
}

void FirmwareConfigManager::processConfigUpload( const SIPRPacket& packet )
{
    ++m_sentCount;

//    LOGP( wxString::Format( _T("m_dirtyCount: %d, m_sentCount: %d, m_burnCount: %d"), m_dirtyCount, m_sentCount,m_burnCount ) );
//    if ( ( m_burnCount > 0 ) && ( m_sentCount == ( m_dirtyCount - m_burnCount ) ) ) {
//        LOG
//        m_pProgressDialog->UpdateMessage( _T("Burning config in progress...") );
//    }

    if ( m_dirtyCount == m_sentCount ) {
        LOGP( _T("Upload finished!") );
//        m_deviceController->stop( VDC_FLAG_NOEVENT );
//        m_deviceController->Disconnect( wxEVT_INPUT, InputEventHandler( FirmwareConfigManager::OnSIPR ), NULL, this );
//        if ( m_deviceInfo.isV3() ) {
            m_pProgressDialog->UpdateMessage( _T("Downloading config in progress...") );
            verifyConfig(false);
//        } else {
//            destroyProgressDialog();
//            showConfigUploadResult();
//            m_deviceController->resume( VDC_FLAG_GET_PAGES );
//        }
    }
}


//void FirmwareConfigManager::OnSIPR( InputEvent& event )
//{


//        case SIPR_CONFIG_ERROR:
//        case SIPR_TIMEOUT:
//        {
///// TODO
////            reset();
////
////            //m_pController->request( new SetStateCommand( ECUInfo::MEGATUNE ) );
////            m_pController->restore();
////            if ( m_pController->isReadSensors() )
////            {
////                m_pController->request( new EnableReadRealTimeCommand( true ) );
////            }
////            stop();
//
//        } break;
//    }
//}


void FirmwareConfigManager::reset()
{
    m_pUploadManager->clear();

    m_iLineCount = 0;
    m_bIsFilteredUpload = false;

    destroyProgressDialog();
}


void FirmwareConfigManager::uploadConfigFromProject( int currentConfigType, int uploadConfigType )
{
    Unidlg2Test dlg( m_parent );

    dlg.setDialogType( _T("config_upload") );
    dlg.SetLabel( _T("Upload Config Files") );
    wxString btn = _T("Upload");
    dlg.setButtonOK( btn );

    Project *prj = ProjectHelper::GetPrjMgr()->GetCurrPrj();

    dlg.setProjectConfigSaver( prj );
    dlg.loadProject( prj->CreateFirmwareConfigItems( dlg.getTypes() ) );

    if ( dlg.ShowModal() == wxID_OK )
    {
        wxString        headerFile  = dlg.getHeaderFile();
        wxArrayString   configFiles = dlg.getConfigFiles();

        m_pUploadManager->clear();
        uploadConfig( configFiles, currentConfigType, uploadConfigType );
        verifyConfig(false);
    }
}



void FirmwareConfigManager::verifyConfigFromProject()
{
    Unidlg2Test dlg( m_parent );

    dlg.setDialogType( _T("config") );
    dlg.SetLabel( _T("Verify Config Files") );
    wxString tmp = _T("Verify");
    dlg.setButtonOK(tmp);

    Project *prj = ProjectHelper::GetPrjMgr()->GetCurrPrj();

    dlg.setProjectConfigSaver(prj);
    dlg.loadProject( prj->CreateFirmwareConfigItems(dlg.getTypes()) );

    if ( dlg.ShowModal() != wxID_OK ) {
        return;
    }

//    wxString headerFile         = dlg.getHeaderFile();
    wxArrayString configFiles   = dlg.getConfigFiles();

    reset();

    m_pUploadManager->createUploadConfig( configFiles );

    createProgressDialog( _T("Download config in progress!") );

    verifyConfig();
}

void FirmwareConfigManager::verifyConfigFromFile( const wxString& filename )
{
    m_deviceInfo = m_deviceController->getDeviceInfo();

    if ( !isEnabled() ) {
        showErrorMessage();
        return;
    }

    if ( !isHeaderFileExist() ) {
        showHeaderFileMissingMessage();
        return;
    }

    if ( filename.IsEmpty() ) {
        if ( !getConfigFile() ) {
            return;
        }
    } else {
        m_cfgfilename = filename;
    }

    //m_deviceController->stop( VDC_FLAG_NOEVENT );

    createUploadConfig();

    createProgressDialog( _T("Download config in progress!") );

    verifyConfig();
}

void FirmwareConfigManager::parseConfigFilterFile( const wxString& name, StringSet& keys )
{
    m_bIsFilteredUpload = true;

    wxTextFile filterFile( name );

    if ( !filterFile.Open( wxConvISO8859_1 ) ) return;

    wxString line = filterFile.GetFirstLine();

    for ( ; !filterFile.Eof(); line = filterFile.GetNextLine() )
    {
        if ( line.Find( _T('=') ) == wxNOT_FOUND )
        {
            keys.insert( line );
        }
        else
        {
            keys.insert( line.BeforeFirst( _T('=') ) );
        }
    }

    filterFile.Close();

    LOGP( wxString::Format( _T("Found %d keys!"), keys.size() ) );
}

bool FirmwareConfigManager::isHeaderFileExist()
{
    FirmwareInfo* fwInfo = &m_deviceInfo.firmwareInfo;
    wxString headerFile = fwInfo->getHeaderFile();

    return ( !headerFile.IsEmpty() );
}

void FirmwareConfigManager::showHeaderFileMissingMessage()
{
    FirmwareInfo* fwInfo = &m_deviceInfo.firmwareInfo;
    wxString fwVersion = fwInfo->getFirmwareVersionAsString();

    wxString msg = _T("Descriptor file has been not found for firmware ")+ fwVersion +_T("\n");
    msg +=_T("This means that you should download firmware: ")+ fwVersion+ _T("by VemsTune!\n\n");
    msg +=_T("1. Go to menu Firmware/Firmware web tool")+ fwVersion+ _T("\n");
    msg +=_T("2. Select the given firmware version and press download!\n");
    msg +=_T("3. When it has been downloaded press right mouse button on it and click Install!\n");
    msg +=_T("Now the descriptor files should be available for VemsTune!\n");

    wxMessageBox( msg );
}

bool FirmwareConfigManager::getConfigFile()
{
    wxString filter = _T("Vems config file (*.vemscfg)|*.vemscfg");
    filter += _T("|mcdmct (*.txt)|*.txt");
    filter += _T("|All files (*.*)|*.*");
    wxString message = _T("Select a config file");
    wxString dir = DirHelper::GetPath(dLOG);

    wxFileDialog dlg(NULL, message,dir ,wxEmptyString, filter, wxCHANGE_DIR);

    if ( dlg.ShowModal() == wxID_OK ) {
        m_cfgfilename = dlg.GetPath();
        return true;
    }

    return false;
}

bool FirmwareConfigManager::getFilterFile()
{
    wxString filter =_T("mcdmct (*.txt)|*.txt");
    filter += _T("|Vems config file (*.vemscfg)|*.vemscfg");
    filter += _T("|All files (*.*)|*.*");
    wxString message = _T("Select a filter file");
    wxString dir = DirHelper::GetPath(dCONFIGFILTERS);

    wxFileDialog dlg( NULL, message, dir, wxEmptyString, filter, wxCHANGE_DIR );

    if ( dlg.ShowModal() == wxID_OK ) {
        m_filterFile = dlg.GetPath();
        return true;
    }

    return false;
}

void FirmwareConfigManager::createUploadConfig()
{
    reset();

    wxArrayString configFiles;
    configFiles.Add( m_cfgfilename );
    m_pUploadManager->createUploadConfig( configFiles );
}

void FirmwareConfigManager::createProgressDialog( const wxString& title, const wxString& message )
{
    m_pProgressDialog = new bcProgressDialog(   title,
                                                message,
                                                100,
                                                NULL,
                                                wxPD_APP_MODAL | wxPD_AUTO_HIDE | wxPD_SMOOTH |
                                                wxPD_ELAPSED_TIME | wxPD_ESTIMATED_TIME | wxPD_REMAINING_TIME );
    m_pProgressDialog->Pulse();
}

void FirmwareConfigManager::destroyProgressDialog()
{
    if ( m_pProgressDialog )
    {
        m_pProgressDialog->Destroy();
        m_pProgressDialog = NULL;
    }
}

bool FirmwareConfigManager::isEnabled() const
{
    return ( m_deviceController->isConnected() && ( m_deviceController->isV3() || m_deviceController->isRound() ) );
}

void FirmwareConfigManager::showErrorMessage() const
{
        UserLog::vLog( _T("ecu_not_found_1") );
        wxString msg = _T("ECU is not detected!");
        wxMessageBox(msg);
}

size_t FirmwareConfigManager::getUploadRequestCount( int currentConfigType, int uploadConfigType ) const
{
    if ( m_deviceInfo.isV3() ) {

        if ( uploadConfigType == CFG_BOTH ) {
            return 2;
        }

        if ( currentConfigType == CFG_UNKNOWN ) {
            return 0;
        }

        return 1;

     } else {
        return 0;
    }
}

void FirmwareConfigManager::setLoadCmd( int currentConfigType, int uploadConfigType )
{
    if ( ( currentConfigType == CFG_A ) && ( uploadConfigType == CFG_B ) ) {
        m_loadCmd = new StringCommand( _T("ManmcL0mbye"), 0, true );
    } else if ( ( currentConfigType == CFG_B ) && ( uploadConfigType == CFG_A ) ) {
        m_loadCmd = new StringCommand( _T("ManmcL1mbye"), 0, true );
    } else {
        m_loadCmd = NULL;
    }
}

void FirmwareConfigManager::connectEventHandlers()
{
    Connect(    wxID_ANY,
                wxEVT_COMMUNICATION,
                CommunicationEventHandler( FirmwareConfigManager::onCommunicationEvent ),
                NULL, this );

    Connect(    VDC_EVENT_DEVICE_DISCONNECTED,
                wxEVT_VEMS_DEVICE_CONTROLLER,
                VemsDeviceControllerEventHandler( FirmwareConfigManager::onDeviceDisconnected ),
                NULL, this );

    m_deviceController->Connect( wxEVT_INPUT, InputEventHandler( FirmwareConfigManager::onSIPR ), NULL, this );
}

void FirmwareConfigManager::disconnectEventHandlers()
{
    Disconnect( wxID_ANY,
                wxEVT_COMMUNICATION,
                CommunicationEventHandler( FirmwareConfigManager::onCommunicationEvent ),
                NULL, this );

    Disconnect( VDC_EVENT_DEVICE_DISCONNECTED,
                wxEVT_VEMS_DEVICE_CONTROLLER,
                VemsDeviceControllerEventHandler( FirmwareConfigManager::onDeviceDisconnected ),
                NULL, this );

    m_deviceController->Disconnect( wxEVT_INPUT, InputEventHandler( FirmwareConfigManager::onSIPR ), NULL, this );
}

void FirmwareConfigManager::onCommunicationEvent( CommunicationEvent& event )
{
    disconnectEventHandlers();
    destroyProgressDialog();
    wxMessageBox( _T("Verify config failed!"), _T("Communication Error!"), wxOK | wxICON_ERROR );
}

void FirmwareConfigManager::onSIPR( InputEvent& event )
{
    SIPRPacket packet = event.getPacket();

    switch ( packet.type() )
    {
        case SIPR_TIMEOUT:
            LOG
            break;

        case SIPR_ACK:
            processConfigUpload( packet );
            break;

        case SIPR_GET_TABLEDATA:
        case SIPR_GET_CONFIG_1:
        case SIPR_GET_CONFIG_2:
            //update model
            //LOG
            event.Skip();
            break;

        case SIPR_LOG_MARKER:
            if ( packet.data() && packet.size() > 0 ) {
                if ( packet.data()[0] == SIPR_REFRESH_STOP ) {
                    //m_deviceController->Disconnect( wxEVT_INPUT, InputEventHandler( FirmwareConfigManager::OnSIPR ), NULL, this );
                    disconnectEventHandlers();
                    destroyProgressDialog();
                    showConfigDownloadResult();
                    if ( m_loadCmd ) {
                        m_deviceController->request( m_loadCmd );
                        m_deviceController->request( new ReadAllPageCommand() );
						m_loadCmd = NULL;
                    }
                    //m_deviceController->resume( VDC_FLAG_GET_PAGES );
                }
            }
            event.Skip();
            break;

//        default:
//            LOGP( wxString::Format( _T("ERROR - Unprocessed SIPRPacket type! (%d)"), packet.type() ) );
//            LOGP( wxString::Format( _T("m_dirtyCount: %d, m_sentCount: %d"), m_dirtyCount, m_sentCount ) );
    }
}

void FirmwareConfigManager::onDeviceDisconnected( VemsDeviceControllerEvent& event )
{
    disconnectEventHandlers();
    destroyProgressDialog();

    wxMessageBox( _T("Verify config failed!"), _T("Device disconnected!"), wxOK | wxICON_ERROR );
}
