#include "ExpressionParser.h"

#include "../../common/types/StringHash.h"
#include "../../common/types/DoubleHash.h"

#include "../../logger.h"

#ifndef NAN
    #define NAN -10000000000
#endif

//

///-----------------------------------
ExpressionParser::ExpressionParser()
{
    reset();

    user_var = NULL;

    m_selfCreatedVList = false;

    createVariableList();
}

ExpressionParser::ExpressionParser( VariableList* varlist )
{
    reset();

    user_var = varlist;

    m_selfCreatedVList = false;
}

ExpressionParser::~ExpressionParser()
{
    if ( m_selfCreatedVList )
    {
        delete user_var;
    }
}

void ExpressionParser::reset()
{
    expr[0] = '\0';
    e = NULL;

    token[0] = '\0';
    token_type = NOTHING;
}

void ExpressionParser::createVariableList()
{
    if ( user_var != NULL && m_selfCreatedVList )
    {
        delete user_var;
    }

    user_var = new VariableList;

    m_selfCreatedVList = true;
}

void ExpressionParser::addConstant( const wxString& var, const double& value )
{
    user_var->add( var, value );
}

void ExpressionParser::setVariableList( VariableList* list )
{
    if ( user_var != NULL && m_selfCreatedVList )
        delete user_var;

    user_var = list;

    m_selfCreatedVList = false;
}

double ExpressionParser::parse( const wxString& expression, bool showExceptions, bool throwExceptions )
{
    size_t len = expression.Length();

    m_exprs = expression;

    char* new_expr = new char[expression.Length()+1];

    memcpy( new_expr, expression.mb_str(),expression.Length() );
    new_expr[len] = '\0';

    try
    {

        // check the length of expr
        if ( strlen( new_expr ) > EXPR_LEN_MAX )
        {
            delete []new_expr;

            if ( throwExceptions )
                throw ExpressionException( row(), col(), ExpressionException::EXPR_TOO_LONG, m_exprs );

        }

        // initialize all variables
        strcpy( expr, new_expr );   // copy the given expression to expr
        delete []new_expr;
        e = expr;                   // let e point to the start of the expression
        result = 0;

        getToken();

        if ( token_type == DELIMITER && *token == '\0' )
        {
            if ( throwExceptions )
                throw ExpressionException( row(), col(), ExpressionException::EMPTY_EXPR, m_exprs );
        }

        result = parse_level1();

        if ( token_type != DELIMITER || *token != '\0' )
        {
            if ( token_type == DELIMITER )
            {
                if ( throwExceptions )
                    throw ExpressionException( row(), col(), ExpressionException::UNKNOWN_OP, m_exprs, token );;
            }
            else
            {
                if ( throwExceptions )
                    throw ExpressionException(row(), col(), ExpressionException::UNEXPECTED_EOEXPR_, m_exprs );
            }
        }

        // add the answer to memory as variable "Ans"
        user_var->add("result", result);
    }
    catch ( ExpressionException e )
    {
        if ( showExceptions )
        {
            LOGP( e.err )
            LOGP( e.getExprErr() )
        }

        if ( throwExceptions )
            throw e;

        return ( NAN );
    }

    return result;
}

void ExpressionParser::getToken()
{
    token_type = NOTHING;

    char* t;

    t = token;

    *t = '\0';

    // skip over whitespaces
    while ( *e == ' ' || *e == '\t' )
        e++;

    // check for end of expression
    if ( *e == '\0' )
    {
        token_type = DELIMITER;
        return;
    }

    // check for minus
    if ( *e == '-' )
    {
        token_type = DELIMITER;
        *t = *e;
        e++;
        t++;

        *t = '\0';

        return;
    }

    // check for parentheses
    if ( *e == '(' || *e == ')' )
    {
        token_type = DELIMITER;
        *t = *e;
        e++;
        t++;
        *t = '\0';
        return;
    }

    // check for operators (DELIMITERs)
    if ( isDelimiter( *e ) )
    {
        token_type = DELIMITER;
        while ( isDelimiter( *e ) )
        {
            *t = *e;

            e++;
            t++;
        }

        *t = '\0';
        return;
    }

    // check for a value
    if ( isDigitDot( *e ) )
    {
        token_type = NUMBER;
        while ( isDigitDot( *e ) )
        {
            *t = *e;
            e++;
            t++;
        }

        // check for scientific notation like "2.3e-4" or "1.23e50"
        if ( toupper( *e ) == 'E' )
        {
            *t = *e;
            e++;
            t++;

            if ( *e == '+' || *e == '-' )
            {
                *t = *e;
                e++;
                t++;
            }

            while ( isDigit( *e ) )
            {
                *t = *e;
                e++;
                t++;
            }
        }

        if ( toupper( *e) == 'X' )
        {
            *t = *e;
            e++;
            t++;

            while ( isHexDigit( *e ) )
            {
                *t = *e;
                e++;
                t++;
            }
        }

        *t = '\0';
        return;
    }

    // check for variables or functions
    if ( isAlpha( *e ) )
    {
        while ( isAlpha( *e ) || isDigit( *e ) )
        {
            *t = *e;
            e++;
            t++;
        }
        *t = '\0';

        // check if this is a variable or a function.
        // a function has a parentesis '(' open after the name
        char* e2 = NULL;
        e2 = e;

        // skip whitespaces
        while ( *e2 == ' ' || *e2 == '\t' )     // space or tab
        {
            e2++;
        }

        if ( *e2 == '(' )
        {
            token_type = FUNCTION;
        }
        else
        {
            token_type = VAR;
        }

        return;
    }

    // something unknown is found, wrong characters -> a syntax error
    token_type = UNKNOWN;

    while ( *e != '\0' )
    {
        *t = *e;
        e++;
        t++;
    }
    *t = '\0';

    throw ExpressionException( row(), col(), ExpressionException::SYNTAX_ERROR, m_exprs );

    return;
}

double ExpressionParser::parse_level1()
{
    if ( token_type == VAR )
    {
        // copy current token
        char* e_now = e;

        TOKENTYPE token_type_now = token_type;

        char token_now[NAME_LEN_MAX + 1];

        strcpy( token_now, token );

        getToken();

        if ( strcmp(token, "= ") == 0 )
        {
            // assignment
            double ans;

            getToken();

            ans = parse_level1b();
            if ( user_var->add( token_now, ans ) == false)
            {
                throw ExpressionException( row(), col(), ExpressionException::DEFINING_VAR_FAILED, m_exprs );
            }

            return ans;
        }
        else
        {
            e = e_now;
            token_type = token_type_now;
            strcpy( token, token_now );
        }
    }

    return parse_level1b();
}

double ExpressionParser::parse_level1b()
{
    int op_id;

    double ans;

    ans = parse_level2();

    op_id = get_operator_id( token );

    while ( op_id == LOGAND || op_id == LOGOR )
    {
        getToken();

        ans = eval_operator( op_id, ans, parse_level2() );

        op_id = get_operator_id( token );
    }

    return ans;
}

double ExpressionParser::parse_level2()
{
    int op_id;

    double ans;

    ans = parse_level3();

    op_id = get_operator_id( token );

    while ( op_id == AND || op_id == OR || op_id == SHL || op_id == SHR  )
    {
        getToken();

        ans = eval_operator( op_id, ans, parse_level3() );

        op_id = get_operator_id( token );
    }

    return ans;
}

double ExpressionParser::parse_level3()
{
    int op_id;

    double ans;

    ans = parse_level4();

    op_id = get_operator_id( token );

    while ( op_id == EQ || op_id == NEQ || op_id == LT || op_id == GT || op_id == LTE || op_id == GTE )
    {
        getToken();

        ans = eval_operator( op_id, ans, parse_level4() );

        op_id = get_operator_id( token );
    }

    return ans;
}

double ExpressionParser::parse_level4()
{
    int op_id;

    double ans;

    ans = parse_level5();

    op_id = get_operator_id( token );

    while ( op_id == PLUS || op_id == MINUS )
    {
        getToken();

        ans = eval_operator( op_id, ans, parse_level5() );

        op_id = get_operator_id( token );
    }

    return ans;
}


double ExpressionParser::parse_level5()
{
    int op_id;

    double ans;

    ans = parse_level8();

    op_id = get_operator_id( token );

    while ( op_id == MUL || op_id == DIV || op_id == MOD || op_id == XOR )
    {
        getToken();

        ans = eval_operator( op_id, ans, parse_level8() );

        op_id = get_operator_id(token);
    }

    return ans;
}

double ExpressionParser::parse_level8()
{
    double ans;

    int op_id = get_operator_id( token );

    if ( op_id == MINUS )
    {
        getToken();
        ans = parse_level9();
        ans = -ans;
    }
    else if ( op_id == LOGNOT )
    {
        getToken();
        ans = parse_level9();
        ans = !ans;
    }
    else
    {
        ans = parse_level9();
    }

    return ans;
}


double ExpressionParser::parse_level9()
{
    char fn_name[NAME_LEN_MAX + 1];

    double ans;

    if ( token_type == FUNCTION )
    {
        strcpy( fn_name, token );

        getToken();

        ans = eval_function( fn_name, parse_level10() );
    }
    else
    {
        ans = parse_level10();
    }

    return ans;
}


double ExpressionParser::parse_level10()
{
    // check if it is a parenthesized expression
    if ( token_type == DELIMITER )
    {
        if ( ( token[0] == '(') && ( token[1] == '\0' ) )
        {
            getToken();

            double ans = parse_level1b();

            if ( token_type != DELIMITER || token[0] != ')' || token[1] || '\0' )
            {
                throw ExpressionException( row(), col(), ExpressionException::PARENTHESIS_MISSING, m_exprs );
            }

            getToken();

            return ans;
        }
    }
    return parse_number();
}

double ExpressionParser::parse_number()
{
    double ans = 0;

    switch ( token_type )
    {
        case NUMBER:
            ans = strtod( token, NULL );

            getToken();

            break;

        case VAR:
            ans = eval_variable( token );

            getToken();

            break;

        default:
            if ( token[0] == '\0' )
                throw ExpressionException( row(), col(), ExpressionException::UNEXPECTED_EOEXPR_, m_exprs );
            else
                throw ExpressionException( row(), col(), ExpressionException::VALUE_EXPECTED, m_exprs );
            break;
    }

    return ans;
}

int ExpressionParser::get_operator_id( const char op_name[] )
{
    if ( !strcmp( op_name, "!") )
        return LOGNOT;

    if ( !strcmp( op_name, "&&" ) )
        return LOGAND;

    if ( !strcmp( op_name, "||" ) )
        return LOGOR;

    if ( !strcmp( op_name, "&" ) )
        return AND;

    if ( !strcmp( op_name, "|" ) )
        return OR;

    if ( !strcmp( op_name, "<<" ) )
        return SHL;

    if ( !strcmp( op_name, ">>" ) )
        return SHR;

    if ( !strcmp( op_name, "==" ) )
        return EQ;

    if ( !strcmp( op_name, "!=" ) )
        return NEQ;

    if ( !strcmp( op_name, "<" ) )
        return LT;

    if ( !strcmp( op_name, ">" ) )
        return GT;

    if ( !strcmp( op_name, "<=" ) )
        return LTE;

    if ( !strcmp( op_name, ">=" ) )
        return GTE;

    if ( !strcmp( op_name, "+" ) )
        return PLUS;
    if ( !strcmp( op_name, "-" ) )
        return MINUS;

    // level 5
    if ( !strcmp( op_name, "*" ) )
        return MUL;
    if ( !strcmp( op_name, "/" ) )
        return DIV;
    if ( !strcmp( op_name, "%" ) )
        return MOD;
    if ( !strcmp( op_name, "^" ) )
        return XOR;

    return -1;
}

double ExpressionParser::eval_operator( const int op_id, const double &lhs, const double &rhs )
{
    if ( lhs != lhs )
        return NAN;
    if ( rhs != rhs )
        return NAN;

    int ilhs = (int)lhs;
    int irhs = (int)rhs;

    switch ( op_id )
    {
        case LOGAND:    return ( ( lhs != 0 ) && ( rhs != 0 ) );
        case LOGOR:     return ( ( lhs != 0 ) || ( rhs != 0 ) );
        case AND:       return ilhs & irhs;
        case OR:        return ilhs | irhs;
        case SHL:       return ilhs << irhs;
        case SHR:       return ilhs >> irhs;

        case EQ:        return lhs == rhs;
        case NEQ:       return lhs != rhs;
        case LT:        return lhs < rhs;
        case GT:        return lhs > rhs;
        case LTE:       return lhs <= rhs;
        case GTE:       return lhs >= rhs;

        case PLUS:      return lhs + rhs;
        case MINUS:     return lhs - rhs;

        // level 5
        case MUL:       return lhs * rhs;
        case DIV:       return lhs / rhs;
        case MOD:       return ilhs % irhs;
        case XOR:       return ilhs ^ irhs;
    }

    throw ExpressionException( row(), col(), ExpressionException::UNKNOWN_OP, m_exprs );
    return 0;
}

double ExpressionParser::eval_function( const char fn_name[], const double &value )
{
    try
    {
        // todo define functions
        // if ( !strcmp( fn_name,"FUNCNAME") ) { return FUNC(value); }
        /*if ( !strcmp( fn_name, "defined") )
            return value==value;*/
    }
    catch (ExpressionException e)
    {
        LOG
    }

    // unknown function
    throw ExpressionException( row(), col(), ExpressionException::UNKNOWN_FUNC, m_exprs, fn_name );
    return 0;
}

double ExpressionParser::eval_variable( const char var_name[] )
{
    // check for user defined variables
    double ans = 0;

    if ( user_var->get_value(var_name, &ans) )
    {
        return ans;
    }

    throw ExpressionException(row(), col(), ExpressionException::UNKNOWN_VAR, m_exprs, var_name );

    return 0;
}


// helpers

inline bool ExpressionParser::isMinus( const char c )
{
    if ( c == 0 )
        return false;
    return c == '-';
}

inline bool ExpressionParser::isWhiteSpace( const char c )
{
    if ( c == 0 )
        return false;
    return c == 32 || c == 9;
}

inline bool ExpressionParser::isDelimiter( const char c )
{
    if ( c == 0 )
        return false;

    return ( strchr( "&|<>=+/*%^!", c ) != 0 );
}

inline bool ExpressionParser::isNotDelimiter( const char c )
{
    if ( c == 0 )
        return false;

    return ( strchr("&|<>=+-/*%^!()", c) != 0 );
}

inline bool ExpressionParser::isAlpha( const char c )
{
    if ( c == 0 )
        return false;

    return ( strchr("ABCDEFGHIJKLMNOPQRSTUVWXYZ_", toupper(c)) != 0 );
}

inline bool ExpressionParser::isDigitDot( const char c )
{
    if ( c == 0 )
        return false;

    return ( strchr("0123456789.", c) != 0 );
}

inline bool ExpressionParser::isDigit( const char c )
{
    if ( c == 0 )
        return false;

    return ( strchr("0123456789", c) != 0 );
}

inline bool ExpressionParser::isHexDigit( const char c )
{
    if ( c == 0 )
        return false;

    return ( strchr("0123456789abcdefABCDEF", c) != 0 );
}
