#include "GPSCommunicationController.h"

#include "GPSDetectorClient.h"
#include "GpsCommClient.h"
#include "../SerialPortHelper.h"
#include "../communication/CommunicationController2.h"
#include "../serialconfig.h"
#include "../logger.h"
#include "../serialporthandler.h"
#include "../ConfigBase.h"
#include "../configmanager/ConfigHelper2.h"
#include "../configmanager/SerialConfigMgr.h"
#include "../events/CommunicationEvent.h"
#include "../events/GPSCommEvent.h"

BEGIN_EVENT_TABLE( GPSCommunicationController, wxEvtHandler )
    EVT_GPS_COMMEVT( GPSCommunicationController::OnGPSEVent )
END_EVENT_TABLE()

GPSCommunicationController::GPSCommunicationController( wxEvtHandler* evtHandler ) : m_evtHandler( evtHandler ), m_state( IDLE )
{
    m_availablePorts = SerialPortHelper::GetCOMPortList();

    m_availableBaudrates = SerialPortHelper::getGPSSerialBaudRates();

    m_communicationClient = NULL;
    m_commbase = NULL;
    m_controller = NULL;

    Connect( wxID_ANY, wxEVT_COMMUNICATION, CommunicationEventHandler( GPSCommunicationController::OnCommunicationEvent ), NULL, this );
}

GPSCommunicationController::~GPSCommunicationController()
{
    if ( m_communicationClient )
        delete m_communicationClient;

    if ( m_commbase )
        delete m_commbase;

    if ( m_controller )
        delete m_controller;

    Disconnect( wxID_ANY, wxEVT_COMMUNICATION, CommunicationEventHandler( GPSCommunicationController::OnCommunicationEvent ), NULL, this );
}

void GPSCommunicationController::handleEvent( CommunicationClientEvent::Type eventType )
{
    // nop
}

void GPSCommunicationController::Abort()
{
    if ( m_state == DETECTING )
        m_aborted = true;
}

void GPSCommunicationController::start()
{
    if ( isRunning() ) {
       m_controller->stop();
    }

    delete m_commbase;
    m_commbase = NULL;

    delete m_communicationClient;
    m_communicationClient = NULL;

    delete m_controller;
    m_controller = NULL;

    // get gps config from config file
    ConfigBase* configBase = ConfigHelper2::getConfig();
    SerialConfigMgr cfgMgr( configBase, _T("/gps") );
    cfgMgr.load( &m_serialConfig );

    if ( !m_serialConfig.isEnabled() ) {
        sendOfflineEvent();
        return;
    }

    m_commbase = new SerialPortHandler( &m_serialConfig );
    m_communicationClient = new GpsCommClient( m_evtHandler );

    m_controller = new CommunicationController2();
    m_controller->setCommunicationBase( m_commbase );
    m_controller->registerClient( m_communicationClient );
    m_controller->setEventHandler( this );

    m_controller->start();

    sendOnlineEvent();
}

void GPSCommunicationController::stop()
{
    if ( isRunning() ) {
        m_controller->stop();
    }

    delete m_commbase;
    m_commbase = NULL;

    delete m_communicationClient;
    m_communicationClient = NULL;

    delete m_controller;
    m_controller = NULL;

    sendOfflineEvent();
}

bool GPSCommunicationController::isRunning()
{
    if ( m_controller ) {
        LOGP( wxString::Format( _T("m_controller->isRunning(): %d"), m_controller->isRunning() ) );
        return m_controller->isRunning();
    } else {
        LOG
        return false;
    }
}

int GPSCommunicationController::getGpsState()
{
    GpsCommClient* client = dynamic_cast<GpsCommClient*>(m_communicationClient);
    if ( client ) {
        return client->getGPSState();
    } else {
        return GPSInfo::NOT_CONNECTED;
    }
}

void GPSCommunicationController::detectGPS()
{
    m_state = DETECTING;
    m_aborted = false;

    if ( m_availablePorts.GetCount() == 0 )
    {
        GPSCommunicationEvent event;
        event.state = GPSCommunicationEvent::INVALID_PORT;

        event.isSearching = true;
        event.searchstate = GPSCommunicationEvent::SEARCH_ABORTED;
        event.userdata = 100.0;

        if ( m_evtHandler )
            m_evtHandler->AddPendingEvent( event );

        m_state = IDLE;

        return;
    }

    m_currentPortIdx = 0;
    m_currentBrIdx = -1;

    m_searchCounter = 0;

    searchNext();
}

void GPSCommunicationController::testGPSPort( const wxString& comport, int baudrate )
{
    m_state = TESTING;
    m_aborted = false;

    if ( m_availablePorts.Index( comport, false ) == wxNOT_FOUND )
    {
        // LOGP( comport + _T(" is not a valid port!"))

        GPSCommunicationEvent event;

        event.state = GPSCommunicationEvent::INVALID_PORT;
        event.portName = comport;
        event.baudRate = baudrate;
        event.isTesting = true;

        if ( m_evtHandler )
            m_evtHandler->AddPendingEvent( event );

        return;
    }


    if ( m_availableBaudrates.Index( baudrate ) == wxNOT_FOUND )
    {
        // LOGP( wxString::Format( _T("%d is not a valid baudrate!") , baudrate ) )

        GPSCommunicationEvent event;

        event.state = GPSCommunicationEvent::INVALID_BAUDRATE;
        event.portName = comport;
        event.baudRate = baudrate;
        event.isTesting = true;

        if ( m_evtHandler )
            m_evtHandler->AddPendingEvent( event );

        return;
    }

    m_currentPortIdx = m_availablePorts.Index( comport );
    m_currentBrIdx = m_availableBaudrates.Index( baudrate );

    reset( comport, baudrate );
}

void GPSCommunicationController::reset( const wxString& comport, int baudrate )
{
    // stop process first
    if ( m_controller && m_controller->isRunning() )if ( m_controller && m_controller->isRunning() )
        m_controller->stop();

    if ( m_commbase )
        delete m_commbase;

    if ( m_communicationClient )
        delete m_communicationClient;

    if ( m_controller )
        delete m_controller;

    SerialConfig* scfg = new SerialConfig;
    scfg->setPortName( comport );
    scfg->setBaudRate( baudrate );

    // LOGP( _T("port=")+comport+wxString::Format(_T(", br=%d"), baudrate ))

    m_commbase = new SerialPortHandler( scfg );

    m_communicationClient = new GPSDetectorClient( this );

//    m_controller = new CommunicationController2( m_communicationClient, m_commbase );
    m_controller = new CommunicationController2();
    m_controller->setCommunicationBase( m_commbase );
    m_controller->registerClient( m_communicationClient );

    m_controller->start();
}

void GPSCommunicationController::searchNext()
{
    if ( m_aborted )
    {
        GPSCommunicationEvent event;
        event.state = GPSCommunicationEvent::NOT_FOUND;
        event.isSearching = true;
        event.searchstate = GPSCommunicationEvent::SEARCH_ABORTED;
        event.userdata = 100.0;

        if ( m_evtHandler )
            m_evtHandler->AddPendingEvent( event );

        return;
    }

    ++m_currentBrIdx;

    if ( m_currentBrIdx >= (signed int)m_availableBaudrates.GetCount() )
    {
        ++m_currentPortIdx;
        m_currentBrIdx = 0;
    }

    // no more ports
    if ( m_currentPortIdx >= (signed int)m_availablePorts.GetCount() )
    {
        GPSCommunicationEvent event;
        event.state = GPSCommunicationEvent::NOT_FOUND;
        event.searchstate = GPSCommunicationEvent::SEARCH_END;
        event.isSearching = true;
        event.userdata = 100.0;

        if ( m_evtHandler )
            m_evtHandler->AddPendingEvent( event );

        return;
    }

    wxString portName = m_availablePorts[m_currentPortIdx];
    int baudrate = m_availableBaudrates[m_currentBrIdx];

    // LOGP( _T("Detecting port ") + portName + wxString::Format( _T(", at br=%d"), baudrate ) )

    // send event about testing port
    if ( m_evtHandler )
    {
        GPSCommunicationEvent event;

        event.isSearching = true;
        event.searchstate = GPSCommunicationEvent::SEARCH_START;
        event.portName = portName;
        event.baudRate = baudrate;

        m_evtHandler->AddPendingEvent( event );
    }
    reset( portName, baudrate );
}

void GPSCommunicationController::OnGPSEVent( GPSCommunicationEvent& event )
{
    GPSDataStruct data = event.gpsdata;

    // add portname & baudrate to the event
    event.portName = m_availablePorts[m_currentPortIdx];
    event.baudRate = m_availableBaudrates[m_currentBrIdx];

    unsigned int totalComb = m_availableBaudrates.GetCount() * m_availablePorts.GetCount();

    //
    if ( m_state == TESTING )
    {
        event.isTesting = true;

        switch ( event.state )
        {
            // timeout, cannot open port
            case GPSCommunicationEvent::CANNOT_OPEN_PORT:
                LOGP(_T("Timeout"));

                if ( m_evtHandler )
                    m_evtHandler->AddPendingEvent( event );

                break;

            // no valid gps data found
            case GPSCommunicationEvent::NOT_FOUND:
                // LOGP( _T("No valid GPS data found"));

                if ( m_evtHandler )
                    m_evtHandler->AddPendingEvent( event );

                break;

            // valid gps data found
            case GPSCommunicationEvent::GPS_FOUND:
                // LOGP( _T("GPS Found") )

                if ( m_evtHandler )
                    m_evtHandler->AddPendingEvent( event );

                break;

            default:
                break;
        }

        m_state = IDLE;
    }

    // detecting
    if ( m_state == DETECTING )
    {
        switch ( event.state )
        {
            // timeout, cannot open port
            case GPSCommunicationEvent::CANNOT_OPEN_PORT:
                // LOGP(_T("Cannot open port"));

                ++m_currentPortIdx;
                m_currentBrIdx = -1;

                m_searchCounter += m_availableBaudrates.GetCount();

                event.userdata = (double)m_searchCounter * 100.0 / (double) totalComb;
                event.isSearching = true;

                if ( m_evtHandler )
                {
                    LOGP(_T("sent"))
                    m_evtHandler->AddPendingEvent( event );
                }


                searchNext();

                break;

            // no valid gps data found
            case GPSCommunicationEvent::NOT_FOUND:
                // LOGP( _T("No valid GPS data found"));

                ++m_searchCounter;

                event.userdata = (double)m_searchCounter * 100.0 / (double) totalComb;
                event.isSearching = true;

                if ( m_evtHandler )
                    m_evtHandler->AddPendingEvent( event );

                searchNext();

                break;

            // valid gps data found
            case GPSCommunicationEvent::GPS_FOUND:
                // LOGP( _T("GPS Found") )

                event.userdata = 100.0;
                event.isSearching = true;

                if ( m_evtHandler )
                    m_evtHandler->AddPendingEvent( event );

                break;

            default:
                break;
        }
    }

}

void GPSCommunicationController::OnCommunicationEvent( CommunicationEvent& event )
{

}

void GPSCommunicationController::sendOfflineEvent()
{
    GPSCommEvent event;
    event.state = GPSInfo::OFFLINE;
    m_evtHandler->AddPendingEvent( event );
}

void GPSCommunicationController::sendOnlineEvent()
{
    GPSCommEvent event;
    event.state = GPSInfo::ONLINE;
    m_evtHandler->AddPendingEvent( event );
}
