#ifndef SDCARDDIALOG_H
#define SDCARDDIALOG_H

#include <wx/wx.h>
#include <wx/grid.h>

#include "../common/types/IntArray.h"
#include "../events/SDCardEvent.h"
#include "../events/VemsDeviceControllerEvent.h"
#include "../ProgressBar.h"
#include <wx/animate.h>
#include <wx/datetime.h>

class VemsDeviceControllerBase;
class SDCardController;
class SDImageHandler;
class SIPRFileLogger;
/**

SDCardSearchPanel - displays progress animation and the result of the searching

**/
class SDCardSearchPanel : public wxPanel {
    public:
        SDCardSearchPanel( wxWindow* parent );

        // state of animation
        // cardstate: result of the search - see implementation
        void         setState( unsigned int i, int cardstate = -1 );

        // returns state
        unsigned int getState() const { return m_state; }

    private:

        wxAnimationCtrl*    m_anim;
        unsigned int        m_state;
        wxStaticText*       m_searchTxt;
};

/**

SDCardDownloadPanel - displays the list of the logfiles on SD Card, a summary & progress bars

**/

class SDCardDownloadPanel : public wxPanel {
    public:
        SDCardDownloadPanel( wxWindow* parent );

        // updates the filelistbox - see MMCFilesDesc
        void            UpdateFiles( const MMCFilesDescA& files );

    private:

        friend class SDCardDialog;

        // control identifiers
        enum {
            idFileListGrid,
            idSummaryGrid,
            idSaveButton,
            idFileListBox,
            idGetSummaryButt,
            idSelectAllButton,
            idCancelButton,
            idMMCCloseButton
        };

        // array of the filedescriptors
        MMCFilesDescA       m_files;

        // state of checkboxes in filelistbox
        bool                m_isAllSelected;

        // summary grid
        wxGrid*             m_summaryGrid;

        // filelist
        wxCheckListBox*     m_fileListBox;

        // select & save all buttons
        wxButton*           m_selallbutton;
        wxButton*           m_saveselbutton;

        // summary & close buttons
        wxButton*           m_summbutt;
        wxButton*           m_closeButton;

        // progressbars
        ProgressBar *       m_summaryProgress;
        ProgressBar*        m_overallProgress;
        ProgressBar*        m_logProgress;

        // display elapsed & estimated time (downloading)
        wxStaticText*       m_elapsedTime;
        wxStaticText*       m_estimatedTime;
        //wxStaticText*       m_remainingTime;

        // input: how many samples to use for the summary
        wxTextCtrl  *       m_sampleCntCtrl;

    private:

        // create GUI elements
        void        buildGUI( wxWindow* parent );

        // enable or disable controls
        void        enableButtons( bool enable = true );

        // update the summary grid
        void        showSummary( const wxVariantHash& data );

        // update elapsed&estimated time
        void        setProgressTime( long elapsed, long estimated );

        // on select all button
        void        OnButton( wxCommandEvent& event );


        DECLARE_EVENT_TABLE()
};

class FirmwareInfo;
class STDataModelBase;
class MTConfigData;


/**

SDCardDialog

**/

class SDCardDialog : public wxDialog
{
    public:
        // ctor - for V3
        SDCardDialog( wxWindow * parent, FirmwareInfo* fwinfo, STDataModelBase* model,
            MTConfigData* mtconfig, VemsDeviceControllerBase* controller );

        // ctor - for image file
        SDCardDialog( wxWindow * parent, FirmwareInfo* fwinfo, STDataModelBase* model,
            MTConfigData* mtconfig, SDImageHandler* sdhandler );

        virtual ~SDCardDialog();

    private:

        // current action
        enum Action { SEARCH, DOWNLOAD, SUMMARY, IDLE };
        Action                  m_currentJob;

        friend class SDCardDownloadPanel;
        friend class SDCardSearchPanel;

        // the panels of the dialog
        SDCardSearchPanel*      m_searchPanel;
        SDCardDownloadPanel*    m_downlPanel;

        // array of file descriptors
        MMCFilesDescA           m_filesDesc;

        // the event & 'action' handlers
        VemsDeviceControllerBase*   m_controller;
        SDCardController*           m_sdCardController;
        SDImageHandler*             m_sdhandler;

        // for creating vemslog
        FirmwareInfo*           m_fwinfo;
        STDataModelBase*        m_model;
        MTConfigData*           m_mtconfig;

        // backup the old event handler
        wxEvtHandler*           m_oldevtHandler;

        // saving files
        // array of the indices of selected files
        IntArray                m_fileindices;
        // how many files were selected
        size_t                  m_numselected;
        // sum of time of all selected files
        size_t                  m_timeAll;
        // sum of time of saved files
        size_t                  m_timeCtr;
        // counter for m_fileindices
        size_t                  m_counter;

        // was cancel button pressed
        bool                    m_isCancelled;

        // elapsed time when save single file
        wxDateTime              m_startTime;

        // start time (saving)
        wxDateTime              m_progTime;
        wxStopWatch             m_sw;

        // old baudrate
        size_t                  m_oldBaudrate;
        bool                    m_resetState;

    private:
        // create gui elements
        void        buildGUI( wxWindow* parent );

        // save files, save buffer to vemslog
        // data is an array of 512-bytes sized blocks
        void        SaveSelectedFiles();
        void        SaveDataToFile( unsigned char* data, size_t len );
        void        SaveTFDataToFile( unsigned char* data, size_t len );
        void        ExportTFDataToLogV3( unsigned char* data, size_t len );

        // control is done by sdcard events - see implementation
        void        OnSDCardEvent( SDCardEvent& event );

        void        onDeviceDisconnected( VemsDeviceControllerEvent& event );
        // on baudratechanged
        void        onBaudRateChanged( VemsDeviceControllerEvent& event );

        /// wxevents - called when pressing a button of the panels
        void        OnSaveSelected( wxCommandEvent& event );
        void        OnCreateSummary( wxCommandEvent& event );
        void        OnCloseButton( wxCommandEvent& event );
        void        OnCancelButton( wxCommandEvent& event );
        ///
        void        OnClose( wxCloseEvent& event );

        // change back to original BR
        void        ResetOriginalState();

        wxString initSIPRFileLogger(SIPRFileLogger& siprFileLogger);
        void initSIPRData(uint8_t* siprdata);

        DECLARE_EVENT_TABLE()
};

/**
    SDCardResultDialog
**/

class SDCardResultDialog : public wxDialog
{
    public:
        SDCardResultDialog( wxWindow* parent, const wxString& message, const wxString& saved_path = wxEmptyString );

    private:
        wxString    m_savedPath;

    private:
        void        OnButtonClick( wxCommandEvent& event );

        DECLARE_EVENT_TABLE()
};

#endif // SDCARDDIALOG_H
