#include "SDCardDirector.h"
#include "../vemsprotocol/VemsDeviceControllerBase.h"
#include "../vemsprotocol/SDCardController.h"
#include "../events/SDCardEvent.h"
#include "../common/VemsContext.h"
#include "../FirmwareVersion.h"
#include "../logger.h"


SDCardDirector::SDCardDirector(  wxWindow* parent, VemsDeviceControllerBase* deviceController )
{
    m_parent            = parent;
    m_deviceController  = deviceController;
    m_task              = TASK_OTHER;

    m_sdCardController = new SDCardController( deviceController, this );

    m_sdcardDlg = new wxProgressDialog( _T(""), _T(""), 100, m_parent );
    m_sdcardDlg->SetSize(250,140);
    m_sdcardDlg->Hide();

    Connect( wxEVT_SDCARD, SDCardEventHandler( SDCardDirector::onSDCardEvent ), NULL, this );
}

SDCardDirector::~SDCardDirector()
{
    ///TODO
    Disconnect( wxEVT_SDCARD, SDCardEventHandler( SDCardDirector::onSDCardEvent ), NULL, this );
}

void SDCardDirector::startLog()
{
    m_deviceController->stop( VDC_FLAG_NOEVENT );

    m_task = TASK_START;

    m_sdCardController->getFwVersion();

    // m_sdCardController->startLog();
}

void SDCardDirector::stopLog()
{
    m_task = TASK_OTHER;
    m_deviceController->stop( VDC_FLAG_NOEVENT );
    m_sdCardController->stopLog();
}

void SDCardDirector::erase()
{
    m_task = TASK_OTHER;
    m_deviceController->stop( VDC_FLAG_NOEVENT );
    m_sdCardController->erase();
}

void SDCardDirector::onSDCardEvent( SDCardEvent& event )
{
    switch ( event.GetType() )
    {
        case SDCardEventTypes::FIRMWARE_VERSION:
        {
            m_deviceController->stop( VDC_FLAG_NOEVENT );
            if ( checkDifferentFw( event.GetString(), event.GetInt() ) )
            {
                switch ( m_task )
                {
                    case TASK_START:
                        LOG
                        m_sdCardController->startLog();
                        break;
                    default:
                        m_deviceController->resume();
                        break;
                }
            }
            else
            {
                m_deviceController->resume();
            }

            break;
        }
        case SDCardEventTypes::COMM_FAILED:
            m_sdcardDlg->Update( 0 ); m_sdcardDlg->Hide();
            wxMessageBox( _T("SDCard - Communication failed"), _T("ERROR!"), wxOK | wxICON_ERROR );
            m_deviceController->resume();
            break;

        case SDCardEventTypes::TIMEOUT:
            m_sdcardDlg->Update( 0 ); m_sdcardDlg->Hide();
            wxMessageBox( _T("SDCard - Timeout"), _T("Timeout!"), wxOK | wxICON_HAND );
            m_deviceController->resume();
            break;

        case SDCardEventTypes::LOGGING_STOPPED:
            wxMessageBox( _T("SDCard - Logging stopped"), _T("Info"), wxOK | wxICON_INFORMATION );
            m_deviceController->resume();
            break;

        case SDCardEventTypes::LOGGING_STARTED:
        {
            m_sdcardDlg->Update( 0 ); m_sdcardDlg->Hide();
            wxMessageBox( _T("SDCard - Logging started"), _T("Info"), wxOK | wxICON_INFORMATION );
            m_deviceController->resume();
            break;
        }

        case SDCardEventTypes::STARTLOG_WAITING:
        {
            if ( m_sdcardDlg->IsShown() == false )
                m_sdcardDlg->Show( true );

            m_sdcardDlg->Update( event.GetInt(), _T("SDCard startlog. Waiting...") );
            break;
        }

        case SDCardEventTypes::STARTLOG_FAILED:
        {
            m_sdcardDlg->Update( 0 ); m_sdcardDlg->Hide();
            wxMessageBox( _T("SDCard - Startlog failed"), _T("Error"), wxOK | wxICON_HAND );
            m_deviceController->resume();
            break;
        }

        case SDCardEventTypes::STOPLOG_FAILED:
        {
            m_sdcardDlg->Update( 0 ); m_sdcardDlg->Hide();
            wxMessageBox( _T("SDCard - Stoplog failed"), _T("Error"), wxOK | wxICON_HAND );
            m_deviceController->resume();
            break;
        }

        case SDCardEventTypes::ERASE_SUCCESS:
        {
            m_sdcardDlg->Update( 0 ); m_sdcardDlg->Hide();
            wxMessageBox( _T("SDCard - SDCard empty"), _T("Info"), wxOK | wxICON_INFORMATION );
            m_deviceController->resume();
            break;
        }

        case SDCardEventTypes::ERASE_FAILED:
        {
            m_sdcardDlg->Update( 0 ); m_sdcardDlg->Hide();
            wxMessageBox( _T("SDCard - Erase failed"), _T("Error"), wxOK | wxICON_ERROR );
            m_deviceController->resume();
            break;
        }

        case SDCardEventTypes::ERASE_WAITING:
        {
            if ( m_sdcardDlg->IsShown() == false )
                m_sdcardDlg->Show( true );

            m_sdcardDlg->Update( event.GetInt(), _T("Checking SD card") );
            break;
        }

        default: break;
    }
}

bool SDCardDirector::checkDifferentFw( const wxString& fwVer, unsigned int fwAsInt )
{
    FirmwareInfo* fwInfo = dynamic_cast<FirmwareInfo*>( VemsContext::getObject( _T("FirmwareInfo") ) );
    if ( fwInfo )
    {
        wxString fws = fwInfo->getFirmwareVersionAsString();
        if ( fws < _T("1.1.87") )
            return true;

        int ecuProhibit;

        fwInfo->getProhibit( ecuProhibit );

        if ( fwAsInt == 0 || ( fwVer == _T("0.0.0") ) )
        {
            // sd card is empty yet
            return true;
        }

        if ( fwAsInt != ecuProhibit )
        {
            wxString msg;
            msg = _T("SD Card firmware version (") + fwVer + wxString::Format(_T(", %d)"), fwAsInt );
            msg += _T(" and V3 firmware version (") + fws + wxString::Format( _T(", %d)"), ecuProhibit );
            msg += _T(" are different\n\nStart logging may corrupt your existing data on the SD Card!\n\n");
            msg += _T("It is recommended to save all data from SD Card first and then erase the card\n");
            msg += _T("NOTE: saving data from SD Card before firmware upgrade is always recommended\n\n");
            msg += _T("Would you like to continue and start logging?");

            int ans = wxMessageBox( msg, _T("Warning"), wxYES_NO | wxICON_WARNING );

            if ( ans == wxYES )
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        else
        {
            return true;
        }
    }
    else
    {
        int ans = wxMessageBox( _T("Could not obtain firmware version. Continue?"), _T("Warning"), wxYES_NO | wxICON_ERROR );
        if ( ans == wxID_YES )
            return true;
        else
            return false;
    }
}
