#include "RTDataMgrBase.h"

#include "RTDataViewerBase.h"
#include "../SymbolMgrBase.h"
#include "../logger.h"
#include "../common/typedef.h"

RTDataMgrBase::RTDataMgrBase(SymbolMgrBase* symbolmgrbase)
{
    m_symbolMgrBase = symbolmgrbase;

    m_isRequireChanged = true;
    m_minRefreshInterval = 0;
}

RTDataMgrBase::~RTDataMgrBase()
{
    //dtor
}

/**
 The user can change the displayed item (symbol) on the
 viewer so that it needs get updated from all registered viewers.
*/
void RTDataMgrBase::getRequiredSymbols()
{
    if( !m_isRequireChanged ) return;

    if( m_RTDataViewerBaseA->GetCount() < 1 ) return;

    int i;
    RTDataViewerBase * rtDataViewer;

    //m_DoubleHash.clear();

    for( i = (m_RTDataViewerBaseA->GetCount()-1); i >= 0; i-- )
    {
        rtDataViewer = m_RTDataViewerBaseA->Item(i);

        if( rtDataViewer == NULL )
        {
            m_RTDataViewerBaseA->RemoveAt(i);
            continue;
        }

        //rtDataViewer->require( &m_DoubleHash );
        rtDataViewer->require();
    }

    setRequestedSymbols();

    m_isRequireChanged = false;
}

/**
    Data has been changed so that needs to notify all viewers
*/
void RTDataMgrBase::notifyViewers()
{
    size_t i;
    RTDataViewerBase * rtDataViewer;

    for(i = 0; i < m_RTDataViewerBaseA->GetCount(); i++)
    {
        rtDataViewer = m_RTDataViewerBaseA->Item(i);
        if( rtDataViewer == NULL ) { continue; }
        //wxMutexGuiEnter(); // BAD SIDE EFFECT (eat CPU) ON SLOW PC
        //Logger::logf( LOG_MSG, _T("RTDataMgrBase::notifyViewers() - before refresh") );

        rtDataViewer->refresh(&m_RegisteredSymbolTable);
        //wxMutexGuiLeave();
    }
}

void RTDataMgrBase::filteredNotifyViewers()
{

    if(m_minRefreshInterval > 0){
        if(m_packetFilterTimer.Time() < m_minRefreshInterval) return;
        m_packetFilterTimer.Start(0);
    }

    notifyViewers();
}

/**
    Invoke:
    getRequiredSymbols()
    refresh
    notifyViewers()
*/
int RTDataMgrBase::updateViewers()
{

    getRequiredSymbols();

    int result = updateViewerModel();

    if(result < 0){
        Logger::logf(LOG_MSG, _T("RTDataMgrBase::updateViewers() refresh failed") );
        return result;
    }

    notifyViewers();



    return 0;
}

void RTDataMgrBase::RegisterViewer( RTDataViewerBase * viewer )
{
    m_RTDataViewerBaseA->Add( viewer );
    requireChanged();
    viewer->setRegistered( true );

    //Logger::logf(LOG_MSG, _T("RTDataMgrBase::RegisterViewer() DoubleHash size: %i"), m_RegisteredSymbolTable.size() );
    //viewer->refresh( &m_RegisteredSymbolTable );
}

void RTDataMgrBase::UnRegisterViewer(RTDataViewerBase * viewer)
{
    if( !viewer ) return;

    int idx = m_RTDataViewerBaseA->Index( viewer );
    if( idx == wxNOT_FOUND ) return;

    viewer->unregisterSymbols();
    viewer->setRegistered(false);
    m_RTDataViewerBaseA->Item( idx ) = NULL;
    requireChanged();
}

size_t RTDataMgrBase::getRegisteredCount()
{
    return m_RTDataViewerBaseA->GetCount();
}

void RTDataMgrBase::requireChanged()
{
    m_isRequireChanged = true;
}

void RTDataMgrBase::UnregisterAll()
{
    for( uint32_t i = 0; i < m_RTDataViewerBaseA->GetCount(); ++i )
    {
        RTDataViewerBase* viewer = m_RTDataViewerBaseA->Item( i );
        UnRegisterViewer( viewer );
    }
}

void RTDataMgrBase::registerSymbol( const wxString& symbolName )
{
    if(m_symbolMgrBase && !m_symbolMgrBase->isExistSymbol(symbolName)) return;

    int count = m_requestedSymbolRegistry[ symbolName ];

    ++count;

    m_requestedSymbolRegistry[ symbolName ] = count;
    //Logger::logf( LOG_MSG, _T("RTDataMgrBase::registerSymbol: ") + symbolName + _T("=%d"), count );
}

void RTDataMgrBase::unregisterSymbol( const wxString& symbolName )
{

    if( m_requestedSymbolRegistry.find( symbolName ) == m_requestedSymbolRegistry.end() ) { return; }

    int count = m_requestedSymbolRegistry[ symbolName ];

    //Logger::logf( LOG_MSG, _T("unregisterSymbol : ") + symbolName + _T(", %d"), count );

    if( count > 0 ) { --count; }

    //Logger::logf( LOG_MSG, _T("RTDataMgrBase::unregisterSymbol: ") + symbolName + _T("=%d"), count );
    m_requestedSymbolRegistry[ symbolName ] = count;

    if( count == 0 ){
        m_RegisteredSymbolTable.erase( symbolName );
    }
}

void RTDataMgrBase::setRequestedSymbols()
{
    String2IntHash::const_iterator CI;

    for( CI = m_requestedSymbolRegistry.begin(); CI != m_requestedSymbolRegistry.end(); ++CI )
    {

        if( CI->second == 0 )
        {
            m_RegisteredSymbolTable.erase( CI->first );
            //Logger::logf( LOG_MSG, _T("RTDataMgrBase::setRequestedSymbols() - unregister : ") + CI->first );
        }
        else
        {
            m_RegisteredSymbolTable[ CI->first ];
//            Logger::logf( LOG_MSG, _T("RTDataMgrBase::setRequestedSymbols() - register : ") + CI->first + _T(", count : %d"), CI->second );
        }
    }

}

void RTDataMgrBase::reset()
{
//    m_requestedSymbolRegistry.clear();
//    m_RegisteredSymbolTable.clear();
}
