#include "RTDataAquisition.h"

#include "RTDataStat.h"

#include "../logger.h"
#include "../VemsTuneController.h"
#include "../events/InputEvent.h"
#include "../loghistory/LogHistoryBase.h"
#include "../stdatamanager/STDataMgrBase.h"
#include "../vemsgauge/VemsTableDesc.h"
#include "../vemsprotocol/SIPRPacket.h"
#include "../vemsprotocol/SIPRTypes.h"

int idStartButton   = wxNewId();
int idStopButton    = wxNewId();
int idClearButton   = wxNewId();
int idAnalyzeButton = wxNewId();

BEGIN_EVENT_TABLE( RTDataAquisition, wxFrame )
    EVT_BUTTON( idStartButton, RTDataAquisition::OnStartStop )
    EVT_BUTTON( idStopButton, RTDataAquisition::OnStartStop )
    EVT_BUTTON( idClearButton, RTDataAquisition::OnClearButton )
    EVT_BUTTON( idAnalyzeButton, RTDataAquisition::OnAnalyzeButton )
END_EVENT_TABLE()

RTDataAquisition::RTDataAquisition( wxWindow* parent, VemsTuneController* vtc, LogHistoryBase* loghistory, VemsTableDesc* tableDesc )
{
    m_statFrame = NULL;

    m_isstarted = false;

    m_parent = parent;

    m_vtc = vtc;
    m_loghistory = loghistory;

    m_stDataMgr = m_loghistory->getSTDataMgr();

    m_symbolMgr = m_loghistory->getSymbolMgr();

    m_tableDesc = tableDesc;

    m_xBinSymbol = m_symbolMgr->getSymbol( tableDesc->xSTSymbol );
    m_yBinSymbol = m_symbolMgr->getSymbol( tableDesc->ySTSymbol );

    m_xRTSymbol = m_symbolMgr->getSymbol( tableDesc->xRTSymbol );
    m_yRTSymbol = m_symbolMgr->getSymbol( tableDesc->yRTSymbol );

    m_cols = (unsigned int) m_xBinSymbol->getShape0();
    m_rows = (unsigned int) m_yBinSymbol->getShape0();

    m_registeredSymbolNames.Add( m_xRTSymbol->getName() );
    m_registeredSymbolNames.Add( m_yRTSymbol->getName() );

    m_symreg = false;

    m_matrix = new VemsMatrix<unsigned int>( m_rows, m_cols, 0 );

    m_startIndex = 0;
    m_endIndex = 0;

    m_tableClosed = false;

    calcLimits();

    buildGUI( );
}

RTDataAquisition::~RTDataAquisition()
{
    unregisterSymbols();
    m_rtDataMgr->UnRegisterViewer( this );
    delete m_matrix;
}

void RTDataAquisition::refresh( DoubleHash* table )
{
    if ( m_isstarted == false )
    {
        return;
    }

    double xsym = 0,  ysym = 0;

    DoubleHash::iterator it;

    it = table->find( m_xRTSymbol->getName() );

    if ( it != table->end() )
    {
        xsym = it->second;
    }

    it = table->find( m_yRTSymbol->getName() );

    if ( it != table->end() )
    {
        ysym = it->second;
    }

    updateTable( (int) xsym, (int) ysym );
}

void RTDataAquisition::require()
{
    if ( m_symreg )
    {
        return;
    }

    for ( size_t i = 0; i < m_registeredSymbolNames.GetCount(); ++i )
    {
        m_rtDataMgr->registerSymbol( m_registeredSymbolNames[i] );
    }

    m_symreg = true;

    requireChanged();
}

void RTDataAquisition::unregisterSymbols()
{
    if ( m_symreg == false )
    {
        return;
    }

    for ( size_t i = 0; i < m_registeredSymbolNames.GetCount(); ++i )
    {
        m_rtDataMgr->unregisterSymbol( m_registeredSymbolNames[i] );
    }

    m_symreg = false;
}

void RTDataAquisition::buildGUI()
{
    Create( m_parent, wxID_ANY, _T("Data aquisition"), wxDefaultPosition, wxDefaultSize,
        wxDEFAULT_FRAME_STYLE | wxFRAME_FLOAT_ON_PARENT );

    SetBackgroundColour( wxSystemSettings::GetColour( wxSYS_COLOUR_3DFACE ) );
    ClearBackground();

    wxBoxSizer* mainsizer = new wxBoxSizer( wxVERTICAL );

    // min #data
    wxBoxSizer* s1 = new wxBoxSizer( wxHORIZONTAL );
    wxStaticText* txt = new wxStaticText( this, -1, _T("Min. number of data") );
    m_minDataTxt = new wxTextCtrl( this, -1, _T("20"), wxDefaultPosition, wxSize(50,20) );
    m_minData = 20;

    s1->Add( txt, 0, wxALL, 5 );
    s1->Add( m_minDataTxt, 0, wxALL, 5 );
    mainsizer->Add( s1, 0, wxALL, 5 );

    // buttons
    s1 = new wxBoxSizer( wxHORIZONTAL );
    m_startButton = new wxButton( this, idStartButton, _T("Start") );
    m_stopButton = new wxButton( this, idStopButton, _T("Stop") );
    m_clearButton = new wxButton( this, idClearButton, _T("Reset") );
    m_analyzeButton = new wxButton( this, idAnalyzeButton, _T("Analyze data") );

    s1->Add( m_startButton, 0, wxALL, 5 );
    s1->Add( m_stopButton, 0, wxALL, 5 );
    s1->Add( m_clearButton, 0, wxALL, 5 );
    s1->Add( m_analyzeButton, 0, wxALL, 5 );
    mainsizer->Add( s1, 1, wxALL, 5 );

    m_stopButton->Disable();
    m_clearButton->Disable();
    m_analyzeButton->Disable();

    // VE table ( grid )
    createGrid();

    mainsizer->Add( m_table, 0, wxALL, 5 );

    ///

    SetSizer( mainsizer );
    mainsizer->Fit( this );
    mainsizer->SetSizeHints( this );
}

void RTDataAquisition::createGrid()
{
    m_table = new wxGrid( this, wxID_ANY );
    m_table->CreateGrid( m_rows, m_cols );

    wxFont headerFont( 8, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD );
    m_table->SetLabelFont( headerFont );
    m_table->SetRowLabelAlignment( wxALIGN_CENTRE, wxALIGN_CENTRE );
    m_table->SetColLabelAlignment( wxALIGN_CENTRE, wxALIGN_CENTRE );

    for ( size_t i = 0; i < m_cols; ++i )
    {
        m_table->SetColLabelValue( i, m_xbins[i] );
        m_table->SetColSize( i, 45 );
    }

    for ( size_t i = 0; i < m_rows; ++i )
    {
        m_table->SetRowLabelValue( i, m_ybins[i] );
    }

    m_table->SetDefaultCellAlignment( wxALIGN_CENTRE, wxALIGN_CENTRE );
}

void RTDataAquisition::calcLimits()
{
    // row names
    int i;
    int prev_value = 0;

    for ( i = (int)( m_rows - 1 ); i >= 0; --i )
    {
        int value = (int) m_stDataMgr->getDataFromViewerModel( *m_yBinSymbol, 0, i );
        if ( i < (int)( m_rows - 1 ) )
        {
            int limit = ( value + prev_value ) / 2;
            m_HorizLimits.Add( limit );
        }

        if ( i == (int)( m_rows - 2 ) )
            m_ymax = prev_value + ( prev_value - value );

        if ( i == 0 )
            m_ymin = value - ( prev_value - value );

        prev_value = value;

        m_ybins.Add( wxString::Format( _T("%d"), value ) );
    }

    // col names
    for ( i = 0; i < (int) m_cols; ++i )
    {
        int value = (int) m_stDataMgr->getDataFromViewerModel( *m_xBinSymbol, 0, i );

        if ( (int)i > 0 )
        {
            int limit = ( value + prev_value ) / 2;
            m_VertLimits.Add( limit );
        }

        if ( i == (int)( m_cols - 1 ) )
            m_xmax = value + ( value - prev_value );

        if ( i == 1 )
            m_xmin = prev_value - ( value - prev_value );

        prev_value = value;

        m_xbins.Add( wxString::Format( _T("%d"), value ) );
    }
}

void RTDataAquisition::OnStartStop( wxCommandEvent& event )
{
    int id = event.GetId();

    if ( id == idStartButton )
    {
        SendLogMarker( true );
        m_startButton->Disable();
        m_stopButton->Enable();
        m_clearButton->Disable();
        m_analyzeButton->Disable();

        m_isstarted = true;

        m_startIndex = m_loghistory->getHistoryCount();

        if ( m_loghistory->getLogFileName() != wxEmptyString )
            m_startIndex = 0;

    }

    if ( id == idStopButton )
    {
        m_isstarted = false;

        SendLogMarker( false );
        m_endIndex = m_loghistory->getHistoryCount();

        if ( m_loghistory->getLogFileName() != wxEmptyString )
            m_endIndex = m_loghistory->getHistoryCount() - 1;

        m_stopButton->Disable();
        m_startButton->Enable();

        bool hasData = ( ( m_endIndex - m_startIndex ) > 0 );
        m_clearButton->Enable( hasData );
        m_analyzeButton->Enable( hasData );
    }
    event.Skip();
}

void RTDataAquisition::OnClearButton( wxCommandEvent& event )
{
    m_analyzeButton->Disable();
    m_clearButton->Disable();

    m_startIndex = 0; m_endIndex = 0;

    event.Skip();
}

void RTDataAquisition::OnAnalyzeButton( wxCommandEvent& event )
{
    if ( m_tableClosed == true )
    {
        m_statFrame = NULL;
    }
    if ( m_statFrame == NULL )
    {
        m_statFrame = new RTDataStat( m_parent, m_loghistory, m_vtc->getLogHistoryObs(), m_tableDesc, true );
        m_rtDataMgr->RegisterViewer( m_statFrame );
        m_statFrame->SetDataMgr( m_vtc->getRTDataMgr() );

        m_statFrame->registerObserver( this );
    }

    m_statFrame->load( m_startIndex, m_endIndex );

    m_statFrame->Show();

    m_tableClosed = false;
}

void RTDataAquisition::updateTable( int xSym, int ySym )
{
    uint8_t i;

    int8_t matcol, matrow;

    matcol = matrow = -1;

    // out of range
    if ( xSym > m_xmax || xSym < m_xmin || ySym > m_ymax || ySym < m_ymin )
    {
        return;
    }

    for ( i = 0; i < m_cols - 1; ++i )
    {
        if ( xSym < m_VertLimits[i])
        {
            matcol = i;
            break;
        }
    }

    for ( i = 0; i < m_rows - 1; ++i )
    {
        if ( ySym > m_HorizLimits[i])
        {
            matrow = i;
            break;
        }
    }

    matcol = ( matcol != -1 ) ? matcol : m_cols - 1;
    matrow = ( matrow != -1 ) ? matrow : m_rows - 1;

    unsigned val = m_matrix->get( matrow, matcol );
    val++;
    m_matrix->set( matrow, matcol, val );

    wxString valStr;
    valStr << val;
    m_table->SetCellValue( matrow, matcol, valStr );

    wxString minDataStr = m_minDataTxt->GetValue();
    minDataStr.ToLong( &m_minData );

    if ( val > (unsigned)m_minData )
    {
        m_table->SetCellBackgroundColour( matrow, matcol, wxColor( 0, 166, 41 ) );
    }
}

void RTDataAquisition::SendLogMarker( bool start )
{
    uint8_t subtype;
    if ( start )
        subtype = SIPR_LOG_MARKER_START;
    else
        subtype = SIPR_LOG_MARKER_STOP;

    SIPRPacket packet( 0, SIPR_LOG_MARKER, 1, &subtype );

    InputEvent event;

    event.setPacket( packet );

    m_vtc->AddPendingEvent( event );
}

void RTDataAquisition::notify( int i, void* ptr )
{
    if ( ptr == m_statFrame )
    {
        if ( i == 1 )
        {
            m_tableClosed = true;
            m_statFrame->unRegisterObserver( this );
            m_rtDataMgr->UnRegisterViewer( m_statFrame );
        }
    }
}
