#ifndef RTPLOTTER_H
#define RTPLOTTER_H

#include <wx/wx.h>
#include <wx/dc.h>
#include <wx/dynarray.h>

#include "RTDataStatMath.h"
#include "../wxRealRect.h"
#include "../common/types/DoubleArray.h"
#include "../common/types/RealPointArray.h"
#include "../common/types/wxColourArray.h"
#include "../common/types/Int2DoubleArrayHash.h"

class wxAgg2;
class Int2IntHash;

class Curve {
    public:
        Curve();
        Curve(const Curve& c);
        ~Curve();
        double*     x;
        double*     y;
        unsigned    n;
        double      xmin, xmax, ymin, ymax;
        double      linewidth;
        wxString    name;
        wxString    descr;
        bool        fillcurve;
        wxColor     color;
        wxColor     fillcolor;
    private:
        bool        copied;
};

WX_DECLARE_OBJARRAY( Curve, CurveA );

/**
 how to draw mouse coordinates to the canvas:
 create an instance of RTPlotter with DC=NULL, parent = (frame) args where (frame) is
 a wxwindow that capture mouse events. in onmouse method call plotter->setmousepos(..)
 (it will refresh the frame) In onpaint call plotter->setdc(..) and the drawing functions
*/
class RTPlotter
{
    public:
        RTPlotter( wxDC* dc, wxWindow* parent = NULL );
        ~RTPlotter();

        void SetDC( wxDC* dc ) { m_dc = dc; }

        // initialization
        void SetViewPortSize( unsigned w, unsigned h );
        void SetViewPortPosition( unsigned x, unsigned y );
        void SetViewPortLimits( float minx, float maxx, float miny, float maxy );
        void SetAxisDivision( int vert = -1, int hor = -1 );
        void SetDisplayValues( bool displayvalues ) { m_displayValues = displayvalues; }

        void SetLineColor( wxColour color ) { m_linecolor = color; }
        void SetLineWidth( double width ) { m_linewidth = width; }
        void SetDivisionColor( const wxColour& color ) { m_divcolor = color; }
        void SetFillColor( wxColour color ) { m_fillcolor = color; }
        void SetBackgroundColor( wxColour color ) { m_bgcolor = color; }
        void SetSelectionColor( wxColour color ) { m_selectionColor = color; }
        void SetAxisLabels( const wxString& xlabel, const wxString& ylabel );
        void SetPlotTitle( const wxString& title );

        // adding external functions, curves
        void AddFunction( const wxString& funcname, dblFunc f );
        void AddCurve( Curve curve );
        void AddCurve( const wxString& funcname, double xmin, double xmax, wxColour color = wxNullColour, const wxString& descr = wxEmptyString, bool fillcurve = false, wxColour fillColor = wxNullColour );
        void AddCurve( unsigned n, double *x, double *y, wxColour color = wxNullColour, const wxString& name = _T("data"), bool fillcurve = false, wxColour fillColor = wxNullColour );
        void AddCurve( const RealPointArray& points, wxColour color = wxNullColour, const wxString& name = _T("data"), bool fillcurve = false, wxColour fillColor = wxNullColour );
        void ClearCurves();

        // drawing functions
        void DrawCurves( bool eraseBg = false, bool autoscaleaxes = false );
        void DrawParametricPlot( dblFunc xt, dblFunc yt, double tmin, double tmax, unsigned points, wxColour color = wxNullColour, const wxString& descr = _T("parametric"), bool fillcurve = false, wxColour fillColor = wxNullColour  );
        void DrawHistogram( unsigned n, double* values, double xmin, double xmax, unsigned nbins, bool relative = false, bool isMax1 = false  );
        void DrawHistogram( const DoubleArray& values, double xmin, double xmax, unsigned int nbins, bool relative = false, bool isMax1 = false );
        void DrawPiechart( DoubleArray& values, const wxColourArr& colors, const wxArrayString& names = wxArrayString() );

        void DrawPolarPlot( dblFunc polarFunc, double tmin, double tmax, unsigned points );

        void DrawText( const wxString& text, unsigned int x, unsigned int y, double fontSize = -1 );

        // special: primary & secondary trigger
        //void DrawTriggers( const DoubleArray& primary, const DoubleArray& secondary, const wxColour& primColor, const wxColour& secColor, double xmin, double xmax );
        void DrawTriggers( Int2DoubleArrayHash& triggers, double xmin, double xmax, const Int2IntHash& triggerFrames );

        // canvas rendering
        void CreateCanvas( double mintextHeight = -1 );
        void Render();
        void RenderToMemoryDC( wxDC* dc );
        void DeleteCanvas();

        // mouse position on canvas
        void SetMousePos( wxPoint mousepos );
        void SetExtraText( const wxString& text = wxEmptyString );
        bool GetXYValue( unsigned scrx, unsigned scry, double& x, double& y);
        void SetSelection( double from = -1, double to = -1 );

        void Render3DFunc(dbl3dFunc func, double xmin, double xmax, double ymin, double ymax );
        void SetCameraPos( double x, double y, double z );

    private:

        enum CoordSystemType { CS_NONE, CS_CARTHESIAN, CS_POLAR, CS_3D };

        wxDC*                   m_dc;
        wxWindow*               m_parent;

        unsigned                m_width, m_height;
        unsigned                m_xpos, m_ypos;
        float                   m_minx, m_miny;
        float                   m_maxx, m_maxy;
        static FunctionA        m_functions;

        CurveA                  m_curves;

        int                     m_vert_div, m_hor_div;

        bool                    m_drawXValues;
        bool                    m_drawYValues;

        double                  m_epsilon;

        wxColour                m_linecolor;
        wxColour                m_fillcolor;
        wxColour                m_bgcolor;
        wxColour                m_selectionColor;
        wxColour                m_divcolor;

        wxRealRect              m_rectSize;

        float                   m_textHeight;

        double                  m_linewidth;
        wxAgg2*                 m_aggcanvas;

        bool                    m_displayValues;
        wxPoint                 m_mousepos;

        wxString                m_xaxislabel;
        wxString                m_yaxislabel;
        wxString                m_plottitle;

        CoordSystemType         m_coordSytem;

        double                  m_cx, m_cy, m_cz;

        wxString                m_xvalFormat;
        wxString                m_yvalFormat;

        wxString                m_extraText;

        double                  m_selectFrom, m_selectTo;

    private:

        void DrawAxes();
        void GetRectSize();
        void RenderValues();
        bool funcExists( const wxString& funcname )
            { FunctionA::iterator it = m_functions.find(funcname); return ( it != m_functions.end() );  }
        void DrawSelection();

};

#endif // RTPLOTTER_H
