#include "TriggerDetect.h"
#include "../logger.h"

/**
 * Detects average, peak of the input "sample" signal.
 * Besides the obvious wav sample processing, it can be useful for any other signal, like
 * the measured sparkphase and dwell signals.
 * The learnt peak and average values can be displayed on a gauge.
*/
TriggerDetect::TriggerDetect()
{
    up_threshold = MIDPOINT;
    lo_threshold = MIDPOINT;
    peak_p = MIDPOINT;
    peak_n = MIDPOINT;
    m_state = 0; // TODO: unknown state ?
    m_transitionCount = 0;

    m_minThreshold = 328; // appr 1% of 32767
    m_overthresholdsamplecnt = 47;
    m_threshold_strategy = 0;
    m_name = _T("unknown");
    m_toward_mid = TOWARD_MID;
    m_toward_peak = TOWARD_PEAK;
    m_toward_average = TOWARD_AVERAGE;

    m_thresh256 = THIRD;
    m_sample_received = 0;
    m_onePerArraySize = 1/(arraysize-2);
    m_highCounter = 0;
    m_lowCounter = 0;
}

TriggerDetect::~TriggerDetect()
{
    //dtor
}

/**
 * set the minimum number of samples the signal must be over threshold (to be detected)
*/
void TriggerDetect::setThresholdMinSamples( uint8_t sampleCount ){
    m_overthresholdsamplecnt = sampleCount;
}

void TriggerDetect::setName(const wxString &name)
{
    m_name = name;
}

/** digital lowpass "RC" filter, tau time constant depends on TOWARD_AVERAGE
 * for each sample it applies  *(65536 - TOWARD_AVERAGE)/ 65536
 * For example with TOWARD_AVERAGE=3
 * log( 1 / e ) / log( 1-3/65536)
 * ans =  21845  ( with 44100 sps that is appr 0.5 seconds ). Verify:
 * (1-3/65536) ** 21845
 * ans = 0.368  (which is 1/e)
 */
uint16_t TriggerDetect::average(uint16_t sample)
{
    double s = (double)sample * m_toward_average;
    s += (double)m_average * (65536.0 - m_toward_average);
    m_average = (uint16_t)(s / 65536.0);
    /*
    uint32_t newVal = sample * m_toward_average;
    uint16_t weight = 0 - m_toward_average;   // weight of old average, 0x1000 - TOWARD_AVERAGE in fact
    uint32_t t32 = weight * m_average;
    // Prevent overflow: >>=1 before addition: (and >>=15 after)
    t32 >>= 8;
    newVal >>= 8;
    newVal += t32;
    if( newVal > (0xFFFF << 8)){
        m_average = 0xFFFF;
    } else {
        m_average = newVal >> 8;
    }
    */
    return m_average;
}

void TriggerDetect::setTowardMid(uint16_t rpm)
{

 m_toward_mid = (rpm >= 32 * TOWARD_MID) ? TOWARD_MID : (rpm >> 5);

}
/** Detects positive and negative peaks;
 * No sudden change: it takes a few samples to converge to peak;
 * Convergence towards mid (fading) is extremely slow (thousands of samples)
 * TOWARD_PEAK / 256 is very fast
 * TOWARD_MID / 65536 is slow
 */
void TriggerDetect::peakDetect(uint16_t sample)
{
    if( sample >= peak_p ){
        uint16_t diff = sample - peak_p;

        uint32_t dmult = diff * m_toward_peak;
        diff = dmult >> 8;
        if( peak_p < 0xFFFF - diff){
            peak_p += diff;
        } else {
            peak_p = 0xFFFF;
        }
        // return; // no other condition can hold but peak_n TOWARD_MID does not hurt
    }else{
        uint16_t diff = peak_p - sample;
        uint32_t dmult = diff * m_toward_mid; // note the >>16 !
        diff = dmult >> 16;
        if( peak_p > diff){
            peak_p -= diff;
        } else {
            peak_p = 0;
        }
    }

    if( sample <= peak_n){
        uint16_t diff = peak_n - sample;
        uint32_t dmult = diff * m_toward_peak;
        diff = dmult >> 8;
        //uint16_t oldpeakn = peak_n;
        if( peak_n > diff ){
            peak_n -= diff;
        } else { // should NEVER happen
            peak_n = 0;
        }
        // Logger::logf(LOG_MSG, m_name + _T(":( peak_n changed: %d => %d )"), oldpeakn, peak_n );

    }else{
        uint16_t diff = sample - peak_n;
        uint32_t dmult = diff * m_toward_mid; // note the >>16 !
        diff = dmult >> 16;
        if( peak_n < 0xFFFF - diff){
            peak_n += diff ;
        } else {
                peak_n = 0xFFFF;
        }
    }
}

/**
 * Set thresholds above and below val
 * TODO: check peak_n < val < peak_p
 * For THRESHOLD_STRATEGY_AROUND_MEAN val = m_average, but beware:
 * the mean value can be
 * close to positive peak (if duty ~ 100%)
 * close to negative peak (if duty ~ 0%)
 * Might be too noise sensitive for such "asymetric" signals
 */
void TriggerDetect::setThresholdsAround( uint16_t val )
{
    uint16_t diff = (peak_p > val) ? peak_p - val : 0;
    uint32_t dmult = diff * m_thresh256;
    diff = (dmult >> 8) & 0xFFFF;
    up_threshold = val + diff;

    diff = (val > peak_n) ? val - peak_n : 0;
    dmult = diff * m_thresh256;
    diff = (dmult >> 8) & 0xFFFF; /// 85/256 = 0,33
    lo_threshold = val - diff;
}
/** THRESHOLD_STRATEGY_BETWEEN_PEAKS
 * thresholds are THIRD-way between peak_p and peak_n
 */
void TriggerDetect::setThresholds( void )
{
    if( m_threshold_strategy == THRESHOLD_STRATEGY_AROUND_MEAN){
        setThresholdsAround( m_average );
        return;
    } else { // 1/2 + 1/2 == 0 ??
        setThresholdsAround( (peak_n >> 1) + (peak_p >> 1) );
    }
}

/** Compares against thresholds.
 * oldstate  => counting => newstate
 * set new state if necessary
 * returns new state
 */
uint8_t TriggerDetect::thresholdDetect(uint16_t sample)
{
    // TODO: sudden upPeak, sudden downPeak (separate counters ?)
    uint8_t prevstate = m_state;
    if ( sample >= up_threshold ){
        if( m_highCounter == m_overthresholdsamplecnt ){
            m_state = 1; // transition to state = 1
            m_highCounter = 0;
        } else {
            m_highCounter ++;
        }
        m_lowCounter = 0;
    } else if ( sample <= lo_threshold ){
        if( m_lowCounter == m_overthresholdsamplecnt ){
            m_state = 0; // transition to state = 0
            m_lowCounter = 0;
        } else {
            m_lowCounter ++;
        }
        m_highCounter = 0;
    } else { // between thresholds
        m_lowCounter = 0;
        m_highCounter = 0;
    }

    if ( m_state != prevstate ){
            // Logger::logf(LOG_MSG, _T("sample=%d, state=%d; "), sample, state);
            m_transitionCount++;
    }

    return m_state;
}

// The most useful interface, does everything:
uint8_t TriggerDetect::feedSample(uint16_t sample)
{
    if( m_sample_received == 0){
        // first sample. Initialize peaks
        peak_n = peak_p = m_average = sample;
        m_sample_received = 1;
        int n;
        for(n = 0; n < arraysize; n++){
            m_prevsample[n] = sample;
        }
    }
    saveLastSamples(sample);
    m_sample = sample; // save example sample
    // average() and peakDetect() useful for any signal (also perfect for measured sparkPhase and dwell)
    average( sample );
    peakDetect( sample );

    // Threshold processing is good to find rising/falling edges in the input wav.
    // Probably useless for the measured sparkPhase and dwell.
    setThresholds();
    thresholdDetect( sample );
    return 0;
}

// TODO: cheat now, only for debug. Not real average, but midpoint between up_threshold and lo_threshold
void TriggerDetect::printData()
{
    Logger::logf(LOG_MSG, m_name + _T(":(ave:%d)%d\t%d\t%d\t%d\t%d(lastsample=%d)"), m_average, peak_n, lo_threshold, up_threshold, peak_p, m_transitionCount, m_sample );
    m_transitionCount =0;
}

void TriggerDetect::resetLocalPeak(uint16_t sample)
{
    peak_n = peak_p = m_average = sample;
}

uint16_t TriggerDetect::diffPeak()
{
     uint16_t diff ;

     diff = peak_p  - peak_n;

     return diff;
}

void TriggerDetect::saveLastSamples(uint16_t currsample)
{

    int n = 0;

    for(n = 0; n < arraysize ; n++){
            m_prevsample[arraysize-n-1] = m_prevsample[arraysize-n-2];
    }
    m_prevsample[0] = currsample;

}

uint32_t TriggerDetect::outlierElimination(void)
{
    uint8_t n = 0;
    uint16_t maxsamplevalue = 0;
    uint16_t minsamplevalue = 0;
    uint32_t average = 0;
  //  Logger::logf(LOG_MSG, _T("TriggerDetect::outlierElimination"));
    //Logger::logf(LOG_MSG, _T("prevsample1 = %i prevsample2 = %i prevsample3 = %iprevsample4 = %i")
    //           ,m_prevsample[0]
    //           ,m_prevsample[1]
    //           ,m_prevsample[2]
    //           ,m_prevsample[3]
    //           );
    for(n = 0; n < arraysize; n++){
        if( n <= 1){
            if( n == 0 ){ // first number
                minsamplevalue = maxsamplevalue = m_prevsample[n];
            }else{ // n = 1
                average -= minsamplevalue;
           }
        }
        if( n >= 1){ //n = 1
            if( m_prevsample[n] < minsamplevalue ){
                 average += minsamplevalue;
                 minsamplevalue = m_prevsample[n];

             } else if( m_prevsample[n] > maxsamplevalue ){
                 average += maxsamplevalue;
                 maxsamplevalue = m_prevsample[n];

             } else {
                 average += m_prevsample[n];
             }
      }
    }
//    Logger::logf(LOG_MSG, _T("sum = %i"),average);
    average = (uint32_t)( average / ( arraysize -2 ) );
//     Logger::logf(LOG_MSG, _T("averag = %i"),average);
return average;
}
///test
void TriggerDetect::loadmprevsample(void)
{

   // m_prevsample[0] = 2;
   // m_prevsample[1] = 1;
   // m_prevsample[2] = 34;
   // m_prevsample[3] = 4;
 //   m_prevsample[4] = 5;
 //   m_prevsample[5] = 22;


}
