#include "STDataIOV3.h"

#include "../common/wxhashdef.h"
#include "STDataModelBase.h"
#include "STDataViewerBase.h"

#include "../Symbol.h"
#include "../vemsprotocol.h"
#include "../logger.h"
#include "../SerialException.h"

#include "../vemsprotocol/ECUInfo.h"
#include "../vemsprotocol/SIPRDescriptor.h"

STDataIOV3::STDataIOV3( const SymbolTable& symbols, SIPRDescriptor* siprDesc, VemsProtocol* vp )
    : m_symbolTable( symbols )
{
    m_pSIPRDescriptor   = siprDesc;
    m_vp                = vp;
}

STDataIOV3::~STDataIOV3() {}

void STDataIOV3::initModel() {}

void STDataIOV3::burnData()
{
    wxString burnCommand = m_pSIPRDescriptor->getBurnCommand();
    m_vp->sendString( burnCommand );
}

void STDataIOV3::readAllData()
{
    int pages = m_pSIPRDescriptor->getPageCount();
    //const Intarr

    m_vp->readString( m_buff, 256, 2);


    for( int i = 0; i < pages; ++i )
    {
        int pageNumber = m_pSIPRDescriptor->getPageNumber( i );

        if ( pageNumber == -1 ) { continue; }

        readPage( pageNumber );
    }

    /*
    for( int i = 1; i <= pages; ++i )
    {
        readPage( i );
    }
    */

    m_vp->readString( m_buff, 256, 2);
}

void STDataIOV3::writeAllData()
{
    for( SymbolTable::const_iterator IT = m_symbolTable.begin(); IT != m_symbolTable.end(); ++IT )
    {
        const Symbol* ptrSymbol = IT->second;

        if( !ptrSymbol->isValid() ) { continue; }

        int pageNo          = ptrSymbol->getPage();
        int symbolOffset    = ptrSymbol->getOffset();
        int dataLength      = ptrSymbol->getSize();

        if ( ptrSymbol->isArray() )
        {

            int dimension   = ptrSymbol->getOrder();
            int xSize       = ptrSymbol->getShape0();
            int ySize       = ptrSymbol->getShape1();
            int pageSize    = xSize;

            if( dimension == 2 ) { pageSize = xSize * ySize; }
            Logger::logf( LOG_WARNING, _T("STDataIOV3::writeAllData(): ")+ ptrSymbol->getName()+_T(" pageSize:%i"),pageSize);
            for( int i = 0; i < pageSize; ++i )
            {
                writeScalarSymbol( pageNo, symbolOffset + i, dataLength );
            }
        }
        else
        {
            writeScalarSymbol( pageNo, symbolOffset, dataLength );
        }
    }

    burnData();
}

//cannot refresh symbol by symbol but whole pages
// need to find different pages
void STDataIOV3::readData( const STDataViewerBaseList* symbolA )
{
    Int2IntHash pagesRefresh;

    for( STDataViewerBaseList::const_iterator it = symbolA->begin(); it != symbolA->end(); it++ )
    {
        wxString symbolName = (*it)->getSymbolName();
        SymbolTable::const_iterator IT = m_symbolTable.find( symbolName );

        if( IT == m_symbolTable.end() ) { continue; }

        Symbol* symbol = IT->second;
        int pageNo = symbol->getPage();
        pagesRefresh[pageNo];
    }

    m_vp->readString( m_buff, 256, 2);
    for( Int2IntHash::iterator it2 = pagesRefresh.begin(); it2 != pagesRefresh.end(); it2++ )
    {
        int pageNo = it2->first;
        readPage(pageNo);
    }
    m_vp->readString( m_buff, 256, 2);

}

void STDataIOV3::writeData( const STDataViewerBaseList* symbolA )
{
    STDataViewerBase* stDataViewer;

    for( STDataViewerBaseList::const_iterator it = symbolA->begin(); it != symbolA->end(); it++ )
    {
        stDataViewer = (*it);
        wxString symbolName = stDataViewer->getSymbolName();

        SymbolTable::const_iterator IT = m_symbolTable.find( symbolName );

        if( IT == m_symbolTable.end() ) { continue; }

        Symbol* symbol = IT->second;

        int pageNo = symbol->getPage();
        int symbolOffset = symbol->getOffset();

        if( symbol->isArray() ){
              // const Int2DoubleHash* dirtyOffsetHash = stDataViewer->getDirtyOffsetA();
              // writeArraySymbol(pageNo, symbolOffset, dirtyOffsetHash);
        }else{
            writeScalarSymbol(pageNo, symbolOffset, symbol->getSize() );
        }

        //stDataViewer->clearDirtyOffsetA();
    }
}

void STDataIOV3::writeArraySymbol(int pageNo, int symbolOffset, const Int2DoubleHash* dirtyOffsetHash)
{
    for(Int2DoubleHash::const_iterator it = dirtyOffsetHash->begin(); it != dirtyOffsetHash->end(); it++){
        int offset = it->first;
        writeScalarSymbol(pageNo, symbolOffset + offset, 1 );
    }
}

void STDataIOV3::writeScalarSymbol(int pageNo, int offset, int size )
{
    wxString valueWriteCommand = m_pSIPRDescriptor->getPageWriteCommand( pageNo );

    uint8_t* buff = m_modelBase->getRawBuffer(pageNo);

    //if ( pageNo == 2 ) { Logger::logf( LOG_MSG, _T("STDataIOV3::writeScalarSymbol(): %i : %u"), offset, buff[offset] ); }

    m_buff[0] =  valueWriteCommand.GetChar(0);

    for( int i = 0; i < size; ++i )
    {
        m_buff[1] =  (uint8_t) offset + i;
        m_buff[2] =  (uint8_t) buff[ offset + i ];
        //Logger::logf(LOG_MSG, _T("STDataIOV3::writeData() refresh pageNo:%d command:%c %X%X"), pageNo, m_buff[0], m_buff[1], m_buff[2]);
        m_vp->sendString(m_buff, 3);
    }
}

void STDataIOV3::readPage( int pageNo)
{
    failSafePageRead(pageNo);
}

void STDataIOV3::failSafePageRead(int pageNo)
{
    wxString pgReadCommand = m_pSIPRDescriptor->getPageReadCommand( pageNo );
    int pageSize = 255;

    uint8_t* buff = m_modelBase->getRawBuffer(pageNo);

    Logger::logf( LOG_WARNING, _T("STDataIOV3::readPage() pageNo: %d (")+ pgReadCommand+_T(")"), pageNo );

    bool tryAgain = false;
    int tryCount = 0;
    int PAGE_MAX_TRY_COUNT = 4;
    wxString tmp;

    if( m_vp->getStoredState() != ECUInfo::MEGATUNE) return;

    try
    {
        /// START ROUND BUG
        if( m_vp->getDevice() == ECUInfo::DEVICE_ROUND){
            m_vp->sendString(_T("q") );
            int result = m_vp->readString( m_buff, 32, 8);
            //m_vp->clear();
        }
        /// END BUG!!!!!!!!!!!!!!

        m_vp->sendString(pgReadCommand);
    }catch (SerialException& se){
        Logger::logf( LOG_WARNING, _T("STDataIOV3::readPage() sendreadCommand Exception: ")+ se.toString() );
    }

    do{
        tryAgain = false;
        try
        {
            int result = m_vp->readString( buff, pageSize, 64);
            if(result != pageSize){
                //Logger::logf( LOG_WARNING, _T("STDataIOV3::readPage() received pageSize:%d expected:%d"), result, pageSize );
                tryCount++;
                tryAgain = true;
                m_vp->clear();
                if(result == 0){
                    Logger::logf( LOG_WARNING, _T("STDataIOV3::readPage() command has been sent again!!"));
                    m_vp->sendString(pgReadCommand);
                }
            }

        }catch (SerialException& se){
            //Logger::logf( LOG_WARNING, _T("STDataIOV3::readPage() Exception: ")+ se.toString() );
            tryCount++;
        }
    } while( tryAgain && tryCount < PAGE_MAX_TRY_COUNT );

    //if( m_vp->getDevice() == ECUInfo::DEVICE_ROUND){ m_vp->readString( m_buff, pageSize, 2); } /// ROUND BUG!!!!!!!!!!!!!!
    //m_vp->readString( m_buff, pageSize, 2);

}
