#include "STArrayViewerPanel.h"
#include "../stdatamanager/STDataMgrBase.h"
#include "../logger.h"


STArrayViewerPanel::STArrayViewerPanel(  wxWindow* parent, const Symbol& s, const wxString& title )
    : wxPanel( parent )
{
    m_title             = title;
    m_symbol            = s;
    m_isSymbolAccepted  = ( m_symbol.getOrder() == 1 );

    init();
}


void STArrayViewerPanel::update()
{
    if ( !m_isSymbolAccepted ) { return; }

    size_t cols = m_symbol.getArrayColSize();

    for ( size_t c = 0; c < cols; ++c )
    {
        double viewerData = m_stDataMgr->getDataFromViewerModel( m_symbol, 0, c );

        m_showCtrls[c]->SetValue( wxString::Format( _T("%.0f"), viewerData ) );
        m_editCtrls[c]->SetValue( wxString::Format( _T("%.0f"), viewerData ) );
    }
}


void STArrayViewerPanel::updateModel()
{
    if ( !m_isSymbolAccepted ) { return; }

    size_t cols = m_symbol.getArrayColSize();

    for ( size_t c = 0; c < cols; ++c )
    {
        double d = 0.0;
        m_editCtrls[c]->GetValue().ToDouble( &d );
        m_stDataMgr->setModelData( m_symbol, 0, c, d );
    }
}


void STArrayViewerPanel::requestWrite()
{
    if ( !m_isSymbolAccepted ) { return; }

    size_t cols = m_symbol.getArrayColSize();

    for ( size_t c = 0; c < cols; ++c )
    {
        m_stDataMgr->requestWrite( m_symbol, 0, c );
    }
}


void STArrayViewerPanel::init()
{
    //display wrong symbol type
    if ( !m_isSymbolAccepted )
    {
        SetSizer( new wxBoxSizer( wxHORIZONTAL ) );
        GetSizer()->Add( new wxStaticText( this, wxID_ANY, m_symbol.getName() + _T(" is not an 1-dimension array!!!") ), 1, wxEXPAND, 0 );
        return;
    }

    //create sizer
    wxStaticBox* pStaticBox =  new wxStaticBox( this, wxID_ANY, m_title );
    wxFont font( pStaticBox->GetFont() );
    font.SetWeight( wxFONTWEIGHT_BOLD );
    //font.SetCol
    pStaticBox->SetOwnFont( font );
    //pStaticBox->SetOwnForegroundColour( *wxBLACK );
    wxStaticBoxSizer* pMainSizer = new wxStaticBoxSizer( pStaticBox, wxVERTICAL );
    SetSizer( pMainSizer );

    //add TextControls
    size_t rowCount = m_symbol.getArrayColSize();

    for ( size_t i = 0; i < rowCount; ++i )
    {
       m_showCtrls.Add( new wxTextCtrl( this, wxID_ANY ) );
       //m_editCtrls.Add( new wxTextCtrl( pStaticBox, wxID_ANY ) );
       m_editCtrls.Add( new wxTextCtrl( this, i ) );

       // connect KillFocus
       // m_editCtrls[ i ]->Connect( wxEVT_KILL_FOCUS, wxFocusEventHandler( STArrayViewerPanel::OnKillFocus ), NULL, this );
       //m_showCtrls[ i ]->SetEditable( false );
       m_showCtrls[ i ]->Disable();
       m_editCtrls[ i ]->SetValidator( wxTextValidator( wxFILTER_NUMERIC ) );

       wxBoxSizer* pLineSizer = new wxBoxSizer( wxHORIZONTAL );
       pLineSizer->Add( m_showCtrls[ i ], 1, wxEXPAND | wxALL, 1 );
       pLineSizer->Add( m_editCtrls[ i ], 1, wxEXPAND | wxALL, 1 );
       pMainSizer->Add( pLineSizer, 1, wxEXPAND, 0 );
    }
}


bool STArrayViewerPanel::verify()
{
    bool bError = false;

    const int MIN_IDX = 0;
    const int MAX_IDX = m_editCtrls.GetCount() - 1;

    double dShowValue = 0.0;
    double dEditValue = 0.0;

    wxString    msg;

    int id = 0;
    for ( ; id < (int)m_editCtrls.GetCount(); ++id )
    {
        m_showCtrls[ id ]->GetValue().ToDouble( &dShowValue );
        m_editCtrls[ id ]->GetValue().ToDouble( &dEditValue );

        if ( id == MIN_IDX && ( dEditValue < dShowValue ) ) // verify min value
        {
            bError = true;
            msg = wxString::Format( _T("Value( %.0f ) is lesser than minimum( %.0f )!!!"), dEditValue, dShowValue );
            break;
        }
        else if ( id == MAX_IDX && ( dEditValue > dShowValue ) ) // verify max value
        {
            bError = true;
            msg = wxString::Format( _T("Value( %.0f ) is greater than maximum( %.0f )!!!"), dEditValue, dShowValue );
            break;
        }

        // verify monotonity
        if ( id < MAX_IDX )
        {
            double dNextEditValue = 0.0;
            m_editCtrls[ id + 1 ]->GetValue().ToDouble( &dNextEditValue );
            if ( dEditValue > dNextEditValue )
            {
                bError = true;
                msg = wxString::Format( _T("Value( %.0f ) is greater than next( %.0f )!!!"), dEditValue, dNextEditValue );
                break;
            }
        }

        if ( id > MIN_IDX )
        {
            double dPrevEditValue = 0.0;
            m_editCtrls[ id - 1 ]->GetValue().ToDouble( &dPrevEditValue );
            if ( dEditValue < dPrevEditValue )
            {
                bError = true;
                msg = wxString::Format( _T("Value( %.0f ) is lesser than previous( %.0f )!!!"), dEditValue, dPrevEditValue );
                break;
            }
        }
    }

    if ( bError )
    {
        m_editCtrls[ id ]->SetValue(  m_showCtrls[ id ]->GetValue() );
        wxMessageBox( msg , _T("TableMapper Error"), wxOK | wxICON_ERROR );
        m_editCtrls[id]->SetFocus();
    }

    return !bError;

}


void STArrayViewerPanel::getOldValues( VemsMatrix<double>& a ) const
{
    get( m_showCtrls, a );
}

void STArrayViewerPanel::getNewValues( VemsMatrix<double>& a ) const
{
    get( m_editCtrls, a );
}

void STArrayViewerPanel::get( const wxTextCtrlArray& ctrls,  VemsMatrix<double>& a ) const
{
    a.setSize( 1, ctrls.GetCount() );

    double d = 0.0;

    for ( size_t i = 0; i < ctrls.GetCount(); ++i )
    {
        ctrls[i]->GetValue().ToDouble( &d );
        a.set( i, d );
    }
}

wxString STArrayViewerPanel::toString()
{
    return _T("STArrayViewerPanel - ") + m_symbol.getName();
}
