#include "GaugeSetHelper.h"

#include <wx/xml/xml.h>
#include <wx/sstream.h>
#include <wx/ffile.h>

#include "GaugeGroupManager.h"
#include "GaugeType.h"
#include "GaugeCarrierBase.h"

#include "../version.h"
#include "../logger.h"

bool GaugeSetHelper::ExportGauges(GaugePropertiesArray& items, const wxString& filename)
{
    wxString xmlContent = ExportGauges(items);

    wxFFile file(filename, _T("w"));

    file.Write(xmlContent);

    file.Close();

    return true;
}

wxString GaugeSetHelper::ExportGauges(GaugePropertiesArray& items)
{
    wxString vemsTuneVer = AutoVersion::getFullVersion();

    //Logger::logf(LOG_MSG,_T("vemsTuneVer: ")+vemsTuneVer);

    wxXmlNode* root = new wxXmlNode(NULL, wxXML_ELEMENT_NODE, _T("GaugeSet"));

    root->AddProperty(_T("program"), AutoVersion::getAppName() + _T(" ") + AutoVersion::getDateTime() );
    root->AddProperty(_T("ver_major"), wxString::Format(_T("%i"),AutoVersion::MAJOR ) );
    root->AddProperty(_T("ver_minor"), wxString::Format(_T("%i"),AutoVersion::MINOR ) );
    root->AddProperty(_T("ver_build"), wxString::Format(_T("%i"),AutoVersion::BUILD ) );

    root->AddProperty(_T("exporting_time"), wxDateTime::Now().FormatISODate() + _T(" ") + wxDateTime::UNow().FormatISOTime() );

    wxXmlNode* childnode = new wxXmlNode(root, wxXML_ELEMENT_NODE, _T("Gauge"));

    root->SetChildren(childnode);

    for(size_t a = 0; a < items.GetCount(); a++) {

        GaugeProperties& prop = items[a];

        wxRect grect = prop.getRect();

//        Logger::logf(LOG_MSG,_T("exporting gauge: type=")+ GaugeType::toString(prop.gaugeType) );

        if(!childnode) {
            childnode = new wxXmlNode(root, wxXML_ELEMENT_NODE, _T("Gauge"));
            root->SetNext(childnode);
        }

        childnode->AddProperty(_T("type"), GaugeType::toString(prop.gaugeType) );
        childnode->AddProperty(_T("name"), prop.symbolName );
        childnode->AddProperty(_T("descriptor"), prop.descName );
        childnode->AddProperty(_T("theme"), prop.themeName );
        childnode->AddProperty(_T("x"), wxString::Format(_T("%i"),grect.x) );
        childnode->AddProperty(_T("y"), wxString::Format(_T("%i"),grect.y) );
        childnode->AddProperty(_T("width"), wxString::Format(_T("%i"),grect.width) );
        childnode->AddProperty(_T("height"), wxString::Format(_T("%i"),grect.height) );

        childnode = NULL;
    }

    wxXmlDocument doc;
    doc.SetRoot(root);

    wxStringOutputStream ostream;

    doc.Save(ostream);

    return ostream.GetString();
}

bool GaugeSetHelper::CheckXmlContentHeader(const wxString& content)
{
    wxStringInputStream istream(content);

    wxXmlDocument doc;

    if(!doc.Load( istream ))
        return false;

    wxXmlNode* rootNode = doc.GetRoot();

    return rootNode->GetName() == _T("GaugeSet") ? true : false;
}

bool GaugeSetHelper::ImportGaugesFromStr(const wxString& xmlContent, GaugeGroupManager* groupmgr, GaugeCarrierBase* carrier)
{
    wxStringInputStream istream(xmlContent);

    wxXmlDocument doc;

    if(!doc.Load( istream ))
        return false;

    wxXmlNode* rootNode = doc.GetRoot();

    if(rootNode->GetName() != _T("GaugeSet")) return false;

    wxString grpName = groupmgr->GetGroupName(carrier);

    if(grpName == wxEmptyString) return false;

    groupmgr->StartTransaction();

    for(wxXmlNode* gaugeNode = rootNode->GetChildren(); gaugeNode; gaugeNode = gaugeNode->GetNext()) {

        GaugeProperties prop;

        long x,y,w,h;


        wxString gtype = gaugeNode->GetPropVal(_T("type"),wxEmptyString);

        Logger::logf(LOG_MSG,_T("import: gtype=")+gtype);

        if( gtype.StartsWith(_T("table")) ) gtype = _T("table");

        prop.gaugeType = GaugeType::getType(gtype);
        prop.symbolName = gaugeNode->GetPropVal(_T("name"),wxEmptyString);
        prop.descName = gaugeNode->GetPropVal(_T("descriptor"),wxEmptyString);
        prop.themeName = gaugeNode->GetPropVal(_T("theme"),wxEmptyString);

        gaugeNode->GetPropVal(_T("x"),wxEmptyString).ToLong(&x);
        gaugeNode->GetPropVal(_T("y"),wxEmptyString).ToLong(&y);
        gaugeNode->GetPropVal(_T("width"),wxEmptyString).ToLong(&w);
        gaugeNode->GetPropVal(_T("height"),wxEmptyString).ToLong(&h);

        prop.setSize(w,h);
        prop.setPos(x,y);

        prop.groupName = grpName;

        Logger::logf(LOG_MSG,_T("importing gauge: ")+gtype);

        bool saveres = groupmgr->SaveGauge(prop);

        if(saveres)
            groupmgr->AddGaugeToGroup(prop, carrier, carrier->IsFloating());

    }

    groupmgr->EndTransaction();

    return true;
}

bool GaugeSetHelper::ImportGauges(const wxString& filename, GaugeGroupManager* groupmgr, GaugeCarrierBase* carrier)
{
    wxFFile file(filename, _T("r"));

    wxString xmlContent;

    file.ReadAll(&xmlContent);

    file.Close();

    return ImportGaugesFromStr(xmlContent,groupmgr,carrier);
}

