#include "VemsExpressionGridWidget.h"
#include "VemsExpressionGridWidgetHelper.h"
#include "GaugeGroupManager.h"
#include "../logger.h"
#include "../DirHelper.h"
#include "../MTConfigHelper.h"
#include "../common/VemsContext.h"
#include "../common/types/IntSet.h"
#include "../vemsdialog/UserDialogBase.h"
#include "../vemsdialog/WindowTipper.h"
#include "../stdatamanager/STDataMgrBase.h"

#include <wx/dcbuffer.h>

#define MAX_FONT_SIZE 24
#define MIN_FONT_SIZE 8
#define TITLE_FONT_SIZE 10

#define DEFAULT_WIDTH 5
#define DEFAULT_PRECISION 1
#define DEFAULT_ALIGN wxALIGN_RIGHT
#define LINK_COLOR (*wxBLUE)


wxString GridCell::getDefaultText() const
{
    if ( variables.empty() ) return text;

    wxString defaultText =  wxString ( _T('0'), width + precision + 1 );

    wxString calculatedText = originText;

    StringSet::const_iterator it = variables.begin();

    for ( ;it != variables.end(); ++it )
    {
        wxString variableName = *it;
        calculatedText.Replace( _T("${") + variableName + _T("}"), defaultText );
    }

    return calculatedText;
}

wxSize GridCell::getMinSize( const VemsExpressionGridWidget& parent, const wxFont& cellFont ) const
{
    return wxSize();
}


BEGIN_EVENT_TABLE( VemsExpressionGridWidget, VemsWidgetBase )
    EVT_PAINT( VemsExpressionGridWidget::OnPaint )
    EVT_SIZE ( VemsExpressionGridWidget::OnSize )
    EVT_LEFT_DOWN( VemsExpressionGridWidget::OnLeftDown )
    EVT_MOTION( VemsExpressionGridWidget::OnMouseMove )
END_EVENT_TABLE()

VemsExpressionGridWidget::VemsExpressionGridWidget( wxWindow* parent, wxWindowID id,GaugeDescManager * gdmgr, GaugeProperties &prop )
    : VemsWidgetBase( parent, id, gdmgr, prop )
{
    Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::Constructor() - BEGIN") );

    m_backbuffer = NULL;

    m_isSymbolMenu = false;


    m_gridDescriptor = NULL;

    SetBackgroundStyle( wxBG_STYLE_CUSTOM );
    SetWindowStyle( wxBORDER_NONE );

    m_borderSize = 2;
    m_isSymbolsRegistered = false;

    //m_fontPointSize = 8;

    m_cellBorderSize = 1;
    m_gridSize = 1;
    m_titleHeight = 20;

    m_cellFont = wxFont( MIN_FONT_SIZE, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL );
    m_titleFont = wxFont( TITLE_FONT_SIZE, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD );

    bool autoSize = prop.getSize() == wxSize( -1, -1 );

    createGrid();

    m_sideRatioRestriction = (double)m_minClientSize.GetWidth()/(double)m_minClientSize.GetHeight();

    calculateSize( autoSize );

    m_isResized = true;
    //SetSize( 400, 400 ); SetSize( )
    Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::Constructor() - END") );

    //m_helpIcon = new wxBitmap( DirHelper::GetPath(dICON) + _T("ul-info-m.png"), wxBITMAP_TYPE_PNG );
    m_winTip = new WindowTipper( this );

    initIcons();
}

VemsExpressionGridWidget::~VemsExpressionGridWidget()
{
    unregisterViewer();
    WX_CLEAR_ARRAY(m_variablePointA);
    WX_CLEAR_ARRAY(m_variableSetA);
    WX_CLEAR_ARRAY(m_helpIcons);
}

wxRegion VemsExpressionGridWidget::GetShape()
{
    return wxRegion(0,0,GetSize().x,GetSize().y);
}

wxRegion VemsExpressionGridWidget::GetBorderRegion()
{
    return wxRegion(0,0,1,1);
}

void VemsExpressionGridWidget::createGrid()
{
    m_gridDescriptor = MTConfigHelper::getMTConfigData()->getExpressionGridDescriptor( GetCurrSymbolName() );

    if ( !m_gridDescriptor ) createDefaultGrid();
    else createGridFromDesc();

    m_colMaxTextWidth.Add( 0, m_cols );
    m_colMaxTextRow.Add( 0, m_cols );
    m_colAvailableTextWidth.Add( 0, m_cols );

    m_rowHeight.Add( 0, m_rows );
    m_rowAvailableTextHeight.Add( 0, m_rows );

    calculateDecorationSize();
    findColsLongestText();
    calculateMinSize();
}


void VemsExpressionGridWidget::createDefaultGrid()
{
    m_rows = 5;
    m_cols = 5;
    m_disableCol = -1;

    m_colsGridPos.Add( 0, m_cols );
    m_rowsGridPos.Add( 0, m_rows );
    m_linesEnabled.Add( 1, m_rows );

    GridCellMatrix::extent_gen extents;
    m_cells.resize(extents[m_rows][m_cols]);

    for ( size_t row = 0; row < m_rows; ++row )
    {
        for ( size_t col = 0; col < m_cols; ++col )
        {
            setCellText( row, col, _T("TEST") );

            m_cells[row][col].width     = DEFAULT_WIDTH;
            m_cells[row][col].precision = DEFAULT_PRECISION;
            m_cells[row][col].align     = DEFAULT_ALIGN;
            m_cells[row][col].disabled  = false;
            m_cells[row][col].linkEnabled  = true;
        }
    }
}

void VemsExpressionGridWidget::createGridFromDesc()
{
    //Logger::logf( LOG_MSG, _T("createGridFromDesc()") );

    size_t descRows = m_gridDescriptor->cells.shape()[0];
    size_t descCols = m_gridDescriptor->cells.shape()[1];
    m_disableCol = m_gridDescriptor->disableCol;

    //Logger::logf( LOG_MSG, _T("descRows: %d, descCols: %d"), descRows, descCols );

    IntSet emptyCols;

    for ( size_t col = 0; col < descCols; ++col )
    {
        bool isColEmpty = true;

        for ( size_t row = 0; row < descRows; ++row )
        {
            ExpressionCellDescriptor& cell = m_gridDescriptor->cells[row][col];

            if ( !cell.text.IsEmpty() )
            {
                isColEmpty = false;
                break;
            }
        }

        if ( isColEmpty ) emptyCols.insert( col );
    }

    m_rows = descRows;
    m_cols = descCols - emptyCols.size();

//    Logger::logf( LOG_MSG, _T("m_rows: %d, m_cols: %d"), m_rows, m_cols );

    m_colsGridPos.Add( 0, m_cols );
    m_rowsGridPos.Add( 0, m_rows );
    m_linesEnabled.Add( 1, m_rows );

    GridCellMatrix::extent_gen extents;
    m_cells.resize(extents[m_rows][m_cols]);


    for ( size_t row = 0; row < descRows; ++row )
    {
        size_t gridCol = 0;

        for ( size_t col = 0; col < descCols; ++col )
        {
            if ( emptyCols.find( col ) != emptyCols.end() ) continue;

            ExpressionCellDescriptor& cell = m_gridDescriptor->cells[row][col];

            size_t found = VemsExpressionGridWidgetHelper::getVariablesFromString( cell.text, m_rtVariables );

            if(found)
            {
                StringSet* variables = new StringSet();
                VemsExpressionGridWidgetHelper::getVariablesFromString( cell.text, *variables );
                m_variablePointA.Add( new wxPoint(row,gridCol) );
                m_variableSetA.Add( variables );
                setCellText( row, gridCol, _T("N/A") );

                VemsExpressionGridWidgetHelper::getVariablesFromString( cell.text, m_cells[row][gridCol].variables );
                m_cells[row][gridCol].originText = cell.text;
            }
            else
            {
                setCellText( row, gridCol, cell.text );
                m_cells[row][gridCol].originText = cell.text;
            }

            m_cells[row][gridCol].width = cell.width == -1 ? DEFAULT_WIDTH : cell.width;
            m_cells[row][gridCol].precision = cell.precision == -1 ? DEFAULT_PRECISION : cell.precision;
            m_cells[row][gridCol].align = cell.align == -1 ? DEFAULT_ALIGN : (wxAlignment)cell.align;
            m_cells[row][gridCol].links = cell.links;
            m_cells[row][gridCol].texts = cell.texts;
            m_cells[row][gridCol].colors = cell.colors;
            m_cells[row][gridCol].disabled  = false;
            m_cells[row][gridCol].help = cell.help;
            m_cells[row][gridCol].linkEnabled = true;

            ++gridCol;
        }
    }
}

void VemsExpressionGridWidget::calculateDecorationSize()
{
    m_decorationsSize.x = 2 * m_borderSize;
    m_decorationsSize.x += m_cols * 2 * m_cellBorderSize;
    m_decorationsSize.x += m_cols > 0 ? ( m_cols - 1 ) * m_gridSize : 0;

    m_decorationsSize.y = 2 * m_borderSize;
    m_decorationsSize.y += m_rows * 2 * m_cellBorderSize;
    m_decorationsSize.y += m_rows > 0 ? ( m_rows - 1 ) * m_gridSize : 0;
    m_decorationsSize.y += m_titleHeight;
}


void VemsExpressionGridWidget::calculatePreferredTextSize()
{
//    Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::calculatePreferredTextSize()") );
//    Logger::logf( LOG_MSG, _T("m_cellFont.GetPointSize(): %d"), m_cellFont.GetPointSize() );

    //calculateDecorationSize();

    m_colMaxTextWidth.Clear();
    m_colMaxTextWidth.Add( 0, m_cols );

    m_textHeight = 0;

    int width   = 0;
    int height  = 0;


    for ( size_t col = 0; col < m_cols; ++col )
    {
        for ( size_t row = 0; row < m_rows; ++row )
        {
            wxString text = m_cells[row][col].getDefaultText();
            GetTextExtent( text, &width, &height, 0, 0, &m_cellFont );

            if ( width > m_colMaxTextWidth[col] ) m_colMaxTextWidth[col] = width;

            if ( height > m_textHeight ) m_textHeight = height;
        }
    }

    int preferredTextWidth = 0;

    for( size_t col = 0; col < m_cols; ++col )
    {
        preferredTextWidth += m_colMaxTextWidth[col];
//        Logger::logf( LOG_MSG, _T("m_colMaxTextWidth[%d] = %d"), col, m_colMaxTextWidth[col] );
    }

    //Logger::logf( LOG_MSG, _T("m_textHeight: %d"), m_textHeight );

    int preferredTextHeight = m_textHeight * m_rows;

    m_preferredTextSize.SetWidth( preferredTextWidth );
    m_preferredTextSize.SetHeight( preferredTextHeight );

//    Logger::logf( LOG_MSG, _T("m_textHeight: %d"), m_textHeight );
//    Logger::logf( LOG_MSG, _T("preferredTextHeight: %d"), preferredTextHeight );
    //Logger::logf( LOG_MSG, _T("m_decorationsSize.GetHeight(): %d"), m_decorationsSize.GetHeight() );

    //m_preferredSize = wxSize( preferredTextWidth, preferredTextHeight );
    m_preferredSize.SetWidth( preferredTextWidth + m_decorationsSize.GetWidth() );
    m_preferredSize.SetHeight( preferredTextHeight + m_decorationsSize.GetHeight() );

//    Logger::logf( LOG_MSG, _T("m_preferredSize: w: %d, h: %d"),
//        m_preferredSize.GetWidth(), m_preferredSize.GetHeight() );
    //Logger::logf( LOG_MSG, _T("m_textHeight: %d"), m_textHeight );
}

void VemsExpressionGridWidget::calculatePreferredGridPosition()
{
//    Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::calculatePreferredGridPosition()") );
//    Logger::logf( LOG_MSG, _T("m_textHeight: %d"), m_textHeight );

    int colPos = m_borderSize;

    for ( size_t col = 0; col < m_cols - 1; ++col )
    {
        colPos += 2 * m_cellBorderSize + m_colMaxTextWidth[col] + m_gridSize;
        m_colsGridPos[col] = colPos;
    }

    int rowPos = m_titleHeight + m_borderSize;

    for ( size_t row = 0; row < m_rows - 1 ; ++row )
    {
        rowPos += 2 * m_cellBorderSize + m_textHeight + m_gridSize;
//        Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::calculatePreferredGridPosition()") );
//        Logger::logf( LOG_MSG, _T("m_cellBorderSize: %d"), m_cellBorderSize );
//        Logger::logf( LOG_MSG, _T("m_textHeight: %d"), m_textHeight );
//        Logger::logf( LOG_MSG, _T("m_gridSize: %d"), m_gridSize );
//        Logger::logf( LOG_MSG, _T("rowPos: %d"), rowPos );
        m_rowsGridPos[row] = rowPos;
    }
}

void VemsExpressionGridWidget::calculatePreferredCellRects()
{
    int rowPos = m_titleHeight + m_borderSize + 1;
    int cellHeight = m_textHeight + 2 * m_cellBorderSize;

    for ( size_t row = 0; row < m_rows ; ++row )
    {
        int colPos = m_borderSize + 1;// + m_cellBorderSize;

        for ( size_t col = 0; col < m_cols; ++col )
        {
            int cellWidth = m_colMaxTextWidth[col] + 2 * m_cellBorderSize;
            m_cells[row][col].rect.x = colPos;
            m_cells[row][col].rect.y = rowPos;
            m_cells[row][col].rect.width = cellWidth;
            m_cells[row][col].rect.height = cellHeight;

            colPos += cellWidth + m_gridSize;
//
//            Logger::logf( LOG_MSG, _T("m_cellRects[%d][%d]"), row, col );
//            Logger::logf( LOG_MSG, _T("x: %d, y: %d, w: %d, h: %d"), colPos, rowPos, cellWidth, cellHeight );

        }

        rowPos += cellHeight + m_gridSize;
    }
}


///TODO - cell types? TEXT, VARIABLE
void VemsExpressionGridWidget::calculatePreferredSize()
{


//    size_t found = VemsExpressionGridWidgetHelper::getVariablesFromString( text, m_rtVariables );
//                if(found){
//                    StringSet* variables = new StringSet();
//                    VemsExpressionGridWidgetHelper::getVariablesFromString( text, *variables );
//                    m_variablePointA.Add( new wxPoint(row,col) );
//                    m_variableSetA.Add( variables );
//                    setCellText( row, col, _T("N/A") );
//                }
}


void VemsExpressionGridWidget::drawTitle( wxDC& dc )
{
    wxPen oldPen = dc.GetPen();
    wxPen pen = dc.GetPen();

    pen.SetColour( m_theme->BorderColor );
    pen.SetWidth( 2 );

	dc.SetPen( pen );

	wxRect rc = wxRect( GetClientSize() );

	dc.DrawLine(rc.x, m_titleHeight, rc.GetRight() + 1, m_titleHeight );

	wxFont oldFont = dc.GetFont();

	dc.SetFont( m_titleFont );
    dc.SetTextForeground( m_theme->BorderColor );

	wxString title = m_gridDescriptor->title.IsEmpty() ? GetCurrSymbolName() : m_gridDescriptor->title;

	dc.DrawText( title, rc.x + m_borderSize + 1, rc.y + m_borderSize );

	dc.SetFont( oldFont );
	dc.SetPen( oldPen );
}

//void VemsExpressionGridWidget::drawBorder( wxDC& dc )
//{
//    wxPen oldPen = dc.GetPen();
//    wxPen pen = dc.GetPen();
//
//    pen.SetColour( m_theme->BorderColor );
//    pen.SetWidth( m_borderSize );
//
//    wxRect rc = wxRect( GetClientSize() );
//
//	dc.SetPen( pen );
//	dc.DrawLine( rc.x + 1, rc.y + 1, rc.GetRight() + 1, rc.y + 1);
//	dc.DrawLine( rc.x + 1, rc.y, rc.x + 1, rc.GetBottom() );
//	dc.DrawLine(rc.GetRight(), rc.GetBottom(), rc.GetRight(), rc.y);
//	dc.DrawLine(rc.GetRight(), rc.GetBottom(), rc.x, rc.GetBottom() );
//
//	dc.SetPen(oldPen);
//
//}
void VemsExpressionGridWidget::drawBorder( wxDC& dc )
{
    if ( !m_borderSize ) return;

    wxPen oldPen = dc.GetPen();
    wxPen pen = dc.GetPen();

    //pen.SetColour( *wxRED );
    pen.SetColour( m_theme->BorderColor );
    pen.SetWidth( m_borderSize );
    pen.SetCap( wxCAP_BUTT );
    pen.SetJoin( wxJOIN_MITER );

    wxRect rect = wxRect( GetClientSize() );

    int offset = m_borderSize / 2;

	dc.SetPen( pen );

	int left = rect.x + offset;
	int top = rect.y + offset;
	int right = rect.width - m_borderSize + offset;
	int bottom = rect.height - m_borderSize + offset;

    dc.DrawLine( rect.x, top, rect.GetWidth(), top );
    dc.DrawLine( rect.x, bottom, rect.GetWidth(), bottom );
	dc.DrawLine( left, rect.y, left, rect.GetHeight() );
	dc.DrawLine( right, rect.y, right, rect.GetHeight() );

	dc.SetPen(oldPen);
}

void VemsExpressionGridWidget::drawGrids( wxDC& dc )
{
    if( m_colsGridPos.GetCount() < 1 || m_rowsGridPos.GetCount() < 1 ) return;

    wxSize clientSize = GetClientSize();

    wxPen oldPen = dc.GetPen();
    wxPen pen = dc.GetPen();
    pen.SetColour( m_theme->BorderColor );
    dc.SetPen( pen );

    for ( size_t col = 0; col < m_colsGridPos.GetCount() - 1; ++col )
    {
        int x = m_colsGridPos[col];
        dc.DrawLine( x, m_titleHeight, x,  clientSize.GetHeight() );
    }

    for( size_t row = 0; row < m_rowsGridPos.GetCount() - 1; ++row )
    {
        int y = m_rowsGridPos[row];
        dc.DrawLine( 0, y, clientSize.GetWidth(), y );
    }

    dc.SetPen(oldPen);
}


void VemsExpressionGridWidget::drawTable( wxDC& dc )
{
    wxFont oldFont = dc.GetFont();
    wxFont newFont = m_cellFont;

    wxPen oldPen = dc.GetPen();
    wxBrush oldBrush = dc.GetBrush();

    for( size_t row = 0; row < m_rows; ++row )
    {
        for ( size_t col = 0; col < m_cols; ++col )
        {
            GridCell& cell = m_cells[row][col];

            //cell->draw( wxDC& dc, WidgetThemeTable* theme );

//            if ( cell.disabled )
//            {
//                dc.SetTextForeground( m_theme->InactiveTextColor );
//                dc.SetBrush( m_theme->InactiveBackgroundColor );
//                dc.SetPen( m_theme->InactiveBackgroundColor );
//            }
//            else
//            {
//                if ( cell.isVariable() )
//                {
//                    dc.SetPen( m_theme->VariableBackgroundColor );
//                    dc.SetBrush( m_theme->VariableBackgroundColor );
//                }
//                else
//                {
//                    dc.SetBrush( m_theme->BackgroundColor );
//                    dc.SetPen( m_theme->BackgroundColor );
//                }
//            }
            setActiveColor( cell );

            dc.SetBrush( cell.bgColor );
            dc.SetPen( cell.bgColor );
            dc.SetTextForeground( cell.fgColor );
//                    dc.SetPen( m_theme->BackgroundColor );


            dc.DrawRectangle( cell.rect.x, cell.rect.y, cell.rect.width, cell.rect.height );

            newFont.SetUnderlined( cell.isLink() );

//            wxColor color = cell.isLink() && !cell.isVariable() ? m_theme->LinkTextColor : m_theme->BorderColor;
//
//            if ( cell.isLink() && !cell.linkEnabled ) {
//                color = m_theme->InactiveTextColor;
//            }

            //dc.SetTextForeground( color );
            dc.SetFont( newFont );

            if ( cell.text == _T("note") )
            {
                wxBitmap* helpIcon = getHelpBmp( cell.rect );

                if ( helpIcon )
                {
                    int x = ( cell.rect.width - helpIcon->GetWidth() ) / 2;
                    int y = ( cell.rect.height - helpIcon->GetHeight() ) / 2;

                    dc.DrawBitmap( *helpIcon, cell.rect.x + x, cell.rect.y + y );
                }
                else
                {
                   dc.DrawLabel( _T("i"), cell.rect, cell.align | wxALIGN_CENTER_VERTICAL );
                }
            }
            else
            {
                dc.DrawLabel( cell.text, cell.rect, cell.align | wxALIGN_CENTER_VERTICAL );
            }
        }
    }

    dc.SetFont( oldFont );
    dc.SetPen( oldPen );
    dc.SetBrush( oldBrush );
}

void VemsExpressionGridWidget::drawVariableCellsOnly( wxDC& dc )
{
    wxFont oldFont = dc.GetFont();
    wxFont newFont = m_cellFont;

    //dc.SetTextForeground( m_theme->BorderColor );

    for( size_t i = 0; i < m_variablePointA.GetCount(); ++i )
    {
        wxPoint* cellPos = m_variablePointA.Item(i);
        int row = cellPos->x;
        int col = cellPos->y;

        GridCell& cell = m_cells[row][col];
        //Logger::logf( LOG_MSG, _T("m_rows: %d, m_cols: %d"), m_rows, m_cols );
        //Logger::logf( LOG_MSG, _T("row: %d, col: %d"), row, col );

        //int align = m_cells[row][col].align;
        //wxString& text = m_cells[row][col].text;
        //wxRect& rect = m_cells[row][col].rect;

        if ( cell.disabled )
        {
            dc.SetTextForeground( m_theme->InactiveTextColor );
            dc.SetBrush( m_theme->InactiveBackgroundColor );
            dc.SetPen( m_theme->InactiveBackgroundColor );
        }
        else
        {
            dc.SetTextForeground( m_theme->BorderColor );
            dc.SetBrush( m_theme->VariableBackgroundColor );
            dc.SetPen( m_theme->VariableBackgroundColor );
        }

        newFont.SetUnderlined( cell.isLink() );

        dc.SetFont( newFont );
        //dc.Clear();
        dc.DrawRectangle( cell.rect.x, cell.rect.y, cell.rect.width, cell.rect.height );
        dc.DrawLabel( cell.text, cell.rect, cell.align | wxALIGN_CENTER_VERTICAL );

    }

    dc.SetFont( oldFont );
}

wxSize VemsExpressionGridWidget::getAvailableTextSize()
{
    int width = GetClientSize().GetWidth();
    width -= 2 * m_borderSize;
    width -= m_cols * 2 * m_cellBorderSize;
    width -= m_cols > 0 ? ( m_cols - 1 ) * m_gridSize : 0;

    int height = GetClientSize().GetHeight();
    height -= 2 * m_borderSize;
    //height -= m_rows * 2 * m_cellBorderSize;
    height -= m_rows > 0 ? ( m_rows - 1 ) * m_gridSize : 0;
    height -= m_titleHeight;

    wxSize size( width, height );

    return size;
}

void VemsExpressionGridWidget::calculateFontSize()
{
    wxPaintDC dc( this );

    wxFont font = wxFont( m_fontPointSize, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL );
    dc.SetFont( font );

    wxString text = getMaxTextSum();

    wxRect clientRect( getAvailableTextSize() );
    wxRect textRect( dc.GetTextExtent( text ) );

    double horizontalRatio = (double)clientRect.width / (double)textRect.width;
    double verticalRatio = (double)clientRect.height / ( (double)textRect.height * m_rows );

#ifndef __WINCE_STANDARDSDK__
    double ratio = std::min( horizontalRatio, verticalRatio );
#else
    double ratio = min( horizontalRatio, verticalRatio );
#endif
    m_fontPointSize = (size_t)((double)m_fontPointSize * ratio);

    if ( m_fontPointSize > MAX_FONT_SIZE )  m_fontPointSize = MAX_FONT_SIZE;
    if ( m_fontPointSize < MIN_FONT_SIZE ) m_fontPointSize = MIN_FONT_SIZE;
}

//void VemsExpressionGridWidget::caluclateGridsPositions()
//{
//    wxPaintDC dc( this );
//
//    wxFont font = wxFont( m_fontPointSize, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL );
//    dc.SetFont( font );
//
//    wxString text = getMaxTextSum();
//    wxSize sumTextSize = dc.GetTextExtent( text );
//
//    wxRect clientRect( getAvailableTextSize() );
//    wxRect textRect = wxRect( sumTextSize );
//
//    double horizontalRatio = (double)clientRect.width / (double)textRect.width;
//    double verticalRatio = (double)clientRect.height / ( (double)textRect.height * m_rows );
//
//
//    int x = m_borderSize;
//    wxSize textSize;
//
//    for ( size_t col = 0; col < m_cols; ++col )
//    {
//        int row = getColumnMaxTextRowIdx( col );
//
//        if ( row == -1 )
//        {
//            m_colsGridPos[col] = 0;
//            continue;
//        }
//
//        m_colsGridPos[col] = x;
//
//        wxString text = m_cells[row][col].text;
//
//        textSize = dc.GetTextExtent( text );
//
//        int textWidth = (int)( (double)textSize.GetWidth() * horizontalRatio );
//
//        for( size_t row = 0; row < m_rows; ++row )
//        {
//            m_cells[row][col].rect.x = x + m_cellBorderSize;
//            m_cells[row][col].rect.width = textWidth;
//        }
//
//        x += textWidth + 2 * m_cellBorderSize;
//
//    }
//
//    int y = m_borderSize + m_titleHeight;
//
//    for( size_t row = 0; row < m_rows; ++row )
//    {
//        m_rowsGridPos[row] = y;
//
//        int textHeight = (int)( (double)textSize.GetHeight() );// * verticalRatio );
//
//        for ( size_t col = 0; col < m_cols; ++col )
//        {
//            m_cells[row][col].rect.y = y; //+ m_cellBorderSize;
//            m_cells[row][col].rect.height = textHeight;
//        }
//
//        y += textHeight; // + m_cellBorderSize * 2;
//    }
//}

//void VemsExpressionGridWidget::calculateSize()
//{
//    calculateFontSize();
//    caluclateGridsPositions();
//}

void VemsExpressionGridWidget::setCellText( size_t row, size_t col, const wxString& text )
{
    if ( row >= m_rows || col >= m_cols ) return;
    m_cells[row][col].text = text;
}


void VemsExpressionGridWidget::OnPaint( wxPaintEvent &WXUNUSED(event) )
{
    //Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::OnPaint()") );

	wxAutoBufferedPaintDC dc( this );
    dc.SetBackground( *wxBLACK );
    dc.Clear();

//    if ( m_isResized )
//    {
//        m_isResized = false;
////        calculateSize();
//    }

    wxSize clientSize = GetClientSize();

////    dc.SetPen( *wxRED_PEN );
////    dc.DrawLine( 0, 0, clientSize.GetWidth(), clientSize.GetHeight() );

    if ( m_isResized )
    {
        calculateSize();
        {//Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::OnPaint()") );
        //Logger::logf( LOG_MSG, _T("create backbuffer - w: %d, h: %d"), clientSize.GetWidth(), clientSize.GetHeight() );

        delete m_backbuffer;
        m_backbuffer = new wxBitmap( clientSize.GetWidth(), clientSize.GetHeight() );
        wxMemoryDC memDC( *m_backbuffer );

        //Logger::logf( LOG_MSG, _T("memDC.IsOk(): %d"), memDC.IsOk() );
        memDC.SetBackground( *wxBLACK );
        memDC.Clear();
        drawBorder( memDC );
        drawTitle( memDC );
        drawGrids( memDC );
        drawTable( memDC );
        }
        dc.DrawBitmap( *m_backbuffer, 0, 0, true );

    }
    else
    {
        //Logger::logf( LOG_MSG, _T("draw backbuffer - w: %d, h: %d"), m_backbuffer->GetWidth(), m_backbuffer->GetHeight() );
        dc.DrawBitmap( *m_backbuffer, 0, 0, true );
        drawVariableCellsOnly( dc );
    }

    m_isResized = false;


    DrawExtraLayer(dc);
}

wxString VemsExpressionGridWidget::getMaxTextSum()
{
    wxString maxTextSum = wxEmptyString;

    for ( size_t col = 0; col < m_cols; ++col )
    {
        int row = getColumnMaxTextRowIdx( col );

        if ( row > -1 ) maxTextSum += m_cells[row][col].text;
    }

    return maxTextSum;
}

int VemsExpressionGridWidget::getColumnMaxTextRowIdx( size_t col )
{
    wxPaintDC dc( this );

    int maxTextRow      = -1;
    int maxTextWidth    = 0;

    for ( size_t row = 0; row < m_rows; ++row )
    {
        if ( col >= m_cols ) return -1;

        wxString text = m_cells[row][col].text;
        int textWidth = dc.GetTextExtent( text ).GetWidth();

        if ( textWidth > maxTextWidth )
        {
            maxTextWidth = textWidth;
            maxTextRow = row;
        }
    }

    return maxTextRow;
}

void VemsExpressionGridWidget::OnSize( wxSizeEvent &event )
{
    Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::OnSize") );
    Logger::logf( LOG_MSG, _T("m_sideRatioRestriction: %.2f"), m_sideRatioRestriction );

    //calculatePreferredTextSize();
//    Logger::logf( LOG_MSG, _T("m_preferredTextSize - w: %d, h: %d"),
//        m_preferredTextSize.GetWidth(), m_preferredTextSize.GetHeight() );

//    int width = 0;
//    int height = 0;
//
//    GetClientSize( &width, &height );
//
//    if ( width > height * m_sideRatio )
//        height = width / m_sideRatio;
//    if ( height > width / m_sideRatio )
//        width = height * m_sideRatio;
//
//    SetSize( width, height );

//    int width, height;
//
//
//	wxWindow* parent = 0;//GetParent();
//
//	if( parent )
//	{
//	    parent->GetSize( &width, &height );
//    }
//    else
//    {
//        GetSize( &width, &height );---------------------------------------------------
//    }
//
//    if( height > width * m_sideRatioRestriction )
//        height = width * m_sideRatioRestriction;
//    if( width * m_sideRatioRestriction > height )
//        width = height / m_sideRatioRestriction;
//
//    SetSize( width, height );

    //calculateSize();
    m_isResized = true;
    Refresh();

    event.Skip();
}

//void VemsExpressionGridWidget::OnLeftDown( wxMouseEvent& event )
//{
//    event.Skip();
//}
//void VemsExpressionGridWidget::OnMouseRightClick(wxMouseEvent& event)
//{
//    //Logger::logf(LOG_MSG,_T("VemsWidgetBase::OnMouseRightClick %i"),GetItemID());
//
//	if (FindFocus() != this)
//	{
//        SetFocus();
//	}
//
//    setMenuItemsState();
//	const wxPoint& pos = event.GetPosition();
//
//    wxMenu * menu = m_popupMenuViewMode;
//
//    if( m_groupManager && m_groupManager->IsEditMode() ) menu = m_popupMenuEditMode;
//
//    PopupMenu( menu , pos.x, pos.y);
//
//	//event.Skip();
//}

void VemsExpressionGridWidget::SetNewTheme( WidgetThemeTable *theme )
{
    m_isResized = true;
    m_theme = theme;
    //UpdateSize();
    Refresh();
}

void VemsExpressionGridWidget::update()
{
    //LOG

    bool isChanged = false;

    Int2StringHash& linesEnableExpr = m_gridDescriptor->linesEnableExpr;

    Int2StringHash::const_iterator it = linesEnableExpr.begin();

    //LOGP( wxString::Format( _T("linesEnableExpr.size() :%d"), linesEnableExpr.size() ) );

    for ( ; it != linesEnableExpr.end(); ++it ) {
        wxString expr = it->second;
        int line = it->first;

        if ( line < 0 || line >= (int)m_linesEnabled.GetCount() ) {
            LOGP( wxString::Format( _T("INVALID LINE IDX: %d"), line ) );
            continue;
        }

        //LOGP( wxString::Format( _T("line: %d, expr: "), line ) + expr );

        bool lineEnabled = m_stDataMgr->calculate( expr );

        //LOGP( wxString::Format( _T("lineEnabled: %d, m_linesEnabled[line]: %d"), lineEnabled, m_linesEnabled[line] ) );

        if ( m_linesEnabled[line] != lineEnabled ) {
            m_linesEnabled[line] = lineEnabled;

            for ( size_t col = 0; col < m_cols; ++col )
            {
                m_cells[line][col].disabled = !lineEnabled;
            }

            isChanged = true;
        }
    }

    if ( setLinksEnabled() ) {
        isChanged = true;
    }

    if ( setActiveTexts() ) {
        isChanged = true;
    }

    if ( isChanged ) {
        m_isResized = true;
        Refresh();
    }
}

void VemsExpressionGridWidget::refresh( DoubleHash* table )
{
    if ( !m_gridDescriptor ) return;

//    Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::refresh()") );

    for(size_t i = 0; i < m_variablePointA.GetCount(); ++i)
    {
        wxPoint* cell = m_variablePointA.Item(i);
        int row = cell->x;
        int col = cell->y;
        //wxArrayString* row = m_gridDescriptor->rows.Item(cell->x);
        //wxString cellStr = row->Item(cell->y);
        wxString text = m_cells[row][col].originText;

        StringSet* varSet = m_variableSetA.Item(i);
        int precision = m_cells[cell->x][cell->y].precision;
        VemsExpressionGridWidgetHelper::replaceRTVariables( *table, *varSet, &text, precision );

        setCellText( row, col, text );
    }

    setDisabledRows();

    Refresh();
}

void VemsExpressionGridWidget::require()
{
    if( m_isSymbolsRegistered ) return;

    StringSet::const_iterator IT = m_rtVariables.begin();
    for(; IT != m_rtVariables.end(); ++IT){
        wxString varName = *IT;
        m_rtDataMgr->registerSymbol( varName );
    }

    m_isSymbolsRegistered = true;
}

void VemsExpressionGridWidget::unregisterSymbols()
{
    if( !m_isSymbolsRegistered ) return;

    StringSet::const_iterator IT = m_rtVariables.begin();
    for(; IT != m_rtVariables.end(); ++IT){
        wxString varName = *IT;
        m_rtDataMgr->unregisterSymbol( varName );
    }
}

void VemsExpressionGridWidget::initIcons()
{
    wxString path = DirHelper::GetPath(dICON);

    m_helpIcons.Add( new wxBitmap( path + _T("ul-info-s.png"), wxBITMAP_TYPE_PNG ) );
    m_helpIcons.Add( new wxBitmap( path + _T("ul-info-m.png"), wxBITMAP_TYPE_PNG ) );
    m_helpIcons.Add( new wxBitmap( path + _T("ul-info-l.png"), wxBITMAP_TYPE_PNG ) );
    m_helpIcons.Add( new wxBitmap( path + _T("ul-info-xl.png"), wxBITMAP_TYPE_PNG ) );
    //m_helpIcons.Add( new wxBitmap( path + _T("ul-info-xxl.png"), wxBITMAP_TYPE_PNG ) );
}

void VemsExpressionGridWidget::createWidgetMenu( wxMenu& popupMenu )
{
    Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::createWidgetMenu()") );

    wxMenu* changeGridMenu = new wxMenu();

    m_menuIDtoGridName.clear();

    const ExpressionGridDescArray& gridDescs = MTConfigHelper::getMTConfigData()->m_expressionGridDescriptors;

    for ( size_t i = 0; i < gridDescs.GetCount(); ++i )
    {
        wxString name = gridDescs[i]->name;

        if ( name.IsSameAs( GetCurrSymbolName() ) ) continue;

        int menuId = wxNewId();

        changeGridMenu->Append( menuId, name );

        m_menuIDtoGridName[menuId] = name;

        Connect( menuId, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler(VemsExpressionGridWidget::OnPopupMenuChangeGrid) );
    }

    popupMenu.AppendSubMenu( changeGridMenu, _T("Change To") );
}

void VemsExpressionGridWidget::OnPopupMenuChangeGrid( wxCommandEvent& event )
{
    Int2StringHash::const_iterator it = m_menuIDtoGridName.find( event.GetId() );

    if ( it == m_menuIDtoGridName.end() ) return;

    SetCurrSymbolName(it->second);

    m_groupManager->SaveSymbolname( m_props.itemID, GetCurrSymbolName() );

    unregisterSymbols();
    clear();
    createGrid();
    calculateSize( true );
    //SetSize( m_preferredSize );
    //SetSize( m_minClientSize );
    //createChangeGridMenuItems();
    m_isSymbolsRegistered = false;
    requireChanged();
    m_isResized = true;
    update();
    Refresh();

}


//void VemsExpressionGridWidget::createChangeGridMenuItems( wxMenu& popup )
//{
////    Int2StringHash::const_iterator it = m_menuIDtoGridName.begin();
////
////    for ( ; it != m_menuIDtoGridName.end(); ++it )
////    {
////        m_popupMenuViewMode->Delete( it->first );
////    }
////
//    m_menuIDtoGridName.clear();
//
//    const ExpressionGridDescArray& gridDescs = MTConfigHelper::getMTConfigData()->m_expressionGridDescriptors;
//
//    //wxArrayString gridNames;
//
//    for ( size_t i = 0; i < gridDescs.GetCount(); ++i )
//    {
//        wxString name = gridDescs[i]->name;
//
//        if ( name.IsSameAs( m_currSymbolName ) ) continue;
//
//        int menuId = wxNewId();
//
//
//        //gridNames.Add( gridDescs[i]->name );
//    }
//
//    //gridNames.Remove( m_currSymbolName );
//
//    for ( size_t i = 0; i < gridNames.GetCount(); ++i )
//    {
//        int id = wxNewId();
//        m_changeGridMenu->Append( id, gridNames[i] );
//        m_menuIDtoGridName[id] = gridNames[i];
//        Connect( id, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler(VemsExpressionGridWidget::OnPopupMenuChangeGrid) );
//    }
//}

void VemsExpressionGridWidget::clear()
{
    m_disableCol = -1;

    GridCellMatrix::extent_gen extents;
    m_cells.resize(extents[1][1]);

    m_colMaxTextWidth.clear();
    m_colMaxTextRow.clear();
    m_colAvailableTextWidth.clear();

    m_rowHeight.clear();
    m_rowAvailableTextHeight.clear();

    m_colsGridPos.clear();
    m_rowsGridPos.clear();
    m_linesEnabled.clear();

    m_rtVariables.clear();

    WX_CLEAR_ARRAY( m_variablePointA );
    WX_CLEAR_ARRAY( m_variableSetA );

    //m_menuIDtoGridName.clear();

    m_gridDescriptor = NULL;

    m_cellFont = wxFont( MIN_FONT_SIZE, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL );
}

void VemsExpressionGridWidget::OnLeftDown(wxMouseEvent& evt)
{
    if( m_groupManager->IsEditMode() )
    {
        LOGP( _T("It does not work in EDIT MODE!") );
        evt.Skip();
        return;
    }

    if(GetShape().Contains(evt.GetX(),evt.GetY()) != wxInRegion )
    {
        //m_parent->ProcessEvent(evt);
        return;
    }

    SetFocus();

    wxPoint mousePos = evt.GetPosition();

    wxPoint cellPosition = getCellPosition( mousePos );

    showLink( cellPosition );

    int row = cellPosition.x;
    int col = cellPosition.y;

    if ( row < 0 || row >= (int)m_rows || col < 0 || col >= (int)m_cols ) return;

    GridCell& cell = m_cells[row][col];

    if ( cell.text == _T("note") && !m_winTip->IsShown())
    {
        m_winTip->SetSize( wxRect( mousePos+GetScreenPosition(), wxDefaultSize ) );
        m_winTip->ShowMsg( cell.help, 1000, false );
    }
    else
    {
        m_winTip->HideWin();
    }
}

void VemsExpressionGridWidget::OnMouseMove(wxMouseEvent& evt)
{
    if( m_groupManager->IsEditMode() ) {
        SetCursor(wxNullCursor);
        evt.Skip();
        return;
    }

    wxPoint mousePos = evt.GetPosition();
    wxPoint cellPosition = getCellPosition( mousePos );

    int row = cellPosition.x;
    int col = cellPosition.y;

    if ( row < 0 || row >= (int)m_rows || col < 0 || col >= (int)m_cols ) return;

    GridCell& cell = m_cells[row][col];

    if ( cell.isLink() )
    {
        SetCursor( wxCursor( wxCURSOR_HAND ) );
    }
    else
    {
        SetCursor( wxNullCursor );
    }

    evt.Skip();
}


wxPoint VemsExpressionGridWidget::getCellPosition( wxPoint mousePosition )
{
    for ( size_t row = 0; row < m_rows; ++row )
    {
        for ( size_t col = 0; col < m_cols; ++col )
        {
            if ( m_cells[row][col].rect.Contains( mousePosition ) ) return wxPoint( row, col );
        }
    }

    return wxPoint( -1, -1 );
}

void VemsExpressionGridWidget::showLink( wxPoint cellPosition )
{
    int row = cellPosition.x;
    int col = cellPosition.y;

    if ( row < 0 || row >= (int)m_rows || col < 0 || col >= (int)m_cols ) return;

    GridCell& cell = m_cells[row][col];

    if ( !cell.isLink() ) {
        return;
    }

    if ( !cell.linkEnabled ) {
        LOGP( _T("Link is disabled!") );
        return;
    }

    StringSet::const_iterator it = cell.dlgSymbols.begin();

    wxString symbolsToMark;

    for ( ; it != cell.dlgSymbols.end(); ++it )
        symbolsToMark += *it + _T(",");

    if ( symbolsToMark != _T("") )
    {
        symbolsToMark.RemoveLast( 1 );
    }

    wxObject* obj = VemsContext::getObject(_T("UserDialogBase") );

    if( obj )
    {
        UserDialogBase* ud = dynamic_cast<UserDialogBase*>( obj );

        if ( ud )
        {
            //ud->setSymbolsToMark( symbolsToMark );
            ud->showDialog( cell.activeLink, symbolsToMark );
    }
        //((UserDialogBase*)obj)->showDialog( cell.activeLink );
    }
    else
    {
        Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::OnLeftDown() - ")
                     _T("UserDialogManager Object not found in VemsContext!") );
    }
}


void VemsExpressionGridWidget::findColsLongestText()
{
    int height  = 0;
    int width   = 0;

    for ( size_t row = 0; row < m_rows; ++row )
    {
        for ( size_t col = 0; col < m_cols; ++col )
        {
            wxString text = m_cells[row][col].getDefaultText();

            GetTextExtent( text, &width, &height );

            if ( width > m_colMaxTextWidth[col] )
            {
                m_colMaxTextWidth[col] = width;
                m_colMaxTextRow[col] = row;
            }
        }
    }

    for ( size_t row = 0; row < m_rows; ++row )
    {
        GetTextExtent( _T("0"), &width, &height );
        m_rowHeight[row] = height;
    }
}

void VemsExpressionGridWidget::calculateMinSize()
{
    int maxTextWidth = 0;

    for ( size_t col = 0; col < m_colMaxTextWidth.GetCount(); ++col )
    {
        maxTextWidth += m_colMaxTextWidth[col];
    }

    int textHeight = 0;

    for ( size_t row = 0; row < m_rowHeight.GetCount(); ++row )
    {
        textHeight += m_rowHeight[row];
    }

    m_minTextSize = wxSize( maxTextWidth, textHeight );

    m_minClientSize = m_minTextSize + m_decorationsSize;

    m_sideRatio = (double)m_minClientSize.GetWidth() / (double)m_minClientSize.GetHeight();
}

void VemsExpressionGridWidget::calculateSize( bool autoSize )
{
    //Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::calculateSize() - autoSize: %d"), autoSize );

    if ( autoSize )
    {
        wxWindow* parent = GetParent();

        SetSize( m_minClientSize );

        if ( GetItemID() == -1 && parent )
        {
            Logger::logf( LOG_MSG, _T("VemsExpressionGridWidget::calculateSize() - set parent size: w: %d, h: %d"),
                        m_minClientSize.GetWidth(), m_minClientSize.GetHeight() );
            parent->SetClientSize( m_minClientSize );
        }

        //SetSize( m_minClientSize );
    }
    //findColsLongestText();
    //calculateMinSize();

    wxSize availableSize = GetClientSize();

   //Logger::logf( LOG_MSG, _T("availableSize: w: %d, h: %d"), availableSize.GetWidth(), availableSize.GetHeight() );

    wxSize avaialableTextSize = availableSize - m_decorationsSize;

    double horizontalScale = (double)avaialableTextSize.GetWidth() / (double)m_minTextSize.GetWidth();
    double verticalScale = (double)avaialableTextSize.GetHeight() / (double)m_minTextSize.GetHeight();

    //Logger::logf( LOG_MSG, _T("horizontalScale: %.2f"), horizontalScale );
    //Logger::logf( LOG_MSG, _T("verticalScale: %.2f"), verticalScale );

///calculate new width
    IntArray ranges;
    int range = 0;

    for ( size_t col = 0; col < m_colMaxTextWidth.GetCount(); ++col )
    {
        range += m_colMaxTextWidth[col];
        ranges.Add( (int)( range * horizontalScale + 0.5 ) );
    }

    for ( size_t col = 0; col < m_colMaxTextWidth.GetCount(); ++col )
    {
        m_colAvailableTextWidth[col] = col == 0 ? ranges[col] : ranges[col] - ranges[col-1];
    }

/// calculate new height

    ranges.clear();
    range = 0;

    for ( size_t row = 0; row < m_rowHeight.GetCount(); ++row )
    {
        range += m_rowHeight[row];
        ranges.Add( (int)( range * verticalScale + 0.5 ) );
    }

    for ( size_t row = 0; row < m_rowHeight.GetCount(); ++row )
    {
        m_rowAvailableTextHeight[row] = row == 0 ? ranges[row] : ranges[row] - ranges[row-1];
    }

/// calculate new font size
    wxFont newFont = m_cellFont;

    int fontSize = (int)(MIN_FONT_SIZE * horizontalScale);

    bool allCellFit = true;

    do {

        allCellFit = true;

        for ( size_t col = 0; col < m_colMaxTextWidth.GetCount(); ++col )
        {
            int width = 0;
            int height = 0;
            size_t row = m_colMaxTextRow[col];

            GetTextExtent( m_cells[row][col].text, &width, &height, 0, 0, &newFont );

            if ( width > m_colAvailableTextWidth[col] || height > m_rowAvailableTextHeight[row] )
            {
                --fontSize;

                if ( fontSize < MIN_FONT_SIZE )
                {
                    fontSize = MIN_FONT_SIZE;
                    allCellFit = true;
                    break;
                }

                newFont.SetPointSize( fontSize );
                allCellFit = false;
                break;
            }
        }
    } while ( !allCellFit );

    m_cellFont.SetPointSize( fontSize );

    caluclateGridsPositions();
}

void VemsExpressionGridWidget::caluclateGridsPositions()
{


    int colPos = 0;

    for ( size_t col = 0; col < m_cols - 1; ++col )
    {
        if ( col == 0 )
        {
            colPos += m_borderSize + 2 * m_cellBorderSize + m_colAvailableTextWidth[col];
        }
        else
        {
            colPos += 2 * m_cellBorderSize + m_colAvailableTextWidth[col] + m_gridSize;
        }

        m_colsGridPos[col] = colPos;
    }

    int rowPos = 0;

    for ( size_t row = 0; row < m_rows - 1 ; ++row )
    {
        if ( row == 0 )
        {
            //rowPos += m_borderSize + 2 * m_cellBorderSize + m_textHeight;
            rowPos += m_titleHeight + m_borderSize + 2 * m_cellBorderSize + m_rowAvailableTextHeight[row];
        }
        else
        {
            //rowPos += 2 * m_cellBorderSize + m_textHeight + m_gridSize;
            rowPos += 2 * m_cellBorderSize + m_rowAvailableTextHeight[row] + m_gridSize;
        }
        m_rowsGridPos[row] = rowPos;
    }


    rowPos = m_borderSize + m_titleHeight;
    int cellWidth = 0;
    //int cellHeight = m_textHeight + 2 * m_cellBorderSize;

    for ( size_t row = 0; row < m_rows ; ++row )
    {
        int cellHeight = m_rowAvailableTextHeight[row] + 2 * m_cellBorderSize;

        colPos = m_borderSize;

        for ( size_t col = 0; col < m_cols; ++col )
        {
            cellWidth = m_colAvailableTextWidth[col] + 2 * m_cellBorderSize;


            m_cells[row][col].rect.x = colPos;
            m_cells[row][col].rect.y = rowPos;
            m_cells[row][col].rect.width = cellWidth;
            m_cells[row][col].rect.height = cellHeight;

            colPos += cellWidth + m_gridSize;

            //Logger::logf( LOG_MSG, _T("m_cellRects[%d][%d]"), row, col );
            //Logger::logf( LOG_MSG, _T("x: %d, y: %d, w: %d, h: %d"), colPos, rowPos, cellWidth, cellHeight );

        }

        rowPos += cellHeight + m_gridSize;
    }
}

void VemsExpressionGridWidget::setDisabledRows()
{
   // LOGP( wxString::Format( _T("m_disableCol: %d"), m_disableCol ) );

    if ( m_disableCol == -1 ) return;

    for ( size_t row = 1; row < m_rows; ++row )
    {
        if ( !m_linesEnabled[row] ) {
            continue;
        }
        //LOGP( wxString::Format( _T("m_cells[%d][%d].isVariable() = %d"), row, m_disableCol, m_cells[row][m_disableCol].isVariable() ) );
        //LOGP( wxString::Format( _T("m_cells[%d][%d].text = ") + m_cells[row][m_disableCol].text, row, m_disableCol ) );
        if ( !m_cells[row][m_disableCol].isVariable() ) continue;

        m_cells[row][m_disableCol].disabled = m_cells[row][m_disableCol].text == m_cells[row-1][m_disableCol].text;
       //Logger::logf( LOG_MSG, _T("m_cells[%d][%d].disabled = %d"), row, m_disableCol, m_cells[row][m_disableCol].disabled );
    }

}

wxBitmap* VemsExpressionGridWidget::getHelpBmp( const wxRect& rect )
{
    if(m_helpIcons.GetCount() < 1) return NULL;

    for ( int i = m_helpIcons.GetCount() - 1;  i >=0 ; --i )
    {
        if ( m_helpIcons[i]->GetWidth() <= rect.width && m_helpIcons[i]->GetHeight() <= rect.height ) {
            return m_helpIcons[i];
        }
    }

    return NULL;
}

bool VemsExpressionGridWidget::setLinksEnabled()
{
    bool isChanged = false;

    for ( size_t row = 0; row < m_rows; ++row )
    {
        for ( size_t col = 0; col < m_cols; ++col )
        {
            GridCell& cell = m_cells[row][col];

            bool linkEnabled = cell.linkEnabled;

            setActiveLink( cell );

            if ( linkEnabled != cell.linkEnabled ) {
                isChanged = true;
            }
        }
    }

    return isChanged;
}

void VemsExpressionGridWidget::setActiveLink( GridCell& cell )
{
    ExpressionCellLinkArray& links = cell.links;

    for ( size_t i = 0; i < links.GetCount(); ++i )
    {
        ExpressionCellLink& link = links[i];

        if ( link.enableExpr.IsEmpty() ) {
            cell.activeLink     = link.dialogId;
            cell.linkEnabled    = true;

            cell.dlgSymbols.clear();
            for ( size_t i = 0; i < link.variables.GetCount(); ++i )
            {
                cell.dlgSymbols.insert( link.variables[i] );
            }

            return;
        }

        bool linkEnabled = m_stDataMgr->calculate( link.enableExpr );

        if ( linkEnabled ) {
            cell.activeLink     = link.dialogId;
            cell.linkEnabled    = true;

            cell.dlgSymbols.clear();
            for ( size_t i = 0; i < link.variables.GetCount(); ++i )
            {
                cell.dlgSymbols.insert( link.variables[i] );
            }

            return;
        } else {
            cell.activeLink     = wxEmptyString;
            cell.linkEnabled    = false;

            cell.dlgSymbols.clear();
        }
    }
}

bool VemsExpressionGridWidget::setActiveTexts()
{
    bool isChanged = false;

    for ( size_t row = 0; row < m_rows; ++row )
    {
        for ( size_t col = 0; col < m_cols; ++col )
        {
            GridCell& cell = m_cells[row][col];

            wxString textBefore = cell.text;
            setActiveText( cell );

            if ( textBefore != cell.text ) {
                isChanged = true;
            }
        }
    }

    return isChanged;
}

void VemsExpressionGridWidget::setActiveText( GridCell& cell )
{
    ExprCellTextArray& texts = cell.texts;

    if ( texts.IsEmpty() ) return;

    cell.text = cell.originText;

    LOGP( cell.text + _T(", ") + cell.originText );

    for ( size_t i = 0; i < texts.GetCount(); ++i )
    {
        ExprCellText& text = texts[i];

        bool textEnabled = m_stDataMgr->calculate( text.enableExpr );

        if ( textEnabled ) {
            cell.text = text.text;
        }
    }
}

void VemsExpressionGridWidget::setActiveColor( GridCell& cell )
{
    ExprCellColorArray& colors = cell.colors;

    if ( cell.disabled ) {
        cell.bgColor = m_theme->InactiveBackgroundColor;
        cell.fgColor = m_theme->InactiveTextColor;
    } else {
        if ( cell.isVariable() ) {
            cell.bgColor = m_theme->VariableBackgroundColor;
        } else {
            cell.bgColor = m_theme->BackgroundColor;
        }
    }

    cell.fgColor = cell.isLink() && !cell.isVariable() ? m_theme->LinkTextColor : m_theme->BorderColor;

    if ( cell.isLink() && !cell.linkEnabled ) {
        cell.fgColor = m_theme->InactiveTextColor;
    }

    if ( colors.IsEmpty() ) return;

    for ( size_t i = 0; i < colors.GetCount(); ++i )
    {
        ExprCellColor& color = colors[i];

        bool colorEnabled = m_stDataMgr->calculate( color.enableExpr );

        if ( !colorEnabled ) continue;

        getColorFromStr( color.bgColor, cell.bgColor );
        getColorFromStr( color.fgColor, cell.fgColor );
    }
}

void VemsExpressionGridWidget::getColorFromStr( const wxString& colorStr, wxColor& color )
{
    if ( !colorStr.StartsWith( _T("0x") ) && !colorStr.StartsWith( _T("0X") ) ) return;

    if ( colorStr.Len() < 8 ) return;

    unsigned long val = 0;

    if ( !colorStr.ToULong( &val, 0 ) ) return;

    unsigned char red, green, blue;

    red     = ( val & 0x00FF0000 ) >> 16;
    green   = ( val & 0x0000FF00 ) >> 8;
    blue    = ( val & 0x000000FF );

    color = wxColor( red, green, blue );
}
