#include "FirmwareData.h"

#include "../logger.h"

FirmwareData::FirmwareData(ECUInfo::Device device)
{
    //ctor
    m_type = FirmwareData::UNKNOWN;
    m_device = device;
    switch(device){
        case ECUInfo::DEVICE_ARM:
        case ECUInfo::DEVICE_V3:
            m_pageSize = 256;
            break;
        case ECUInfo::DEVICE_ROUND:
            m_pageSize = 128;
            break;
        default:
            m_pageSize = 256;
    }

}

FirmwareData::~FirmwareData()
{
    //dtor
}


void FirmwareData::setPageCheckSum(uint32_t idx, uint32_t chk)
{
   m_pageCheckSum[idx] = chk;
}

void FirmwareData::setPageData(uint32_t idx, uint32_t offset, uint8_t * buff, uint32_t len)
{
    if( m_pageBufferHash.find(idx) == m_pageBufferHash.end() ){
            m_pageBufferHash[idx] = new uint8_t [m_pageSize];
            m_intArray.Add(idx);
            m_PageSet.insert(idx);
            //init buff with FF
            uint8_t *b = m_pageBufferHash[idx];
            for(uint32_t i = 0; i < m_pageSize; i++){
                b[i] = 0xFF;
            }
    }

    if(offset + len > m_pageSize ){
        Logger::logf(LOG_MSG, _T("FirmwareData::setPageData() Warning: using len= %i Out of Bounds (%i < %i + %i)!"), m_pageSize - offset, m_pageSize, offset, len);
        len = m_pageSize - offset;
    }

    uint8_t * ptr = m_pageBufferHash[idx];
    uint8_t * curPtr = ptr+offset;
    memcpy( curPtr, buff, len);
}

uint8_t * FirmwareData::getPageData(uint32_t idx, bool& found)
{
    if( m_pageBufferHash.find(idx) == m_pageBufferHash.end() ){
        found = false;
        return NULL;
    }
    found = true;
    uint8_t * ptr = m_pageBufferHash[idx];
    return ptr;
}

void FirmwareData::printFirmwareData()
{
    uint32_t pagesNum = getPages();

    int dataInLine = 16;
    size_t i,j;
    wxString idxStr, line;
    for(i = 0; i < pagesNum; i++)
    {
        uint32_t idx = getPageIndexbyOrderIdx(i);
        bool found;
        uint8_t *buf = getPageData(idx, found);

        idxStr= wxString::Format(_T("%00i "),idx);
        wxString hexData;
        for(j=0; j < m_pageSize; j++ ){
           hexData += wxString::Format(_T("%02X "), buf[j] );
           if( ((j+1) % dataInLine) == 0){
                line = wxString::Format(_T("0x%04X ")+ hexData, getPageAddress(idx)+ j-(dataInLine-1) );
                Logger::logf(LOG_MSG, idxStr + line);
                hexData = wxEmptyString;
           }

        }

    }
}

/*
uint16_t * FirmwareData::getPageDataW(uint32_t idx, bool& found)
{
    return (uint16_t*) getPageData(idx, found);
}
*/

uint32_t FirmwareData::getPageCheckSum(uint32_t idx)
{
    if(m_type == FirmwareData::CHEX){ return m_pageCheckSum[idx]; }

    uint16_t checksum1 = 0;
    switch(m_device){
        case ECUInfo::DEVICE_ARM:
            checksum1 = idx; // ARM
            break;
        case ECUInfo::DEVICE_V3:
           checksum1 = getPageAddressW(idx); // V3,
            break;
        case ECUInfo::DEVICE_ROUND:
        default:
            Logger::logf(LOG_ERROR, _T("FirmwareData::getPageCheckSum() INVALID DEVICE"));
    }


    uint16_t checksum2 = checksum1;
    uint16_t dataW;
    uint32_t i;
    bool found;
    uint8_t * pageData = getPageData(idx, found);

    for (i = 0; i< getPageSize(); i+=2)
    {
        dataW = getLittleEndianWordValue(pageData,i);
        //if(idx == 503) Logger::logf(LOG_ERROR, _T("FirmwareData::getPageCheckSum() idx:%i, dataW:%04X"),i, dataW );
        checksum1 += dataW;
        checksum2 ^= dataW;
    }

    //fprintf( stdout, "Firmware::getCheckSum(file): %04X, %04X \n", checksum1, checksum2);

    return (checksum1 << 16) + checksum2;
    //pageBuff->checkSum =
}

/*
void FirmwareData::setChecksumBuff(uint32_t idx, uint8_t * checksumStr)
{
    uint32_t value = getPageCheckSum(idx);
	checksumStr[0] = (value >> 24) & 0xFF;
	checksumStr[1] = (value >> 16) & 0xFF;
	checksumStr[2] = (value >>  8) & 0xFF;
	checksumStr[3] = value & 0xFF;
}
*/
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
uint16_t FirmwareData::getLittleEndianWordValue(uint8_t *buf, uint32_t i)
{
    return buf[i]+ 256 * buf[i+1] ;
    //return buf[i+1]+ 256 * buf[i] ;
}

void FirmwareData::addPatch(FirmwareData* firmwarePatch)
{
    bool found;
    for(uint32_t i =0; i < firmwarePatch->getPages(); i++){
        uint32_t idx = firmwarePatch->getPageIndexbyOrderIdx(i);
        Logger::logf(LOG_MSG, _T("FirmwareData::addPatch(): patch page No: %d"), idx );
        uint8_t * buff = firmwarePatch->getPageData(idx, found);
        setPageData(idx, 0, buff, firmwarePatch->getPageSize() );
    }

}

wxString FirmwareData::getTypeAsString()
{
    switch(m_type){
        case FirmwareData::HEX:
            return _T("Intel HEX");
        case FirmwareData::CHEX:
            return _T("Crypto CHEX");
        default:
            return _T("Unknown");
    }
    return _T("Unknown");
}

void FirmwareData::setV3MarkerWord()
{
    bool found = false;
    uint8_t *pageData = getPageData(503, found);
    if(!found) return;
    Logger::logf(LOG_MSG, _T("FirmwareData::setV3MarkerWord() at page 503 !!!"));

    //uint32_t chkOrig = getPageCheckSum(503);
    //Logger::logf(LOG_MSG, _T("FirmwareData::setV3MarkerWord() chksumOrig:%08X"), chkOrig);
    //pageData[510] = 0xA5;
    //pageData[511] = 0x5A;
    //uint32_t chkA55A = getPageCheckSum(503);
    //Logger::logf(LOG_MSG, _T("FirmwareData::setV3MarkerWord() chkA55A:%08X"), chkA55A);

    pageData[254] = 0x5A;
    pageData[255] = 0xA5;
    //uint32_t chk5AA5 = getPageCheckSum(503);
    //Logger::logf(LOG_MSG, _T("FirmwareData::setV3MarkerWord() chk5AA5:%08X"), chk5AA5);

}
 void FirmwareData::writeToIhex(void)
 {
    wxFile verificationIhexFile( _T("/tmp/verification.phex"), wxFile::write);

    uint32_t pageIdx;
    for (size_t j=0;j<getPages();j++){
        pageIdx = getPageIndexbyOrderIdx(j);
        uint16_t baseAddress = getPageAddress(pageIdx);
        uint8_t l;
        uint16_t pagesize16 = getPageSize() >> 4; // the number of lines (16-byte each)
        uint8_t i, val;
        char filebuff[45];
        bool found;


        uint8_t* buf = getPageData(pageIdx,found);
        if ( !found ) {
            printf("pageIdx invalid%d\n",pageIdx);
            return;
        }

        uint16_t pageOffset = 0;
        for( l = 0; l< pagesize16; l++){
            uint16_t lineAddress = baseAddress +  pageOffset;
            uint8_t chksum;
            // TODO: length is always 0x10 wired in !!
            sprintf( filebuff,":10%04X00", lineAddress );
            chksum = (0x10 + (lineAddress >> 8)) + (uint8_t)lineAddress;

            for ( i=0; i<=15; i++ ){
                val = buf[  pageOffset ];
                sprintf(filebuff+9+(i*2),"%02X", val);
                chksum += val;
                pageOffset ++;
            }
            chksum = 256 - chksum; // sum of all hex bytes must be zero

            sprintf(filebuff+41,"%02X\r\n", chksum );
            verificationIhexFile.Write(filebuff,45 );

        }
    }
    verificationIhexFile.Close();
 }
