#ifndef MEGATUNEPROTOCOL_H
#define MEGATUNEPROTOCOL_H

#include "ProtocolBase.h"
#include "../common/types/Int2IntHash.h"
#include "../common/types/IntSet.h"

class SIPRPacket;
class SIPRDescriptor;

///-----------------------------------------------------------------------------------------------MegatuneCommand
class PageRequest
{
    public:
        PageRequest()
        {
            active = false;
        }

        virtual ~PageRequest() {}

        bool    active;
        IntSet  requestedPages;
};

WX_DECLARE_LIST( PageRequest, PageRequestList );


///-----------------------------------------------------------------------------------------------MegatuneCommand
class MegatuneCommand
{
    public:
            MegatuneCommand(    int             commandType,
                                const uint8_t*  data,               // data must exist, data copied
                                uint8_t         size,               // must greater than > 0
                                int             expectedRxSize,
                                int             timeoutCount,
                                int             clearCount,
                                int             pageNo          = -1 );

            MegatuneCommand(    int             commandType,
                                const           wxString& command,
                                int             expectedRxSize,
                                int             timeoutCount,
                                int             clearCount,
                                int             pageNo          = -1 );

            ~MegatuneCommand() { delete[] m_pData; }

                            //int pageOffset = -1,
                            //uint8_t value = 0 );
        // for sync pages
        void setExpectedRxSize( int size ) { m_iExpectedRxSize = size; }

        int             getExpectedRxSize() const { return m_iExpectedRxSize; }
        int             getTimeoutCount() const { return m_iTimeoutCount; }
        int             getClearCount() const { return m_iClearCount; }
        const uint8_t*  getData() const { return m_pData; }
        uint8_t         getSize() const { return m_size; }
        int             getType() const { return m_iType; }
        int             getPageNumber() const { return m_iPageNumber; }

    private:
        uint8_t*    m_pData;
        uint8_t     m_size;
        int         m_iExpectedRxSize;
        int         m_iTimeoutCount;
        int         m_iClearCount;
        int         m_iPageNumber;
        int         m_iType;

    private:
        void    init(   int             type,
                        const uint8_t*  data,
                        uint8_t         size,
                        int             expectedRxSize,
                        int             timeoutCount,
                        int             clearCount,
                        int             pageNo );

        void    init(   int             type,
                        const wxString& command,
                        int             expectedRxSize,
                        int             timeoutCount,
                        int             clearCount,
                        int             pageNo );
};

WX_DECLARE_LIST( MegatuneCommand, MegatuneCommandList ); //store pointers


///-----------------------------------------------------------------------------------------------MegatuneProtocol
class MegatuneProtocol : public ProtocolBase
{
    public:
        MegatuneProtocol(   SIPRDescriptor* siprDesc,
                            int sensorType,
                            wxEvtHandler* eventHandler,
                            int timeout = 10,
                            int rtErrorCount = 3 );

        virtual ~MegatuneProtocol();

        // ProtocolBase virtual functions ( communication )
        int rx( uint8_t* rxBuffer, int size );
        int tx( uint8_t* txBuffer, int& size );

        // ProtocolBase virtual functions ( request )
        void request( VemsCommand* command );
        int type() const { return ECUInfo::MEGATUNE; }

        // ProtocolBase virtual functions ( for sync )
        bool hasError() { return ( m_state == TIME_OUT ); }
        bool needClear();
        void reset();
        void disable();
        void enable();
        bool isDisabled();
        bool isIdle() { return m_state == IDLE; }

    private:
//        static const int SYNC_RT_MAX_TRY_COUNT = 10;

        enum Mode       { NORMAL, SYNC };
        enum State      { IDLE, WAIT, CLEAR, TIME_OUT };
        enum SyncState
        {
            NEED_RT_SIZE,
            GET_RT_SIZE,
            NEED_TEST_RT_SIZE,
            TEST_RT_SIZE,
            HAVE_RT_SIZE,

            NEED_PAGE_SIZE,
            GET_PAGE_SIZE,
            NEED_TEST_PAGE_SIZE,
            TEST_PAGE_SIZE,
            HAVE_PAGE_SIZE
        };

        enum { SYNC_PAGE_TRY_COUNT_MAX = 3 };
        enum { READ_PAGE_TRY_COUNT_MAX = 3 };

    private:
        MegatuneCommandList     m_txQueue;
        SIPRDescriptor*         m_pSIPRDescriptor;
        wxEvtHandler*           m_pEventHandler;
        wxString                m_strReadRealtimeCommand;
        int                     m_iRealtimeLength;
        int                     m_iSensorType;

        // rx variables
        int                     m_iTimeoutCount;
        int                     m_iClearCount;
        int                     m_iRxDataLength;
        uint8_t                 m_rxBuffer[512];
        int                     m_realtimeErrorCount;
        int                     m_realtimeCRCErrorCount;
        int                     m_realtimeCount;
        int                     m_maxRealtimeErrorCount;

        // sync variables
        int                     m_iNumberOfPages;
        int                     m_iCurrentPageIndex;
        Int2IntHash             m_SizeOfPages;
        int                     m_missingPages;

        // state variables
        bool                    m_isReadSensors;        // for emulate always read_sensor command
        int                     m_mode;
        int                     m_state;
        int                     m_syncState;

        MegatuneCommand*        m_pCurrentCommand;
        uint8_t                 m_buffer[261];

        bool                    m_bNeedDisable;
        bool                    m_bIsDisabled;
        bool                    m_bSyncStarted;
        wxDateTime              m_realtimeErrorCountTime;
        int                     m_realtimeErrorCountPerMinute;
        int                     m_maxRealtimeErrorCountPerMinute;
        bool                    m_isRTErrorMessageSent;
        wxDateTime              m_startTime;
        FirmwareVersion         m_rtCRCCheckMinimumVersion;
        PageRequestList         m_readPageRequests;
        int                     m_syncPageTryCount;
        int                     m_readPageTryCount;

    private:
        void                    rxWaitState( uint8_t* rxBuffer, int size );
        void                    rxWaitReadState( uint8_t* rxBuffer, int size );
        void                    rxWaitEmptyState( uint8_t* rxBuffer, int size );
        void                    rxTimeoutNormalMode( uint8_t* rxBuffer, int size );
        void                    rxDataReceivedNormalMode( uint8_t* rxBuffer, int size );
        bool                    checkPageCrc();
        void                    rxDataReceivedSyncMode( uint8_t* rxBuffer, int size );
        void                    rxTimeoutSyncMode( uint8_t* rxBuffer, int size );
        void                    rxTimeoutSyncModeGetRtSize( uint8_t* rxBuffer, int size );
        void                    rxTimeoutSyncModeTestRtSize( uint8_t* rxBuffer, int size );
        void                    rxTimeoutSyncModeGetPageSize( uint8_t* rxBuffer, int size );
        void                    rxTimeoutSyncModeTestPageSize( uint8_t* rxBuffer, int size );
        void                    rxClearState( uint8_t* rxBuffer, int size );

        int                     txSyncMode( uint8_t* txBuffer, int& size );
        void                    txSyncModeNeedRtSize( uint8_t* txBuffer, int& size );
        void                    txSyncModeNeedTestRtSize( uint8_t* txBuffer, int& size );
        void                    txSyncModeHaveRtSize( uint8_t* txBuffer, int& size );
        int                     txSyncModeNeedPageSize( uint8_t* txBuffer, int& size );
        void                    txSyncModeNeedTestPageSize( uint8_t* txBuffer, int& size );
        int                     txSyncModeHavePageSize( uint8_t* txBuffer, int& size );
        void                    txWait( uint8_t* txBuffer, int& size );
        void                    txIdle( uint8_t* txBuffer, int& size );

        bool                    checkRtCrc();
        void                    sendRtTimePacket();
        void                    sendRtPacket();
        void                    notify();
        void                    resetRxProperties();
        void                    sendEvent( const SIPRPacket& packet );
        void                    sendEvent( int type );
        void                    sendSIPRGetTableEvent( int megatunePageNo );
        void                    setReadRealTimeCommand();
        MegatuneCommand*        createReadPageCommand( int pageNo );
        bool                    isPageCrcEnabled();
};

#endif // MEGATUNEPROTOCOL_H
