#include "SDCardController.h"
#include "VemsDeviceControllerBase.h"
#include "SIPRTypes.h"
#include "../logger.h"
#include "../Interruptable.h"
#include "../communication/clients/DeviceModeSelector.h"
#include "../communication/clients/BaudRateSelector.h"
#include "../communication/clients/SDCardFileSearch.h"
#include "../communication/clients/SDCardFileDownloader.h"
#include "../communication/clients/SDCardFileSummary.h"
#include "../communication/clients/SDCardV3Logger.h"
#include "../events/SDCardEvent.h"
#include "../configmanager/ConfigHelper2.h"
#include "../configmanager/CommunicationConfig.h"
#include "../configmanager/ConfigMgrCommunication.h"
#include "../ConfigBase.h"
#include "../CommunicationBase.h"

IMPLEMENT_DYNAMIC_CLASS( SDCardController, wxEvtHandler )


SDCardController::SDCardController()
{
    m_eventHandler              = NULL;
    m_deviceController          = NULL;
    Connect( wxEVT_SDCARD, SDCardEventHandler( SDCardController::OnSDCardEvent ), NULL, this );

    m_task = IDLE;
}

SDCardController::SDCardController( VemsDeviceControllerBase* deviceController, wxEvtHandler* eventHandler )
{
    m_deviceController  = deviceController;
    m_eventHandler      = eventHandler;
    Connect( wxEVT_SDCARD, SDCardEventHandler( SDCardController::OnSDCardEvent ), NULL, this );

    m_task = IDLE;
}

SDCardController::~SDCardController()
{
    Disconnect( wxEVT_SDCARD, SDCardEventHandler( SDCardController::OnSDCardEvent ), NULL, this );
}

void SDCardController::setDeviceController( VemsDeviceControllerBase* deviceController )
{
    m_deviceController = deviceController;
}

void SDCardController::setEventHandler( wxEvtHandler* eventHandler  )
{
    m_eventHandler = eventHandler;
}

void SDCardController::search()
{
    ///TODO - check ?
    //    if ( !isConnected() || !isV3() || isBootMode() )
//    {
//        return false;
//    }

    m_task = SEARCH;

    if ( isBaudrateSelected() ) {
        LOG
        if ( m_deviceController->isManMode() ) {
            m_deviceController->registerClient( new SDCardFileSearch( this ) );
        } else {
            registerDeviceModeSelector();
        }
    } else {
            LOG
            m_deviceController->registerClient( new BaudRateSelector(   this, this,
                                                                        m_deviceController->getCommBase(),
                                                                        m_deviceController->getSupportedBaudrates(),
                                                                        m_deviceController->getDeviceMode(),
                                                                        getCurrentBaudrate(),
                                                                        getConfigBaudrate() ) );
    }

    m_deviceController->startCommControllerNow();
}

void SDCardController::download( const MMCFilesDesc& desc )
{
    m_task = DOWNLOAD;

    m_deviceController->registerClient( new SDCardFileDownloader( desc, this ) );

    m_deviceController->startCommControllerNow();
}

void SDCardController::createSummary( const MMCFilesDesc& desc )
{
    LOG

    m_task = CREATE_SUMMARY;

    m_deviceController->registerClient( new SDCardFileSummary( desc, this ) );

    m_deviceController->startCommControllerNow();

    LOG
}

void SDCardController::abort()
{
    CommunicationControllerClient* commClient = m_deviceController->getCommClient();

    // try to cast to Interruptable
    Interruptable* client = dynamic_cast<Interruptable*>( commClient );

    if ( client )
    {
        client->Abort();
    }
    else
    {
        LOGP( _T("Abort SD Process - cannot cast client to Interruptable!") )
    }
}

void SDCardController::startLog()
{
    m_task = START_LOG;

    if ( m_deviceController->isManMode() ) {
        m_deviceController->registerClient( new SDCardV3Logger( (SDLogCommand)0, this ) );
    } else {
        registerDeviceModeSelector();
    }

    m_deviceController->startCommControllerNow();
}

void SDCardController::getFwVersion()
{
    m_task = SD_FWVER;

    if ( m_deviceController->isManMode() )
    {
        m_deviceController->registerClient( new SDCardV3Logger( (SDLogCommand) SD_GET_SD_FIRMWARE , this ) );
    }
    else
    {
        registerDeviceModeSelector();
    }

    m_deviceController->startCommControllerNow();
}

void SDCardController::stopLog()
{
    m_task = STOP_LOG;

    if ( m_deviceController->isManMode() ) {
        LOG
        m_deviceController->registerClient( new SDCardV3Logger( (SDLogCommand)1, this ) );
    } else {
//        LOG
        registerDeviceModeSelector();
    }

    m_deviceController->startCommControllerNow();
}

void SDCardController::erase()
{
    m_task = ERASE;

    if ( m_deviceController->isManMode() ) {
        m_deviceController->registerClient( new SDCardV3Logger( (SDLogCommand)2, this ) );
    } else {
        registerDeviceModeSelector();
    }

    m_deviceController->startCommControllerNow();
}

void SDCardController::OnDeviceModeSelected( ECUInfo::EcuMode deviceMode )
{
    switch ( m_task )
    {
        case SEARCH:
            m_deviceController->registerClient( new SDCardFileSearch( this ) );
            break;

        case START_LOG:
            m_deviceController->registerClient( new SDCardV3Logger( (SDLogCommand)0, this ) );
            break;

        case STOP_LOG:
//            LOG
            m_deviceController->registerClient( new SDCardV3Logger( (SDLogCommand)1, this ) );
            break;

        case ERASE:
            m_deviceController->registerClient( new SDCardV3Logger( (SDLogCommand)2, this ) );
            break;

        case SD_FWVER:
            LOG
            m_deviceController->registerClient( new SDCardV3Logger( (SDLogCommand)SD_GET_SD_FIRMWARE, this ) );
            break;

        default:
            break;
    }

    m_deviceController->startCommControllerNow();
}

void SDCardController::OnBaudrateChanged( size_t baudrate )
{
    LOGP( wxString::Format( _T("baudrate") ) );
    m_deviceController->setCommBaudrate( baudrate );

    if ( m_deviceController->isManMode() ) {
        m_deviceController->registerClient( new SDCardFileSearch( this ) );
    } else {
        registerDeviceModeSelector();
    }

    m_deviceController->startCommControllerNow();
}

void SDCardController::OnTimeout()
{
    LOG
//    CommunicationEvent event( CommunicationEvent::TIMEOUT );
//    event.SetEventObject( this );
//    m_eventHandler->AddPendingEvent( event );
}

void SDCardController::registerDeviceModeSelector()
{
    m_deviceController->registerClient( new DeviceModeSelector( this, this,
                                                                m_deviceController->getCommBase(),
                                                                m_deviceController->getSupportedBaudrates(),
                                                                m_deviceController->getDeviceMode(),
                                                                ECUInfo::MAN,
                                                                &m_deviceMode ) );
}

void SDCardController::OnSDCardEvent( SDCardEvent& event )
{
    m_eventHandler->AddPendingEvent( event );
}

bool SDCardController::isBaudrateSelected()
{
    return ( m_deviceController->getCommBase()->getBaudRate() == getConfigBaudrate() );
}

size_t SDCardController::getCurrentBaudrate()
{
    return m_deviceController->getCommBase()->getBaudRate();
}

size_t SDCardController::getConfigBaudrate()
{
    CommunicationConfig commConfig;
    ConfigBase* cfg = ConfigHelper2::getConfig(ConfigHelper2::CONFIG_MAIN);
    ConfigMgrCommunication cfgMgr(cfg);
    cfgMgr.load(&commConfig);

    LOGP( wxString::Format( _T("%d"), commConfig.sdcardBaudrate ) );

    return (size_t)commConfig.sdcardBaudrate;
}
