#include "ContentLoaderManager.h"

#include "../logger.h"
#include "../common/types/IntArray.h"


ContentLoaderManager::ContentLoaderManager(int id, ContentLoaderBase * loader, wxEvtHandler* resultHandler, int threadCount, int timebasems)
    :m_mainTimer(this,idTimer)
{
    //ctor
    m_id = id;

    m_maxThreadCount = threadCount;

    m_timeBase = timebasems;



    Connect(idTimer, wxEVT_TIMER, wxTimerEventHandler(ContentLoaderManager::OnTimer), NULL, this);

    m_loaderProvider = loader;

    m_resultHandler = resultHandler;

    m_tryCount = 5;
    m_timeout = 5000;

    Connect(EVT_DOWNLOADING_STATUS, DownloadingStatusEventHandler(ContentLoaderManager::OnContentProgressEnd), NULL, this );

    m_mainTimer.Start(m_timeBase);
}

ContentLoaderManager::~ContentLoaderManager()
{
    Disconnect(EVT_DOWNLOADING_STATUS, DownloadingStatusEventHandler(ContentLoaderManager::OnContentProgressEnd), NULL, this );

    //dtor
    m_mainTimer.Stop();

    for(DownloaderProcessHash::iterator it = m_processes.begin(); it != m_processes.end(); it++) {
        DownloaderProcess * proc = it->second;
        if(proc->isRunning())
            proc->kill();

        if(!proc->isRunning())
            delete proc;
    }

    m_processes.clear();

    //delete m_mainTimer;
}

void ContentLoaderManager::OnContentProgressEnd(DownloadingStatusEvent& evt)
{
    wxVariantHash result = evt.GetProperties();

    if(result.find(_T("resultType")) == result.end() || result.find(_T("id")) == result.end() ) {
        Logger::logf(LOG_MSG,_T("ContentLoaderManager::notify FAIL!"));
        return;
    }

    int resultType = result[_T("resultType")].GetLong();

    int pid = result[_T("id")].GetLong();

    DownloaderProcess* proc = NULL;

    if(m_processes.find(pid) != m_processes.end()) {
        proc = m_processes[pid];
    }

//    if(m_processes.find(pid) == m_processes.end()) {
//        Logger::logf(LOG_MSG,_T("ContentLoaderManager::notify FAIL!"));
//        return;
//    }



    DownloadingStatusEvent event;
    event.SetId(m_id);
    event.SetProperties(result);

    switch(resultType) {
        case CLR_DEFAULT_STATE: {
            long trycnt = result[_T("tryCount")].GetLong();
            if( trycnt >= m_tryCount ) break;
            result[_T("tryCount")] = ++trycnt;
            AddQuery(result);
            }break;
        case CLR_NO_ERROR:
        case CLR_WRONG_URL:
        case CLR_WRONG_FORMAT:
        case CLR_FILE_NOT_FOUND:
            m_resultHandler->AddPendingEvent(event);
            RemoveProcess(pid);
            break;

        case CLR_NO_CONNECTION:
            if(proc && proc->GetTryCount() <= m_tryCount){
                if(proc->isRunning())
                    proc->stop();

                if(!proc->isRunning())
                    proc->start();
                else
                    Logger::logf(LOG_MSG,_T("OnContentProgressEnd state: OMFG!"));
            } else
                RemoveProcess(pid);
            break;
    }

}

void ContentLoaderManager::ThrowRemoveThread(int pid)
{
    wxCommandEvent evt(wxEVT_COMMAND_ENTER);
    evt.SetInt(pid);
    AddPendingEvent(evt);
}

void ContentLoaderManager::notify( wxVariantHash& hash )
{
    wxVariantHash rhash = hash;

    DownloadingStatusEvent event;
    event.SetId(m_id);
    event.SetProperties(rhash);

    AddPendingEvent(event);
}

void ContentLoaderManager::RemoveProcess(int pid, bool isKill)
{
    if( m_processes.find(pid) == m_processes.end() ) return;

    DownloaderProcess * proc = m_processes[pid];

    if(proc->isRunning()){
        if(isKill)
            proc->kill();
        else
            proc->stop();
    }

    if(!proc->isRunning())
        delete proc;

    m_processes.erase(pid);
}

void ContentLoaderManager::AddQuery(wxVariantHash parameters)
{
    m_queue.Add(parameters);
}

void ContentLoaderManager::AddFile(const wxString& path)
{
    wxVariantHash p;
    p[_T("filepath")] = path;

    m_queue.Add(p);
}


void ContentLoaderManager::AddURL(const wxString& url)
{
    wxVariantHash p;
    p[_T("url")] = url;

    m_queue.Add(p);
}

void ContentLoaderManager::OnTimer(wxTimerEvent& evt)
{
    IntArray remprocs;

    for(DownloaderProcessHash::iterator it = m_processes.begin(); it != m_processes.end(); it++) {
        DownloaderProcess * proc = it->second;
        if( proc->IsTimeOuted(m_timeout) ) remprocs.Add(it->first);
    }

    for(size_t a=0; a<remprocs.GetCount(); a++ ) RemoveProcess( remprocs[a], true );

    while( m_queue.GetCount() > 0 && m_processes.size() < m_maxThreadCount ) {
        int id = wxNewId();

        DownloaderProcess * proc = new DownloaderProcess(id, m_loaderProvider, m_queue[0]);
        proc->registerObserver(this);
        m_queue.RemoveAt(0);

        m_processes[id] = proc;

        proc->start();
    }

}

/***************************************************************************************************
                DownloaderProcess
*/

DownloaderProcess::DownloaderProcess(int id, ContentLoaderBase* loader, wxVariantHash params)
{
    //Logger::logf(LOG_MSG,_T("DownloaderProcess Ctor pid=%i"),wxThread::GetCurrentId());

    m_tryCount = 0;
    m_loader = loader;
    m_params = params;
    m_id = id;
    m_startTime = wxDateTime::UNow();

    m_params[_T("id")] = (long)m_id;

    m_params[_T("resultType")] = (long)CLR_DEFAULT_STATE;
}


bool DownloaderProcess::IsTimeOuted(int timeoutMS)
{
    wxTimeSpan span = wxDateTime::UNow().Subtract( m_startTime );

    if( span.GetMilliseconds().ToLong() > timeoutMS ) return true;

    return false;
}

bool DownloaderProcess::process()
{
    m_tryCount++;
    m_startTime = wxDateTime::UNow();

    m_loader->Load(m_params);

    return false;
}

void DownloaderProcess::finished()
{
    //Logger::logf(LOG_MSG,_T("finished()"));
    observerNotify(m_params);
}
