#include "OSMTileManager.h"

#include <wx/filename.h>
#include <wx/dir.h>

#include "../web/ImageContentLoaderFile.h"
#include "../web/ImageContentLoaderURL.h"
#include "../web/ContentLoaderManager.h"
#include "../logger.h"
#include "../DirHelper.h"

IMPLEMENT_DYNAMIC_CLASS(TileDesc, wxObject);
IMPLEMENT_VARIANT_OBJECT_SHALLOWCMP(TileDesc);

OSMTileManager::OSMTileManager(const wxString& tileserver, const wxString& cachedir)
{
    //ctor

    m_webDownMgr = new ContentLoaderManager(idWebDownloader, new ImageContentLoaderURL(),this, 3 );
    m_cacheMgr = new ContentLoaderManager(idCacheLoader, new ImageContentLoaderFile(),this, 3, 10 );
    m_cacheMgr->SetTryCount(2);
    m_cacheMgr->SetTimeoutMs(1000);


    Connect(EVT_DOWNLOADING_STATUS, DownloadingStatusEventHandler(OSMTileManager::OnLoadedFile), NULL, this);

    m_cacheDir = cachedir;
    m_tileServer = tileserver;

    if( !wxFileName::DirExists(m_cacheDir) ) DirHelper::MkDir(m_cacheDir);

    if( m_cacheDir.Last() != DirHelper::Separ() ) m_cacheDir += DirHelper::Separ();

}

OSMTileManager::~OSMTileManager()
{
    Disconnect(EVT_DOWNLOADING_STATUS, DownloadingStatusEventHandler(OSMTileManager::OnLoadedFile), NULL, this);
    //dtor
    delete m_webDownMgr;
    delete m_cacheMgr;
}

void OSMTileManager::Request(TileDesc desc, bool checkInHist)
{
    if( !desc.IsOk() ) return;

    if(checkInHist && CheckReqTilesList(desc)) return;

    m_reqTilesHistory.Add(desc);

    if( !desc.IsOk() ) return;

    wxString path = m_cacheDir + desc.GetAsString() + _T(".png");

    //Logger::logf(LOG_MSG,_T("tilemgr req tile: ")+desc.GetAsString());

    wxVariantHash hash;
    hash[_T("tile")] << desc;

    if(wxFileName::FileExists(path) && wxFileName::GetSize(path).ToULong() > 0) {
        hash[_T("filepath")] = path;
        m_cacheMgr->AddQuery(hash);
    } else {
        hash[_T("url")] = m_tileServer + _T("/") + desc.GetAsString() + _T(".png");
        m_webDownMgr->AddQuery(hash);
    }

}

bool OSMTileManager::CheckReqTilesList(TileDesc& desc)
{
    for(size_t a = 0; a < m_reqTilesHistory.GetCount(); a++) {
        if(desc == m_reqTilesHistory[a]) return true;
    }
    return false;
}

void OSMTileManager::OnLoadedFile(DownloadingStatusEvent& evt)
{
    wxVariantHash props = evt.GetProperties();

    int id = evt.GetId();

    if( props.find(_T("resultType")) == props.end() ) return;

    if(props.find(_T("tile")) == props.end()) {
        LOGP(_T("no tile"));
        return;
    }

    int result = props[_T("resultType")].GetLong();

    TileDesc desc;
    desc << props[_T("tile")];

    if(props.find(_T("image")) == props.end()) {
        LOGP(wxString::Format(_T("no image res=%i"),result));
        if(result != CLR_WRONG_FORMAT)
            Request(desc, false);
        return;
    }

    if(id == idWebDownloader) SaveTile( props );

    if(result != CLR_NO_ERROR) {
        LOGP(_T("wrong tile"));
        return;
    }

    observerNotify( props );
}

void OSMTileManager::SaveTile(wxVariantHash& hash)
{
    if(hash.find(_T("tile")) == hash.end()) {
        LOGP(_T("no tile"));
        return;
    }

    if(hash.find(_T("image")) == hash.end()) {
        LOGP(_T("no image"));
        return;
    }

    TileDesc desc;
    desc << hash[_T("tile")];

    wxImage img;
    img << hash[_T("image")];

    if( !img.IsOk() ) return;

    wxString path = m_cacheDir + wxString::Format(_T("%i%s%i"),desc.zoom,DirHelper::Separ().c_str(),desc.x);

    DirHelper::CheckDir( path );

    img.SaveFile( path + DirHelper::Separ() + wxString::Format(_T("%i.png"),desc.y), wxBITMAP_TYPE_PNG );
}

wxString OSMTileManager::GetCacheSizeStr()
{
    return wxFileName::GetHumanReadableSize(wxDir::GetTotalSize(m_cacheDir));
}

size_t OSMTileManager::GetCacheSize()
{
    return wxDir::GetTotalSize(m_cacheDir).ToULong();
}

/*********************************************************
                    TileDesc
*/

bool TileDesc::operator!=(const TileDesc& tile) const
{
    if( (*this) == tile )
        return false;
    else
        return true;
}

bool TileDesc::operator==(const TileDesc& tile) const
{
    if( tile.x == x && tile.y == y && tile.zoom == zoom )
        return true;
    else
        return false;
}
