#include "wxSqliteHelper.h"

#include <wx/filename.h>
#include <wx/ffile.h>
#include <wx/tokenzr.h>
#include <wx/wxsqlite3.h>

#include "common/VemsMessageHashHelper.h"
#include "common/types/StringHash.h"
#include "common/typedef.h"

#include "UserLog.h"
#include "logger.h"

#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(ColumnDescArray);

bool wxSqliteHelper::CheckDBFile(const wxString& filepath)
{
    wxSQLite3Database db;
    db.Open(filepath);

    if( db.IsOpen() == false ) return false;

    bool res = true;

    try {
        db.ExecuteQuery(_T("SELECT * FROM sqlite_master WHERE type='table';"));
    } catch ( wxSQLite3Exception& se ) {
        res = false;
    }

    db.Close();

    return res;
}

void wxSqliteHelper::EscapeString(wxString &str)
{
    str.Replace(_T("\\"),_T("\\\\"));
    str.Replace(_T("'"),_T("\\'"));
    str.Replace(_T("\""),_T("\\\""));
}

wxString wxSqliteHelper::EscapeString(const wxString &str)
{
    wxString ret = str;
    EscapeString(ret);
    return ret;
}

bool wxSqliteHelper::Exist(const wxString &dbpath)
{
    wxFileName fn(dbpath);

    if(!fn.FileExists()) {
        UserLog::uLog(ULOG_ERROR, _T("File not found! ") + dbpath );
        return false;
    }

    if( fn.GetExt() != _T("db") && fn.GetExt() != _T("s3db") ) {
        UserLog::uLog(ULOG_ERROR, _T("Wrong database file! It must be *.db or *.s3db"));
        return false;
    }

    return true;
}

bool wxSqliteHelper::TablesCheck( const wxArrayString &tablenames, const wxString dbfilename )
{
    bool res = true;

    wxSQLite3Database db;
    db.Open( dbfilename );

    StringHash dbe_hash;

    dbe_hash[_T("file")] = dbfilename;

    for( uint8_t a = 0; a < tablenames.GetCount(); a++){
        if( !db.TableExists( tablenames[a] ) ) {
            dbe_hash[_T("table")] = tablenames[a];
            UserLog::vLog( VemsMessageHashHelper::GetVemsMsg(_T("db_table_error_1"),&dbe_hash) );
            res = false;
        }
    }
    db.Close();
    return res;
}



bool wxSqliteHelper::ColumnExists(wxSQLite3Database *db, const wxString &tableName, const wxString &columnName, bool isCreate, const wxString &colType)
{
    bool res = true;

    if( !(db->TableExists(tableName)) ) return false;

    try{
        db->GetMetaData(wxEmptyString, tableName, columnName, NULL, NULL);
    } catch ( wxSQLite3Exception &exc ) {
        res = false;
        if(isCreate) {
            try{
                db->ExecuteUpdate( _T("alter table ") + tableName + _T(" add column ") + columnName + _T(" ") + colType + _T(";") );
            } catch (wxSQLite3Exception &exc) {
                Logger::logf(LOG_MSG,_T("wxSqliteHelper::ColumnExists() ") + exc.GetMessage() );
            }
        }
    }

    return res;
}

void wxSqliteHelper::SQLLineToVariantHash(wxSQLite3ResultSet& res, wxVariantHash& hash)
{
    for(int a = 0; a < res.GetColumnCount(); a++) {

        switch( res.GetColumnType(a) ){
            case WXSQLITE_INTEGER:
                hash[ res.GetColumnName(a) ] = (long)(res.GetInt(a)); break;
            case WXSQLITE_FLOAT:
                hash[ res.GetColumnName(a) ] = res.GetDouble(a); break;
            case WXSQLITE_TEXT:
                hash[ res.GetColumnName(a) ] = res.GetString(a); break;
            case WXSQLITE_NULL:
                hash[ res.GetColumnName(a) ] = wxNullVariant; break;
            default:
                break;
        }
    }
}

wxString wxSqliteHelper::ConvertWxTypesToSQLTypes( const wxString &wxdatatype)
{
    wxString field = _T("unknown");

    if( wxdatatype == _T("string") )
        field = _T("varchar(255)");         /// if you want to use longer string... you dont want!

    else if( wxdatatype == _T("double") )
        field = _T("float");

    else if( wxdatatype == _T("long") )
        field = _T("integer");

    else if( wxdatatype == _T("bool") )
        field = _T("boolean");

    return field;
}


bool wxSqliteHelper::DatabaseCheck(wxSQLite3Database& db, ColumnDescArrayHash& desc, const wxString& dbname)
{
    for( ColumnDescArrayHash::iterator it = desc.begin(); it != desc.end(); it++ ) {

        wxString tableName = it->first;
        ColumnDescArray& coloumarray = it->second;

        if( coloumarray.GetCount() == 0 ) continue;

        size_t a = 0;

        if(db.TableExists( tableName, dbname ) == false) {

            //wxString qs = _T("create table ") + ( dbname==wxEmptyString ? _T("") : (dbname+_T(".")) ) + tableName + _T(" ( ");
			wxString qs = _T("create table ");
			qs += ( dbname==wxEmptyString ? _T("") : (dbname+_T(".")) ) + tableName + _T(" ( ");

            qs += coloumarray[0].GetColDefStr(); /// sqlite cant create empty table

            qs += _T(" );");
            db.ExecuteUpdate( qs );
            a++;
        }


        for( ; a < coloumarray.GetCount(); a++){

            ColumnDesc& col = coloumarray[a];

            wxString columnName = col.Name;
            wxString colType = col.Type;

            try{
                db.GetMetaData(dbname, tableName, columnName, NULL, NULL);
            } catch ( wxSQLite3Exception &exc ) {
                try{

                    //wxString qs = _T("alter table ") + ( dbname==wxEmptyString ? _T("") : (dbname+_T(".")) ) + tableName;
					wxString qs = _T("alter table ");
					qs += ( dbname==wxEmptyString ? _T("") : (dbname+_T(".")) ) + tableName;

                    qs += _T(" add column ") + col.GetColDefStr() + _T(" ;");

                    db.ExecuteUpdate( qs );

                } catch (wxSQLite3Exception &exc) {
                    Logger::logf(LOG_MSG,_T("wxSqliteHelper::DatabaseCheck exception: ") + exc.GetMessage() );
                }
            }

        }

    }

    return true;

}

bool wxSqliteHelper::DatabaseCheck(const wxString& filename, ColumnDescArrayHash& desc, const wxString& dbname)
{
    wxSQLite3Database db;

    db.Open(filename);

    bool res = DatabaseCheck(db,desc,dbname);

    db.Close();

    return res;
}


void wxSqliteHelper::ExecuteFromFile(wxSQLite3Database *db, const wxString& filepath)
{
    wxFFile file(filepath);
    wxString content;

    file.ReadAll(&content);

    file.Close();

    wxArrayString arr = wxStringTokenize(content,_T("\n"));

    db->Begin();

    try {
        for(size_t a = 0; a < arr.GetCount(); a++) db->ExecuteUpdate( arr[a] );
    } catch ( wxSQLite3Exception& ex ) {
        Logger::logf(LOG_MSG,_T("wxSqliteHelper::ExecuteFromFile fail")+ex.GetMessage());
    }

    db->Commit();


}
